Package: kindling
Type: Package
Title: Higher-Level Interface of 'torch' Package to Auto-Train Neural
        Networks
Version: 0.3.1
Authors@R: c(
    person("Joshua", "Marie", 
           email = "joshua.marie.k@gmail.com", 
           role = c("aut", "cre")),
    person("Antoine", "Soetewey", 
           email = "ant.soetewey@gmail.com", 
           role = "aut",
           comment = c(ORCID = "0000-0001-8159-0804"))
  )
Description: Provides a higher-level interface to the 'torch' package for defining,
    training, and fine-tuning neural networks through code generation.
    The package supports several architectures, including feedforward
    (multi-layer perceptron) and recurrent neural networks (RNN, LSTM,
    GRU), while reducing boilerplate 'torch' code. Model training
    methods also bridge to machine learning frameworks in R,
    particularly the 'tidymodels' ecosystem, including 'parsnip' model
    specifications, workflows, recipes, and tuning tools.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: purrr, torch, rlang, cli, glue, vctrs, parsnip (>= 1.0.0),
        tibble, tidyr, dplyr, stats, NeuralNetTools, ggplot2, tune,
        dials, hardhat, lifecycle, coro
Suggests: testthat (>= 3.0.0), magrittr, box, vip, recipes, workflows,
        rsample, yardstick, mlbench, modeldata, knitr, rmarkdown,
        DiceDesign, lhs, sfd, covr
Config/testthat/edition: 3
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
URL: https://kindling.joshuamarie.com,
        https://github.com/joshuamarie/kindling
BugReports: https://github.com/joshuamarie/kindling/issues
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-07-02 03:22:59 UTC; DESKTOP
Author: Joshua Marie [aut, cre],
  Antoine Soetewey [aut] (ORCID: <https://orcid.org/0000-0001-8159-0804>)
Maintainer: Joshua Marie <joshua.marie.k@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-02 06:30:02 UTC
Built: R 4.5.2; ; 2026-07-02 10:35:40 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$",ffnn_fit)
S3method("$",layer_pr)
S3method("$",nn_fit)
S3method("$",rnn_fit)
S3method(autoplot,nn_fit)
S3method(garson,ffnn_fit)
S3method(grid_depth,default)
S3method(grid_depth,list)
S3method(grid_depth,model_spec)
S3method(grid_depth,param)
S3method(grid_depth,parameters)
S3method(grid_depth,workflow)
S3method(olden,ffnn_fit)
S3method(plot,nn_fit)
S3method(predict,ffnn_fit)
S3method(predict,nn_fit)
S3method(predict,nn_fit_ds)
S3method(predict,nn_fit_tab)
S3method(predict,rnn_fit)
S3method(print,early_stop_spec)
S3method(print,ffnn_fit)
S3method(print,layer_index_pr)
S3method(print,layer_input_pr)
S3method(print,layer_is_output_pr)
S3method(print,layer_output_pr)
S3method(print,layer_pr)
S3method(print,mlp_kindling)
S3method(print,nn_arch)
S3method(print,nn_fit)
S3method(print,rnn_fit)
S3method(print,rnn_kindling)
S3method(print,train_nnsnip)
S3method(train_nn,data.frame)
S3method(train_nn,dataset)
S3method(train_nn,default)
S3method(train_nn,formula)
S3method(train_nn,matrix)
S3method(translate,mlp_kindling)
S3method(translate,rnn_kindling)
S3method(translate,train_nnsnip)
S3method(tunable,mlp_kindling)
S3method(tunable,rnn_kindling)
S3method(tunable,train_nnsnip)
S3method(update,mlp_kindling)
S3method(update,rnn_kindling)
S3method(update,train_nnsnip)
export(.i)
export(.in)
export(.is_output)
export(.layer)
export(.out)
export(act_funs)
export(activations)
export(args)
export(autoplot_diagnostics)
export(bias)
export(bidirectional)
export(early_stop)
export(ffnn)
export(ffnn_generator)
export(ffnn_wrapper)
export(garson)
export(grid_depth)
export(hidden_neurons)
export(mlp_kindling)
export(n_hlayers)
export(new_act_fn)
export(nn_arch)
export(nn_module_generator)
export(olden)
export(optimizer)
export(output_activation)
export(plot_diagnostics)
export(rnn)
export(rnn_generator)
export(rnn_kindling)
export(rnn_wrapper)
export(table_summary)
export(train_nn)
export(train_nn_wrapper)
export(train_nnsnip)
export(validation_split)
importFrom(NeuralNetTools,garson)
importFrom(NeuralNetTools,olden)
importFrom(cli,cat_line)
importFrom(cli,cli_abort)
importFrom(cli,rule)
importFrom(cli,style_bold)
importFrom(cli,style_italic)
importFrom(ggplot2,autoplot)
importFrom(glue,glue)
importFrom(lifecycle,badge)
importFrom(parsnip,translate)
importFrom(purrr,imap)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,set_names)
importFrom(purrr,transpose)
importFrom(rlang,":=")
importFrom(rlang,as_string)
importFrom(rlang,call2)
importFrom(rlang,call_name)
importFrom(rlang,enquo)
importFrom(rlang,enquos)
importFrom(rlang,eval_tidy)
importFrom(rlang,exec)
importFrom(rlang,expr)
importFrom(rlang,f_rhs)
importFrom(rlang,is_call)
importFrom(rlang,is_formula)
importFrom(rlang,is_installed)
importFrom(rlang,new_function)
importFrom(rlang,quo_get_expr)
importFrom(rlang,quo_is_null)
importFrom(rlang,sym)
importFrom(stats,reorder)
importFrom(stats,setNames)
importFrom(stats,update)
importFrom(tune,tunable)
importFrom(vctrs,new_vctr)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# kindling <img src="man/figures/logo.png" align="right" alt="" width="120"/>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/kindling)](https://CRAN.R-project.org/package=kindling)
[![R-CMD-check](https://github.com/joshuamarie/kindling/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/joshuamarie/kindling/actions/workflows/R-CMD-check.yaml)
[![CRAN
Downloads](https://cranlogs.r-pkg.org/badges/kindling)](https://CRAN.R-project.org/package=kindling)
<!-- [![CRAN Downloads Total](https://cranlogs.r-pkg.org/badges/grand-total/kindling?color=brightgreen)](https://CRAN.R-project.org/package=kindling) -->
[![Codecov test
coverage](https://codecov.io/gh/joshuamarie/kindling/graph/badge.svg)](https://app.codecov.io/gh/joshuamarie/kindling)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
<!-- badges: end -->

## Package overview

Title: ***Higher-Level Interface of ‘torch’ Package to Auto-Train Neural
Networks***

Whether you’re generating neural network architecture expressions or
directly fitting/training models, `{kindling}` minimizes boilerplate
code while preserving `{torch}`. Since this package uses `{torch}` as
its backend, GPU acceleration is supported.

`{kindling}` also bridges the gap between `{torch}` and `{tidymodels}`.
It works seamlessly with `{parsnip}`, `{recipes}`, and `{workflows}` to
bring deep learning into your existing `{tidymodels}` modeling pipeline.
This enables a streamlined interface for building, training, and tuning
deep learning models within the familiar `{tidymodels}` ecosystem.

### Main Features

<!-- -   Seamless integration with `parsnip` through `set_engine("kindling")` -->

- Code generation of `{torch}` expression

- Multiple architectures available

  - Base models interface: feedforward networks (MLP/DNN/FFNN) and
    recurrent variants (RNN, LSTM, GRU)
  - Generalized neural network trainer that has the same topology as
    MLPs

- Native support for R ML workflows and pipelines (currently
  `{tidymodels}`; `{mlr3}` planned)

- Fine-grained control over network depth, layer sizes, and activation
  functions

- GPU acceleration support via `{torch}` tensors

## Installation

You can install `{kindling}` on CRAN:

``` r
install.packages('kindling')
```

Or install the development version from GitHub:

``` r
# install.packages("pak")
pak::pak("joshuamarie/kindling")
## devtools::install_github("joshuamarie/kindling")
```

## Learn more

- [Getting Started with
  kindling](https://kindling.joshuamarie.com/articles/kindling.html)
- [Tuning
  Capabilities](https://kindling.joshuamarie.com/articles/tuning-capabilities.html)
- [Custom Activation
  Function](https://kindling.joshuamarie.com/articles/custom-act-fn.html)
- [Special Cases: Linear and Logistic
  Regression](https://kindling.joshuamarie.com/articles/special-cases.html)
- [Similar Packages and
  Comparison](https://kindling.joshuamarie.com/articles/similar-packages.html)

## References

Falbel D, Luraschi J (2023). *torch: Tensors and Neural Networks with
‘GPU’ Acceleration*. R package version 0.13.0,
<https://torch.mlverse.org>, <https://github.com/mlverse/torch>.

Wickham H (2019). *Advanced R*, 2nd edition. Chapman and Hall/CRC. ISBN
978-0815384571, <https://adv-r.hadley.nz/>.

Goodfellow I, Bengio Y, Courville A (2016). *Deep Learning*. MIT Press.
<https://www.deeplearningbook.org/>.

## Citation

If you use `{kindling}` in a publication, please cite it. Run
`citation("kindling")` in R to get the current citation, or see the
[CITATION
file](https://github.com/joshuamarie/kindling/blob/main/inst/CITATION).

## License

MIT + file LICENSE

## Code of Conduct

Please note that the kindling project is released with a [Contributor
Code of
Conduct](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

box::use(
    kindling[train_nn, nn_arch, ffnn],
    torch[
        torch_randn, torch_zeros, torch_exp, 
        nn_module, nn_parameter, nn_linear
    ]
)

rbf_gaussian = function(alpha) {
    torch_exp(-1 * alpha$pow(2))
}

rbf_layer = nn_module(
    "RBFLayer",
    initialize = function(in_features, out_features, basis_func = \(alpha) torch_exp(-alpha)) {
        self$in_features = in_features
        self$out_features = out_features
        self$centres = nn_parameter(torch_randn(out_features, in_features) * 0.1)
        self$log_sigma = nn_parameter(torch_zeros(out_features))
        self$basis_func = basis_func
    },
    
    forward = function(input) {
        batch_size = input$size(1)
        
        x = input$unsqueeze(2)
        c = self$centres$unsqueeze(1)
        distances = (x - c)$pow(2)$sum(3)
        
        # Use log_sigma for numerical stability
        # beta = 1 / (2 * exp(2 * log_sigma)) = exp(-2 * log_sigma) / 2
        beta = torch_exp(-2 * self$log_sigma)$unsqueeze(1) / 2
        
        self$basis_func(distances * beta)
    }
)

model_rbf = train_nn(
    Species ~ .,
    data = iris,
    hidden_neurons = c(32, 16),
    epochs = 150,
    batch_size = 32,
    learn_rate = 0.01,
    loss = "cross_entropy",
    validation_split = 0.2,
    arch = nn_arch(
        nn_name = "RBFNet",
        nn_layer = ~ rbf_layer,
        out_nn_layer = ~ torch::nn_linear,
        layer_arg_fn = ~ if (.is_output) {
            list(.in, .out)
        } else {
            list(in_features = .in, out_features = .out)
        }
    )
)

model_nn = train_nn(
    Species ~ .,
    data = iris,
    hidden_neurons = c(32, 16),
    epochs = 150,
    batch_size = 32,
    learn_rate = 0.01,
    loss = "cross_entropy",
    validation_split = 0.2
)

model_nn2 = ffnn(
    Species ~ .,
    data = iris,
    hidden_neurons = c(32, 16),
    epochs = 150,
    batch_size = 32,
    learn_rate = 0.01,
    loss = "cross_entropy",
    validation_split = 0.2
)

preds_rbf = predict(model_rbf, newdata = iris)
preds_nn = predict(model_nn, newdata = iris)
preds_nn2 = predict(model_nn2, newdata = iris)

cat("\n\nPrediction of RBF Layer:\n\n")
table(actual = iris$Species, prediction = preds_rbf)

cat("\n\nPrediction of MLP:\n\n")
table(actual = iris$Species, prediction = preds_nn)

cat("\n\nPrediction of MLP (from `ffnn()` function):\n\n")
table(actual = iris$Species, prediction = preds_nn2)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
    eval = identical(Sys.getenv("NOT_CRAN"), "true"), 
    fig.width = 7,
    fig.height = 5,
    warning = FALSE,
    message = FALSE
)

# Sys.setenv("_R_USE_PIPEBIND_" = TRUE)

## -----------------------------------------------------------------------------
# box::use(
#     kindling[train_nn, act_funs, args],
#     recipes[
#         recipe, step_dummy, step_normalize,
#         all_nominal_predictors, all_numeric_predictors
#     ],
#     rsample[initial_split, training, testing],
#     yardstick[metric_set, rmse, rsq, accuracy, mn_log_loss],
#     dplyr[mutate, select],
#     tibble[tibble]
# )

## ----linear-data--------------------------------------------------------------
# set.seed(42)
# split = initial_split(mtcars, prop = 0.8)
# train = training(split)
# test = testing(split)
# 
# rec = recipe(mpg ~ ., data = train) |>
#     step_normalize(all_numeric_predictors())

## ----linear-fit---------------------------------------------------------------
# lm_nn = train_nn(
#     mpg ~ .,
#     data = train,
#     hidden_neurons = c(),
#     loss = torch::nnf_l1_loss,
#     optimizer = "rmsprop",
#     learn_rate = 0.01,
#     epochs = 200,
#     verbose = FALSE
# )
# 
# lm_nn

## ----linear-eval--------------------------------------------------------------
# preds = predict(lm_nn, newdata = test)
# 
# tibble(
#     truth = test$mpg,
#     estimate = preds
# ) |>
#     metric_set(rmse, rsq)(truth = truth, estimate = estimate)

## ----linear-compare-----------------------------------------------------------
# lm_fit = lm(mpg ~ ., data = train)
# 
# tibble(
#     truth = test$mpg,
#     estimate = predict(lm_fit, newdata = test)
# ) |>
#     metric_set(rmse, rsq)(truth = truth, estimate = estimate)

## ----binary-data--------------------------------------------------------------
# data("Sonar", package = "mlbench")
# 
# sonar = Sonar
# set.seed(42)
# split_s = initial_split(sonar, prop = 0.8, strata = Class)
# train_s = training(split_s)
# test_s = testing(split_s)
# 
# rec_s = recipe(Class ~ ., data = train_s) |>
#     step_normalize(all_numeric_predictors())

## ----binary-fit---------------------------------------------------------------
# logit_nn = train_nn(
#     Class ~ .,
#     data = train_s,
#     hidden_neurons = c(),
#     loss = "cross_entropy",
#     optimizer = "adam",
#     learn_rate = 0.01,
#     epochs = 200,
#     verbose = FALSE
# )
# 
# logit_nn

## ----binary-eval--------------------------------------------------------------
# preds_s = predict(logit_nn, newdata = test_s, type = "response")
# 
# tibble(
#     truth = test_s$Class,
#     estimate = preds_s
# ) |>
#     accuracy(truth = truth, estimate = estimate)

## ----logit-compare------------------------------------------------------------
# box::use(nnet[multinom])
# 
# glm_fit = glm(Class ~ ., data = train_s, family = binomial())
# 
# tibble(
#     truth = test_s$Class,
#     estimate = {
#         as.factor({
#             preds = predict(glm_fit, newdata = test_s, type = "response")
#             ifelse(preds < 0.5, "M", "R")
#         })
#     }
# ) |>
#     accuracy(truth = truth, estimate = estimate)

box::use(
    kindling[nn_module_generator, act_funs, args]
)

nn_module_generator(
    nn_name = "CNN1DClassifier",
    nn_layer = "nn_conv1d",
    layer_arg_fn = ~ if (.is_output) {
        list(.in, .out)
    } else {
        list(
            in_channels = .in,
            out_channels = .out,
            kernel_size = 3L,
            stride = 1L,
            padding = 1L 
        )
    },
    after_output_transform = ~ .$mean(dim = 2),
    last_layer_args = list(kernel_size = 1, stride = 2),
    hd_neurons = c(16, 32, 64),
    no_x = 1,
    no_y = 10,
    activations = "relu"
)

nn_module_generator(
    nn_name = "CNN1DClassifier",
    nn_layer = "nn_conv1d",
    layer_arg_fn = ~ if (.is_output) {
        list(.in, .out, kernel_size = 1, stride = 2)
    } else {
        list(
            in_channels = .in,
            out_channels = .out,
            kernel_size = 3L,
            stride = 1L,
            padding = 1L 
        )
    },
    after_output_transform = ~ .$mean(dim = 2),
    # last_layer_args = list(kernel_size = 1, stride = 2),
    hd_neurons = c(16, 32, 64),
    no_x = 1,
    no_y = 10,
    activations = "relu"
)

box::use(
    kindling[ffnn, act_funs, args],
    NeuralNetTools[garson, olden],
    vip[vip, vi]
)

cat("\n\n=== Example 1: FFNN Classification ===\n")

model_iris = ffnn(
    Species ~ .,
    data = iris,
    hidden_neurons = c(64, 32),
    activations = "relu",
    epochs = 100,
    verbose = FALSE,
    cache_weights = TRUE
)

cat("\nGarson's Algorithm:\n")

imp_garson = garson(model_iris, bar_plot = FALSE)
imp_garson

cat("\nOlden's Algorithm:\n")
imp_olden = olden(model_iris, bar_plot = FALSE)
imp_olden

cat("\nvip Integration:\n")
imp_vip = vi(model_iris)
imp_vip

cat("\n\n=== Example 2: FFNN Regression (mtcars dataset) ===\n")

model_mtcars = ffnn(
    mpg ~ cyl + disp + hp + drat + wt + qsec,
    data = mtcars,
    hidden_neurons = c(32, 16),
    activations = "relu",
    epochs = 150,
    loss = "mse",
    verbose = FALSE,
    cache_weights = TRUE
)

cat("\nVariable Importance (Olden's method):\n")
imp_reg = olden(model_mtcars, bar_plot = FALSE)
imp_reg

cat("\nVariable Importance (Garson's method):\n")
imp_reg = garson(model_mtcars, bar_plot = FALSE)
imp_reg

cat("\nvip Integration:\n")
imp_vip = vi(model_iris)
imp_vip

box::use(
    kindling[
        mlp_kindling, hidden_neurons,
        activations, output_activation,
        grid_depth
    ],
    parsnip[mlp, fit, augment, set_engine, set_mode],
    recipes[
        recipe, step_YeoJohnson,
        step_normalize, all_numeric_predictors
    ],
    workflows[workflow, add_recipe, add_model],
    rsample[
        initial_split, training, testing,
        vfold_cv
    ],
    tune[
        tune_grid, tune,
        select_best, finalize_workflow
    ],
    dials[
        grid_random, penalty, hidden_units
    ],
    yardstick[
        metric_set, accuracy, roc_auc
    ],
    dplyr[
        select, bind_rows, mutate
    ]
)

data(Ionosphere, package = "mlbench")

set.seed(123)

# =========================================================
# 1. TRAIN / TEST SPLIT
# =========================================================

split =
    initial_split(
        Ionosphere,
        prop = 0.8,
        strata = Class
    )

train = training(split)
test = testing(split)

folds =
    vfold_cv(
        train,
        v = 5,
        strata = Class
    )

# =========================================================
# 2. RECIPE
# =========================================================

data_rec =
    train |>
    select(!(1:2)) |>
    recipe(Class ~ .) |>
    step_YeoJohnson(all_numeric_predictors()) |>
    step_normalize(all_numeric_predictors())

# =========================================================
# 3A. KINDLING MODEL
# =========================================================

kindling_spec =
    mlp_kindling(
        mode = "classification",
        hidden_neurons = tune(),
        activations = tune(),
        output_activation = tune()
    )

kindling_wf =
    workflow() |>
    add_recipe(data_rec) |>
    add_model(kindling_spec)

kindling_grid =
    grid_depth(
        hidden_neurons(c(16L, 64L)),
        activations(c("relu", "elu", "softshrink(lambd = 0.2)")),
        output_activation(c("sigmoid", "relu")),
        n_hlayer = 2,
        size = 10,
        type = "latin_hypercube"
    )

kindling_tune =
    tune_grid(
        kindling_wf,
        resamples = folds,
        grid = kindling_grid,
        metrics = metric_set(accuracy, roc_auc)
    )

best_kindling =
    select_best(
        kindling_tune,
        metric = "roc_auc"
    )

final_kindling_wf =
    finalize_workflow(
        kindling_wf,
        best_kindling
    )

final_kindling_model =
    fit(
        final_kindling_wf,
        data = train
    )
final_kindling_model

kindling_train_metrics =
    final_kindling_model |>
    augment(new_data = train) |>
    metric_set(accuracy, roc_auc)(
        truth = Class,
        estimate = .pred_class,
        .pred_good
    )

kindling_test_metrics =
    final_kindling_model |>
    augment(new_data = test) |>
    metric_set(accuracy, roc_auc)(
        truth = Class,
        estimate = .pred_class,
        .pred_good
    )

# =========================================================
# 3B. BRULEE VIA PARSNIP MLP
# =========================================================

brulee_spec =
    mlp(
        hidden_units = tune(),
        penalty = tune(),
        epochs = 100
    ) |>
    set_mode("classification") |>
    set_engine("brulee")

brulee_wf =
    workflow() |>
    add_recipe(data_rec) |>
    add_model(brulee_spec)

brulee_grid =
    grid_random(
        hidden_units(c(32L, 128L)),
        penalty(),
        size = 10
    )

brulee_tune =
    tune_grid(
        brulee_wf,
        resamples = folds,
        grid = brulee_grid,
        metrics = metric_set(accuracy, roc_auc)
    )

best_brulee =
    select_best(
        brulee_tune,
        metric = "roc_auc"
    )

final_brulee_wf =
    finalize_workflow(
        brulee_wf,
        best_brulee
    )

final_brulee_model =
    fit(
        final_brulee_wf,
        data = train
    )
final_brulee_model

brulee_train_metrics =
    final_brulee_model |>
    augment(new_data = train) |>
    metric_set(accuracy, roc_auc)(
        truth = Class,
        estimate = .pred_class,
        .pred_good
    )

brulee_test_metrics =
    final_brulee_model |>
    augment(new_data = test) |>
    metric_set(accuracy, roc_auc)(
        truth = Class,
        estimate = .pred_class,
        .pred_good
    )

# =========================================================
# 4. FINAL COMPARISON
# =========================================================

train_comparison = 
    bind_rows(
        kindling_train_metrics |>
            mutate(model = "kindling"),
        brulee_train_metrics |>
            mutate(model = "brulee")
    )

test_comparison =
    bind_rows(
        kindling_test_metrics |>
            mutate(model = "kindling"),
        brulee_test_metrics |>
            mutate(model = "brulee")
    )

train_comparison
test_comparison

box::use(
    kindling[
        mlp_kindling, hidden_neurons, activations, output_activation, grid_depth
    ],
    parsnip[fit, augment],
    recipes[recipe],
    workflows[workflow, add_recipe, add_model],
    rsample[vfold_cv],
    tune[tune_grid, tune, select_best, finalize_workflow, collect_metrics],
    dials[grid_random],
    yardstick[accuracy, roc_auc, metric_set, metrics]
)

mlp_tune_spec = mlp_kindling(
    mode = "classification",
    hidden_neurons = tune(),
    activations = tune(),
    output_activation = tune()
)

iris_folds = vfold_cv(iris, v = 3)
nn_wf = workflow() |>
    add_recipe(recipe(Species ~ ., data = iris)) |>
    add_model(mlp_tune_spec)

nn_grid = grid_random(
    hidden_neurons(c(32L, 128L)),
    activations(c("relu", "elu")),
    output_activation(c("sigmoid", "linear")),
    size = 10
)

nn_grid_depth = grid_depth(
    hidden_neurons(c(32L, 128L)),
    activations(c("relu", "elu")),
    output_activation(c("sigmoid", "linear")),
    n_hlayer = 2,
    size = 10,
    type = "latin_hypercube"
)

nn_tunes = tune::tune_grid(
    nn_wf,
    iris_folds,
    grid = nn_grid_depth
    # metrics = metric_set(accuracy, roc_auc)
)

best_nn = select_best(nn_tunes)
final_nn = finalize_workflow(nn_wf, best_nn)

# collect_metrics(final_nn)
final_nn_model = fit(final_nn, data = iris)
final_nn_model

final_nn_model |>
    augment(new_data = iris) |>
    metrics(truth = Species, estimate = .pred_class)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
    fig.width = 7,
    fig.height = 5
)

## ----comparison_table, echo=FALSE---------------------------------------------
comparison = data.frame(
    Feature = c(
        "**Primary Focus**",
        "**Design Philosophy**",
        "**Architectures**",
        "**Code Generation**",
        "**tidymodels Integration**",
        "**Formula Syntax**",
        "**Layer-specific Activations**",
        "**GPU Support**",
        "**Explainability/xAI**",
        "**Statistical Inference**",
        "**Custom Loss Functions**",
        "**For whom?**"
    ),
    kindling = c(
        "Architectural versatility & flexibility, statistical modelling, and code generation",
        "Three-level API (code gen, training, ML framework (currently tidymodels) integration)",
        "Versatile — Feedforward Neural Networks (DNN/FFNN/MLP), Recurrent Neural Networks (RNN, LSTM, GRU), and more (in the future)",
        "Yes (inspect & modify torch code)",
        "Full (parsnip models & tuning)",
        "Yes",
        "Yes",
        "Yes",
        "Garson's & Olden's algorithms, vip integration, and more in the future",
        "Not yet implemented",
        "Yes",
        "Wanted versatile architectures (more in the future), fine-grained control, tidymodels users"
    ),
    brulee = c(
        "Production-ready statistical models",
        "Batteries-included with sensible defaults",
        "MLP, Linear/Logistic/Multinomial regression",
        "No",
        "Full (official tidymodels package)",
        "Yes",
        "No",
        "Yes",
        "Limited",
        "No",
        "No",
        "Wants standard supervised learning, stable production models"
    ),
    cito = c(
        "Statistical inference & interpretation",
        "User-friendly with comprehensive xAI pipeline",
        "Fully-connected networks, CNNs",
        "No",
        "No (standalone package)",
        "Yes",
        "No",
        "Yes (CPU, GPU, MacOS)",
        "Extensive (PDP, ALE, variable importance, etc.)",
        "Yes (confidence intervals, p-values via bootstrap)",
        "Yes",
        "Do ecological modeling, interpretable models, statistical inference"
    ),
    luz = c(
        "Training loop abstraction",
        "High-level API reducing boilerplate",
        "Any torch nn_module",
        "No",
        "No (standalone package)",
        "No (uses torch modules directly)",
        "No (also uses torch modules directly)",
        "Yes (automatic device placement)",
        "No",
        "No",
        "Yes",
        "Wants custom architectures, users needing human-friendly training loop control"
    ),
    stringsAsFactors = FALSE
)

knitr::kable(
    comparison, 
    col.names = c("Feature", "kindling", "brulee", "cito", "luz"),
    label = "Table of comparison"
)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
    # eval = identical(Sys.getenv("BUILD_VIGNETTES"), "true"), 
    eval = identical(Sys.getenv("NOT_CRAN"), "true"), 
    fig.width = 7,
    fig.height = 5,
    warning = FALSE,
    message = FALSE
)

## -----------------------------------------------------------------------------
# # library(kindling)
# # library(tidymodels)
# # library(modeldata)
# 
# box::use(
#     kindling[mlp_kindling, act_funs, args, hidden_neurons, activations, grid_depth],
#     dplyr[select, ends_with, mutate, slice_sample],
#     tidyr[drop_na],
#     rsample[initial_split, training, testing, vfold_cv],
#     recipes[
#         recipe, step_dummy, step_normalize,
#         all_nominal_predictors, all_numeric_predictors
#     ],
#     modeldata[penguins],
#     parsnip[tune, set_mode, fit, augment],
#     workflows[workflow, add_recipe, add_model],
#     dials[learn_rate],
#     tune[tune_grid, show_best, collect_metrics, select_best, finalize_workflow, last_fit],
#     yardstick[metric_set, rmse, rsq],
#     ggplot2[autoplot]
# )

## ----spec---------------------------------------------------------------------
# spec = mlp_kindling(
#     hidden_neurons = tune(),
#     activations = tune(),
#     epochs = 50,
#     learn_rate = tune()
# ) |>
#     set_mode("regression")

## ----data---------------------------------------------------------------------
# penguins_clean = penguins |>
#     drop_na() |>
#     select(body_mass_g, ends_with("_mm"), sex, species) |>
#     mutate(body_mass_kg = body_mass_g / 1000) |>
#     slice_sample(n = 30, by = species)
# 
# set.seed(123)
# split = initial_split(penguins_clean, prop = 0.8, strata = species)
# train = training(split)
# test = testing(split)
# folds = vfold_cv(train, v = 5, strata = body_mass_kg)
# 
# 
# rec = recipe(body_mass_kg ~ ., data = train) |>
#     step_dummy(all_nominal_predictors()) |>
#     step_normalize(all_numeric_predictors())

## ----grid---------------------------------------------------------------------
# set.seed(42)
# depth_grid = grid_depth(
#     hidden_neurons(c(16, 32)),
#     activations(c("relu", "elu", "softshrink(lambd = 0.2)")),
#     learn_rate(),
#     n_hlayer = 1:3,
#     size = 10,
#     type = "latin_hypercube"
# )
# 
# depth_grid

## ----tune---------------------------------------------------------------------
# wflow = workflow() |>
#     add_recipe(rec) |>
#     add_model(spec)
# 
# tune_res = tune_grid(
#     wflow,
#     resamples = folds,
#     grid = depth_grid,
#     metrics = metric_set(rmse, rsq)
# )

## ----results------------------------------------------------------------------
# collect_metrics(tune_res)
# show_best(tune_res, metric = "rmse", n = 5)

## ----final--------------------------------------------------------------------
# best_params = select_best(tune_res, metric = "rmse")
# final_wflow = wflow |>
#     finalize_workflow(best_params)
# 
# final_model = fit(final_wflow, data = train)
# final_model

## ----eval---------------------------------------------------------------------
# final_model |>
#     augment(new_data = test) |>
#     metric_set(rmse, rsq)(
#         truth = body_mass_kg,
#         estimate = .pred
#     )

## ----parametric---------------------------------------------------------------
# spec_manual = mlp_kindling(
#     hidden_neurons = c(50, 15),
#     activations = act_funs(
#         softshrink[lambd = 0.5],
#         relu
#     ),
#     epochs = 150,
#     learn_rate = 0.01
# ) |>
#     set_mode("regression")

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    eval = identical(Sys.getenv("NOT_CRAN"), "true"),
    fig.width = 7,
    fig.height = 5,
    warning = FALSE,
    message = FALSE
)

fansi::set_knit_hooks(knitr::knit_hooks)

## ----eval = FALSE-------------------------------------------------------------
# # install.packages("pak")
# pak::pak("joshuamarie/kindling")
# ## devtools::install_github("joshuamarie/kindling")

## -----------------------------------------------------------------------------
# library(kindling)

## ----collapse=FALSE, comment=""-----------------------------------------------
# ffnn_generator(
#     nn_name = "MyFFNN",
#     hd_neurons = c(64, 32, 16),
#     no_x = 10,
#     no_y = 1,
#     activations = 'relu'
# )

## ----comment="", collapse=FALSE-----------------------------------------------
# model = ffnn(
#     Species ~ .,
#     data = iris,
#     hidden_neurons = c(10, 15, 7),
#     activations = act_funs(relu, softshrink[lambd = 0.5], elu),
#     loss = "cross_entropy",
#     epochs = 100
# )
# 
# model

## -----------------------------------------------------------------------------
# predict(model) |>
#     (\(x) table(actual = iris$Species, predicted = x))()

## -----------------------------------------------------------------------------
# sample_iris = dplyr::slice_sample(iris, n = 10, by = Species)
# 
# predict(model, newdata = sample_iris) |>
#     (\(x) table(actual = sample_iris$Species, predicted = x))()

## -----------------------------------------------------------------------------
# # library(kindling)
# # library(parsnip)
# # library(yardstick)
# box::use(
#     kindling[mlp_kindling, rnn_kindling, act_funs, args],
#     parsnip[fit, augment],
#     yardstick[metrics]
# )
# data(Ionosphere, package = "mlbench")
# 
# ionosphere_data = Ionosphere[, -2]
# 
# # Train a feedforward network with parsnip
# mlp_kindling(
#     mode = "classification",
#     hidden_neurons = c(128, 64),
#     activations = act_funs(relu, softshrink[lambd = 0.5]),
#     epochs = 100
# ) |>
#     fit(Class ~ ., data = ionosphere_data) |>
#     augment(new_data = ionosphere_data) |>
#     metrics(truth = Class, estimate = .pred_class)
# 
# # Or try a recurrent architecture (demonstrative example with tabular data)
# rnn_kindling(
#     mode = "classification",
#     hidden_neurons = c(128, 64),
#     activations = act_funs(relu, elu),
#     epochs = 100,
#     rnn_type = "gru"
# ) |>
#     fit(Class ~ ., data = ionosphere_data) |>
#     augment(new_data = ionosphere_data) |>
#     metrics(truth = Class, estimate = .pred_class)

## ----collapse=FALSE, comment=""-----------------------------------------------
# # library(tidymodels)
# box::use(
#     kindling[
#         mlp_kindling, hidden_neurons, activations, output_activation, grid_depth
#     ],
#     parsnip[fit, augment],
#     recipes[recipe],
#     workflows[workflow, add_recipe, add_model],
#     rsample[vfold_cv],
#     tune[tune_grid, tune, select_best, finalize_workflow],
#     dials[grid_random],
#     yardstick[accuracy, roc_auc, metric_set, metrics]
# )
# 
# mlp_tune_spec = mlp_kindling(
#     mode = "classification",
#     hidden_neurons = tune(),
#     activations = tune(),
#     output_activation = tune()
# )
# 
# iris_folds = vfold_cv(iris, v = 3)
# nn_wf = workflow() |>
#     add_recipe(recipe(Species ~ ., data = iris)) |>
#     add_model(mlp_tune_spec)
# 
# nn_grid_depth = grid_depth(
#     hidden_neurons(c(32L, 128L)),
#     activations(c("relu", "elu")),
#     output_activation(c("sigmoid", "linear")),
#     n_hlayer = 2,
#     size = 10,
#     type = "latin_hypercube"
# )
# 
# # This is supported but limited to 1 hidden layer only
# ## nn_grid = grid_random(
# ##     hidden_neurons(c(32L, 128L)),
# ##     activations(c("relu", "elu")),
# ##     output_activation(c("sigmoid", "linear")),
# ##     size = 10
# ## )
# 
# nn_tunes = tune::tune_grid(
#     nn_wf,
#     iris_folds,
#     grid = nn_grid_depth
# )
# 
# best_nn = select_best(nn_tunes)
# best_nn
# final_nn = finalize_workflow(nn_wf, best_nn)
# final_nn_model = fit(final_nn, data = iris)
# final_nn_model
# 
# final_nn_model |>
#     augment(new_data = iris) |>
#     metrics(truth = Species, estimate = .pred_class)

## -----------------------------------------------------------------------------
# garson(model, bar_plot = FALSE)

## -----------------------------------------------------------------------------
# olden(model, bar_plot = FALSE)

## ----fig.height=6, fig.width=8------------------------------------------------
# box::use(
#     vip[vi, vip]
# )
# 
# vi(model) |>
#     vip()

#' Base models for Neural Network Training in kindling
#'
#' @section FFNN:
#' Train a feed-forward neural network using the torch package.
#'
#' @param formula Formula. Model formula (e.g., y ~ x1 + x2).
#' @param data Data frame. Training data.
#' @param hidden_neurons Integer vector. Number of neurons in each hidden layer.
#' @param activations Activation function specifications. See `act_funs()`.
#' @param output_activation Optional. Activation for output layer.
#' @param bias Logical. Use bias weights. Default `TRUE`.
#' @param epochs Integer. Number of training epochs. Default `100`.
#' @param batch_size Integer. Batch size for training. Default `32`.
#' @param penalty Numeric. Regularization penalty (lambda). Default `0` (no regularization).
#' @param mixture Numeric. Elastic net mixing parameter (0-1). Default `0`.
#' @param learn_rate Numeric. Learning rate for optimizer. Default `0.001`.
#' @param optimizer Character. Optimizer type ("adam", "sgd", "rmsprop"). Default `"adam"`.
#' @param optimizer_args Named list. Additional arguments passed to the optimizer. Default `list()`.
#' @param loss Character. Loss function ("mse", "mae", "cross_entropy", "bce"). Default `"mse"`.
#' @param validation_split Numeric. Proportion of data for validation (0-1). Default `0`.
#' @param device Character. Device to use ("cpu", "cuda", "mps"). Default `NULL` (auto-detect).
#' @param verbose Logical. Print training progress. Default `FALSE`.
#' @param cache_weights Logical. Cache weight matrices for faster variable importance. Default `FALSE`.
#' @param ... Additional arguments. Can be used to pass `x` and `y` for direct interface.
#' @param x When not using formula: predictor data (data.frame or matrix).
#' @param y When not using formula: outcome data (vector, factor, or matrix).
#'
#' @return An object of class "ffnn_fit" containing the trained model and metadata.
#'
#' @examples
#' \donttest{
#' if (torch::torch_is_installed()) {
#'     # Formula interface (original)
#'     model_reg = ffnn(
#'         Sepal.Length ~ .,
#'         data = iris[, 1:4],
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 50
#'     )
#'
#'     # XY interface (new)
#'     model_xy = ffnn(
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 50,
#'         x = iris[, 2:4],
#'         y = iris$Sepal.Length
#'     )
#' }
#' }
#'
#' @name kindling-basemodels
#' @export
ffnn = 
    function(
        formula = NULL,
        data = NULL,
        hidden_neurons,
        activations = NULL,
        output_activation = NULL,
        bias = TRUE,
        epochs = 100,
        batch_size = 32,
        penalty = 0,
        mixture = 0,
        learn_rate = 0.001,
        optimizer = "adam",
        optimizer_args = list(),
        loss = "mse",
        validation_split = 0,
        device = NULL,
        verbose = FALSE,
        cache_weights = FALSE,
        ...,
        x = NULL,
        y = NULL
    ) {
    dots = list(...)
    
    act_specs = eval_act_funs({{ activations }}, {{ output_activation }})
    activations = act_specs$activations
    output_activation = act_specs$output_activation
    
    # Starting from 0.2.0
    # Use 'hardhat' package instead
    if (!is.null(x) || !is.null(y)) {
        if (is.null(x) || is.null(y)) {
            cli::cli_abort("Both {.arg x} and {.arg y} must be provided when not using formula interface.")
        }
        if (!is.null(formula) || !is.null(data)) {
            cli::cli_warn("Both formula/data and x/y provided. Using x/y interface.")
        }
        
        processed = hardhat::mold(x, y)
        
    } else if (!is.null(formula)) {
        if (is.null(data)) {
            cli::cli_abort("{.arg data} must be provided when using formula interface.")
        }
        
        processed = hardhat::mold(formula, data)
        
    } else {
        cli::cli_abort("Must provide either {.arg formula} and {.arg data}, or {.arg x} and {.arg y}.")
    }
    
    predictors = processed$predictors
    outcomes = processed$outcomes
    
    if (!is.matrix(predictors)) {
        predictors = as.matrix(predictors)
    }
    
    if (is.data.frame(outcomes)) {
        if (ncol(outcomes) == 1) {
            outcomes = outcomes[[1]]
        } else {
            outcomes = as.matrix(outcomes)
        }
    } else if (ncol(outcomes) == 1) {
        outcomes = outcomes[[1]]
    }
    
    fit = ffnn_impl(
        x = predictors,
        y = outcomes,
        hidden_neurons = hidden_neurons,
        activations = activations,
        output_activation = output_activation,
        bias = bias,
        epochs = epochs,
        batch_size = batch_size,
        penalty = penalty,
        mixture = mixture,
        learn_rate = learn_rate,
        optimizer = optimizer,
        optimizer_args = optimizer_args,
        loss = loss,
        validation_split = validation_split,
        device = device,
        verbose = verbose,
        cache_weights = cache_weights
    )
    
    fit$blueprint = processed$blueprint
    if (!is.null(formula)) {
        fit$formula = formula
    }
    
    fit
}

#' FFNN Implementation
#' @keywords internal
ffnn_impl = 
    function(
        x,
        y,
        hidden_neurons,
        activations = NULL,
        output_activation = NULL,
        bias = TRUE,
        epochs = 100,
        batch_size = 32,
        penalty = 0,
        mixture = 0,
        learn_rate = 0.001,
        optimizer = "adam",
        optimizer_args = list(),
        loss = "mse",
        validation_split = 0,
        device = NULL,
        verbose = FALSE,
        cache_weights = FALSE
    ) {
    if (!requireNamespace("torch", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg torch} is required but not installed.")
    }
    
    # Device selection
    if (is.null(device)) {
        device = get_default_device()
    } else {
        device = validate_device(device)
    }
    
    if (verbose) {
        cli::cli_alert_info("Using device: {device}")
    }
    
    validate_regularization(penalty, mixture)
    
    feature_names = colnames(x)
    if (is.null(feature_names)) {
        feature_names = paste0("V", seq_len(ncol(x)))
    }
    
    response_name = if (is.null(names(y))) "y" else names(y)[1]
    is_classification = is.factor(y) || is.character(y)
    
    if (is_classification) {
        if (is.character(y)) y = as.factor(y)
        y_levels = levels(y)
        n_classes = length(y_levels)
        y_numeric = as.integer(y)
        no_y = n_classes
        
        if (loss == "mse") {
            loss = "cross_entropy"
            if (verbose) {
                cli::cli_alert("Auto-detected classification task. Using cross_entropy loss.")
            }
        }
    } else {
        y_levels = NULL
        n_classes = NULL
        y_numeric = if (is.matrix(y)) y else as.numeric(y)
        no_y = if (is.matrix(y)) ncol(y) else 1L
    }
    
    no_x = ncol(x)
    n_obs = nrow(x)
    
    if (validation_split > 0 && validation_split < 1) {
        n_val = floor(n_obs * validation_split)
        val_idx = sample(n_obs, n_val)
        train_idx = setdiff(seq_len(n_obs), val_idx)
        
        x_train = x[train_idx, , drop = FALSE]
        y_train = y_numeric[train_idx]
        x_val = x[val_idx, , drop = FALSE]
        y_val = y_numeric[val_idx]
    } else {
        x_train = x
        y_train = y_numeric
        x_val = NULL
        y_val = NULL
    }
    
    x_train_t = torch::torch_tensor(x_train, dtype = torch::torch_float32(), device = device)
    
    if (is_classification) {
        y_train_t = torch::torch_tensor(y_train, dtype = torch::torch_long(), device = device)
    } else {
        y_train_t = torch::torch_tensor(
            if (is.matrix(y_train)) y_train else matrix(y_train, ncol = 1),
            dtype = torch::torch_float32(),
            device = device
        )
    }
    
    if (!is.null(x_val)) {
        x_val_t = torch::torch_tensor(x_val, dtype = torch::torch_float32(), device = device)
        if (is_classification) {
            y_val_t = torch::torch_tensor(y_val, dtype = torch::torch_long(), device = device)
        } else {
            y_val_t = torch::torch_tensor(
                if (is.matrix(y_val)) y_val else matrix(y_val, ncol = 1),
                dtype = torch::torch_float32(),
                device = device
            )
        }
    }
    
    model_expr = ffnn_generator(
        nn_name = "FFNN",
        hd_neurons = hidden_neurons,
        no_x = no_x,
        no_y = no_y,
        activations = activations,
        output_activation = output_activation,
        bias = bias
    )
    model = eval(model_expr)()
    model$to(device = device)
    
    validate_optimizer(tolower(optimizer))
    optimizer_fn = get(paste0("optim_", tolower(optimizer)), envir = asNamespace("torch"))
    opt = do.call(
        optimizer_fn,
        c(list(params = model$parameters, lr = learn_rate), optimizer_args)
    )
    
    loss_fn = switch(
        tolower(loss),
        mse = function(input, target) torch::nnf_mse_loss(input, target),
        mae = function(input, target) torch::nnf_l1_loss(input, target),
        cross_entropy = function(input, target) torch::nnf_cross_entropy(input, target),
        bce = function(input, target) torch::nnf_binary_cross_entropy_with_logits(input, target),
        cli::cli_abort("Unknown loss function: {loss}")
    )
    
    loss_history = numeric(epochs)
    val_loss_history = if (!is.null(x_val)) numeric(epochs) else NULL
    n_batches = ceiling(nrow(x_train) / batch_size)
    
    for (epoch in seq_len(epochs)) {
        model$train()
        epoch_loss = 0
        idx = sample(nrow(x_train))
        
        for (batch in seq_len(n_batches)) {
            start_idx = (batch - 1) * batch_size + 1
            end_idx = min(batch * batch_size, nrow(x_train))
            batch_idx = idx[start_idx:end_idx]
            
            x_batch = x_train_t[batch_idx, ]
            y_batch = y_train_t[batch_idx]
            
            opt$zero_grad()
            y_pred = model(x_batch)
            loss = loss_fn(y_pred, y_batch)
            reg_loss = regularizer(model, penalty, mixture)
            total_loss = loss + reg_loss
            total_loss$backward()
            opt$step()
            
            epoch_loss = epoch_loss + total_loss$item()
        }
        
        loss_history[epoch] = epoch_loss / n_batches
        
        if (!is.null(x_val)) {
            model$eval()
            torch::with_no_grad({
                y_val_pred = model(x_val_t)
                val_loss = loss_fn(y_val_pred, y_val_t)
                val_loss_history[epoch] = val_loss$item()
            })
        }
        
        if (verbose && (epoch %% max(1, epochs %/% 10) == 0 || epoch == epochs)) {
            msg = sprintf("Epoch %d/%d - Loss: %.4f", epoch, epochs, loss_history[epoch])
            if (!is.null(val_loss_history)) {
                msg = paste0(msg, sprintf(" - Val Loss: %.4f", val_loss_history[epoch]))
            }
            message(msg)
        }
    }
    
    model$eval()
    fitted_tensor = torch::with_no_grad({
        model(torch::torch_tensor(x, dtype = torch::torch_float32(), device = device))
    })
    
    if (is_classification) {
        fitted_probs = torch::nnf_softmax(fitted_tensor, dim = 2)
        fitted_classes = torch::torch_argmax(fitted_probs, dim = 2)
        fitted_values = as.integer(fitted_classes$cpu())
        fitted_values = factor(fitted_values, levels = seq_along(y_levels), labels = y_levels)
    } else {
        fitted_values = as.matrix(fitted_tensor$cpu())
        if (no_y == 1L) fitted_values = as.vector(fitted_values)
    }
    
    cached_weights = NULL
    if (cache_weights) {
        n_hidden = length(hidden_neurons)
        input_layer = model$fc1
        W_input = as.matrix(input_layer$weight$cpu())
        output_layer = model$out
        W_output = as.matrix(output_layer$weight$cpu())
        
        intermediate_weights = list()
        if (n_hidden > 1) {
            for (i in seq_len(n_hidden - 1)) {
                layer = model[[paste0("fc", i + 1)]]
                intermediate_weights[[i]] = as.matrix(layer$weight$cpu())
            }
        }
        
        cached_weights = list(
            input = W_input,
            output = W_output,
            intermediate = intermediate_weights
        )
    }
    
    structure(
        list(
            model = model,
            fitted = fitted_values,
            loss_history = loss_history,
            val_loss_history = val_loss_history,
            n_epochs = epochs,
            hidden_neurons = hidden_neurons,
            activations = activations,
            output_activation = output_activation,
            penalty = penalty,
            mixture = mixture,
            feature_names = feature_names,
            response_name = response_name,
            no_x = no_x,
            no_y = no_y,
            is_classification = is_classification,
            y_levels = y_levels,
            n_classes = n_classes,
            device = device,
            cached_weights = cached_weights
        ),
        class = "ffnn_fit"
    )
}

#' Predict method for kindling basemodel fits
#'
#' @param object An object of class `"ffnn_fit"` or `"rnn_fit"`.
#' @param newdata Data frame. New data for predictions. If `NULL`, uses the
#'   original training data (if available).
#' @param new_data Alternative to newdata (for consistency with hardhat).
#' @param type Character. Type of prediction:
#'   * `"response"` (default) – predicted values or predicted classes
#'   * `"prob"` – class probabilities (only for classification models)
#' @param ... Currently unused.
#'
#' @return
#' * For **regression** models: a numeric vector (single output) or matrix
#'   (multiple outputs) of predicted values.
#' * For **classification** models:
#'   * `type = "response"`: a factor vector of predicted class labels
#'   * `type = "prob"`: a numeric matrix of class probabilities, with columns
#'     named after the class levels.
#'
#' @name predict-basemodel
#' @keywords internal
#' @export
predict.ffnn_fit = function(object, newdata = NULL, new_data = NULL, type = "response", ...) {
    if (!requireNamespace("torch", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg torch} is required but not installed.")
    }
    
    if (!is.null(new_data) && is.null(newdata)) {
        newdata = new_data
    }
    
    device = object$device
    
    if (is.null(newdata)) {
        if (type == "prob" && object$is_classification) {
            cli::cli_abort("Cannot compute probabilities without {.arg newdata}. Use the fitted values instead.")
        }
        return(object$fitted)
    }
    
    if (!is.null(object$blueprint)) {
        processed = hardhat::forge(newdata, object$blueprint)
        x_new = processed$predictors
    } else {
        x_new = newdata
    }
    
    if (!is.matrix(x_new)) {
        x_new = as.matrix(x_new)
    }
    
    x_new_t = torch::torch_tensor(x_new, dtype = torch::torch_float32(), device = device)
    
    object$model$eval()
    pred_tensor = torch::with_no_grad({
        object$model(x_new_t)
    })
    
    if (object$is_classification) {
        probs = torch::nnf_softmax(pred_tensor, dim = 2)
        
        if (type == "prob") {
            prob_matrix = as.matrix(probs$cpu())
            colnames(prob_matrix) = object$y_levels
            return(prob_matrix)
        } else {
            pred_classes = torch::torch_argmax(probs, dim = 2)
            predictions = as.integer(pred_classes$cpu())
            predictions = factor(predictions,
                                 levels = seq_along(object$y_levels),
                                 labels = object$y_levels)
            return(predictions)
        }
    } else {
        predictions = as.matrix(pred_tensor$cpu())
        if (object$no_y == 1L) predictions = as.vector(predictions)
        return(predictions)
    }
}

#' @keywords internal
#' @export
`$.ffnn_fit` = function(x, name) {
    if (name %in% names(x)) {
        return(x[[name]])
    }
    
    attr(x, name, exact = TRUE)
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("FFNN regularization parameters are validated", {
    skip_if_no_torch()
    
    # Negative penalty should error
    expect_error(
        ffnn(
            Sepal.Length ~ ., 
            data = iris[1:50, ], 
            hidden_neurons = 10, 
            penalty = -0.01, 
            epochs = 5
        ),
        "non-negative"
    )
    
    # Invalid mixture (> 1)
    expect_error(
        ffnn(
            Sepal.Length ~ ., 
            data = iris[1:50, ], 
            hidden_neurons = 10, 
            penalty = 0.01, 
            mixture = 1.5, 
            epochs = 5
        ),
        "between 0 and 1"
    )
    
    # Invalid mixture (< 0)
    expect_error(
        ffnn(
            Sepal.Length ~ ., 
            data = iris[1:50, ], 
            hidden_neurons = 10, 
            penalty = 0.01, 
            mixture = -0.1, 
            epochs = 5
        ),
        "between 0 and 1"
    )
})

test_that("FFNN trains with L1 regularization (Lasso)", {
    skip_if_no_torch()
    
    data = iris[1:50, 1:4]
    
    # Pure L1 (LASSO regression)
    model = ffnn(
        Sepal.Length ~ ., 
        data = data, 
        hidden_neurons = c(16, 8), 
        penalty = 0.01,
        mixture = 1,
        epochs = 10,
        verbose = FALSE
    )
    
    expect_s3_class(model, "ffnn_fit")
    expect_equal(model$penalty, 0.01)
    expect_equal(model$mixture, 1)
    expect_length(model$fitted, nrow(data))
})

test_that("FFNN trains with L2 regularization (Ridge)", {
    skip_if_no_torch()
    
    # Pure L2 (Ridge regression)
    model = ffnn(
        Sepal.Length ~ ., 
        data = iris[1:50, 1:4], 
        hidden_neurons = 16, 
        penalty = 0.001,
        mixture = 0,      
        epochs = 10,
        verbose = FALSE
    )
    
    expect_s3_class(model, "ffnn_fit")
    expect_equal(model$penalty, 0.001)
    expect_equal(model$mixture, 0)
    expect_true(length(model$loss_history) == 10)
})

test_that("FFNN trains with elastic net regularization", {
    skip_if_no_torch()
    
    # Elastic net
    # With mixture = 0.5
    model = ffnn(
        Species ~ ., 
        data = iris[1:50, ], 
        hidden_neurons = 16, 
        penalty = 0.005,
        mixture = 0.5,    
        epochs = 10,
        verbose = FALSE
    )
    
    expect_s3_class(model, "ffnn_fit")
    expect_equal(model$penalty, 0.005)
    expect_equal(model$mixture, 0.5)
    expect_true(model$is_classification)
})

test_that("RNN trains with regularization", {
    skip_if_no_torch()
    
    model = rnn(
        Sepal.Length ~ ., 
        data = iris[1:50, 1:4], 
        hidden_neurons = c(16, 8), 
        rnn_type = "lstm",
        penalty = 0.01,
        mixture = 0,    
        epochs = 5,
        verbose = FALSE
    )
    
    expect_s3_class(model, "rnn_fit")
    expect_equal(model$penalty, 0.01)
    expect_equal(model$mixture, 0)
})

test_that("Regularization reduces to no penalty when penalty = 0", {
    skip_if_no_torch()
    
    model_no_reg = ffnn(
        Sepal.Length ~ ., 
        data = iris[1:30, 1:4], 
        hidden_neurons = 10, 
        penalty = 0,
        epochs = 10,
        verbose = FALSE
    )
    
    model_with_reg = ffnn(
        Sepal.Length ~ ., 
        data = iris[1:30, 1:4], 
        hidden_neurons = 10, 
        penalty = 0,
        mixture = 0,
        epochs = 10,
        verbose = FALSE
    )
    
    expect_s3_class(model_no_reg, "ffnn_fit")
    expect_s3_class(model_with_reg, "ffnn_fit")
    expect_equal(model_no_reg$penalty, 0)
    expect_equal(model_with_reg$penalty, 0)
})

test_that("Different mixture values produce different models", {
    skip_if_no_torch()
    
    set.seed(123)
    model_l2 = ffnn(
        Sepal.Length ~ ., 
        data = iris[1:30, 1:4], 
        hidden_neurons = 10, 
        penalty = 0.01,
        mixture = 0,      # L2
        epochs = 20,
        verbose = FALSE
    )
    
    set.seed(123)
    model_l1 = ffnn(
        Sepal.Length ~ ., 
        data = iris[1:30, 1:4], 
        hidden_neurons = 10, 
        penalty = 0.01,
        mixture = 1,      # L1
        epochs = 20,
        verbose = FALSE
    )
    
    expect_false(isTRUE(all.equal(model_l2$fitted, model_l1$fitted)))
})

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("ffnn works with optimizer_args", {
    skip_if_no_torch()
    
    model1 = ffnn(
        Sepal.Length ~ .,
        data = iris[1:50, 1:4],
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5,
        optimizer = "adam",
        optimizer_args = list(weight_decay = 0.01),
        verbose = FALSE
    )
    
    expect_s3_class(model1, "ffnn_fit")
    expect_true(length(model1$loss_history) == 5)
    
    # Test with momentum (SGD)
    model2 = ffnn(
        Species ~ .,
        data = iris[1:50, ],
        hidden_neurons = 16,
        activations = "relu",
        epochs = 5,
        optimizer = "sgd",
        learn_rate = 0.01,
        optimizer_args = list(momentum = 0.9),
        verbose = FALSE
    )
    
    expect_s3_class(model2, "ffnn_fit")
    expect_true(model2$is_classification)
    
    # Test with multiple optimizer args
    model3 = ffnn(
        Sepal.Length ~ .,
        data = iris[1:50, 1:4],
        hidden_neurons = 8,
        epochs = 3,
        optimizer = "adam",
        optimizer_args = list(weight_decay = 0.001, amsgrad = TRUE),
        verbose = FALSE
    )
    
    expect_s3_class(model3, "ffnn_fit")
})

test_that("rnn works with optimizer_args", {
    skip_if_no_torch()
    
    # Test LSTM with weight_decay
    model1 = rnn(
        Sepal.Length ~ .,
        data = iris[1:50, 1:4],
        hidden_neurons = c(16, 8),
        rnn_type = "lstm",
        epochs = 5,
        optimizer = "adam",
        optimizer_args = list(weight_decay = 0.01),
        verbose = FALSE
    )
    
    expect_s3_class(model1, "rnn_fit")
    expect_equal(model1$rnn_type, "lstm")
    
    # Test GRU with momentum
    model2 = rnn(
        Species ~ .,
        data = iris[1:50, ],
        hidden_neurons = 16,
        rnn_type = "gru",
        epochs = 5,
        optimizer = "sgd",
        optimizer_args = list(momentum = 0.9, dampening = 0.1),
        verbose = FALSE
    )
    
    expect_s3_class(model2, "rnn_fit")
    expect_true(model2$is_classification)
    expect_equal(model2$rnn_type, "gru")
})

test_that("optimizer_args defaults to empty list", {
    skip_if_no_torch()
    
    # Should work without optimizer_args
    model = ffnn(
        Sepal.Length ~ Sepal.Width,
        data = iris[1:30, ],
        hidden_neurons = 8,
        epochs = 3,
        verbose = FALSE
    )
    
    expect_s3_class(model, "ffnn_fit")
})

test_that("invalid optimizer_args throw errors", {
    skip_if_no_torch()
    
    expect_error(
        ffnn(
            Sepal.Length ~ .,
            data = iris[1:30, 1:4],
            hidden_neurons = 8,
            epochs = 2,
            optimizer = "adam",
            optimizer_args = list(invalid_param = 999)
        )
    )
})

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("garson method works for ffnn_fit", {
    skip_if_no_torch()
    skip_if_not_installed("NeuralNetTools")
    
    model = ffnn(
        Species ~ .,
        data = iris,
        hidden_neurons = c(32, 16),
        activations = "relu",
        epochs = 20,
        verbose = FALSE,
        cache_weights = TRUE
    )
    
    imp = garson(model, bar_plot = FALSE)
    
    expect_s3_class(imp, c("garson", "data.frame"))
    expect_equal(nrow(imp), 4)
    expect_true(all(c("x_names", "y_names", "rel_imp") %in% names(imp)))
    expect_equal(sum(imp$rel_imp), 100, tolerance = 1e-6)
    expect_true(all(imp$rel_imp >= 0))
})

test_that("olden method works for ffnn_fit", {
    skip_if_no_torch()
    skip_if_not_installed("NeuralNetTools")
    
    model = ffnn(
        mpg ~ cyl + disp + hp + wt,
        data = mtcars,
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 20,
        loss = "mse",
        verbose = FALSE,
        cache_weights = TRUE
    )
    
    imp = olden(model, bar_plot = FALSE)
    
    expect_s3_class(imp, c("olden", "data.frame"))
    expect_equal(nrow(imp), 4)
    expect_true(all(c("x_names", "y_names", "rel_imp") %in% names(imp)))
    expect_true(all(imp$x_names %in% c("cyl", "disp", "hp", "wt")))
})

test_that("vi_model method works with ffnn_fit", {
    skip_if_no_torch()
    skip_if_not_installed("vip")
    
    model = ffnn(
        Species ~ .,
        data = iris,
        hidden_neurons = c(16),
        activations = "relu",
        epochs = 15,
        verbose = FALSE,
        cache_weights = TRUE
    )
    
    # Test with olden
    imp_olden = vip::vi_model(model, type = "olden")
    expect_true(all(c("Variable", "Importance") %in% names(imp_olden)))
    expect_equal(nrow(imp_olden), 4)

    # Test with garson
    imp_garson = vip::vi_model(model, type = "garson")
    expect_equal(nrow(imp_garson), 4)
    expect_true(all(imp_garson$Importance >= 0))
})

test_that("vip integration works", {
    skip_if_no_torch()
    skip_if_not_installed("vip")
    
    model = ffnn(
        Species ~ Sepal.Length + Sepal.Width,
        data = iris,
        hidden_neurons = c(8),
        activations = "relu",
        epochs = 10,
        verbose = FALSE,
        cache_weights = TRUE
    )
    
    imp = vip::vi(model)
    
    expect_equal(nrow(imp), 2)
    expect_true("Variable" %in% names(imp))
    expect_true("Importance" %in% names(imp))
})

test_that("variable importance works without cached weights", {
    skip_if_no_torch()
    skip_if_not_installed("NeuralNetTools")
    
    model = ffnn(
        Species ~ .,
        data = iris,
        hidden_neurons = c(16),
        activations = "relu",
        epochs = 10,
        verbose = FALSE,
        cache_weights = FALSE
    )
    
    expect_null(model$cached_weights)
    
    imp = garson(model, bar_plot = FALSE)
    expect_s3_class(imp, "garson")
    expect_equal(nrow(imp), 4)
})

test_that("variable importance handles multi-layer networks", {
    skip_if_no_torch()
    skip_if_not_installed("NeuralNetTools")
    
    model = ffnn(
        Species ~ .,
        data = iris,
        hidden_neurons = c(32, 16, 8),
        activations = "relu",
        epochs = 15,
        verbose = FALSE,
        cache_weights = TRUE
    )
    
    imp_garson = garson(model, bar_plot = FALSE)
    imp_olden = olden(model, bar_plot = FALSE)
    
    expect_equal(nrow(imp_garson), 4)
    expect_equal(nrow(imp_olden), 4)
    expect_equal(sum(imp_garson$rel_imp), 100, tolerance = 1e-6)
})

test_that("garson bar_plot = TRUE runs without error", {
    skip_if_no_torch()
    skip_if_not_installed("NeuralNetTools")
    skip_if_not_installed("ggplot2")
    
    model = ffnn(
        Species ~ .,
        data = iris,
        hidden_neurons = c(16),
        activations = "relu",
        epochs = 10,
        verbose = FALSE,
        cache_weights = TRUE
    )
    
    pdf(NULL)
    expect_no_error(garson(model, bar_plot = TRUE))
    dev.off()
})

test_that("olden bar_plot = TRUE runs without error", {
    skip_if_no_torch()
    skip_if_not_installed("NeuralNetTools")
    skip_if_not_installed("ggplot2")
    
    model = ffnn(
        Species ~ .,
        data = iris,
        hidden_neurons = c(16),
        activations = "relu",
        epochs = 10,
        verbose = FALSE,
        cache_weights = TRUE
    )
    
    pdf(NULL)
    expect_no_error(olden(model, bar_plot = TRUE))
    dev.off()
})

#' Generalized Neural Network Module Expression Generator
#' 
#' @description
#' `r lifecycle::badge("experimental")`
#' 
#' `nn_module_generator()` is a generalized function that generates neural network 
#' module expressions for various architectures. It provides a flexible framework for creating
#' custom neural network modules by parameterizing layer types, construction arguments, and
#' forward pass behavior.
#'
#' While designed primarily for `{torch}` modules, it can work with custom layer implementations
#' from the current environment, including user-defined layers like RBF networks, custom
#' attention mechanisms, or other novel architectures.
#'
#' This function serves as the foundation for specialized generators like `ffnn_generator()`
#' and `rnn_generator()`, but can be used directly to create custom architectures.
#'
#' @param nn_name Character string specifying the name of the generated neural network module class.
#'   Default is `"nnModule"`.
#'
#' @param nn_layer The type of neural network layer to use. Can be specified as:
#'   - `NULL` (default): Uses `nn_linear()` from `{torch}`
#'   - Character string: e.g., `"nn_linear"`, `"nn_gru"`, `"nn_lstm"`, `"some_custom_layer"`
#'   - Named function: A function object that constructs the layer
#'   - Anonymous function: e.g., `\() nn_linear()` or `function() nn_linear()`
#'   
#'   The layer constructor is first searched in the current environment, then in parent
#'   environments, and finally falls back to the `{torch}` namespace. This allows you to
#'   use custom layer implementations alongside standard torch layers.
#' 
#' @param out_nn_layer Default `NULL`. If supplied, it forces to be the neural network layer to be used 
#'   on the last layer. Can be specified as:
#'   - Character string, e.g. `"nn_linear"`, `"nn_gru"`, `"nn_lstm"`, `"some_custom_layer"`
#'   - Named function: A function object that constructs the layer
#'   - Formula interface, e.g. `~torch::nn_linear`, `~some_custom_layer`
#'   
#'   Internally, it almost works the same as `nn_layer` parameter. 
#' 
#' @param nn_layer_args Named list of additional arguments passed to the layer constructor
#'   specified by `nn_layer`. These arguments are applied to all layers. For layer-specific
#'   arguments, use `layer_arg_fn`. Default is an empty list.
#'
#' @param layer_arg_fn Optional function or formula that generates layer-specific construction arguments.
#'   Can be specified as:
#'   - **Formula**: `~ list(input_size = .in, hidden_size = .out)` where `.in`, `.out`, `.i`, and `.is_output` are available
#'   - **Function**: `function(i, in_dim, out_dim, is_output)` with signature as before
#'   
#'   The formula/function should return a named list of arguments to pass to the layer constructor.
#'   Available variables in formula context:
#'   
#'   - `.i` or `i`: Integer, the layer index (1-based)
#'   - `.in` or `in_dim`: Integer, input dimension for this layer
#'   - `.out` or `out_dim`: Integer, output dimension for this layer  
#'   - `.is_output` or `is_output`: Logical, whether this is the final output layer
#'   
#'   If `NULL`, defaults to FFNN-style arguments: `list(in_dim, out_dim, bias = bias)`.
#'
#' @param forward_extract Optional formula or function that processes layer outputs in the forward pass.
#'   Useful for layers that return complex structures (e.g., RNNs return `list(output, hidden)`).
#'   Can be specified as:
#'   
#'   - **Formula**: `~ .[[1]]` or `~ .$output` where `.` represents the layer output
#'   - **Function**: `function(expr)` that accepts/returns a language object
#'   
#'   Common patterns: 
#'   
#'   - Extract first element: `~ .[[1]]`
#'   - Extract named element: `~ .$output`
#'   - Extract with method: `~ .$get_output()`
#'   
#'   If `NULL`, layer outputs are used directly.
#'
#' @param before_output_transform Optional formula or function that transforms input before the output layer.
#'   This is applied after the last hidden layer (and its activation) but before the output layer.
#'   Can be specified as:
#'   
#'   - **Formula**: `~ .[, .$size(2), ]` where `.` represents the current tensor
#'   - **Function**: `function(expr)` that accepts/returns a language object
#'   
#'   Common patterns:
#'   
#'   - Extract last timestep: `~ .[, .$size(2), ]`
#'   - Flatten: `~ .$flatten(start_dim = 1)`
#'   - Global pooling: `~ .$mean(dim = 2)`
#'   - Extract token: `~ .[, 1, ]`
#'   
#'   If `NULL`, no transformation is applied.
#' 
#' @param after_output_transform Optional formula or function that transforms the output after the output layer.
#'   This is applied after `self$out(x)` (the final layer) but before returning the result.
#'   Can be specified as:
#'   
#'   - **Formula**: `~ .$mean(dim = 2)` where `.` represents the output tensor
#'   - **Function**: `function(expr)` that accepts/returns a language object
#'   
#'   Common patterns:
#'   
#'   - Global average pooling: `~ .$mean(dim = 2)`
#'   - Squeeze dimensions: `~ .$squeeze()`
#'   - Reshape output: `~ .$view(c(-1, 10))`
#'   - Extract specific outputs: `~ .[, , 1:5]`
#'   
#'   If `NULL`, no transformation is applied.
#' 
#' @param last_layer_args Optional named list or formula specifying additional arguments 
#'   for the output layer only. These arguments are appended to the output layer constructor
#'   after the arguments from `layer_arg_fn`. Can be specified as:
#'   
#'   - **Formula**: `~ list(kernel_size = 2L, bias = FALSE)` 
#'   - **Named list**: `list(kernel_size = 2L, bias = FALSE)`
#'   
#'   This is useful when you need to override or add specific parameters to the final layer
#'   without affecting hidden layers. For example, in CNNs you might want a different kernel
#'   size for the output layer, or in RNNs you might want to disable bias in the final linear
#'   projection. Arguments in `last_layer_args` will override any conflicting arguments from
#'   `layer_arg_fn` when `.is_output = TRUE`. Default is an empty list.
#' 
#' @param hd_neurons Integer vector specifying the number of neurons (hidden units) in each 
#'   hidden layer. The length determines the number of hidden layers in the network.
#'   Must contain at least one element.
#'
#' @param no_x Integer specifying the number of input features (input dimension).
#'
#' @param no_y Integer specifying the number of output features (output dimension).
#'
#' @param activations Activation function specifications for hidden layers. Can be:
#'   - `NULL`: No activation functions applied
#'   - Character vector: e.g., `c("relu", "sigmoid", "tanh")`
#'   - `activation_spec` object: Created using `act_funs()`, which allows
#'     specifying custom arguments. See examples.
#'   
#'   If a single activation is provided, it will be replicated across all hidden layers.
#'   Otherwise, the length should match the number of hidden layers.
#'
#' @param output_activation Optional activation function for the output layer.
#'   Same format as `activations`, but should specify only a single activation.
#'   Common choices include `"softmax"` for classification or `"sigmoid"` for 
#'   binary outcomes. Default is `NULL` (no output activation).
#'
#' @param bias Logical indicating whether to include bias terms in layers.
#'   Default is `TRUE`. Note that this is passed to `layer_arg_fn` if provided,
#'   so custom layer argument functions should handle this parameter appropriately.
#'
#' @param eval Logical indicating whether to evaluate the generated expression immediately.
#'   If `TRUE`, returns an instantiated `nn_module` class that can be called directly
#'   (e.g., `model()`). If `FALSE` (default), returns the unevaluated language expression
#'   that can be inspected or evaluated later with `eval()`. Default is `FALSE`.
#'
#'  
#' @param .env Default is [parent.frame()]. The environment in which the generated expression is to be evaluated
#'
#' @param ... Additional arguments passed to layer constructors or for future extensions.
#'
#' @return 
#' If `eval = FALSE` (default): A language object (unevaluated expression) representing 
#' a `torch::nn_module` definition. This expression can be evaluated with `eval()` to 
#' create the module class, which can then be instantiated with `eval(result)()` to 
#' create a model instance.
#' 
#' If `eval = TRUE`: An instantiated `nn_module` class constructor that can be called 
#' directly to create model instances (e.g., `result()`).
#'
#' @examples
#' \dontrun{
#' \donttest{
#' if (torch::torch_is_installed()) {
#'     # Basic usage with formula interface
#'     nn_module_generator(
#'         nn_name = "MyGRU",
#'         nn_layer = "nn_gru",
#'         layer_arg_fn = ~ if (.is_output) {
#'             list(.in, .out)
#'         } else {
#'             list(input_size = .in, hidden_size = .out, 
#'                  num_layers = 1L, batch_first = TRUE)
#'         },
#'         forward_extract = ~ .[[1]],
#'         before_output_transform = ~ .[, .$size(2), ],
#'         hd_neurons = c(128, 64, 32),
#'         no_x = 20,
#'         no_y = 5,
#'         activations = "relu"
#'     )
#'     
#'     # LSTM with cleaner syntax
#'     nn_module_generator(
#'         nn_name = "MyLSTM",
#'         nn_layer = "nn_lstm",
#'         layer_arg_fn = ~ list(
#'             input_size = .in,
#'             hidden_size = .out,
#'             batch_first = TRUE
#'         ),
#'         forward_extract = ~ .[[1]],
#'         before_output_transform = ~ .[, .$size(2), ],
#'         hd_neurons = c(64, 32),
#'         no_x = 10,
#'         no_y = 2
#'     )
#'     
#'     # CNN with global average pooling
#'     nn_module_generator(
#'         nn_name = "SimpleCNN",
#'         nn_layer = "nn_conv1d",
#'         layer_arg_fn = ~ list(
#'             in_channels = .in,
#'             out_channels = .out,
#'             kernel_size = 3L,
#'             padding = 1L
#'         ),
#'         before_output_transform = ~ .$mean(dim = 2),
#'         hd_neurons = c(16, 32, 64),
#'         no_x = 1,
#'         no_y = 10,
#'         activations = "relu"
#'     )
#'     
#'     # CNN with after_output_transform (pooling applied AFTER output layer)
#'     nn_module_generator(
#'         nn_name = "CNN1DClassifier",
#'         nn_layer = "nn_conv1d",
#'         layer_arg_fn = ~ if (.is_output) {
#'             list(.in, .out)
#'         } else {
#'             list(
#'                 in_channels = .in,
#'                 out_channels = .out,
#'                 kernel_size = 3L,
#'                 stride = 1L,
#'                 padding = 1L 
#'             )
#'         },
#'         after_output_transform = ~ .$mean(dim = 2),
#'         last_layer_args = list(kernel_size = 1, stride = 2),
#'         hd_neurons = c(16, 32, 64),
#'         no_x = 1,
#'         no_y = 10,
#'         activations = "relu"
#'     )
#'     
#' } else {
#'   message("torch not installed - skipping examples")
#' }
#' }
#' }
#'
#' @importFrom rlang new_function call2 expr sym f_rhs is_formula
#' @importFrom purrr map map2
#' @importFrom glue glue
#' @importFrom cli cli_abort
#'
#' @export
nn_module_generator = 
    function(
        nn_name = "nnModule",
        nn_layer = NULL,     
        out_nn_layer = NULL,       
        nn_layer_args = list(),    
        layer_arg_fn = NULL,       
        forward_extract = NULL,
        before_output_transform = NULL,
        after_output_transform = NULL,
        last_layer_args = list(), 
        hd_neurons,
        no_x,
        no_y,
        activations = NULL,
        output_activation = NULL,
        bias = TRUE,
        eval = FALSE, 
        .env = parent.frame(), 
        ...
    ) {
    if (is.null(nn_layer)) nn_layer = "torch::nn_linear"
    
    if (missing(hd_neurons) || is.null(hd_neurons) || length(hd_neurons) == 0L) {
        hd_neurons = integer(0) 
    }
    
    if (missing(no_x) || missing(no_y)) {
        cli::cli_abort("Both {.arg no_x} and {.arg no_y} must be specified.")
    }
    
    act_specs = eval_act_funs({{ activations }}, {{ output_activation }})
    activations = act_specs$activations
    output_activation = act_specs$output_activation
    
    # ---- INPUT PROCESSING ----
    layer_arg_fn = formula_to_function(
        layer_arg_fn,
        default_fn = function(i, in_dim, out_dim, is_output) {
            list(in_dim, out_dim, bias = bias)
        },
        arg_names = c("i", "in_dim", "out_dim", "is_output"),
        alias_map = list(
            i = ".i",
            in_dim = ".in", 
            out_dim = ".out",
            is_output = ".is_output"
        )
    )
    
    forward_extract = formula_to_expr_transformer(forward_extract)
    before_output_transform = formula_to_expr_transformer(before_output_transform)
    after_output_transform = formula_to_expr_transformer(after_output_transform)
    
    # ---- Process 1: Architecture setup ----
    nodes = c(no_x, hd_neurons, no_y)
    n_layers = length(nodes) - 1L
    n_hidden = length(hd_neurons)
    
    # ---- Process 2: Tweak activations input ----
    activation_spec = parse_activation_spec(activations, n_hidden)
    activation_calls = process_activations(activation_spec, prefix = "nnf_")
    
    if (!is.null(output_activation)) {
        output_spec = parse_activation_spec(output_activation, 1L)
        output_call = process_activations(output_spec, prefix = "nnf_")[[1]]
    } else {
        output_call = NULL
    }
    
    all_activation_calls = c(activation_calls, list(output_call))
    
    nn_layer_name = layer_to_name(nn_layer)
    
    # ---- Build initialize() ----
    init_body = map(seq_len(n_layers), function(i) {
        is_output = (i == n_layers)
        layer_base_name = if (startsWith(nn_layer_name, "nn_")) {
            substring(nn_layer_name, 4)
        } else {
            nn_layer_name
        }
        layer_name = if (is_output) "out" else glue("{layer_base_name}_{i}")
        in_dim = nodes[i]
        out_dim = nodes[i + 1]
        
        layer_args = layer_arg_fn(i, in_dim, out_dim, is_output)
        
        current_layer = if (is_output && !is.null(out_nn_layer)) {
            out_nn_layer
        } else {
            nn_layer
        }
        
        additional_args = if (is_output && !is.null(last_layer_args)) {
            if (rlang::is_formula(last_layer_args)) {
                eval(rlang::f_rhs(last_layer_args))
            } else if (is.list(last_layer_args)) {
                last_layer_args
            } else {
                list()
            }
        } else {
            list()
        }
        
        layer_expr = if (is.function(current_layer)) {
            rlang::enexpr(current_layer)
        } else if (is.character(current_layer)) {
            rlang::parse_expr(current_layer)
        } else if (rlang::is_formula(current_layer)) {
            rlang::f_rhs(current_layer)
        } else if (is.symbol(current_layer) || is.call(current_layer)) {
            current_layer
        } else { 
            cli::cli_abort("{.arg {out_nn_layer}} must be a string, symbol, or function, got {class(current_layer)[1]}")
        }
        
        layer_call = call2(
            layer_expr,
            !!!c(layer_args, nn_layer_args, additional_args)
        )
        
        call2("=", call2("$", expr(self), sym(layer_name)), layer_call)
    })
    
    init = new_function(
        args = pairlist(),
        body = call2("{", !!!init_body)
    )
    
    # ---- Build forward() ----
    forward_body_exprs = map(seq_len(n_layers), function(i) {
        is_output = (i == n_layers)
        is_last_hidden = (i == n_layers - 1L)
        layer_base_name = if (startsWith(nn_layer_name, "nn_")) {
            substring(nn_layer_name, 4)
        } else {
            nn_layer_name
        }
        layer_name = if (is_output) "out" else glue("{layer_base_name}_{i}")
        act_call_fn = all_activation_calls[[i]]
        
        layer_expr = call2(call2("$", expr(self), sym(layer_name)), expr(x))
        if (!is.null(forward_extract) && !is_output) {
            layer_expr = forward_extract(layer_expr)
        }
        
        line1 = call2("=", expr(x), layer_expr)
        
        out = list(line1)
        
        if (!is.null(act_call_fn)) {
            line2 = call2("=", expr(x), act_call_fn(expr(x)))
            out = c(out, list(line2))
        }
        
        # Apply before_output_transform after last hidden layer
        # This happens AFTER the last hidden layer's activation
        if (is_last_hidden && !is.null(before_output_transform)) {
            transform_line = call2("=", expr(x), before_output_transform(expr(x)))
            out = c(out, list(transform_line))
        }
        
        # Apply after_output_transform after output layer
        # This happens AFTER self$out(x)
        if (is_output && !is.null(after_output_transform)) {
            transform_line = call2("=", expr(x), after_output_transform(expr(x)))
            out = c(out, list(transform_line))
        }
        
        out
    })
    
    forward_body_exprs = c(
        unlist(forward_body_exprs, recursive = FALSE),
        list(expr(x)) 
    )
    
    forward = new_function(
        args = list(x = expr()),
        body = call2("{", !!!forward_body_exprs)
    )
    
    # ---- Build final nn_module call ----
    full_call = call2(
        expr(nn_module),
        nn_name,
        initialize = init,
        forward = forward,
        .ns = "torch"
    )
    
    if (eval) eval(full_call) else rlang::new_quosure(full_call, env = .env)
}

#' Formula to Function with Named Arguments
#' 
#' @param formula_or_fn A formula or function
#' @param default_fn Default function if `formula_or_fn` is `NULL`
#' @param arg_names Character vector of formal argument names
#' @param alias_map Named list mapping arg_names to formula aliases (e.g., list(in_dim = ".in"))
#' 
#' @return A function
#' @keywords internal
formula_to_function = function(formula_or_fn, default_fn = NULL, arg_names = NULL, alias_map = NULL) {
    if (is.null(formula_or_fn)) {
        return(default_fn)
    }
    
    if (rlang::is_formula(formula_or_fn)) {
        rhs = rlang::f_rhs(formula_or_fn)
        args_list = setNames(rep(list(rlang::missing_arg()), length(arg_names)), arg_names)
        
        if (!is.null(alias_map)) {
            alias_assignments = lapply(names(alias_map), function(arg) {
                alias = alias_map[[arg]]
                call2("=", sym(alias), sym(arg))
            })
            body_expr = call2("{", !!!alias_assignments, rhs)
        } else {
            body_expr = rhs
        }
        
        fn = rlang::new_function(
            args = args_list,
            body = body_expr,
            env = rlang::f_env(formula_or_fn)
        )
        return(fn)
    }
    
    if (is.function(formula_or_fn)) {
        return(formula_or_fn)
    }
    
    cli::cli_abort("Expected a formula or function, got {class(formula_or_fn)[1]}")
}


#' Convert Formula to Expression Transformer
#' 
#' @param formula_or_fn A formula like `~ .[[1]]` or a function that transforms expressions
#' 
#' @return A function that takes an expression and returns a transformed expression, or NULL
#' @keywords internal
formula_to_expr_transformer = function(formula_or_fn) {
    if (is.null(formula_or_fn)) {
        return(NULL)
    }
    
    if (rlang::is_formula(formula_or_fn)) {
        rhs = rlang::f_rhs(formula_or_fn)
        
        return(function(expr) {
            substitute_dot(rhs, expr)
        })
    }
    
    if (is.function(formula_or_fn)) {
        return(formula_or_fn)
    }
    
    cli::cli_abort("Expected a formula or function, got {class(formula_or_fn)[1]}")
}


#' Recursively Substitute . with Expression
#' 
#' @param expr Expression containing `.` placeholders
#' @param replacement Expression to substitute for `.`
#' 
#' @return Modified expression
#' @keywords internal
substitute_dot = function(expr, replacement) {
    if (is.symbol(expr) && identical(expr, quote(.))) {
        return(replacement)
    }
    
    if (is.call(expr)) {
        expr[] = lapply(expr, function(e) substitute_dot(e, replacement))
        return(expr)
    }
    
    expr
}

#' Layer argument pronouns for formula-based specifications
#' 
#' @description
#' These pronouns provide a cleaner, more readable way to reference layer parameters
#' in formula-based specifications for `nn_module_generator()` and related functions.
#' They work similarly to `rlang::.data` and `rlang::.env`.
#' 
#' @details
#' Available pronouns:
#' 
#' - `.layer`: Access all layer parameters as a list-like object
#' - `.i`: Layer index (1-based integer)
#' - `.in`: Input dimension for the layer
#' - `.out`: Output dimension for the layer
#' - `.is_output`: Logical indicating if this is the output layer
#' 
#' These pronouns can be used in formulas passed to:
#' 
#' - `layer_arg_fn` parameter
#' - Custom layer configuration functions
#' 
#' @section Usage:
#' 
#' ``` r
#' # Using individual pronouns
#' layer_arg_fn = ~ list(
#'     input_size = .in,
#'     hidden_size = .out,
#'     num_layers = if (.i == 1) 2L else 1L
#' )
#' 
#' # Using .layer pronoun (alternative syntax)
#' layer_arg_fn = ~ list(
#'     input_size = .layer$ind,
#'     hidden_size = .layer$out,
#'     is_first = .layer$i == 1
#' )
#' ```
#' 
#' @name layer_prs
#' @aliases .layer .i .in .out .is_output
NULL

#' @rdname layer_prs
#' @export
.layer = structure(
    list(),
    class = c("layer_pr", "list")
)

#' @rdname layer_prs
#' @export
.i = structure(
    list(),
    class = c("layer_index_pr", "layer_pr", "list")
)

#' @rdname layer_prs
#' @export
.in = structure(
    list(),
    class = c("layer_input_pr", "layer_pr", "list")
)

#' @rdname layer_prs
#' @export
.out = structure(
    list(),
    class = c("layer_output_pr", "layer_pr", "list")
)

#' @rdname layer_prs
#' @export
.is_output = structure(
    list(),
    class = c("layer_is_output_pr", "layer_pr", "list")
)

#' "Layer" attributes
#' 
#' @param x The .layer itself
#' @param name It could be the following: 
#' -  `i`: Layer index (1-based integer)
#' -  `ind`: Input dimension for the layer
#' -  `out`: Output dimension for the layer
#' -  `is_output`: Logical indicating if this is the output layer
#' 
#' @return A pronoun, it returns nothing
#' 
#' @name layer-attributes
#' @keywords internal
#' @export
`$.layer_pr` = function(x, name) {
    if (inherits(x, "layer_index_pr"))  return(quote(i))
    if (inherits(x, c("layer_input_pr", "layer_output_pr", "layer_is_output_pr"))) {
        pr = switch(
            class(x),
            layer_input_pr  = quote(in_dim),
            layer_output_pr = quote(out_dim),
            layer_is_output_pr = quote(is_output)
        )
        
        return(pr)
    }
    
    switch(
        name,
        i = quote(i),
        ind = quote(in_dim),
        out = quote(out_dim),
        is_output = quote(is_output),
        cli::cli_abort("Unknown layer pronoun field: {name}")
    )
}

#' Print method for the pronouns
#' 
#' @param x An object of class "ffnn_fit"
#' @param ... Additional arguments (unused)
#'
#' @return No return value, prints out the type of pronoun to be used
#' 
#' @section For `.layer`: 
#' It displays what fields to be accessed by `$`.
#' 
#' @rdname print-layer_pronoun
#' @keywords internal
#' @export
print.layer_pr = function(x, ...) {
    cat("<layer pronoun>\n")
    cat("Fields: i, ind, out, is_output\n")
    invisible(x)
}

#' @rdname print-layer_pronoun
#' @keywords internal
#' @export
print.layer_index_pr = function(x, ...) {
    cat("<layer index pronoun>\n")
    invisible(x)
}

#' @rdname print-layer_pronoun
#' @keywords internal
#' @export
print.layer_input_pr = function(x, ...) {
    cat("<layer input dimension pronoun>\n")
    invisible(x)
}

#' @rdname print-layer_pronoun
#' @keywords internal
#' @export
print.layer_output_pr = function(x, ...) {
    cat("<layer output dimension pronoun>\n")
    invisible(x)
}

#' @rdname print-layer_pronoun
#' @keywords internal
#' @export
print.layer_is_output_pr = function(x, ...) {
    cat("<layer is_output flag pronoun>\n")
    invisible(x)
}

layer_to_name = function(x) {
    if (is.character(x)) return(x)
    if (rlang::is_formula(x)) {
        rhs = rlang::f_rhs(x)
        if (is.symbol(rhs)) return(as.character(rhs))
        if (is.call(rhs) && identical(rhs[[1L]], quote(`::`)))
            return(as.character(rhs[[3L]]))
    }
    if (is.function(x)) {
        nm = deparse(substitute(x))   
        return(nm)
    }
    if (is.symbol(x)) return(as.character(x))
    
    "layer"
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("nn_module_generator() errors when no_x is missing", {
    skip_if_no_torch()
    expect_error(nn_module_generator(hd_neurons = 4, no_y = 2))
})

test_that("nn_module_generator() errors when no_y is missing", {
    skip_if_no_torch()
    expect_error(nn_module_generator(hd_neurons = 4, no_x = 3))
})

test_that("nn_module_generator() returns a quosure when eval = FALSE", {
    skip_if_no_torch()
    out = nn_module_generator(hd_neurons = 8, no_x = 4, no_y = 1)
    expect_true(rlang::is_quosure(out))
})

test_that("nn_module_generator() returns nn_module class when eval = TRUE", {
    skip_if_no_torch()
    out = nn_module_generator(hd_neurons = 8, no_x = 4, no_y = 1, eval = TRUE)
    expect_true(inherits(out, "nn_module_generator"))
})

test_that("nn_module_generator() works with no hidden layers", {
    skip_if_no_torch()
    expect_no_error(nn_module_generator(hd_neurons = c(), no_x = 4, no_y = 1))
    expect_no_error(nn_module_generator(hd_neurons = NULL, no_x = 4, no_y = 1))
    expect_no_error(nn_module_generator(no_x = 4, no_y = 1))
})

#' @importFrom cli rule style_bold cat_line style_italic

concat = function(x) {
    if (is.null(x)) return("No act function applied")
    
    if (inherits(x, "custom_activation")) {
        attr(x, "act_name") 
    } else if (inherits(x, "parameterized_activation")) {
        fname = attr(x, "act_name")
        params = paste(
            names(x),
            format(unlist(x), trim = TRUE),
            sep = " = ",
            collapse = ", "
        )
        glue::glue("{fname}({params})")
    } else {
        as.character(x)
    }
}

#' Ordinal Suffixes Generator
#'
#' @description
#' This function is originally from `numform::f_ordinal()`.
#'
#' @param x Vector of numbers. Could be a string equivalent
#' 
#' @return Returns a string vector with ordinal suffixes.
#'
#' @references
#'
#' Rinker, T. W. (2021). numform: A publication style number and plot formatter
#' version 0.7.0. \url{https://github.com/trinker/numform}
#'
#' @section This is how you use it:
#'
#' ``` r
#' kindling:::ordinal_gen(1:10)
#' ```
#'
#' Note: This is not exported into public namespace.
#' So please, refer to `numform::f_ordinal()` instead.
#'
#' @rdname ordinal_gen
ordinal_gen = function(x) {
    if (is.numeric(x) && any(x < 1))
        warning("Values below 1 found.\nMay yield incorrect results")
    
    x = as.character(x)
    
    regs = c(
        th = "^1[1-9]$|[0456789]$",
        st = "(?<!^1)1$",
        nd = "(?<!^1)2$",
        rd = "(?<!^1)3$"
    )
    
    for (i in seq_along(regs)) {
        locs = grepl(regs[i], x, perl = TRUE)
        x[locs] = paste0(x[locs], names(regs)[i])
    }
    
    x
}

#' Print method for ffnn_fit objects
#'
#' @param x An object of class "ffnn_fit"
#' @param ... Additional arguments (unused)
#'
#' @return No return value, called for side effects (printing model summary)
#'
#' @keywords internal
#' @export
print.ffnn_fit = function(x, ...) {
    # ---Title section---
    title = "Feedforward Neural Networks (MLP)"
    title_block = rule(center = title, line = "=")
    title_block = style_bold(title_block)
    
    cat_line("\n", title_block, "\n")
    
    hidden_units_str = if (is.null(x$hidden_neurons)) {
        "Not specified"
    } else {
        paste(as.character(x$hidden_neurons), collapse = ", ")
    }
    
    summary_data = data.frame(
        type = c(
            "NN Model Type",
            "Number of Epochs",
            "Hidden Layer Units",
            "Number of Hidden Layers",
            "Pred. Type",
            "n_predictors",
            "n_response",
            "reg.", 
            "Device"
        ),
        res = c(
            "FFNN",
            as.character(x$n_epochs),
            hidden_units_str,
            as.character(length(x$hidden_neurons)),
            if (x$is_classification) "classification" else "regression",
            as.character(x$no_x),
            as.character(x$no_y),
            if (x$penalty == 0) "None" else glue("[\u03BB = {x$penalty}, \u03B1 = {x$mixture}]"), 
            x$device
        ),
        stringsAsFactors = FALSE
    )
    
    inner_acts = if (is.list(x$activations)) {
        acts = vapply(x$activations, concat, character(1))
        if (length(acts) == 1L && length(x$hidden_neurons) > 1L) {
            rep(acts, length(x$hidden_neurons))
        } else {
            acts
        }
    } else if (!is.null(x$activations)) {
        if (length(x$activations) == 1) {
            rep(as.character(x$activations), length(x$hidden_neurons))
        } else {
            as.character(x$activations)
        }
    } else {
        # When no activations specified 
        # replicate "None" for each hidden layer
        rep("None", length(x$hidden_neurons))
    }
    outer_acts = concat(x$output_activation)
    
    act_data = data.frame(
        layer = c(
            paste(
                ordinal_gen(seq_along(inner_acts)), "Layer",
                paste0("{", x$hidden_neurons, "}")
            ),
            "Output Activation"
        ),
        infos = c(inner_acts, outer_acts),
        stringsAsFactors = FALSE
    )
    
    # ---Display summary table---
    heading1 = rule(left = "FFNN Model Summary", line = "-")
    heading1_block = style_italic(heading1)
    cat_line("\n", heading1_block, "\n\n")
    table_summary(summary_data, l = 5, center_table = TRUE, style = list(sep = ":  "))
    cat("\n\n")
    
    # ---Activation function summary---
    heading2 = rule(left = "Activation function", line = "-")
    heading2_block = style_italic(heading2)
    cat_line("\n", heading2_block, "\n\n")
    table_summary(act_data, l = 5, center_table = TRUE, style = list(sep = ":  "))
    
    invisible(x)
}

#' Print method for rnn_fit objects
#'
#' @param x An object of class "rnn_fit"
#' @param ... Additional arguments (unused)
#'
#' @return No return value, called for side effects (printing model summary)
#'
#' @keywords internal
#' @export
print.rnn_fit = function(x, ...) {
    rnn_type = x$rnn_type
    
    # ---Title section---
    title = switch(
        rnn_type,
        "rnn" = "Recurrent Neural Networks",
        "lstm" = "Long Short-Term Memory (RNN)",
        "gru" = "Gated Recurrent Unit (RNN)"
    )
    title_block = rule(center = title, line = "=")
    title_block = style_bold(title_block)
    
    cat_line("\n", title_block, "\n")
    
    hidden_units_str = if (is.null(x$hidden_neurons)) {
        "Not specified"
    } else {
        paste(as.character(x$hidden_neurons), collapse = ", ")
    }
    
    summary_data = data.frame(
        type = c(
            "NN Model Type",
            "RNN Type",
            "Bidirectional",
            "Number of Epochs",
            "Hidden Layer Units",
            "Number of Hidden Layers",
            "Pred. Type",
            "n_predictors",
            "n_response",
            "reg.", 
            "Device"
        ),
        res = c(
            "RNN",
            toupper(rnn_type),
            if (x$bidirectional) "Yes" else "No",
            as.character(x$n_epochs),
            hidden_units_str,
            as.character(length(x$hidden_neurons)),
            if (x$is_classification) "classification" else "regression",
            as.character(x$no_x),
            as.character(x$no_y),
            if (x$penalty == 0) "None" else glue("[\u03BB = {x$penalty}, \u03B1 = {x$mixture}]"), 
            x$device
        ),
        stringsAsFactors = FALSE
    )
    
    inner_acts = if (is.list(x$activations)) {
        acts = vapply(x$activations, concat, character(1))
        if (length(acts) == 1L && length(x$hidden_neurons) > 1L) {
            rep(acts, length(x$hidden_neurons))
        } else {
            acts
        }
    } else if (!is.null(x$activations)) {
        if (length(x$activations) == 1) {
            rep(as.character(x$activations), length(x$hidden_neurons))
        } else {
            as.character(x$activations)
        }
    } else {
        rep("None", length(x$hidden_neurons))
    }
    outer_acts = concat(x$output_activation)
    
    act_data = data.frame(
        layer = c(
            paste(
                ordinal_gen(seq_along(inner_acts)), "Layer",
                paste0("{", x$hidden_neurons, "}")
            ),
            "Output Activation"
        ),
        infos = c(inner_acts, outer_acts),
        stringsAsFactors = FALSE
    )
    
    # ---Display summary table---
    heading1 = rule(left = "RNN Model Summary", line = "-")
    heading1_block = style_italic(heading1)
    cat_line("\n", heading1_block, "\n\n")
    table_summary(summary_data, l = 7, center_table = TRUE, style = list(sep = ":  "))
    cat("\n\n")
    
    # ---Activation function summary---
    heading2 = rule(left = "Activation function", line = "-")
    heading2_block = style_italic(heading2)
    cat_line("\n", heading2_block, "\n\n")
    table_summary(act_data, l = 5, center_table = TRUE, style = list(sep = ":  "))
    cat("\n")
    
    invisible(x)
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

# ---- ordinal_gen() ----

test_that("ordinal_gen() handles standard cases", {
    expect_equal(ordinal_gen(1), "1st")
    expect_equal(ordinal_gen(2), "2nd")
    expect_equal(ordinal_gen(3), "3rd")
    expect_equal(ordinal_gen(4), "4th")
})

test_that("ordinal_gen() handles teens (11th, 12th, 13th)", {
    expect_equal(ordinal_gen(11), "11th")
    expect_equal(ordinal_gen(12), "12th")
    expect_equal(ordinal_gen(13), "13th")
})

test_that("ordinal_gen() warns on values below 1", {
    expect_warning(ordinal_gen(0), "Values below 1")
})

# ---- print.nn_fit() ----

test_that("print.nn_fit() runs without error for regression", {
    skip_if_no_torch()
    fit = train_nn(
        x = Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 3
    )
    expect_output(print(fit))
})

test_that("print.nn_fit() runs without error for classification", {
    skip_if_no_torch()
    fit = train_nn(
        x = Species ~ .,
        data = iris,
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 3
    )
    expect_output(print(fit))
})

test_that("print.nn_fit() shows regularization info when penalty > 0", {
    skip_if_no_torch()
    fit = train_nn(
        x = Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = 16L,
        activations = "relu",
        epochs = 3,
        penalty = 0.01,
        mixture = 0.5
    )
    expect_output(print(fit), "\u03BB")
})

test_that("print.nn_fit() handles NULL hidden_neurons (no hidden layers)", {
    skip_if_no_torch()
    fit = train_nn(
        x = Sepal.Length ~ .,
        data = iris[, 1:4],
        epochs = 3
    )
    expect_output(print(fit), "Number of Hidden Layers\\s*:\\s*0")
})

test_that("print.nn_fit() shows arch info when nn_arch is provided", {
    skip_if_no_torch()
    arch = nn_arch(nn_name = "mlp_model")
    fit = train_nn(
        x = Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(16, 8),
        activations = "relu",
        architecture = arch,
        epochs = 3
    )
    expect_output(print(fit), "mlp_model")
})

# ---- print.ffnn_fit() ----

test_that("print.ffnn_fit() runs without error for regression", {
    skip_if_no_torch()
    fit = ffnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 3
    )
    expect_output(print(fit), "Feedforward Neural Network")
})

test_that("print.ffnn_fit() runs without error for classification", {
    skip_if_no_torch()
    fit = ffnn(
        Species ~ .,
        data = iris,
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 3
    )
    expect_output(print(fit), "classification")
})

test_that("print.ffnn_fit() shows regularization when penalty > 0", {
    skip_if_no_torch()
    fit = ffnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = 16L,
        activations = "relu",
        epochs = 3,
        penalty = 0.01,
        mixture = 0.5
    )
    expect_output(print(fit), "\u03BB")
})

test_that("print.ffnn_fit() shows 'Not specified' when hidden_neurons is NULL", {
    skip_if_no_torch()
    fit = ffnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = 8,
        epochs = 3
    )
    fit$hidden_neurons = NULL
    expect_output(print(fit), "Not specified")
})

# ---- print.rnn_fit() ----

test_that("print.rnn_fit() runs without error for lstm regression", {
    skip_if_no_torch()
    fit = rnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(16, 8),
        rnn_type = "lstm",
        activations = "relu",
        epochs = 3
    )
    expect_output(print(fit), "LSTM")
})

test_that("print.rnn_fit() runs without error for gru", {
    skip_if_no_torch()
    fit = rnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = 16L,
        rnn_type = "gru",
        epochs = 3
    )
    expect_output(print(fit), "GRU")
})

test_that("print.rnn_fit() runs without error for plain rnn", {
    skip_if_no_torch()
    fit = rnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = 16L,
        rnn_type = "rnn",
        epochs = 3
    )
    expect_output(print(fit), "Recurrent Neural Network")
})

test_that("print.rnn_fit() shows bidirectional status", {
    skip_if_no_torch()
    fit_bi = rnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = 16L,
        rnn_type = "lstm",
        bidirectional = TRUE,
        epochs = 3
    )
    fit_uni = rnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = 16L,
        rnn_type = "lstm",
        bidirectional = FALSE,
        epochs = 3
    )
    expect_output(print(fit_bi),  "Yes")
    expect_output(print(fit_uni), "No")
})

test_that("print.rnn_fit() shows regularization when penalty > 0", {
    skip_if_no_torch()
    fit = rnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = 16L,
        rnn_type = "lstm",
        epochs = 3,
        penalty = 0.01,
        mixture = 0.5
    )
    expect_output(print(fit), "\u03BB")
})

#' Custom Activation Function Constructor
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Wraps a user-supplied function into a validated custom activation,
#' ensuring it accepts and returns a `torch_tensor`. Performs an eager
#' dry-run probe at *definition time* so errors surface early, and
#' wraps the function with a *call-time* type guard for safety.
#'
#' @param fn A function taking a single tensor argument and returning a tensor.
#'   E.g. `\(x) torch::torch_tanh(x)`.
#' @param probe Logical. If `TRUE` (default), runs a dry-run with a small
#'   dummy tensor at definition time to catch obvious errors early.
#' @param .name A string to be stored in an attribute. Nothing special, except it is 
#'   used when displaying the info of a trained neural network model. 
#'   Default is `"<custom>"`. 
#'
#' @return An object of class `c("custom_activation", "parameterized_activation")`,
#'   compatible with `act_funs()`.
#'
#' @examples
#' \dontrun{
#' \donttest{
#' act_funs(relu, elu, new_act_fn(\(x) torch::torch_tanh(x)))
#' act_funs(new_act_fn(\(x) torch::nnf_silu(x)))
#' }
#' }
#' 
#' @export
new_act_fn = function(fn, probe = TRUE, .name = "<custom>") {
    if (!is.function(fn)) {
        cli::cli_abort(c(
            "{.arg fn} must be a function.",
            i = "Use a lambda like {.code \\(x) torch::torch_tanh(x)}."
        ), class = "custom_activation_type_error")
    }
    
    fn_formals = formals(fn)
    if (length(fn_formals) < 1L) {
        cli_abort(c(
            "{.arg fn} must accept at least one argument (the input tensor).",
            i = "Use a lambda like {.code \\(x) torch::torch_tanh(x)}."
        ), class = "custom_activation_arity_error")
    }
    
    # ---- Dry-run probe at definition time ----
    if (probe) {
        if (!requireNamespace("torch", quietly = TRUE)) {
            cli::cli_warn(c(
                "{.pkg torch} is not installed; skipping dry-run probe for {.fn new_act_fn}.",
                i = "Type safety will only be enforced at call time."
            ))
        } else {
            # Prepare a tiny 1-D tensor to validate
            dummy = torch::torch_zeros(2L)  
            out = torch::with_no_grad(
                tryCatch(
                    fn(dummy),
                    error = function(e) {
                        cli_abort(c(
                            "Dry-run of custom activation function failed.",
                            x = "{e$message}",
                            i = "Ensure {.arg fn} accepts a {.cls torch_tensor} and returns one."
                        ), class = "custom_activation_probe_error")
                    }
                )
            )
            .assert_tensor_output(out, context = "Dry-run")
        }
    }
    
    # ---- Build the call-time type-guarded wrapper ----
    guarded_fn = function(x) {
        out = fn(x)
        .assert_tensor_output(out, context = "Custom activation output")
        out
    }
    
    structure(
        list(),
        act_fn = guarded_fn, 
        act_name = .name,
        class = c("custom_activation", "parameterized_activation")
    )
}

#' New custom activation function validator
#' @keywords internal
#' @noRd
.assert_tensor_output = function(x, context = "Output") {
    if (!inherits(x, "torch_tensor")) {
        cli_abort(c(
            "{context} must be a {.cls torch_tensor}.",
            x = "Got {.cls {class(x)}}.",
            i = "Ensure your function returns the result of a {.pkg torch} operation."
        ), class = "custom_activation_output_error")
    }
    invisible(x)
}

#' Tunable hyperparameters for `kindling` models
#'
#' These parameters extend the **dials** framework to support hyperparameter
#' tuning of neural networks built with the `{kindling}` package. They control
#' network architecture, activation functions, optimization, and training
#' behavior.
#'
#' @section Architecture Strategy:
#' Since tidymodels tuning works with independent parameters, we use a simplified
#' approach where:
#' - `hidden_neurons` specifies a single value that will be used for ALL layers
#' - `activations` specifies a single activation that will be used for ALL layers
#' - `n_hlayers` controls the depth
#'
#' For more complex architectures with different neurons/activations per layer,
#' users should manually specify these after tuning or use custom tuning logic.
#'
#' @section Parameters:
#'
#' \describe{
#'   \item{`n_hlayers`}{Number of hidden layers in the network.}
#'   \item{`hidden_neurons`}{Number of units per hidden layer (applied to all layers).}
#'   \item{`activation`}{Single activation function applied to all hidden layers.}
#'   \item{`output_activation`}{Activation function for the output layer.}
#'   \item{`optimizer`}{Optimizer algorithm.}
#'   \item{`bias`}{Whether to include bias terms in layers.}
#'   \item{`validation_split`}{Proportion of training data held out for validation.}
#'   \item{`bidirectional`}{Whether RNN layers are bidirectional.}
#' }
#' 
#' @return
#' Each function returns a `dials` parameter object:
#' \describe{
#'     \item{`n_hlayers()`}{A quantitative parameter for the number of hidden layers}
#'     \item{`hidden_neurons()`}{A quantitative parameter for hidden units per layer}
#'     \item{`activations()`}{A qualitative parameter for activation function names}
#'     \item{`output_activation()`}{A qualitative parameter for output activation}
#'     \item{`optimizer()`}{A qualitative parameter for optimizer type}
#'     \item{`bias()`}{A qualitative parameter for bias inclusion}
#'     \item{`validation_split()`}{A quantitative parameter for validation proportion}
#'     \item{`bidirectional()`}{A qualitative parameter for bidirectional RNN}
#' }
#'
#' @examples
#' \donttest{
#' library(dials)
#' library(tune)
#'
#' # Create a tuning grid
#' grid = grid_regular(
#'     n_hlayers(range = c(1L, 4L)),
#'     hidden_neurons(range = c(32L, 256L)),
#'     activations(c('relu', 'elu', 'selu')),
#'     levels = c(4, 5, 3)
#' )
#'
#' # Use in a model spec
#' mlp_spec = mlp_kindling(
#'     mode = "classification",
#'     hidden_neurons = tune(),
#'     activations = tune(),
#'     epochs = tune(),
#'     learn_rate = tune()
#' )
#' }
#'
#' @name dials-kindling
#' @keywords internal
NULL

#' @section Number of Hidden Layers:
#' Controls the depth of the network. When tuning, this will determine how many
#' layers are created, each with `hidden_neurons` units and `activations` function.
#'
#' @param range A two-element integer vector with the default lower and upper bounds.
#' @param trans An optional transformation (e.g., `transform_log()`); `NULL` for none.
#'
#' @rdname dials-kindling
#' @export
n_hlayers = function(range = c(1L, 2L), trans = NULL) {
    dials::new_quant_param(
        type = "integer",
        range = range,
        inclusive = c(TRUE, TRUE),
        trans = trans,
        label = c(n_hlayers = "Number of Hidden Layers"),
        finalize = NULL
    )
}

#' @section Hidden Units per Layer:
#' Specifies the number of units per hidden layer.
#'
#' @param range A two-element integer vector with the default lower and upper bounds.
#' @param disc_values `NULL` (default) or an integer vector of specific possible disc_values 
#'     (e.g., `c(32L, 64L, 128L, 256L)`). When provided, tuning will be restricted to 
#'     these discrete values. The range is automatically derived from these values if not explicitly given.
#'     The `trans` parameter would still be ignored by this parameter when supplied. 
#' @param trans An optional transformation; `NULL` for none. 
#'
#' @rdname dials-kindling
#' @export
hidden_neurons = function(range = c(8L, 512L), disc_values = NULL, trans = NULL) {
    param = if (!is.null(disc_values)) {
        
        if (any(disc_values <= 0L) || anyNA(disc_values)) {
            rlang::abort("`disc_values` must be positive integers with no missing values.")
        }
        
        dials::new_quant_param(
            type = "integer",
            # range = range,
            values = disc_values, 
            inclusive = c(TRUE, TRUE),
            trans = NULL,
            label = c(hidden_neurons = "Hidden Units per Layer"),
            finalize = NULL
        )
        # attr(param, "discrete_values") = disc_values
    } else {
        dials::new_quant_param(
            type = "integer",
            range = range,
            inclusive = c(TRUE, TRUE),
            trans = trans,
            label = c(hidden_neurons = "Hidden Units per Layer"),
            finalize = NULL
        )
    }
    
    param
}

#' @section Activation Function (Hidden Layers):
#' Activation functions for hidden layers.
#'
#' @param values Character vector of possible activation names.
#'
#' @rdname dials-kindling
#' @export
activations = function(values = c(
    "relu", "relu6", "elu", "selu", "celu",
    "leaky_relu", "gelu", "softplus", "softshrink",
    "softsign", "tanhshrink", "hardtanh",
    "hardshrink", "hardswish", "hardsigmoid",
    "silu", "mish", "logsigmoid"
)) {
    dials::new_qual_param(
        type = "character",
        values = values,
        label = c(activations = "Activation Function (Hidden Layers)"),
        finalize = NULL
    )
}

#' @section Output Activation Function:
#' Activation function applied to the output layer. Values must correspond to
#' `torch::nnf_*` functions.
#'
#' @param values Character vector of possible output activation names.
#'
#' @rdname dials-kindling
#' @export
output_activation = function(values = c(
    "relu", "elu", "selu", "softplus",
    "softmax", "log_softmax", "logsigmoid",
    "hardtanh", "hardsigmoid", "silu"
)) {
    dials::new_qual_param(
        type = "character",
        values = values,
        label = c(output_activation = "Output Activation Function"),
        finalize = NULL
    )
}

#' @section Optimizer Type:
#' The optimization algorithm used during training.
#'
#' @param values Character vector of supported optimizers.
#'
#' @rdname dials-kindling
#' @export
optimizer = function(values = c("adam", "sgd", "rmsprop", "adamw")) {
    dials::new_qual_param(
        type = "character",
        values = values,
        label = c(optimizer = "Optimizer Type"),
        finalize = NULL
    )
}

#' @section Include Bias Terms:
#' Whether layers should include bias parameters.
#'
#' @param values Logical vector of possible values.
#'
#' @rdname dials-kindling
#' @export
bias = function(values = c(TRUE, FALSE)) {
    dials::new_qual_param(
        type = "logical",
        values = values,
        label = c(bias = "Include Bias Terms"),
        finalize = NULL
    )
}

#' @section Validation Split Proportion:
#' Fraction of the training data to use as a validation set during training.
#'
#' @param range A two-element numeric vector with the default lower and upper bounds.
#' @param trans An optional transformation; `NULL` for none.
#'
#' @rdname dials-kindling
#' @export
validation_split = function(range = c(0, 0.5), trans = NULL) {
    dials::new_quant_param(
        type = "double",
        range = range,
        inclusive = c(TRUE, TRUE),
        trans = trans,
        label = c(validation_split = "Validation Split Proportion"),
        finalize  = NULL
    )
}

#' @section Bidirectional RNN:
#' Whether recurrent layers should process sequences in both directions.
#'
#' @param values Logical vector of possible values.
#'
#' @rdname dials-kindling
#' @export
bidirectional = function(values = c(TRUE, FALSE)) {
    dials::new_qual_param(
        type = "logical",
        values = values,
        label = c(bidirectional = "Bidirectional RNN"),
        finalize = NULL
    )
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

# ---- grid_depth() ----

test_that("grid_depth() works with param objects (grid_depth.param)", {
    grid = grid_depth(
        hidden_neurons(c(32L, 128L)),
        activations(c("relu", "elu")),
        n_hlayer = 2L,
        type = "random",
        size = 5
    )
    expect_s3_class(grid, "tbl_df")
    expect_equal(nrow(grid), 5)
    expect_true("hidden_neurons" %in% names(grid))
    expect_true("activations" %in% names(grid))
    expect_true(all(lengths(grid$hidden_neurons) == 2L))
})

test_that("grid_depth() with n_hlayer range produces variable-length list cols", {
    grid = grid_depth(
        hidden_neurons(c(32L, 128L)),
        activations(c("relu", "elu")),
        n_hlayer = 2:3,
        type = "random",
        size = 20
    )
    depths = lengths(grid$hidden_neurons)
    expect_true(all(depths %in% 2:3))
})

test_that("grid_depth() type = 'regular' works", {
    grid = grid_depth(
        hidden_neurons(c(32L, 128L)),
        activations(c("relu", "elu")),
        n_hlayer = 2L,
        type = "regular",
        levels = 2
    )
    expect_s3_class(grid, "tbl_df")
    expect_gt(nrow(grid), 0)
})

test_that("grid_depth() type = 'latin_hypercube' works", {
    skip_if_not_installed("lhs")
    skip_if_no_torch()
    grid = grid_depth(
        hidden_neurons(c(32L, 128L)),
        activations(c("relu", "elu")),
        dials::epochs(c(50L, 200L)),
        n_hlayer = 2L,
        type = "latin_hypercube",
        size = 8
    )
    expect_equal(nrow(grid), 8)
    expect_true("epochs" %in% names(grid))
})

test_that("grid_depth() works from a list", {
    params = list(
        hidden_neurons(c(32L, 128L)),
        activations(c("relu", "elu"))
    )
    grid = grid_depth(params, n_hlayer = 2L, type = "random", size = 5)
    expect_s3_class(grid, "tbl_df")
    expect_equal(nrow(grid), 5)
})

test_that("grid_depth() works from a parameters object", {
    params = dials::parameters(
        hidden_neurons(c(32L, 128L)),
        activations(c("relu", "elu"))
    )
    grid = grid_depth(params, n_hlayer = 2L, type = "random", size = 5)
    expect_s3_class(grid, "tbl_df")
    expect_equal(nrow(grid), 5)
})

test_that("grid_depth() with n_hlayer = 1 returns scalar columns", {
    grid = grid_depth(
        hidden_neurons(c(32L, 128L)),
        activations(c("relu", "elu")),
        n_hlayer = 1L,
        type = "random",
        size = 5
    )
    expect_type(grid$hidden_neurons, "integer")
    expect_type(grid$activations, "character")
})

test_that("grid_depth() with n_hlayer as param object works", {
    grid = grid_depth(
        hidden_neurons(c(32L, 128L)),
        activations(c("relu", "elu")),
        n_hlayer = n_hlayers(range = c(2L, 3L)),
        type = "random",
        size = 10
    )
    expect_s3_class(grid, "tbl_df")
    expect_equal(nrow(grid), 10)
})

test_that("grid_depth() errors on unsupported class", {
    expect_error(grid_depth(42), class = "rlang_error")
})

test_that("grid_depth() works from a workflow", {
    skip_if_not_installed("workflows")
    skip_if_not_installed("tune")
    skip_if_no_torch()
    
    wf = workflows::workflow() |>
        workflows::add_model(
            mlp_kindling(
                hidden_neurons = tune::tune(), 
                activations = tune::tune(),
                mode = "regression"
            )
        ) |>
        workflows::add_formula(Sepal.Length ~ .)
    
    grid = grid_depth(wf, hidden_neurons(c(32L, 128L)), n_hlayer = 2L, type = "random", size = 5)
    expect_s3_class(grid, "tbl_df")
    expect_equal(nrow(grid), 5)
})

# ---- dials parameter constructors ----

test_that("hidden_neurons() with disc_values restricts to those values", {
    vals = c(32L, 64L, 128L)
    p = hidden_neurons(disc_values = vals)
    grid = dials::grid_random(p, size = 20)
    expect_true(all(grid$hidden_neurons %in% vals))
})

test_that("hidden_neurons() errors on non-positive disc_values", {
    expect_error(hidden_neurons(disc_values = c(-1L, 64L)), class = "rlang_error")
})

test_that("hidden_neurons() errors on NA in disc_values", {
    expect_error(hidden_neurons(disc_values = c(32L, NA_integer_)), class = "rlang_error")
})

test_that("hidden_neurons() with disc_values works inside grid_depth()", {
    grid = grid_depth(
        hidden_neurons(disc_values = c(32L, 64L, 128L)),
        activations(c("relu", "elu")),
        n_hlayer = 2L,
        type = "random",
        size = 5
    )
    expect_s3_class(grid, "tbl_df")
    expect_equal(nrow(grid), 5)
})

test_that("output_activation() values are usable in a grid", {
    grid = dials::grid_random(
        output_activation(c("relu", "softmax")),
        size = 4
    )
    expect_true(all(grid$output_activation %in% c("relu", "softmax")))
})

test_that("optimizer() values are usable in a grid", {
    grid = dials::grid_random(
        optimizer(c("adam", "sgd", "adamw")),
        size = 6
    )
    expect_true(all(grid$optimizer %in% c("adam", "sgd", "adamw")))
})

test_that("bias() samples only TRUE/FALSE", {
    grid = dials::grid_random(bias(), size = 10)
    expect_true(all(grid$bias %in% c(TRUE, FALSE)))
})

test_that("validation_split() stays within range", {
    grid = dials::grid_random(validation_split(range = c(0.1, 0.3)), size = 10)
    expect_true(all(grid$validation_split >= 0.1 & grid$validation_split <= 0.3))
})

test_that("bidirectional() samples only TRUE/FALSE", {
    grid = dials::grid_random(bidirectional(), size = 10)
    expect_true(all(grid$bidirectional %in% c(TRUE, FALSE)))
})

test_that("n_hlayers() respects custom range", {
    p = n_hlayers(range = c(1L, 5L))
    r = dials::range_get(p)
    expect_equal(r$lower, 1L)
    expect_equal(r$upper, 5L)
})

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("Tuning mlp_kindling with grid_depth works", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("workflows")
    skip_if_not_installed("tune")
    skip_if_not_installed("rsample")
    skip_if_not_installed("recipes")
    skip_if_not_installed("dials")
    skip_if_no_torch()
    
    mlp_spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = tune::tune(),
        activations = tune::tune(),
        output_activation = tune::tune(),
        epochs = 10
    )
    
    iris_recipe = recipes::recipe(Species ~ ., data = iris)
    wf = workflows::workflow() |>
        workflows::add_recipe(iris_recipe) |>
        workflows::add_model(mlp_spec)
    
    set.seed(123)
    folds = rsample::vfold_cv(iris, v = 2)
    
    grid = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        output_activation(c("sigmoid", "linear")),
        n_hlayer = 2,
        size = 3,
        type = "random"
    )
    
    expect_s3_class(grid, "tbl_df")
    expect_equal(nrow(grid), 3)
    expect_true("hidden_neurons" %in% names(grid))
    expect_true("activations" %in% names(grid))
    expect_true("output_activation" %in% names(grid))
    expect_true(is.list(grid$hidden_neurons))
    expect_true(is.list(grid$activations))
    expect_equal(length(grid$hidden_neurons[[1]]), 2)
    expect_equal(length(grid$activations[[1]]), 2)
    
    out = tune::tune_grid(
        wf,
        resamples = folds,
        grid = grid,
        control = tune::control_grid(save_pred = FALSE)
    )
    
    expect_s3_class(out, "tune_results")
    expect_true(nrow(tune::collect_metrics(out)) > 0)
})

test_that("Tuning mlp_kindling with latin_hypercube grid works", {
    skip_if_not_installed("lhs")
    skip_if_not_installed("parsnip")
    skip_if_not_installed("workflows")
    skip_if_not_installed("tune")
    skip_if_no_torch()
    
    mlp_spec = mlp_kindling(
        mode = "regression",
        hidden_neurons = tune::tune(),
        activations = tune::tune(),
        epochs = 5
    )
    
    mtcars_recipe = recipes::recipe(mpg ~ ., data = mtcars)
    wf = workflows::workflow() |>
        workflows::add_recipe(mtcars_recipe) |>
        workflows::add_model(mlp_spec)
    
    set.seed(456)
    folds = rsample::vfold_cv(mtcars, v = 2)
    
    grid = grid_depth(
        hidden_neurons(c(8L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = 2:3,
        size = 4,
        type = "latin_hypercube"
    )
    
    expect_equal(nrow(grid), 4)
    expect_true(all(purrr::map_int(grid$hidden_neurons, length) %in% 2:3))
    
    results = tune::tune_grid(
        wf,
        resamples = folds,
        grid = grid,
        control = tune::control_grid(save_pred = FALSE)
    )
    
    expect_s3_class(results, "tune_results")
    
    # ---SELECT BEST---
    best = tune::select_best(results, metric = "rmse")
    expect_s3_class(best, "tbl_df")
    expect_true("hidden_neurons" %in% names(best))
})

test_that("Tuning rnn_kindling with grid_depth works", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("workflows")
    skip_if_not_installed("tune")
    skip_if_not_installed("rsample")
    skip_if_not_installed("recipes")
    skip_if_no_torch()
    
    rnn_spec = rnn_kindling(
        mode = "classification",
        hidden_neurons = tune::tune(),
        activations = tune::tune(),
        epochs = 5
    )
    
    iris_recipe = recipes::recipe(Species ~ ., data = iris)
    wf = workflows::workflow() |>
        workflows::add_recipe(iris_recipe) |>
        workflows::add_model(rnn_spec)
    
    set.seed(789)
    folds = rsample::vfold_cv(iris, v = 2)
    
    grid = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = 2,
        size = 3,
        type = "random"
    )
    
    expect_s3_class(grid, "tbl_df")
    expect_equal(nrow(grid), 3)
    expect_true("hidden_neurons" %in% names(grid))
    expect_true("activations" %in% names(grid))
    
    # --RNN type shouldn't be tunable, thus it won't exist
    expect_false("rnn_type" %in% names(grid))
    
    OUT = tune::tune_grid(
        wf,
        resamples = folds,
        grid = grid,
        control = tune::control_grid(save_pred = FALSE)
    )
    
    expect_s3_class(OUT, "tune_results")
    expect_true(nrow(tune::collect_metrics(OUT)) > 0)
})

test_that("grid_depth handles different n_hlayer specifications", {
    skip_if_no_torch()
    
    grid1 = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = 2,
        size = 5,
        type = "random"
    )
    
    expect_true(any(purrr::map_int(grid1$hidden_neurons, length) == 2))
    expect_true(any(purrr::map_int(grid1$activations, length) == 2))
    
    # ---Multiple depths---
    grid2 = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = 1:3,
        size = 9,
        type = "random"
    )
    
    depths = purrr::map_int(grid2$hidden_neurons, length)
    expect_true(all(depths %in% 1:3))
})

test_that("grid_depth works with workflow method", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("workflows")
    skip_if_no_torch()
    
    mlp_spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = tune::tune(),
        activations = tune::tune(),
        epochs = tune::tune()
    )
    
    wf = workflows::workflow() |>
        workflows::add_recipe(recipes::recipe(Species ~ ., data = iris)) |>
        workflows::add_model(mlp_spec)
    
    grid = grid_depth(
        wf,
        n_hlayer = 2,
        size = 5,
        type = "random"
    )
    
    expect_s3_class(grid, "tbl_df")
    expect_true("hidden_neurons" %in% names(grid))
    expect_true("activations" %in% names(grid))
    expect_true("epochs" %in% names(grid))
})

test_that("grid_depth handles scalar parameters correctly", {
    skip_if_no_torch()
    
    grid = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        
        # `epochs` belong to `{dials}`, not in `{kindling}`
        dials::epochs(c(10L, 50L)),
        
        # `learn_rate` belong to `{dials}`, not in `{kindling}`
        dials::learn_rate(c(0.001, 0.1)),
        
        n_hlayer = 2,
        size = 5,
        type = "random"
    )
    
    expect_equal(nrow(grid), 5)
    expect_true("epochs" %in% names(grid))
    expect_true("learn_rate" %in% names(grid))
})

test_that("finalize_workflow works with grid_depth results", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("workflows")
    skip_if_not_installed("tune")
    skip_if_not_installed("rsample")
    skip_if_no_torch()
    
    mlp_spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = tune::tune(),
        activations = tune::tune(),
        epochs = 10
    )
    
    wf = workflows::workflow() |>
        workflows::add_recipe(recipes::recipe(Species ~ ., data = iris)) |>
        workflows::add_model(mlp_spec)
    
    set.seed(321)
    folds = rsample::vfold_cv(iris, v = 2)
    
    grid = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = 2,
        size = 3,
        type = "random"
    )
    
    results = tune::tune_grid(wf, folds, grid = grid)
    best = tune::select_best(results, metric = "accuracy")
    
    final_wf = tune::finalize_workflow(wf, best)
    
    expect_s3_class(final_wf, "workflow")
    
    # ---Final parameters stores into a list—expect no failures---
    expect_error({
        final_nn_model = parsnip::fit(final_wf, data = iris)
        final_nn_model
    }, NA)
})

# ============================================================================
# NEW TESTS FOR n_hlayers PARAMETER TUNING
# ============================================================================

test_that("grid_depth accepts n_hlayers() parameter object", {
    skip_if_no_torch()
    
    grid = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = n_hlayers(range = c(2L, 4L)),
        size = 6,
        type = "random"
    )
    
    expect_s3_class(grid, "tbl_df")
    expect_equal(nrow(grid), 6)
    expect_true("hidden_neurons" %in% names(grid))
    expect_true("activations" %in% names(grid))
    
    # Check that depths vary within the specified range
    depths = purrr::map_int(grid$hidden_neurons, length)
    expect_true(all(depths >= 2 & depths <= 4))
    expect_true(all(depths %in% 2:4))
})

test_that("grid_depth with n_hlayers() parameter in regular grid", {
    skip_if_no_torch()
    
    grid = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = n_hlayers(range = c(1L, 3L)),
        type = "regular",
        levels = 2
    )
    
    expect_s3_class(grid, "tbl_df")
    expect_true(nrow(grid) > 0)
    
    # Check that depths are within range
    depths = purrr::map_int(grid$hidden_neurons, length)
    expect_true(all(depths >= 1 & depths <= 3))
})

test_that("grid_depth with n_hlayers() in latin_hypercube", {
    skip_if_not_installed("lhs")
    skip_if_no_torch()
    
    grid = grid_depth(
        hidden_neurons(c(8L, 64L)),
        activations(c("relu", "elu", "selu")),
        n_hlayer = n_hlayers(range = c(2L, 5L)),
        size = 10,
        type = "latin_hypercube"
    )
    
    expect_equal(nrow(grid), 10)
    depths = purrr::map_int(grid$hidden_neurons, length)
    expect_true(all(depths >= 2 & depths <= 5))
    
    # Check activations match neuron depths
    activation_depths = purrr::map_int(grid$activations, length)
    expect_equal(depths, activation_depths)
})

test_that("Tuning mlp_kindling with n_hlayers parameter works end-to-end", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("workflows")
    skip_if_not_installed("tune")
    skip_if_not_installed("rsample")
    skip_if_not_installed("recipes")
    skip_if_no_torch()
    
    mlp_spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = tune::tune(),
        activations = tune::tune(),
        epochs = 10
    )
    
    iris_recipe = recipes::recipe(Species ~ ., data = iris)
    wf = workflows::workflow() |>
        workflows::add_recipe(iris_recipe) |>
        workflows::add_model(mlp_spec)
    
    set.seed(999)
    folds = rsample::vfold_cv(iris, v = 2)
    
    # Use n_hlayers() parameter to tune depth
    grid = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = n_hlayers(range = c(1L, 3L)),
        size = 6,
        type = "random"
    )
    
    expect_equal(nrow(grid), 6)
    
    depths = purrr::map_int(grid$hidden_neurons, length)
    expect_true(length(unique(depths)) > 1)
    
    results = tune::tune_grid(
        wf,
        resamples = folds,
        grid = grid,
        control = tune::control_grid(save_pred = FALSE)
    )
    
    expect_s3_class(results, "tune_results")
    expect_true(nrow(tune::collect_metrics(results)) > 0)
    
    best = tune::select_best(results, metric = "accuracy")
    expect_s3_class(best, "tbl_df")
    expect_true("hidden_neurons" %in% names(best))
    expect_true(is.list(best$hidden_neurons))
    
    final_wf = tune::finalize_workflow(wf, best)
    expect_s3_class(final_wf, "workflow")
})

test_that("grid_depth with n_hlayers() in parameters object", {
    skip_if_no_torch()
    
    params = dials::parameters(
        hidden_neurons(c(16L, 64L)),
        activations(c("relu", "elu")),
        n_hlayers(range = c(2L, 4L))
    )
    
    grid = grid_depth(
        params,
        size = 8,
        type = "random"
    )
    
    expect_equal(nrow(grid), 8)
    expect_true("hidden_neurons" %in% names(grid))
    expect_true("activations" %in% names(grid))
    
    depths = purrr::map_int(grid$hidden_neurons, length)
    expect_true(all(depths >= 2 & depths <= 4))
})

test_that("grid_depth with n_hlayers() and scalar parameters", {
    skip_if_no_torch()
    
    grid = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        dials::epochs(c(10L, 50L)),
        dials::learn_rate(c(0.001, 0.1)),
        n_hlayer = n_hlayers(range = c(1L, 3L)),
        size = 10,
        type = "random"
    )
    
    expect_equal(nrow(grid), 10)
    expect_true(all(c("hidden_neurons", "activations", "epochs", "learn_rate") %in% names(grid)))
    
    depths = purrr::map_int(grid$hidden_neurons, length)
    expect_true(all(depths >= 1 & depths <= 3))
})

test_that("n_hlayers parameter respects custom ranges", {
    skip_if_no_torch()
    
    grid1 = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu")),
        n_hlayer = n_hlayers(range = c(1L, 2L)),
        size = 10,
        type = "random"
    )
    
    depths1 = purrr::map_int(grid1$hidden_neurons, length)
    expect_true(all(depths1 %in% 1:2))
    
    grid2 = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu")),
        n_hlayer = n_hlayers(range = c(3L, 6L)),
        size = 10,
        type = "random"
    )
    
    depths2 = purrr::map_int(grid2$hidden_neurons, length)
    expect_true(all(depths2 >= 3 & depths2 <= 6))
})

test_that("grid_depth backward compatibility with integer vector n_hlayer", {
    skip_if_no_torch()
    
    grid1 = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = 2,
        size = 5,
        type = "random"
    )
    
    depths1 = purrr::map_int(grid1$hidden_neurons, length)
    expect_true(all(depths1 == 2))
    
    grid2 = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "elu")),
        n_hlayer = 2:4,
        size = 9,
        type = "random"
    )
    
    depths2 = purrr::map_int(grid2$hidden_neurons, length)
    expect_true(all(depths2 %in% 2:4))
})

test_that("Tuning rnn_kindling with n_hlayers parameter", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("workflows")
    skip_if_not_installed("tune")
    skip_if_not_installed("rsample")
    skip_if_no_torch()
    
    rnn_spec = rnn_kindling(
        mode = "classification",
        hidden_neurons = tune::tune(),
        activations = tune::tune(),
        epochs = 5
    )
    
    wf = workflows::workflow() |>
        workflows::add_recipe(recipes::recipe(Species ~ ., data = iris)) |>
        workflows::add_model(rnn_spec)
    
    set.seed(777)
    folds = rsample::vfold_cv(iris, v = 2)
    
    grid = grid_depth(
        hidden_neurons(c(16L, 32L)),
        activations(c("relu", "hardtanh")),
        n_hlayer = n_hlayers(range = c(1L, 3L)),
        size = 5,
        type = "random"
    )
    
    depths = purrr::map_int(grid$hidden_neurons, length)
    expect_true(all(depths >= 1 & depths <= 3))
    
    results = tune::tune_grid(
        wf,
        resamples = folds,
        grid = grid,
        control = tune::control_grid(save_pred = FALSE)
    )
    
    expect_s3_class(results, "tune_results")
    expect_true(nrow(tune::collect_metrics(results)) > 0)
})

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("hidden_neurons() with discrete values works in tuning workflow", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("workflows")
    skip_if_not_installed("tune")
    skip_if_not_installed("rsample")
    skip_if_not_installed("recipes")
    skip_if_no_torch()
    
    mlp_spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = tune::tune(),
        activations = tune::tune(),
        epochs = 10
    )
    
    iris_recipe = recipes::recipe(Species ~ ., data = iris)
    wf = workflows::workflow() |>
        workflows::add_recipe(iris_recipe) |>
        workflows::add_model(mlp_spec)
    
    set.seed(123)
    folds = rsample::vfold_cv(iris, v = 2)
    
    grid = grid_depth(
        hidden_neurons(disc_values = c(32, 64, 128)),
        activations(c("relu", "elu")),
        n_hlayer = 2,
        size = 6,
        type = "random"
    )
    
    expect_equal(nrow(grid), 6)
    all_neurons = unlist(grid$hidden_neurons)
    expect_true(all(all_neurons %in% c(32, 64, 128)))
    
    results = tune::tune_grid(
        wf,
        resamples = folds,
        grid = grid,
        control = tune::control_grid(save_pred = FALSE)
    )
    
    expect_s3_class(results, "tune_results")
    expect_true(nrow(tune::collect_metrics(results)) > 0)
    
    best = tune::select_best(results, metric = "accuracy")
    expect_s3_class(best, "tbl_df")
    expect_true("hidden_neurons" %in% names(best))
    
    best_neurons = unlist(best$hidden_neurons)
    expect_true(all(best_neurons %in% c(32, 64, 128)))
    
    final_wf = tune::finalize_workflow(wf, best)
    expect_s3_class(final_wf, "workflow")
    
    final_fit = parsnip::fit(final_wf, data = iris)
    expect_s3_class(final_fit, "workflow")
})

#' Prepare arguments for kindling models
#' @keywords internal
prepare_kindling_args = function(args) {
    args = lapply(args, function(val) {
        if (rlang::is_quosure(val)) {
            return(rlang::eval_tidy(val))
        }
        if (rlang::is_formula(val) && is.null(rlang::f_lhs(val))) {
            return(eval(rlang::f_rhs(val), envir = rlang::f_env(val)))
        }
        val
    })
    
    args = Filter(Negate(is.null), args)
    
    if ("hidden_neurons" %in% names(args) && is.list(args$hidden_neurons)) {
        if (length(args$hidden_neurons) == 1) {
            args$hidden_neurons = args$hidden_neurons[[1]]
        }
    }
    
    if ("activations" %in% names(args) && is.list(args$activations)) {
        if (length(args$activations) == 1) {
            args$activations = args$activations[[1]]
        }
    }
    
    args
}

#' Register kindling engines with parsnip
#'
#' @description
#' This function registers the kindling engine for MLP, RNN, and train_nn models
#' with parsnip. It should be called when the package is loaded.
#'
#' @keywords internal
make_kindling = function() {
    model_env = parsnip::get_model_env()
    
    # ===================================================
    # ---- MLP (Multi-Layer Perceptron) Registration ----
    # ===================================================
    
    if ("mlp_kindling" %in% model_env$models) {
        if (interactive() || identical(Sys.getenv("DEVTOOLS_LOAD"), "true")) {
            if (exists("mlp_kindling", envir = model_env)) {
                rm(list = "mlp_kindling", envir = model_env)
            }
            model_env$models = setdiff(model_env$models, "mlp_kindling")
        } else {
            if ("rnn_kindling" %in% model_env$models) {
                return(invisible(TRUE))
            }
        }
    }
    
    parsnip::set_new_model("mlp_kindling")
    
    parsnip::set_model_mode(model = "mlp_kindling", mode = "regression")
    parsnip::set_model_mode(model = "mlp_kindling", mode = "classification")
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "hidden_neurons",
        original = "hidden_neurons",
        func = list(pkg = "kindling", fun = "hidden_neurons"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "activations",
        original = "activations",
        func = list(pkg = "kindling", fun = "activations"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "output_activation",
        original = "output_activation",
        func = list(pkg = "kindling", fun = "output_activation"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "bias",
        original = "bias",
        func = list(pkg = "kindling", fun = "bias"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "epochs",
        original = "epochs",
        func = list(pkg = "dials", fun = "epochs"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "batch_size",
        original = "batch_size",
        func = list(pkg = "dials", fun = "batch_size"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "penalty",
        original = "penalty",
        func = list(pkg = "dials", fun = "penalty"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "mixture",
        original = "mixture",
        func = list(pkg = "dials", fun = "mixture"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "learn_rate",
        original = "learn_rate",
        func = list(pkg = "dials", fun = "learn_rate"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "optimizer",
        original = "optimizer",
        func = list(pkg = "kindling", fun = "optimizer"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "mlp_kindling",
        eng = "kindling",
        parsnip = "validation_split",
        original = "validation_split",
        func = list(pkg = "kindling", fun = "validation_split"),
        has_submodel = FALSE
    )
    
    # ---- Non-tunable engine args ----
    
    parsnip::set_model_engine(
        model = "mlp_kindling",
        mode = "regression",
        eng = "kindling"
    )
    
    parsnip::set_model_engine(
        model = "mlp_kindling",
        mode = "classification",
        eng = "kindling"
    )
    
    parsnip::set_dependency(
        model = "mlp_kindling",
        eng = "kindling",
        pkg = "kindling"
    )
    
    
    parsnip::set_fit(
        model = "mlp_kindling",
        mode = "regression",
        eng = "kindling",
        value = list(
            interface = "formula",
            protect = c("formula", "data"),
            func = c(pkg = "kindling", fun = "ffnn_wrapper"),
            defaults = list()
        )
    )
    
    parsnip::set_fit(
        model = "mlp_kindling",
        mode = "classification",
        eng = "kindling",
        value = list(
            interface = "formula",
            protect = c("formula", "data"),
            func = c(pkg = "kindling", fun = "ffnn_wrapper"),
            defaults = list(loss = "cross_entropy")
        )
    )
    
    parsnip::set_encoding(
        model = "mlp_kindling",
        mode = "regression",
        eng = "kindling",
        options = list(
            predictor_indicators = "none",
            compute_intercept = FALSE,
            remove_intercept = TRUE,
            allow_sparse_x = FALSE
        )
    )
    
    parsnip::set_encoding(
        model = "mlp_kindling",
        mode = "classification",
        eng = "kindling",
        options = list(
            predictor_indicators = "none",
            compute_intercept = FALSE,
            remove_intercept = TRUE,
            allow_sparse_x = FALSE
        )
    )
    
    parsnip::set_pred(
        model = "mlp_kindling",
        mode = "regression",
        eng = "kindling",
        type = "numeric",
        value = list(
            pre = NULL,
            post = NULL,
            func = c(fun = "predict"),
            args = list(
                object = rlang::expr(object$fit),
                newdata = rlang::expr(new_data),
                type = "response"
            )
        )
    )
    
    parsnip::set_pred(
        model = "mlp_kindling",
        mode = "classification",
        eng = "kindling",
        type = "class",
        value = list(
            pre = NULL,
            post = NULL,
            func = c(fun = "predict"),
            args = list(
                object = rlang::expr(object$fit),
                newdata = rlang::expr(new_data),
                type = "response"
            )
        )
    )
    
    parsnip::set_pred(
        model = "mlp_kindling",
        mode = "classification",
        eng = "kindling",
        type = "prob",
        value = list(
            pre = NULL,
            post = function(result, object) {
                tibble::as_tibble(result)
            },
            func = c(fun = "predict"),
            args = list(
                object = rlang::expr(object$fit),
                newdata = rlang::expr(new_data),
                type = "prob"
            )
        )
    )
    
    # =====================================================
    # ---- RNN (Recurrent Neural Network) Registration ----
    # =====================================================
    
    if ("rnn_kindling" %in% model_env$models) {
        if (interactive() || identical(Sys.getenv("DEVTOOLS_LOAD"), "true")) {
            if (exists("rnn_kindling", envir = model_env)) {
                rm(list = "rnn_kindling", envir = model_env)
            }
            model_env$models = setdiff(model_env$models, "rnn_kindling")
        } else {
            return(invisible(TRUE))
        }
    }
    
    parsnip::set_new_model("rnn_kindling")
    
    parsnip::set_model_mode(model = "rnn_kindling", mode = "regression")
    parsnip::set_model_mode(model = "rnn_kindling", mode = "classification")
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "hidden_neurons",
        original = "hidden_neurons",
        func = list(pkg = "kindling", fun = "hidden_neurons"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "activations",
        original = "activations",
        func = list(pkg = "kindling", fun = "activations"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "output_activation",
        original = "output_activation",
        func = list(pkg = "kindling", fun = "output_activation"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "bias",
        original = "bias",
        func = list(pkg = "kindling", fun = "bias"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "bidirectional",
        original = "bidirectional",
        func = list(pkg = "kindling", fun = "bidirectional"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "dropout",
        original = "dropout",
        func = list(pkg = "dials", fun = "dropout"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "epochs",
        original = "epochs",
        func = list(pkg = "dials", fun = "epochs"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "batch_size",
        original = "batch_size",
        func = list(pkg = "dials", fun = "batch_size"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "penalty",
        original = "penalty",
        func = list(pkg = "dials", fun = "penalty"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "mixture",
        original = "mixture",
        func = list(pkg = "dials", fun = "mixture"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "learn_rate",
        original = "learn_rate",
        func = list(pkg = "dials", fun = "learn_rate"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "optimizer",
        original = "optimizer",
        func = list(pkg = "kindling", fun = "optimizer"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "rnn_kindling",
        eng = "kindling",
        parsnip = "validation_split",
        original = "validation_split",
        func = list(pkg = "kindling", fun = "validation_split"),
        has_submodel = FALSE
    )
    
    # ---- Non-tunable engine args ----
    
    parsnip::set_model_engine(
        model = "rnn_kindling",
        mode = "regression",
        eng = "kindling"
    )
    
    parsnip::set_model_engine(
        model = "rnn_kindling",
        mode = "classification",
        eng = "kindling"
    )
    
    parsnip::set_dependency(
        model = "rnn_kindling",
        eng = "kindling",
        pkg = "kindling"
    )
    
    
    parsnip::set_fit(
        model = "rnn_kindling",
        mode = "regression",
        eng = "kindling",
        value = list(
            interface = "formula",
            protect = c("formula", "data"),
            func = c(pkg = "kindling", fun = "rnn_wrapper"),
            defaults = list()
        )
    )
    
    parsnip::set_fit(
        model = "rnn_kindling",
        mode = "classification",
        eng = "kindling",
        value = list(
            interface = "formula",
            protect = c("formula", "data"),
            func = c(pkg = "kindling", fun = "rnn_wrapper"),
            defaults = list(loss = "cross_entropy")
        )
    )
    
    parsnip::set_encoding(
        model = "rnn_kindling",
        mode = "regression",
        eng = "kindling",
        options = list(
            predictor_indicators = "none",
            compute_intercept = FALSE,
            remove_intercept = TRUE,
            allow_sparse_x = FALSE
        )
    )
    
    parsnip::set_encoding(
        model = "rnn_kindling",
        mode = "classification",
        eng = "kindling",
        options = list(
            predictor_indicators = "none",
            compute_intercept = FALSE,
            remove_intercept = TRUE,
            allow_sparse_x = FALSE
        )
    )
    
    parsnip::set_pred(
        model = "rnn_kindling",
        mode = "regression",
        eng = "kindling",
        type = "numeric",
        value = list(
            pre = NULL,
            post = NULL,
            func = c(fun = "predict"),
            args = list(
                object = rlang::expr(object$fit),
                newdata = rlang::expr(new_data),
                type = "response"
            )
        )
    )
    
    parsnip::set_pred(
        model = "rnn_kindling",
        mode = "classification",
        eng = "kindling",
        type = "class",
        value = list(
            pre = NULL,
            post = NULL,
            func = c(fun = "predict"),
            args = list(
                object = rlang::expr(object$fit),
                newdata = rlang::expr(new_data),
                type = "response"
            )
        )
    )
    
    parsnip::set_pred(
        model = "rnn_kindling",
        mode = "classification",
        eng = "kindling",
        type = "prob",
        value = list(
            pre = NULL,
            post = function(result, object) {
                tibble::as_tibble(result)
            },
            func = c(fun = "predict"),
            args = list(
                object = rlang::expr(object$fit),
                newdata = rlang::expr(new_data),
                type = "prob"
            )
        )
    )
    
    # ===================================
    # ---- train_nnsnip Registration ----
    # ===================================
    
    if ("train_nnsnip" %in% model_env$models) {
        if (interactive() || identical(Sys.getenv("DEVTOOLS_LOAD"), "true")) {
            if (exists("train_nnsnip", envir = model_env)) {
                rm(list = "train_nnsnip", envir = model_env)
            }
            model_env$models = setdiff(model_env$models, "train_nnsnip")
        } else {
            return(invisible(TRUE))
        }
    }
    
    parsnip::set_new_model("train_nnsnip")
    
    parsnip::set_model_mode(model = "train_nnsnip", mode = "regression")
    parsnip::set_model_mode(model = "train_nnsnip", mode = "classification")
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "hidden_neurons",
        original = "hidden_neurons",
        func = list(pkg = "kindling", fun = "hidden_neurons"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "activations",
        original = "activations",
        func = list(pkg = "kindling", fun = "activations"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "output_activation",
        original = "output_activation",
        func = list(pkg = "kindling", fun = "output_activation"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "bias",
        original = "bias",
        func = list(pkg = "kindling", fun = "bias"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "epochs",
        original = "epochs",
        func = list(pkg = "dials", fun = "epochs"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "batch_size",
        original = "batch_size",
        func = list(pkg = "dials", fun = "batch_size"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "penalty",
        original = "penalty",
        func = list(pkg = "dials", fun = "penalty"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "mixture",
        original = "mixture",
        func = list(pkg = "dials", fun = "mixture"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "learn_rate",
        original = "learn_rate",
        func = list(pkg = "dials", fun = "learn_rate"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "optimizer",
        original = "optimizer",
        func = list(pkg = "kindling", fun = "optimizer"),
        has_submodel = FALSE
    )
    
    parsnip::set_model_arg(
        model = "train_nnsnip",
        eng = "kindling",
        parsnip = "validation_split",
        original = "validation_split",
        func = list(pkg = "kindling", fun = "validation_split"),
        has_submodel = FALSE
    )
    
    # ---- Non-tunable engine args ----
    
    parsnip::set_model_engine(
        model = "train_nnsnip",
        mode = "regression",
        eng = "kindling"
    )
    
    parsnip::set_model_engine(
        model = "train_nnsnip",
        mode = "classification",
        eng = "kindling"
    )
    
    parsnip::set_dependency(
        model = "train_nnsnip",
        eng = "kindling",
        pkg = "kindling"
    )
    
    
    parsnip::set_fit(
        model = "train_nnsnip",
        mode = "regression",
        eng = "kindling",
        value = list(
            interface = "formula",
            protect = c("formula", "data"),
            func = c(pkg = "kindling", fun = "train_nn_wrapper"),
            defaults = list()
        )
    )
    
    parsnip::set_fit(
        model = "train_nnsnip",
        mode = "classification",
        eng = "kindling",
        value = list(
            interface = "formula",
            protect = c("formula", "data"),
            func = c(pkg = "kindling", fun = "train_nn_wrapper"),
            defaults = list(loss = "cross_entropy")
        )
    )
    
    parsnip::set_encoding(
        model = "train_nnsnip",
        mode = "regression",
        eng = "kindling",
        options = list(
            predictor_indicators = "none",
            compute_intercept = FALSE,
            remove_intercept = TRUE,
            allow_sparse_x = FALSE
        )
    )
    
    parsnip::set_encoding(
        model = "train_nnsnip",
        mode = "classification",
        eng = "kindling",
        options = list(
            predictor_indicators = "none",
            compute_intercept = FALSE,
            remove_intercept = TRUE,
            allow_sparse_x = FALSE
        )
    )
    
    parsnip::set_pred(
        model = "train_nnsnip",
        mode = "regression",
        eng = "kindling",
        type = "numeric",
        value = list(
            pre = NULL,
            post = NULL,
            func = c(fun = "predict"),
            args = list(
                object = rlang::expr(object$fit),
                newdata = rlang::expr(new_data),
                type = "response"
            )
        )
    )
    
    parsnip::set_pred(
        model = "train_nnsnip",
        mode = "classification",
        eng = "kindling",
        type = "class",
        value = list(
            pre = NULL,
            post = NULL,
            func = c(fun = "predict"),
            args = list(
                object = rlang::expr(object$fit),
                newdata = rlang::expr(new_data),
                type = "response"
            )
        )
    )
    
    parsnip::set_pred(
        model = "train_nnsnip",
        mode = "classification",
        eng = "kindling",
        type = "prob",
        value = list(
            pre = NULL,
            post = function(result, object) {
                tibble::as_tibble(result)
            },
            func = c(fun = "predict"),
            args = list(
                object = rlang::expr(object$fit),
                newdata = rlang::expr(new_data),
                type = "prob"
            )
        )
    )
    
    invisible(TRUE)
}

#' Functions to generate `nn_module` (language) expression
#'
#' @name nn_gens
#'
#' @section Feed-Forward Neural Network Module Generator:
#' The `ffnn_generator()` function generates a feed-forward neural network (FFNN) module expression
#' from the `torch` package in R. It allows customization of the FFNN architecture,
#' including the number of hidden layers, neurons, and activation functions.
#'
#' @param nn_name Character. Name of the generated FFNN module class. Default is `"DeepFFN"`.
#' @param hd_neurons Integer vector. Number of neurons in each hidden layer.
#' @param no_x Integer. Number of input features.
#' @param no_y Integer. Number of output features.
#' @param activations Activation function specifications for each hidden layer.
#' Can be:
#' - `NULL`: No activation functions.
#' - Character vector: e.g., `c("relu", "sigmoid")`.
#' - List: e.g., `act_funs(relu, elu, softshrink = args(lambd = 0.5))`.
#' - `activation_spec` object from `act_funs()`.
#'
#' If the length of `activations` is `1L`, this will be the activation throughout the architecture.
#'
#' @param output_activation Optional. Activation function for the output layer.
#' Same format as `activations` but should be a single activation.
#' @param bias Logical. Whether to use bias weights. Default is `TRUE`.
#'
#' @return A `torch` module expression representing the FFNN.
#'
#' @details The generated FFNN module will have the specified number of hidden layers,
#' with each layer containing the specified number of neurons. Activation functions
#' can be applied after each hidden layer as specified.
#' This can be used for both classification and regression tasks.
#'
#' The generated module properly namespaces all torch functions to avoid
#' polluting the global namespace.
#'
#' @examples
#' \donttest{
#' # FFNN
#' if (torch::torch_is_installed()) {
#'     # Generate an MLP module with 3 hidden layers
#'     ffnn_mod = ffnn_generator(
#'         nn_name = "MyFFNN",
#'         hd_neurons = c(64, 32, 16),
#'         no_x = 10,
#'         no_y = 1,
#'         activations = 'relu'
#'     )
#'
#'     # Evaluate and instantiate
#'     model = eval(ffnn_mod)()
#'
#'     # More complex: With different activations
#'     ffnn_mod2 = ffnn_generator(
#'         nn_name = "MyFFNN2",
#'         hd_neurons = c(128, 64, 32),
#'         no_x = 20,
#'         no_y = 5,
#'         activations = act_funs(
#'             relu,
#'             selu,
#'             sigmoid
#'         )
#'     )
#'
#'     # Even more complex: Different activations and customized argument
#'     # for the specific activation function
#'     ffnn_mod2 = ffnn_generator(
#'         nn_name = "MyFFNN2",
#'         hd_neurons = c(128, 64, 32),
#'         no_x = 20,
#'         no_y = 5,
#'         activations = act_funs(
#'             relu,
#'             selu,
#'             softshrink = args(lambd = 0.5)
#'         )
#'     )
#'
#'     # Customize output activation (softmax is useful for classification tasks)
#'     ffnn_mod3 = ffnn_generator(
#'         hd_neurons = c(64, 32),
#'         no_x = 10,
#'         no_y = 3,
#'         activations = 'relu',
#'         output_activation = act_funs(softmax = args(dim = 2L))
#'     )
#' } else {
#'     message("Torch not fully installed — skipping example")
#' }
#' }
#'
#' @importFrom rlang new_function call2 expr sym
#' @importFrom purrr map map2
#' @importFrom glue glue
#' @importFrom cli cli_abort
#'
#' @export
ffnn_generator = function(nn_name = "DeepFFN",
                          hd_neurons,
                          no_x,
                          no_y,
                          activations = NULL,
                          output_activation = NULL,
                          bias = TRUE) {

    if (missing(hd_neurons) || is.null(hd_neurons) || length(hd_neurons) == 0L) {
        hd_neurons = integer(0) 
    }
    
    nodes = c(no_x, hd_neurons, no_y)
    n_layers = length(nodes) - 1L
    n_hidden = length(hd_neurons)

    if (n_layers < 1L) {
        cli_abort(c(
            "{.arg hd_neurons} must contain at least one hidden layer size.",
            "i" = "You provided {.val {hd_neurons}} (length {n_hidden})."
        ), class = "nn_module_error")
    }

    act_specs = eval_act_funs({{ activations }}, {{ output_activation }})
    activations = act_specs$activations
    output_activation = act_specs$output_activation
    activation_spec = parse_activation_spec(activations, n_hidden)
    activation_calls = process_activations(activation_spec, prefix = "nnf_")

    if (!is.null(output_activation)) {
        output_spec = parse_activation_spec(output_activation, 1L)
        output_call = process_activations(output_spec, prefix = "nnf_")[[1]]
    } else {
        output_call = NULL
    }

    all_activation_calls = c(activation_calls, list(output_call))

    # ---- Build initialize() ----
    init_body = map2(
        .x = seq_len(n_layers),
        .y = map2(nodes[-length(nodes)], nodes[-1], c),
        .f = function(i, dims) {
            layer_name = if (i == n_layers) "out" else glue("fc{i}")
            call2(
                "=",
                call2("$", expr(self), sym(layer_name)),
                call2(
                    sym("nn_linear"),
                    !!!dims,
                    bias = bias,
                    .ns = "torch"
                )
            )
        }
    )

    init = new_function(
        args = pairlist(),
        body = call2("{", !!!init_body)
    )

    # ---- Build forward() ----
    forward_body_exprs = map(seq_len(n_layers), function(i) {
        layer_name = if (i == n_layers) "out" else glue("fc{i}")
        act_call_fn = all_activation_calls[[i]]

        line1 = call2(
            "=", expr(x),
            call2(call2("$", expr(self), sym(layer_name)), expr(x))
        )

        if (is.null(act_call_fn)) {
            list(line1)
        } else {
            line2 = call2("=", expr(x), act_call_fn(expr(x)))
            list(line1, line2)
        }
    })

    forward_body_exprs = c(unlist(forward_body_exprs, recursive = FALSE), list(expr(x)))

    forward = new_function(
        args = list(x = expr()),
        body = call2("{", !!!forward_body_exprs)
    )

    call2(
        sym("nn_module"),
        nn_name, initialize = init, forward = forward,
        .ns = "torch"
    )
}

#' Check RNN Type Validity
#'
#' @param rnn_type Character. The RNN type to validate.
#' @param hd_neurons Integer vector. Hidden neurons (for error context).
#'
#' @importFrom cli cli_abort
#'
#' @noRd
check_rnn_type = function(rnn_type, hd_neurons) {
    valid_types = c("rnn", "lstm", "gru")

    rnn_type = tolower(rnn_type)

    if (!rnn_type %in% valid_types) {
        cli_abort(c(
            "{.arg rnn_type} must be one of {.val {valid_types}}.",
            x = "You provided {.val {rnn_type}}."
        ), class = "rnn_type_error")
    }

    n_rnn_layers = length(hd_neurons)
    if (n_rnn_layers == 0) {
        cli_abort(c(
            "{.arg hd_neurons} must contain at least one hidden layer size.",
            i = "RNNs require at least one recurrent layer."
        ), class = "rnn_module_error")
    }

    invisible(NULL)
}

#' @rdname nn_gens
#'
#' @section Recurrent Neural Network Module Generator:
#' The `rnn_generator()` function generates a recurrent neural network (RNN) module expression
#' from the `torch` package in R. It allows customization of the RNN architecture,
#' including the number of hidden layers, neurons, RNN type, activation functions,
#' and other parameters.
#'
#' @param nn_name Character. Name of the generated RNN module class. Default is `"DeepRNN"`.
#' @param hd_neurons Integer vector. Number of neurons in each hidden RNN layer.
#' @param no_x Integer. Number of input features.
#' @param no_y Integer. Number of output features.
#' @param rnn_type Character. Type of RNN to use. Must be one of `"rnn"`, `"lstm"`, or `"gru"`. Default is `"lstm"`.
#' @param activations Activation function specifications for each hidden layer.
#' Can be:
#' - `NULL`: No activation functions.
#' - Character vector: e.g., `c("relu", "sigmoid")`.
#' - List: e.g., `act_funs(relu, elu, softshrink = args(lambd = 0.5))`.
#' - `activation_spec` object from `act_funs()`.
#'
#' If the length of `activations` is `1L`, this will be the activation throughout the architecture.
#'
#' @param output_activation Optional. Activation function for the output layer.
#' Same format as `activations` but should be a single activation.
#' @param bias Logical. Whether to use bias weights. Default is `TRUE`
#' @param bidirectional Logical. Whether to use bidirectional RNN layers. Default is `TRUE`.
#' @param dropout Numeric. Dropout rate between RNN layers. Default is `0`.
#' @param ... Additional arguments (currently unused).
#'
#' @return A `torch` module expression representing the RNN.
#'
#' @details The generated RNN module will have the specified number of recurrent layers,
#' with each layer containing the specified number of hidden units. Activation functions
#' can be applied after each RNN layer as specified. The final output is taken from the
#' last time step and passed through a linear layer.
#'
#' The generated module properly namespaces all torch functions to avoid
#' polluting the global namespace.
#'
#' @examples
#' \donttest{
#' ## RNN
#' if (torch::torch_is_installed()) {
#'     # Basic LSTM with 2 layers
#'     rnn_mod = rnn_generator(
#'         nn_name = "MyLSTM",
#'         hd_neurons = c(64, 32),
#'         no_x = 10,
#'         no_y = 1,
#'         rnn_type = "lstm",
#'         activations = 'relu'
#'     )
#'
#'     # Evaluate and instantiate
#'     model = eval(rnn_mod)()
#'
#'     # GRU with different activations
#'     rnn_mod2 = rnn_generator(
#'         nn_name = "MyGRU",
#'         hd_neurons = c(128, 64, 32),
#'         no_x = 20,
#'         no_y = 5,
#'         rnn_type = "gru",
#'         activations = act_funs(relu, elu, relu),
#'         bidirectional = FALSE
#'     )
#'
#' } else {
#'     message("Torch not fully installed — skipping example")
#' }
#' }
#'
#' \dontrun{
#' ## Parameterized activation and dropout
#' # (Will throw an error due to `nnf_tanh()` not being available in `{torch}`)
#' # rnn_mod3 = rnn_generator(
#' #     hd_neurons = c(100, 50, 25),
#' #     no_x = 15,
#' #     no_y = 3,
#' #     rnn_type = "lstm",
#' #     activations = act_funs(
#' #         relu,
#' #         leaky_relu = args(negative_slope = 0.01),
#' #         tanh
#' #     ),
#' #     bidirectional = TRUE,
#' #     dropout = 0.3
#' # )
#' }
#'
#' @importFrom rlang new_function call2 expr sym
#' @importFrom purrr map map2
#' @importFrom glue glue
#' @importFrom cli cli_abort
#'
#' @export
rnn_generator = function(nn_name = "DeepRNN",
                         hd_neurons,
                         no_x,
                         no_y,
                         rnn_type = "lstm",
                         bias = TRUE,
                         activations = NULL,
                         output_activation = NULL,
                         bidirectional = TRUE,
                         dropout = 0,
                         ...) {
    
    if (missing(hd_neurons) || is.null(hd_neurons) || length(hd_neurons) == 0L) {
        hd_neurons = integer(0) 
    }

    check_rnn_type(rnn_type, hd_neurons)

    n_rnn_layers = length(hd_neurons)

    act_specs = eval_act_funs({{ activations }}, {{ output_activation }})
    activations = act_specs$activations
    output_activation = act_specs$output_activation
    activation_spec = parse_activation_spec(activations, n_rnn_layers)
    activation_calls = process_activations(activation_spec, prefix = "nnf_")

    if (!is.null(output_activation)) {
        output_spec = parse_activation_spec(output_activation, 1L)
        output_call = process_activations(output_spec, prefix = "nnf_")[[1]]
    } else {
        output_call = NULL
    }

    # ---- Build initialize() ----
    input_sizes = c(no_x, hd_neurons[-n_rnn_layers] * (if (bidirectional) 2L else 1L))

    rnn_layers = map2(seq_len(n_rnn_layers), input_sizes, function(i, input_size) {
        layer_name = glue("rnn{i}")
        hidden_size = hd_neurons[i]
        layer_dropout = if (i < n_rnn_layers && dropout > 0) dropout else 0

        rnn_fn_name = paste0("nn_", rnn_type)
        rnn_call = call2(
            sym(rnn_fn_name),
            input_size = input_size,
            hidden_size = hidden_size,
            num_layers = 1L,
            bias = bias,
            batch_first = TRUE,
            bidirectional = bidirectional,
            dropout = layer_dropout,
            .ns = "torch"
        )

        call2("=", call2("$", expr(self), sym(layer_name)), rnn_call)
    })

    final_hidden = hd_neurons[n_rnn_layers] * (if (bidirectional) 2L else 1L)

    output_layer = call2(
        "=", call2("$", expr(self), sym("out")),
        call2(sym("nn_linear"), final_hidden, no_y, .ns = "torch")
    )

    init_body = c(rnn_layers, list(output_layer))

    init = new_function(
        args = pairlist(),
        body = call2("{", !!!init_body)
    )

    # ---- Build forward() ----
    rnn_forward_exprs = map(seq_len(n_rnn_layers), function(i) {
        layer_name = glue("rnn{i}")
        act_call_fn = activation_calls[[i]]

        rnn_call_expr = call2(
            "=", expr(x),
            call2("[[",
                  call2(call2("$", expr(self), sym(layer_name)), expr(x)),
                  1L)
        )

        if (is.null(act_call_fn)) {
            list(rnn_call_expr)
        } else {
            activation_expr = call2("=", expr(x), act_call_fn(expr(x)))
            list(rnn_call_expr, activation_expr)
        }
    })

    slice_expr = call2(
        "=", expr(x),
        call2(
            "[", expr(x), expr(),
            call2(call2("$", expr(x), sym("size")), 2L),
            expr()
        )
    )

    output_expr = call2(
        "=", expr(x), call2(call2("$", expr(self), sym("out")), expr(x))
    )

    if (!is.null(output_call)) {
        output_activation_expr = call2("=", expr(x), output_call(expr(x)))
        forward_body_exprs = c(
            unlist(rnn_forward_exprs, recursive = FALSE),
            list(slice_expr, output_expr, output_activation_expr, expr(x))
        )
    } else {
        forward_body_exprs = c(
            unlist(rnn_forward_exprs, recursive = FALSE),
            list(slice_expr, output_expr, expr(x))
        )
    }

    forward = new_function(
        args = alist(x = ),
        body = call2("{", !!!forward_body_exprs)
    )

    call2(
        sym("nn_module"),
        nn_name,
        initialize = init,
        forward = forward,
        .ns = "torch"
    )
}

#' Parsnip Interface of `train_nn()`
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `train_nnsnip()` defines a neural network model specification that can be used
#' for classification or regression. It integrates with the tidymodels ecosystem
#' and uses [train_nn()] as the fitting backend, supporting any architecture
#' expressible via [nn_arch()] — feedforward, recurrent, convolutional, and beyond.
#'
#' @param mode A single character string for the type of model. Possible values
#'   are "unknown", "regression", or "classification".
#' @param engine A single character string specifying what computational engine
#'   to use for fitting. Currently only "kindling" is supported.
#' @param hidden_neurons An integer vector for the number of units in each hidden
#'   layer. Can be tuned.
#' @param activations A character vector of activation function names for each
#'   hidden layer (e.g., "relu", "tanh", "sigmoid"). Can be tuned.
#' @param output_activation A character string for the output activation function.
#'   Can be tuned.
#' @param bias Logical for whether to include bias terms. Can be tuned.
#' @param epochs An integer for the number of training iterations. Can be tuned.
#' @param penalty A number for the regularization penalty (lambda). Default `0`
#'   (no regularization). Higher values increase regularization strength. Can be tuned.
#' @param mixture A number between 0 and 1 for the elastic net mixing parameter.
#'   Default `0` (pure L2/Ridge regularization).
#'   - `0`: Pure L2 regularization (Ridge)
#'   - `1`: Pure L1 regularization (Lasso)
#'   - `0 < mixture < 1`: Elastic net (combination of L1 and L2)
#'   Only relevant when `penalty > 0`. Can be tuned.
#' @param batch_size An integer for the batch size during training. Can be tuned.
#' @param learn_rate A number for the learning rate. Can be tuned.
#' @param optimizer A character string for the optimizer type ("adam", "sgd",
#'   "rmsprop"). Can be tuned.
#' @param validation_split A number between 0 and 1 for the proportion of data
#'   used for validation. Can be tuned.
#'
#' @param optimizer_args A named list of additional arguments passed to the
#'   optimizer. Cannot be tuned — pass via `set_engine()`.
#' @param loss A character string or a valid `{torch}` function for the loss function ("mse", "mae",
#'   "cross_entropy", "bce"). Cannot be tuned — pass via `set_engine()`.
#' @param architecture An [nn_arch()] object for a custom architecture. Cannot
#'   be tuned — pass via `set_engine()`.
#' @param flatten_input Logical or `NULL`. Controls input flattening. Cannot be
#'   tuned — pass via `set_engine()`.
#' @param early_stopping An [early_stop()] object or `NULL`. Cannot be tuned —
#'   pass via `set_engine()`.
#' @param device A character string for the device to use ("cpu", "cuda", "mps").
#'   If `NULL`, auto-detects the best available device. Cannot be tuned — pass
#'   via `set_engine()`.
#' @param verbose Logical for whether to print training progress. Default `FALSE`.
#'   Cannot be tuned — pass via `set_engine()`.
#' @param cache_weights Logical. If `TRUE`, stores trained weight matrices in
#'   the returned object. Cannot be tuned — pass via `set_engine()`.
#'
#' @details
#' This function creates a model specification for a neural network that can be
#' used within tidymodels workflows. The underlying engine is [train_nn()], which
#' is architecture-agnostic: when `architecture = NULL` it falls back to a
#' standard feed-forward network, but any architecture expressible via [nn_arch()]
#' can be used instead. The model supports:
#'
#' - Configurable hidden layers and activation functions (default MLP path)
#' - Custom architectures via [nn_arch()] (recurrent, convolutional, etc.)
#' - GPU acceleration (CUDA, MPS, or CPU)
#' - Hyperparameter tuning integration
#' - Both regression and classification tasks
#'
#' When using the default MLP path (no custom `architecture`), `hidden_neurons`
#' accepts an integer vector where each element represents the number of neurons
#' in that hidden layer. For example, `hidden_neurons = c(128, 64, 32)` creates
#' a network with three hidden layers. Pass an [nn_arch()] object via
#' `set_engine()` to use a custom architecture instead.
#'
#' The `device` parameter controls where computation occurs:
#' - `NULL` (default): Auto-detect best available device (CUDA > MPS > CPU)
#' - `"cuda"`: Use NVIDIA GPU
#' - `"mps"`: Use Apple Silicon GPU
#' - `"cpu"`: Use CPU only
#'
#' When tuning, you can use special tune tokens:
#' - For `hidden_neurons`: use `tune("hidden_neurons")` with a custom range
#' - For `activation`: use `tune("activation")` with values like "relu", "tanh"
#'
#' @return A model specification object with class `train_nnsnip`.
#'
#' @examples
#' \donttest{
#' if (torch::torch_is_installed()) {
#'     box::use(
#'         recipes[recipe],
#'         workflows[workflow, add_recipe, add_model],
#'         tune[tune],
#'         parsnip[fit]
#'     )
#'
#'     # Model spec
#'     nn_spec = train_nnsnip(
#'         mode = "classification",
#'         hidden_neurons = c(30, 5),
#'         activations = c("relu", "elu"),
#'         epochs = 100
#'     )
#'
#'     wf = workflow() |>
#'         add_recipe(recipe(Species ~ ., data = iris)) |>
#'         add_model(nn_spec)
#'
#'     fit_wf = fit(wf, data = iris)
#' } else {
#'     message("Torch not fully installed — skipping example")
#' }
#' }
#'
#' @export
train_nnsnip = 
    function(
        mode = "unknown",
        engine = "kindling",
        # ---- Tunable parameters ----
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = NULL,
        epochs = NULL,
        batch_size = NULL,
        penalty = NULL,
        mixture = NULL,
        learn_rate = NULL,
        optimizer = NULL,
        validation_split = NULL,
        # ---- Non-tunable parameters (engine-level, set via set_engine()) ----
        optimizer_args = NULL,
        loss = NULL,
        architecture = NULL,
        flatten_input = NULL,
        early_stopping = NULL,
        device = NULL,
        verbose = NULL,
        cache_weights = NULL
    )
{
    
    if (!requireNamespace("parsnip", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg parsnip} is required but not installed.")
    }
    
    # ---- "Tunable" args ----
    args = list(
        hidden_neurons = rlang::enquo(hidden_neurons),
        activations = rlang::enquo(activations),
        output_activation = rlang::enquo(output_activation),
        bias = rlang::enquo(bias),
        epochs = rlang::enquo(epochs),
        batch_size = rlang::enquo(batch_size),
        penalty = rlang::enquo(penalty),
        mixture = rlang::enquo(mixture),
        learn_rate = rlang::enquo(learn_rate),
        optimizer = rlang::enquo(optimizer),
        validation_split = rlang::enquo(validation_split)
    )
    
    # ---- "Non-tunable" engine args ----
    eng_args = list(
        optimizer_args = rlang::enquo(optimizer_args),
        loss = rlang::enquo(loss),
        architecture = rlang::enquo(architecture),
        flatten_input = rlang::enquo(flatten_input),
        early_stopping = rlang::enquo(early_stopping),
        device = rlang::enquo(device),
        verbose = rlang::enquo(verbose),
        cache_weights = rlang::enquo(cache_weights)
    )
    
    parsnip::new_model_spec(
        "train_nnsnip",
        args = args,
        eng_args = eng_args,
        mode = mode,
        user_specified_mode = !missing(mode),
        method = NULL,
        engine = engine,
        user_specified_engine = !missing(engine)
    )
}


#' @export
print.train_nnsnip = function(x, ...) {
    cat("Kindling Neural Network Model Specification (", x$mode, ")\n\n", sep = "")
    parsnip::model_printer(x, ...)
    invisible(x)
}


#' @export
#' @importFrom stats update
update.train_nnsnip = 
    function(
        object,
        parameters = NULL,
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = NULL,
        epochs = NULL,
        batch_size = NULL,
        penalty = NULL,
        mixture = NULL,
        learn_rate = NULL,
        optimizer = NULL,
        optimizer_args = NULL,
        loss = NULL,
        validation_split = NULL,
        device = NULL,
        verbose = NULL,
        fresh = FALSE,
        ...
    )
{
    
    args = list(
        hidden_neurons = rlang::enquo(hidden_neurons),
        activations = rlang::enquo(activations),
        output_activation = rlang::enquo(output_activation),
        bias = rlang::enquo(bias),
        epochs = rlang::enquo(epochs),
        batch_size = rlang::enquo(batch_size),
        penalty = rlang::enquo(penalty),
        mixture = rlang::enquo(mixture),
        learn_rate = rlang::enquo(learn_rate),
        optimizer = rlang::enquo(optimizer),
        optimizer_args = rlang::enquo(optimizer_args),
        loss = rlang::enquo(loss),
        validation_split = rlang::enquo(validation_split),
        device = rlang::enquo(device),
        verbose = rlang::enquo(verbose)
    )
    
    parsnip::update_spec(
        object = object,
        parameters = parameters,
        args_enquo_list = args,
        fresh = fresh,
        cls = "train_nnsnip",
        ...
    )
}


#' @export
#' @importFrom parsnip translate
translate.train_nnsnip = function(x, engine = x$engine, ...) {
    if (is.null(engine)) {
        cli::cli_abort("Please set an engine with `set_engine()`.")
    }
    
    x = parsnip::translate.default(x, engine, ...)
    x
}


#' @export
#' @importFrom tune tunable
tunable.train_nnsnip = function(x, ...) {
    tibble::tibble(
        name = c(
            "hidden_neurons", "activations", "output_activation", "bias",
            "epochs", "batch_size", "penalty", "mixture",
            "learn_rate", "optimizer", "validation_split"
        ),
        call_info = list(
            list(pkg = "kindling", fun = "hidden_neurons"),
            list(pkg = "kindling", fun = "activations"),
            list(pkg = "kindling", fun = "output_activation"),
            list(pkg = "kindling", fun = "bias"),
            list(pkg = "dials", fun = "epochs"),
            list(pkg = "dials", fun = "batch_size"),
            list(pkg = "dials", fun = "penalty"),
            list(pkg = "dials", fun = "mixture"),
            list(pkg = "dials", fun = "learn_rate"),
            list(pkg = "kindling", fun = "optimizer"),
            list(pkg = "kindling", fun = "validation_split")
        ),
        source = "model_spec",
        component = "train_nnsnip",
        component_id = "main"
    )
}

#' kindling-tidymodels wrapper
#'
#' @param formula A formula specifying the model (e.g., `y ~ x1 + x2`)
#' @param data A data frame containing the training data
#' @param ... Additional arguments passed to the underlying training function
#'
#' @return
#' `train_nn_wrapper()` returns an `"nn_fit_tab"` object. See [train_nn()] for details.
#'
#' @details
#' This wrapper function is designed to interface with the `{tidymodels}`
#' ecosystem, particularly for use with [tune::tune_grid()] and workflows.
#' It handles the conversion of tuning parameters (especially list-column
#' parameters from [grid_depth()]) into the format expected by [train_nn()].
#'
#' @rdname kindling-nn-wrappers
#' @section MLP Wrapper for `{tidymodels}` interface:
#' Internal wrapper — use `mlp_kindling()` + `fit()` instead.
#'
#' @keywords internal
#' @export
train_nn_wrapper = function(formula, data, ...) {
    dots = list(...)
    dots = prepare_kindling_args(dots)
    
    do.call(
        train_nn,
        c(list(x = formula, data = data), dots)
    )
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

# ---- Classification: spec and fit -------------------------------------------

test_that("train_nnsnip() returns correct class and mode - classification", {
    skip_if_not_installed("parsnip")
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5
    )
    
    expect_s3_class(spec, "train_nnsnip")
    expect_equal(spec$mode, "classification")
    expect_equal(spec$engine, "kindling")
})

test_that("train_nnsnip() fits classification workflow without error", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5,
        verbose = FALSE
    )
    
    expect_no_error({
        fitted = parsnip::fit(
            spec,
            Species ~ .,
            data = iris
        )
    })
    
    expect_s3_class(fitted, "model_fit")
    expect_s3_class(fitted$fit, "nn_fit_tab")
})

test_that("train_nnsnip() classification fit sets is_classification = TRUE", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5,
        verbose = FALSE
    )
    
    fitted = parsnip::fit(spec, Species ~ ., data = iris)
    
    expect_true(fitted$fit$is_classification)
    expect_equal(fitted$fit$y_levels, levels(iris$Species))
})

# ---- Regression: spec and fit -----------------------------------------------

test_that("train_nnsnip() returns correct class and mode - regression", {
    skip_if_not_installed("parsnip")
    
    spec = train_nnsnip(
        mode = "regression",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5
    )
    
    expect_s3_class(spec, "train_nnsnip")
    expect_equal(spec$mode, "regression")
})

test_that("train_nnsnip() fits regression workflow without error", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "regression",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5,
        verbose = FALSE
    )
    
    expect_no_error({
        fitted = parsnip::fit(
            spec,
            Sepal.Length ~ .,
            data = iris[, 1:4]
        )
    })
    
    expect_s3_class(fitted, "model_fit")
    expect_s3_class(fitted$fit, "nn_fit_tab")
    expect_false(fitted$fit$is_classification)
})

# ---- Predictions: classification --------------------------------------------

test_that("train_nnsnip() predict type = 'class' returns correct format", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(10),
        epochs = 5,
        verbose = FALSE
    )
    
    fitted = parsnip::fit(spec, Species ~ ., data = iris[1:100, ])
    preds = predict(fitted, new_data = iris[101:110, ])
    
    expect_s3_class(preds, "tbl_df")
    expect_true(".pred_class" %in% names(preds))
    expect_equal(nrow(preds), 10)
    expect_s3_class(preds$.pred_class, "factor")
    expect_equal(levels(preds$.pred_class), levels(iris$Species))
})

test_that("train_nnsnip() predict type = 'prob' returns correct format", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(10),
        epochs = 5,
        verbose = FALSE
    )
    
    fitted = parsnip::fit(spec, Species ~ ., data = iris[1:100, ])
    preds = predict(fitted, new_data = iris[101:110, ], type = "prob")
    
    expect_s3_class(preds, "tbl_df")
    expect_equal(nrow(preds), 10)
    expect_true(all(grepl("^\\.pred_", names(preds))))
    expect_true(all(abs(rowSums(preds) - 1) < 1e-6))
})

# ---- Predictions: regression ------------------------------------------------

test_that("train_nnsnip() predict returns correct format - regression", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "regression",
        hidden_neurons = c(10),
        epochs = 5,
        verbose = FALSE
    )
    
    fitted = parsnip::fit(
        spec,
        Sepal.Length ~ Petal.Length + Petal.Width,
        data = iris[1:50, ]
    )
    
    preds = predict(fitted, new_data = iris[51:60, ])
    
    expect_s3_class(preds, "tbl_df")
    expect_true(".pred" %in% names(preds))
    expect_equal(nrow(preds), 10)
    expect_type(preds$.pred, "double")
})

# ---- Engine args passthrough ------------------------------------------------

test_that("validation_split is passed through and produces val_loss_history", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5,
        validation_split = 0.2,
        verbose = FALSE
    )
    
    fitted = parsnip::fit(spec, Species ~ ., data = iris)
    
    expect_false(is.null(fitted$fit$val_loss_history))
    expect_length(fitted$fit$val_loss_history, fitted$fit$n_epochs)
})

test_that("penalty and mixture are passed through correctly", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5,
        penalty = 0.01,
        mixture = 0.5,
        verbose = FALSE
    )
    
    fitted = parsnip::fit(spec, Species ~ ., data = iris)
    
    expect_equal(fitted$fit$penalty, 0.01)
    expect_equal(fitted$fit$mixture, 0.5)
})

test_that("optimizer arg is passed through without error", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5,
        optimizer = "sgd",
        verbose = FALSE
    )
    
    expect_no_error(parsnip::fit(spec, Species ~ ., data = iris))
})

# ---- Early stopping ---------------------------------------------------------

test_that("early_stopping halts training before max epochs", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 200,
        validation_split = 0.2,
        early_stopping = early_stop(patience = 3, min_delta = 1e10),
        verbose = FALSE
    )
    
    fitted = parsnip::fit(spec, Species ~ ., data = iris)
    
    expect_lt(fitted$fit$n_epochs, 200)
    expect_false(is.na(fitted$fit$stopped_epoch))
})

test_that("without early_stopping all epochs are run", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5,
        verbose = FALSE
    )
    
    fitted = parsnip::fit(spec, Species ~ ., data = iris)
    
    expect_equal(fitted$fit$n_epochs, 5)
    expect_true(is.na(fitted$fit$stopped_epoch))
})

# ---- Custom architecture ----------------------------------------------------

test_that("nn_arch object is passed through and stored in fit", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()
    
    arch = nn_arch(nn_name = "CustomMLP")
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5,
        architecture = arch,
        verbose = FALSE
    )
    
    fitted = parsnip::fit(spec, Species ~ ., data = iris)
    
    expect_false(is.null(fitted$fit$arch))
    expect_equal(fitted$fit$arch$nn_name, "CustomMLP")
})

# ---- update() ---------------------------------------------------------------

test_that("update.train_nnsnip() updates args and preserves others", {
    skip_if_not_installed("parsnip")
    
    spec = train_nnsnip(
        mode = "classification",
        hidden_neurons = c(16, 8),
        activations = "relu",
        epochs = 5
    )
    
    spec2 = update(spec, epochs = 20, learn_rate = 0.01)
    
    expect_equal(rlang::quo_get_expr(spec2$args$epochs), 20)
    expect_equal(rlang::quo_get_expr(spec2$args$learn_rate), 0.01)
    expect_equal(rlang::quo_get_expr(spec2$args$hidden_neurons), quote(c(16, 8)))
})

# ---- tunable() --------------------------------------------------------------

test_that("tunable.train_nnsnip() returns expected parameter names", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("tune")
    
    spec = train_nnsnip(mode = "classification")
    tbl = tune::tunable(spec)
    
    expect_s3_class(tbl, "tbl_df")
    expect_setequal(tbl$name, c(
        "hidden_neurons", "activations", "output_activation", "bias",
        "epochs", "batch_size", "penalty", "mixture",
        "learn_rate", "optimizer", "validation_split"
    ))
})

# ---- prepare_kindling_args() ------------------------------------------------

test_that("prepare_kindling_args() evaluates one-sided formulas and drops NULLs", {
    args = list(epochs = ~50L, loss = ~"cross_entropy", device = ~NULL)
    out = prepare_kindling_args(args)
    
    expect_equal(out$epochs, 50L)
    expect_equal(out$loss, "cross_entropy")
    expect_false("device" %in% names(out))
})

test_that("prepare_kindling_args() evaluates quosures", {
    args = list(
        epochs = rlang::quo(100L),
        optimizer = rlang::quo("adam")
    )
    out = prepare_kindling_args(args)
    
    expect_equal(out$epochs, 100L)
    expect_equal(out$optimizer, "adam")
})

test_that("prepare_kindling_args() unwraps single-element hidden_neurons list", {
    args = list(hidden_neurons = list(c(32L, 16L)))
    out = prepare_kindling_args(args)
    
    expect_equal(out$hidden_neurons, c(32L, 16L))
})

test_that("prepare_kindling_args() passes plain values through unchanged", {
    args = list(epochs = 50L, learn_rate = 0.001)
    out = prepare_kindling_args(args)
    
    expect_equal(out$epochs, 50L)
    expect_equal(out$learn_rate, 0.001)
})

#' Multi-Layer Perceptron (Feedforward Neural Network) via kindling
#'
#' `mlp_kindling()` defines a feedforward neural network model that can be used
#' for classification or regression. It integrates with the tidymodels ecosystem
#' and uses the torch backend via kindling.
#'
#' @param mode A single character string for the type of model. Possible values
#'   are "unknown", "regression", or "classification".
#' @param engine A single character string specifying what computational engine
#'   to use for fitting. Currently only "kindling" is supported.
#' @param hidden_neurons An integer vector for the number of units in each hidden
#'   layer. Can be tuned.
#' @param activations A character vector of activation function names for each
#'   hidden layer (e.g., "relu", "tanh", "sigmoid"). Can be tuned.
#' @param output_activation A character string for the output activation function.
#'   Can be tuned.
#' @param bias Logical for whether to include bias terms. Can be tuned.
#' @param epochs An integer for the number of training iterations. Can be tuned.
#' @param penalty A number for the regularization penalty (lambda). Default `0`
#'   (no regularization). Higher values increase regularization strength. Can be tuned.
#' @param mixture A number between 0 and 1 for the elastic net mixing parameter.
#'   Default `0` (pure L2/Ridge regularization).
#'   - `0`: Pure L2 regularization (Ridge)
#'   - `1`: Pure L1 regularization (Lasso)
#'   - `0 < mixture < 1`: Elastic net (combination of L1 and L2)
#'   Only relevant when `penalty > 0`. Can be tuned.
#' @param batch_size An integer for the batch size during training. Can be tuned.
#' @param learn_rate A number for the learning rate. Can be tuned.
#' @param optimizer A character string for the optimizer type ("adam", "sgd",
#'   "rmsprop"). Can be tuned.
#' @param validation_split A number between 0 and 1 for the proportion of data
#'   used for validation. Can be tuned.
#'
#' @param optimizer_args A named list of additional arguments passed to the
#'   optimizer. Cannot be tuned — pass via `set_engine()`.
#' @param loss A character string for the loss function ("mse", "mae",
#'   "cross_entropy", "bce"). Cannot be tuned — pass via `set_engine()`.
#' @param architecture An [nn_arch()] object for a custom architecture. Cannot
#'   be tuned — pass via `set_engine()`.
#' @param flatten_input Logical or `NULL`. Controls input flattening. Cannot be
#'   tuned — pass via `set_engine()`.
#' @param early_stopping An [early_stop()] object or `NULL`. Cannot be tuned —
#'   pass via `set_engine()`.
#' @param device A character string for the device ("cpu", "cuda", "mps").
#'   Cannot be tuned — pass via `set_engine()`.
#' @param verbose Logical for whether to print training progress. Cannot be
#'   tuned — pass via `set_engine()`.
#' @param cache_weights Logical. If `TRUE`, stores trained weight matrices in
#'   the returned object. Cannot be tuned — pass via `set_engine()`.
#'
#' @details
#' This function creates a model specification for a feedforward neural network
#' that can be used within tidymodels workflows. The model supports:
#'
#' - Multiple hidden layers with configurable units
#' - Various activation functions per layer
#' - GPU acceleration (CUDA, MPS, or CPU)
#' - Hyperparameter tuning integration
#' - Both regression and classification tasks
#'
#' Parameters that cannot be tuned (`architecture`, `flatten_input`,
#' `early_stopping`, `device`, `verbose`, `cache_weights`, `optimizer_args`,
#' `loss`) must be set via `set_engine()`, not as arguments to `mlp_kindling()`.
#'
#' @return A model specification object with class `mlp_kindling`.
#'
#' @examples
#' \donttest{
#' if (torch::torch_is_installed()) {
#'     box::use(
#'         recipes[recipe],
#'         workflows[workflow, add_recipe, add_model],
#'         tune[tune],
#'         parsnip[fit]
#'     )
#'
#'     # library(recipes)
#'     # library(workflows)
#'     # library(parsnip)
#'     # library(tune)
#'
#'     # Model specs
#'     mlp_spec = mlp_kindling(
#'         mode = "classification",
#'         hidden_neurons = c(128, 64, 32),
#'         activation = c("relu", "relu", "relu"),
#'         epochs = 100
#'     )
#'
#'     # If you want to tune
#'     mlp_tune_spec = mlp_kindling(
#'         mode = "classification",
#'         hidden_neurons = tune(),
#'         activation = tune(),
#'         epochs = tune(),
#'         learn_rate = tune()
#'     )
#'      wf = workflow() |>
#'         add_recipe(recipe(Species ~ ., data = iris)) |>
#'         add_model(mlp_spec)
#'
#'      fit_wf = fit(wf, data = iris)
#' } else {
#'     message("Torch not fully installed — skipping example")
#' }
#' }
#'
#' @export
mlp_kindling = 
    function(
        mode = "unknown",
        engine = "kindling",
        # ---- Tunable parameters ----
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = NULL,
        epochs = NULL,
        batch_size = NULL,
        penalty = NULL,
        mixture = NULL,
        learn_rate = NULL,
        optimizer = NULL,
        validation_split = NULL,
        # ---- Non-tunable parameters (engine-level, set via set_engine()) ----
        optimizer_args = NULL,
        loss = NULL,
        architecture = NULL,
        flatten_input = NULL,
        early_stopping = NULL,
        device = NULL,
        verbose = NULL,
        cache_weights = NULL
    )
{
    
    if (!requireNamespace("parsnip", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg parsnip} is required but not installed.")
    }
    
    # ---- "Tunable" args ----
    args = list(
        hidden_neurons = rlang::enquo(hidden_neurons),
        activations = rlang::enquo(activations),
        output_activation = rlang::enquo(output_activation),
        bias = rlang::enquo(bias),
        epochs = rlang::enquo(epochs),
        batch_size = rlang::enquo(batch_size),
        penalty = rlang::enquo(penalty),
        mixture = rlang::enquo(mixture),
        learn_rate = rlang::enquo(learn_rate),
        optimizer = rlang::enquo(optimizer),
        validation_split = rlang::enquo(validation_split)
    )
    
    # ---- "Non-tunable" engine args ----
    eng_args = list(
        optimizer_args = rlang::enquo(optimizer_args),
        loss = rlang::enquo(loss),
        architecture = rlang::enquo(architecture),
        flatten_input = rlang::enquo(flatten_input),
        early_stopping = rlang::enquo(early_stopping),
        device = rlang::enquo(device),
        verbose = rlang::enquo(verbose),
        cache_weights = rlang::enquo(cache_weights)
    )
    
    parsnip::new_model_spec(
        "mlp_kindling",
        args = args,
        eng_args = eng_args,
        mode = mode,
        user_specified_mode = !missing(mode),
        method = NULL,
        engine = engine,
        user_specified_engine = !missing(engine)
    )
}


#' @export
print.mlp_kindling = function(x, ...) {
    cat("Kindling Multi-Layer Perceptron Model Specification (", x$mode, ")\n\n", sep = "")
    parsnip::model_printer(x, ...)
    invisible(x)
}


#' @export
#' @importFrom stats update
update.mlp_kindling = 
    function(
        object,
        parameters = NULL,
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = NULL,
        epochs = NULL,
        batch_size = NULL,
        penalty = NULL,
        mixture = NULL,
        learn_rate = NULL,
        optimizer = NULL,
        validation_split = NULL,
        optimizer_args = NULL,
        loss = NULL,
        architecture = NULL,
        flatten_input = NULL,
        early_stopping = NULL,
        device = NULL,
        verbose = NULL,
        cache_weights = NULL,
        fresh = FALSE,
        ...
    )
{
    
    args = list(
        hidden_neurons = rlang::enquo(hidden_neurons),
        activations = rlang::enquo(activations),
        output_activation = rlang::enquo(output_activation),
        bias = rlang::enquo(bias),
        epochs = rlang::enquo(epochs),
        batch_size = rlang::enquo(batch_size),
        penalty = rlang::enquo(penalty),
        mixture = rlang::enquo(mixture),
        learn_rate = rlang::enquo(learn_rate),
        optimizer = rlang::enquo(optimizer),
        validation_split = rlang::enquo(validation_split)
    )
    
    parsnip::update_spec(
        object = object,
        parameters = parameters,
        args_enquo_list = args,
        fresh = fresh,
        cls = "mlp_kindling",
        ...
    )
}


#' @export
#' @importFrom parsnip translate
translate.mlp_kindling = function(x, engine = x$engine, ...) {
    if (is.null(engine)) {
        cli::cli_abort("Please set an engine with `set_engine()`.")
    }
    x = parsnip::translate.default(x, engine, ...)
    x
}


#' @export
#' @importFrom tune tunable
tunable.mlp_kindling = function(x, ...) {
    tibble::tibble(
        name = c(
            "hidden_neurons", "activations", "output_activation", "bias",
            "epochs", "batch_size", "penalty", "mixture",
            "learn_rate", "optimizer", "validation_split"
        ),
        call_info = list(
            list(pkg = "kindling", fun = "hidden_neurons"),
            list(pkg = "kindling", fun = "activations"),
            list(pkg = "kindling", fun = "output_activation"),
            list(pkg = "kindling", fun = "bias"),
            list(pkg = "dials", fun = "epochs"),
            list(pkg = "dials", fun = "batch_size"),
            list(pkg = "dials", fun = "penalty"),
            list(pkg = "dials", fun = "mixture"),
            list(pkg = "dials", fun = "learn_rate"),
            list(pkg = "kindling", fun = "optimizer"),
            list(pkg = "kindling", fun = "validation_split")
        ),
        source = "model_spec",
        component = "mlp_kindling",
        component_id = "main"
    )
}

#' Basemodels-tidymodels wrappers
#'
#' @param formula A formula specifying the model (e.g., `y ~ x1 + x2`)
#' @param data A data frame containing the training data
#' @param ... Additional arguments passed to the underlying training function
#'
#' @return
#' * `ffnn_wrapper()` returns an object of class `"ffnn_fit"` containing the trained feedforward neural network model and metadata. See [ffnn()] for details.
#' * `rnn_wrapper()` returns an object of class `"rnn_fit"` containing the trained recurrent neural network model and metadata. See [rnn()] for details.
#'
#' @details
#' These wrapper functions are designed to interface with the `{tidymodels}` 
#' ecosystem, particularly for use with [tune::tune_grid()] and workflows.
#' They handle the conversion of tuning parameters (especially list-column 
#' parameters from [grid_depth()]) into the format expected by [ffnn()] and [rnn()].
#' 
#' @rdname kindling-nn-wrappers
#' @section FFNN (MLP) Wrapper for `{tidymodels}` interface:
#' This is a function to interface into `{tidymodels}`
#' (do not use this, use `kindling::ffnn()` instead).
#' 
#' @keywords internal
#' @export
ffnn_wrapper = function(formula, data, ...) {
    dots = list(...)
    dots = prepare_kindling_args(dots)
    
    do.call(
        ffnn,
        c(list(formula = formula, data = data), dots)
    )
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("mlp_kindling works with parsnip fit - regression", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "regression",
        hidden_neurons = c(10, 5),
        activations = "relu",
        epochs = 5,
        verbose = FALSE
    )

    expect_error({
        fitted = parsnip::fit(
            spec,
            Sepal.Length ~ Petal.Length + Petal.Width,
            data = iris[1:50, ]
        )
    }, NA)

    expect_s3_class(fitted, "model_fit")
    expect_s3_class(fitted$fit, "ffnn_fit")
})

test_that("mlp_kindling works with parsnip fit - classification", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = c(10, 5),
        activations = "relu",
        epochs = 5,
        verbose = FALSE
    )

    expect_error({
        fitted = parsnip::fit(
            spec,
            Species ~ .,
            data = iris[1:50, ]
        )
    }, NA)

    expect_s3_class(fitted, "model_fit")
    expect_s3_class(fitted$fit, "ffnn_fit")
})

test_that("rnn_kindling works with parsnip fit - classification", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = rnn_kindling(
        mode = "classification",
        hidden_neurons = c(10),
        rnn_type = "lstm",
        activations = "relu",
        epochs = 5,
        verbose = FALSE
    )

    expect_error({
        fitted = parsnip::fit(
            spec,
            Species ~ .,
            data = iris[1:50, ]
        )
    }, NA)

    expect_s3_class(fitted, "model_fit")
})

test_that("mlp predictions return correct format - regression", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "regression",
        hidden_neurons = c(10),
        epochs = 5,
        verbose = FALSE
    )

    fitted = parsnip::fit(
        spec,
        Sepal.Length ~ Petal.Length + Petal.Width,
        data = iris[1:50, ]
    )

    preds = predict(fitted, new_data = iris[51:60, ])

    expect_s3_class(preds, "tbl_df")
    expect_true(".pred" %in% names(preds))
    expect_equal(nrow(preds), 10)
    expect_type(preds$.pred, "double")
})

test_that("mlp predictions return correct format - classification class", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = c(10),
        epochs = 5,
        verbose = FALSE
    )

    fitted = parsnip::fit(
        spec,
        Species ~ .,
        data = iris[1:100, ]
    )

    preds = predict(fitted, new_data = iris[101:110, ])

    expect_s3_class(preds, "tbl_df")
    expect_true(".pred_class" %in% names(preds))
    expect_equal(nrow(preds), 10)
    expect_s3_class(preds$.pred_class, "factor")
})

test_that("'tanh' should still throws an error from `activations` in `mlp_kindling()`", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    expect_error({
        mlp_kindling(
            mode = "classification",
            hidden_neurons = 15,
            activations = act_funs(tanh),
            epochs = 5,
            verbose = FALSE
        ) |>
        parsnip::fit(Species ~ ., data = iris)
    })
})

test_that("mlp predictions return correct format - classification prob", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = c(10),
        epochs = 5,
        verbose = FALSE
    )

    fitted = parsnip::fit(
        spec,
        Species ~ .,
        data = iris[1:100, ]
    )

    preds = predict(fitted, new_data = iris[101:110, ], type = "prob")

    expect_s3_class(preds, "tbl_df")
    expect_equal(nrow(preds), 10)
    expect_true(all(grepl("^\\.pred_", names(preds))))
    expect_true(all(abs(rowSums(preds) - 1) < 1e-6))
})

test_that("mlp_kindling creates valid model specification", {
    skip_if_not_installed("parsnip")

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 50
    )

    expect_s3_class(spec, "mlp_kindling")
    expect_s3_class(spec, "model_spec")
    expect_equal(spec$mode, "classification")
    expect_equal(spec$engine, "kindling")
})

test_that("mlp_kindling defaults to unknown mode", {
    skip_if_not_installed("parsnip")

    spec = mlp_kindling()

    expect_equal(spec$mode, "unknown")
    expect_false(spec$user_specified_mode)
})

test_that("mlp_kindling accepts all valid parameters", {
    skip_if_not_installed("parsnip")

    spec = mlp_kindling(
        mode = "regression",
        hidden_neurons = c(128, 64, 32),
        activations = c("relu", "elu", "selu"),
        output_activation = "sigmoid",
        bias = TRUE,
        epochs = 100,
        batch_size = 32,
        learn_rate = 0.001,
        optimizer = "adam",
        loss = "mse",
        validation_split = 0.2,
        device = "cpu",
        verbose = FALSE
    )

    expect_s3_class(spec, "mlp_kindling")
    expect_equal(spec$mode, "regression")
})

test_that("mlp_kindling print method works", {
    skip_if_not_installed("parsnip")

    spec = mlp_kindling(mode = "classification")

    expect_output(print(spec), "Kindling Multi-Layer Perceptron")
    expect_output(print(spec), "classification")
})

test_that("mlp_kindling update method works", {
    skip_if_not_installed("parsnip")

    spec = mlp_kindling(
        mode = "classification",
        epochs = 50
    )

    updated_spec = update(spec, epochs = 100, learn_rate = 0.01)

    expect_s3_class(updated_spec, "mlp_kindling")
})

test_that("mlp_kindling update method works", {
    skip_if_not_installed("parsnip")

    spec = mlp_kindling(
        mode = "classification",
        epochs = 50
    )

    updated_spec = update(spec, epochs = 100, learn_rate = 0.01)

    expect_s3_class(updated_spec, "mlp_kindling")
})

test_that("mlp_kindling supports tune placeholders", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("tune")

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = tune::tune(),
        epochs = tune::tune(),
        learn_rate = tune::tune()
    )

    expect_s3_class(spec, "mlp_kindling")
})

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("mlp_kindling with multiple activation functions works", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = c(20, 10),
        activations = c("relu", "elu"),
        epochs = 5,
        verbose = FALSE
    )

    fitted = parsnip::fit(
        spec,
        Species ~ .,
        data = iris[1:100, ]
    )

    expect_s3_class(fitted, "model_fit")

    preds = predict(fitted, new_data = iris[101:110, ])
    expect_equal(nrow(preds), 10)
})

test_that("mlp_kindling handles single hidden layer and accepts using `list()`", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = list(20),
        epochs = 5,
        verbose = FALSE
    )

    fitted = parsnip::fit(spec, Species ~ ., data = iris[1:100, ])
    preds = predict(fitted, new_data = iris[101:110, ])

    expect_equal(nrow(preds), 10)
})

test_that("mlp_kindling handles deep networks", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "regression",
        hidden_neurons = c(64, 32, 16, 8),
        activations = "relu",
        epochs = 5,
        verbose = FALSE
    )

    expect_error({
        fitted = parsnip::fit(
            spec,
            Sepal.Length ~ .,
            data = iris[1:100, ]
        )
    }, NA)
})

test_that("mlp_kindling handles deep neural networks and accepts both using `list()` and a stringed argument for the activation function", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = list(5, 10, 7),
        activations = list('relu', 'softshrink(lambd = 0.5)', 'celu(alpha = 0.8)'),
        epochs = 5,
        verbose = FALSE
    )

    fitted = parsnip::fit(spec, Species ~ ., data = iris[1:100, ])
    preds = predict(fitted, new_data = iris[101:110, ])

    expect_no_warning(fitted)
    expect_no_error(fitted)
    expect_no_warning(preds)
    expect_no_error(preds)
    expect_equal(nrow(preds), 10)
})

test_that("predictions work with single observation", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = 10,
        epochs = 5,
        verbose = FALSE
    )

    fitted = parsnip::fit(spec, Species ~ ., data = iris[1:100, ])
    preds = predict(fitted, new_data = iris[101, ])

    expect_equal(nrow(preds), 1)
    expect_s3_class(preds$.pred_class, "factor")
})

test_that("augment method works correctly", {
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = list(10),
        epochs = 5,
        verbose = FALSE
    )

    fitted = parsnip::fit(spec, Species ~ ., data = iris[1:100, ])
    augmented = parsnip::augment(fitted, new_data = iris[101:110, ])

    expect_s3_class(augmented, "tbl_df")
    expect_equal(nrow(augmented), 10)
    expect_true(".pred_class" %in% names(augmented))
    expect_true("Species" %in% names(augmented))
    expect_true("Sepal.Length" %in% names(augmented))
})

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("mlp_kindling works within workflows", {
    skip_if_not_installed("workflows")
    skip_if_not_installed("recipes")
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = c(10),
        epochs = 5,
        verbose = FALSE
    )

    rec = recipes::recipe(Species ~ ., data = iris)

    wf = workflows::workflow() |>
        workflows::add_recipe(rec) |>
        workflows::add_model(spec)

    expect_s3_class(wf, "workflow")

    fitted_wf = workflows::fit(wf, data = iris[1:100, ])

    expect_s3_class(fitted_wf, "workflow")

    preds = predict(fitted_wf, new_data = iris[101:110, ])

    expect_s3_class(preds, "tbl_df")
    expect_equal(nrow(preds), 10)
})

test_that("rnn_kindling works within workflows", {
    skip_if_not_installed("workflows")
    skip_if_not_installed("recipes")
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = rnn_kindling(
        mode = "classification",
        hidden_neurons = c(10),
        rnn_type = "gru",
        epochs = 5,
        verbose = FALSE
    )

    rec = recipes::recipe(Species ~ ., data = iris)

    wf = workflows::workflow() |>
        workflows::add_recipe(rec) |>
        workflows::add_model(spec)

    expect_s3_class(wf, "workflow")

    fitted_wf = workflows::fit(wf, data = iris[1:100, ])

    expect_s3_class(fitted_wf, "workflow")
})

test_that("workflow with preprocessing recipe works", {
    skip_if_not_installed("workflows")
    skip_if_not_installed("recipes")
    skip_if_not_installed("parsnip")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "regression",
        hidden_neurons = c(10),
        epochs = 5,
        verbose = FALSE
    )

    rec = recipes::recipe(Sepal.Length ~ ., data = iris) |>
        recipes::step_normalize(recipes::all_numeric_predictors()) |>
        recipes::step_dummy(recipes::all_factor_predictors())

    wf = workflows::workflow() |>
        workflows::add_recipe(rec) |>
        workflows::add_model(spec)

    fitted_wf = workflows::fit(wf, data = iris[1:100, ])
    preds = predict(fitted_wf, new_data = iris[101:110, ])

    expect_s3_class(preds, "tbl_df")
    expect_equal(nrow(preds), 10)
    expect_true(all(is.finite(preds$.pred)))
})

test_that("mlp_kindling is properly registered with parsnip", {
    skip_if_not_installed("parsnip")

    models = parsnip::get_model_env()$models

    expect_true("mlp_kindling" %in% models)
})

test_that("rnn_kindling is properly registered with parsnip", {
    skip_if_not_installed("parsnip")

    models = parsnip::get_model_env()$models

    expect_true("rnn_kindling" %in% models)
})

test_that("mlp_kindling supports both regression and classification modes", {
    skip_if_not_installed("parsnip")

    spec_class = mlp_kindling(mode = "classification")
    spec_reg = mlp_kindling(mode = "regression")

    expect_equal(spec_class$mode, "classification")
    expect_equal(spec_reg$mode, "regression")
})

test_that("rnn_kindling supports both regression and classification modes", {
    skip_if_not_installed("parsnip")

    spec_class = rnn_kindling(mode = "classification")
    spec_reg = rnn_kindling(mode = "regression")

    expect_equal(spec_class$mode, "classification")
    expect_equal(spec_reg$mode, "regression")
})

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("classification metrics work with mlp predictions", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("yardstick")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = c(20),
        epochs = 10,
        verbose = FALSE
    )

    fitted = parsnip::fit(
        spec,
        Species ~ .,
        data = iris
    )

    results = parsnip::augment(fitted, new_data = iris)

    metrics = yardstick::metrics(
        results,
        truth = Species,
        estimate = .pred_class
    )

    expect_s3_class(metrics, "tbl_df")
    expect_true("accuracy" %in% metrics$.metric)
    expect_true("kap" %in% metrics$.metric)
})

test_that("regression metrics work with mlp predictions", {
    skip_if_not_installed("parsnip")
    skip_if_not_installed("yardstick")
    skip_if_no_torch()

    spec = mlp_kindling(
        mode = "regression",
        hidden_neurons = c(20),
        epochs = 10,
        verbose = FALSE
    )

    fitted = parsnip::fit(
        spec,
        Sepal.Length ~ .,
        data = iris
    )

    results = parsnip::augment(fitted, new_data = iris)

    metrics = yardstick::metrics(
        results,
        truth = Sepal.Length,
        estimate = .pred
    )

    expect_s3_class(metrics, "tbl_df")
    expect_true("rmse" %in% metrics$.metric)
    expect_true("rsq" %in% metrics$.metric)
    expect_true("mae" %in% metrics$.metric)
})

#' Recurrent Neural Network via kindling
#'
#' `rnn_kindling()` defines a recurrent neural network model that can be used
#' for classification or regression on sequential data. It integrates with the
#' tidymodels ecosystem and uses the torch backend via kindling.
#'
#' @inheritParams mlp_kindling
#' @param rnn_type A character string for the type of RNN cell ("rnn", "lstm",
#'   "gru"). Cannot be tuned — pass via `set_engine()`.
#' @param bidirectional A logical indicating whether to use bidirectional RNN.
#'   Can be tuned.
#' @param dropout A number between 0 and 1 for dropout rate between layers.
#'   Can be tuned.
#'
#' @details
#' This function creates a model specification for a recurrent neural network
#' that can be used within tidymodels workflows. The model supports:
#'
#' - Multiple RNN types: basic RNN, LSTM, and GRU
#' - Bidirectional processing
#' - Dropout regularization
#' - GPU acceleration (CUDA, MPS, or CPU)
#' - Hyperparameter tuning integration
#' - Both regression and classification tasks
#'
#' The `device` parameter controls where computation occurs:
#' - `NULL` (default): Auto-detect best available device (CUDA > MPS > CPU)
#' - `"cuda"`: Use NVIDIA GPU
#' - `"mps"`: Use Apple Silicon GPU
#' - `"cpu"`: Use CPU only
#'
#' @return A model specification object with class `rnn_kindling`.
#'
#' @examples
#' \donttest{
#' if (torch::torch_is_installed()) {
#'     box::use(
#'         recipes[recipe],
#'         workflows[workflow, add_recipe, add_model],
#'         parsnip[fit]
#'     )
#'
#'     # Model specs
#'     rnn_spec = rnn_kindling(
#'         mode = "classification",
#'         hidden_neurons = c(64, 32),
#'         rnn_type = "lstm",
#'         activation = c("relu", "elu"),
#'         epochs = 100,
#'         bidirectional = TRUE
#'     )
#'
#'     wf = workflow() |>
#'         add_recipe(recipe(Species ~ ., data = iris)) |>
#'         add_model(rnn_spec)
#'
#'     fit_wf = fit(wf, data = iris)
#'     fit_wf
#' } else {
#'     message("Torch not fully installed — skipping example")
#' }
#' }
#'
#' @export
rnn_kindling = 
    function(
        mode = "unknown",
        engine = "kindling",
        # ---- Tunable parameters ----
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = NULL,
        bidirectional = NULL,
        dropout = NULL,
        epochs = NULL,
        batch_size = NULL,
        penalty = NULL,
        mixture = NULL,
        learn_rate = NULL,
        optimizer = NULL,
        validation_split = NULL,
        # ---- Non-tunable parameters (engine-level, set via set_engine()) ----
        rnn_type = NULL,
        optimizer_args = NULL,
        loss = NULL,
        early_stopping = NULL,
        device = NULL,
        verbose = NULL,
        cache_weights = NULL
    )
{
    
    if (!requireNamespace("parsnip", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg parsnip} is required but not installed.")
    }
    
    # ---- "Tunable" args ----
    args = list(
        hidden_neurons = rlang::enquo(hidden_neurons),
        activations = rlang::enquo(activations),
        output_activation = rlang::enquo(output_activation),
        bias = rlang::enquo(bias),
        bidirectional = rlang::enquo(bidirectional),
        dropout = rlang::enquo(dropout),
        epochs = rlang::enquo(epochs),
        batch_size = rlang::enquo(batch_size),
        penalty = rlang::enquo(penalty),
        mixture = rlang::enquo(mixture),
        learn_rate = rlang::enquo(learn_rate),
        optimizer = rlang::enquo(optimizer),
        validation_split = rlang::enquo(validation_split)
    )
    
    # ---- "Non-tunable" engine args ----
    eng_args = list(
        rnn_type = rlang::enquo(rnn_type),
        optimizer_args = rlang::enquo(optimizer_args),
        loss = rlang::enquo(loss),
        early_stopping = rlang::enquo(early_stopping),
        device = rlang::enquo(device),
        verbose = rlang::enquo(verbose),
        cache_weights = rlang::enquo(cache_weights)
    )
    
    parsnip::new_model_spec(
        "rnn_kindling",
        args = args,
        eng_args = eng_args,
        mode = mode,
        user_specified_mode = !missing(mode),
        method = NULL,
        engine = engine,
        user_specified_engine = !missing(engine)
    )
}

#' @export
print.rnn_kindling = function(x, ...) {
    cat("Recurrent Neural Network Model Specification (", x$mode, ")\n\n", sep = "")
    parsnip::model_printer(x, ...)
    invisible(x)
}

#' @export
#' @importFrom stats update
update.rnn_kindling = 
    function(
        object,
        parameters = NULL,
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = NULL,
        bidirectional = NULL,
        dropout = NULL,
        epochs = NULL,
        batch_size = NULL,
        penalty = NULL,
        mixture = NULL,
        learn_rate = NULL,
        optimizer = NULL,
        validation_split = NULL,
        rnn_type = NULL,
        optimizer_args = NULL,
        loss = NULL,
        early_stopping = NULL,
        device = NULL,
        verbose = NULL,
        cache_weights = NULL,
        fresh = FALSE,
        ...
    )
{
    
    args = list(
        hidden_neurons = rlang::enquo(hidden_neurons),
        activations = rlang::enquo(activations),
        output_activation = rlang::enquo(output_activation),
        bias = rlang::enquo(bias),
        bidirectional = rlang::enquo(bidirectional),
        dropout = rlang::enquo(dropout),
        epochs = rlang::enquo(epochs),
        batch_size = rlang::enquo(batch_size),
        penalty = rlang::enquo(penalty),
        mixture = rlang::enquo(mixture),
        learn_rate = rlang::enquo(learn_rate),
        optimizer = rlang::enquo(optimizer),
        validation_split = rlang::enquo(validation_split)
    )
    
    parsnip::update_spec(
        object = object,
        parameters = parameters,
        args_enquo_list = args,
        fresh = fresh,
        cls = "rnn_kindling",
        ...
    )
}

#' @export
#' @importFrom parsnip translate
translate.rnn_kindling = function(x, engine = x$engine, ...) {
    if (is.null(engine)) {
        cli::cli_abort("Please set an engine with `set_engine()`.")
    }
    
    x = parsnip::translate.default(x, engine, ...)
    x
}

#' @export
#' @importFrom tune tunable
tunable.rnn_kindling = function(x, ...) {
    tibble::tibble(
        name = c(
            "hidden_neurons", "activations", "output_activation",
            "bias", "bidirectional", "dropout", "epochs", "batch_size",
            "penalty", "mixture", "learn_rate", "optimizer",
            "validation_split"
        ),
        call_info = list(
            list(pkg = "kindling", fun = "hidden_neurons"),
            list(pkg = "kindling", fun = "activations"),
            list(pkg = "kindling", fun = "output_activation"),
            list(pkg = "kindling", fun = "bias"),
            list(pkg = "kindling", fun = "bidirectional"),
            list(pkg = "dials", fun = "dropout"),
            list(pkg = "dials", fun = "epochs"),
            list(pkg = "dials", fun = "batch_size"),
            list(pkg = "dials", fun = "penalty"),
            list(pkg = "dials", fun = "mixture"),
            list(pkg = "dials", fun = "learn_rate"),
            list(pkg = "kindling", fun = "optimizer"),
            list(pkg = "kindling", fun = "validation_split")
        ),
        source = "model_spec",
        component = "rnn_kindling",
        component_id = "main"
    )
}

#' @rdname kindling-nn-wrappers
#' @section RNN Wrapper for `{tidymodels}` interface:
#' Internal wrapper — use `rnn_kindling()` + `fit()` instead.
#'
#' @keywords internal
#' @export
rnn_wrapper = function(formula, data, ...) {
    dots = list(...)
    dots = prepare_kindling_args(dots)
    
    do.call(
        rnn,
        c(list(formula = formula, data = data), dots)
    )
}

test_that("rnn_kindling creates valid model specification", {
    skip_if_not_installed("parsnip")

    spec = rnn_kindling(
        mode = "classification",
        hidden_neurons = c(64, 32),
        rnn_type = "lstm",
        activations = "relu",
        epochs = 50
    )

    expect_s3_class(spec, "rnn_kindling")
    expect_s3_class(spec, "model_spec")
    expect_equal(spec$mode, "classification")
    expect_equal(spec$engine, "kindling")
})

test_that("rnn_kindling accepts all RNN-specific parameters", {
    skip_if_not_installed("parsnip")

    spec = rnn_kindling(
        mode = "regression",
        hidden_neurons = c(128, 64),
        rnn_type = "gru",
        activations = c("relu", "elu"),
        bidirectional = TRUE,
        dropout = 0.3,
        epochs = 100
    )

    expect_s3_class(spec, "rnn_kindling")
    expect_equal(spec$mode, "regression")
})

test_that("rnn_kindling print method works", {
    skip_if_not_installed("parsnip")

    spec = rnn_kindling(mode = "classification", rnn_type = "lstm")

    expect_output(print(spec), "Recurrent Neural Network")
    expect_output(print(spec), "classification")
})

test_that("rnn_kindling update method works", {
    skip_if_not_installed("parsnip")

    spec = rnn_kindling(
        mode = "classification",
        rnn_type = "lstm",
        epochs = 50
    )

    updated_spec = update(spec, epochs = 100, dropout = 0.2)

    expect_s3_class(updated_spec, "rnn_kindling")
})

#' @importFrom stats reorder setNames
#' @importFrom rlang :=
#' @importFrom lifecycle badge
#' @importFrom ggplot2 autoplot
NULL

# To avoid R CMD check notes about undefined global variables
utils::globalVariables(c(
    "x_names", "rel_imp", "object", "new_data", "nn_module", "fit_class",
    "epoch", "loss", "set",
    "fitted_val", "residuals", "actual_val",
    "actual", "predicted", "n", "prop"
))

#' Print method for nn_fit objects
#'
#' @param x An object of class `"nn_fit"`
#' @param ... Additional arguments (unused)
#'
#' @return No return value, called for side effects (printing model summary)
#'
#' @keywords internal
#' @export
print.nn_fit = function(x, ...) {
    # ---- Title ----
    title = "Generalized Neural Network"
    title_block = style_bold(rule(center = title, line = "="))
    cat_line("\n", title_block, "\n")
    
    hidden_units_str = if (is.null(x$hidden_neurons)) {
        "Not specified"
    } else {
        paste(as.character(x$hidden_neurons), collapse = ", ")
    }
    
    nn_model_type = if (!is.null(x$arch)) {
        nn_layer = x$arch$nn_layer %||% "nn_linear"
        if (inherits(nn_layer, "formula")) {
            nn_layer = paste(deparse(nn_layer), collapse = " ")
        }
        glue("{x$arch$nn_name} ({nn_layer})")
    } else {
        "FFNN"
    }
    
    # ---- Table 1: Model Summary ----
    summary_data = data.frame(
        type = c(
            "NN Model Type",
            "Number of Epochs",
            "Hidden Layer Units",
            "Number of Hidden Layers",
            "Pred. Type",
            "n_predictors",
            "n_response",
            "reg.",
            "Device"
        ),
        res = c(
            nn_model_type,
            as.character(x$n_epochs),
            hidden_units_str,
            as.character(length(x$hidden_neurons)),
            if (x$is_classification) "classification" else "regression",
            as.character(x$no_x),
            as.character(x$no_y),
            if (x$penalty == 0) "None" else glue("[\u03BB = {x$penalty}, \u03B1 = {x$mixture}]"),
            x$device
        ),
        stringsAsFactors = FALSE
    )
    
    heading1 = style_italic(rule(left = "Model Summary", line = "-"))
    cat_line("\n", heading1, "\n\n")
    table_summary(summary_data, l = 5, center_table = TRUE, style = list(sep = ":  "))
    cat("\n\n")
    
    # ---- Table 2: Activation Functions ----
    inner_acts = if (is.list(x$activations)) {
        vapply(x$activations, concat, character(1))
    } else if (!is.null(x$activations)) {
        if (length(x$activations) == 1) {
            rep(as.character(x$activations), length(x$hidden_neurons))
        } else {
            as.character(x$activations)
        }
    } else {
        rep("None", length(x$hidden_neurons))
    }
    outer_acts = concat(x$output_activation)
    
    act_data = data.frame(
        layer = c(
            paste(
                ordinal_gen(seq_along(inner_acts)), "Layer",
                paste0("{", x$hidden_neurons, "}")
            ),
            "Output Activation"
        ),
        infos = c(inner_acts, outer_acts),
        stringsAsFactors = FALSE
    )
    
    heading2 = style_italic(rule(left = "Activation Functions", line = "-"))
    cat_line("\n", heading2, "\n\n")
    table_summary(act_data, l = 5, center_table = TRUE, style = list(sep = ":  "))
    cat("\n\n")
    
    # ---- Table 3: Architecture Spec ----
    flag = function(val) if (!is.null(val)) "yes" else "N/A"
    arch_str = function(val, default = "N/A") {
        if (is.null(val)) return(default)
        if (inherits(val, "formula")) return(paste(deparse(val), collapse = " "))
        if (is.function(val) || is.list(val)) return("yes")
        paste(as.character(val), collapse = ", ")
    }
    arch = x$arch
    arch_data = data.frame(
        type = c(
            "nn_layer",
            "out_nn_layer",
            "nn_layer_args",
            "layer_arg_fn",
            "forward_extract",
            "before_output_transform",
            "after_output_transform",
            "last_layer_args",
            "input_transform"
        ),
        res = if (is.null(arch)) {
            rep("N/A", 9L)
        } else {
            c(
                arch_str(arch$nn_layer, default = "nn_linear (default)"),
                arch_str(arch$out_nn_layer),
                if (length(arch$nn_layer_args) > 0) "yes" else "N/A",
                flag(arch$layer_arg_fn),
                flag(arch$forward_extract),
                flag(arch$before_output_transform),
                flag(arch$after_output_transform),
                if (length(arch$last_layer_args) > 0) "yes" else "N/A",
                flag(arch$input_transform)
            )
        },
        stringsAsFactors = FALSE
    )
    
    heading3 = style_italic(rule(left = "Architecture Spec", line = "-"))
    cat_line("\n", heading3, "\n\n")
    table_summary(arch_data, l = 5, center_table = TRUE, style = list(sep = ":  "))
    cat("\n")
    
    invisible(x)
}

#' Activation Functions Specification Helper
#'
#' This function is a DSL function, kind of like `ggplot2::aes()`, that helps to
#' specify activation functions for neural network layers. It validates that
#' activation functions exist in `torch` and that any parameters match the
#' function's formal arguments.
#'
#' @param ... Activation function specifications. Can be:
#' - Bare symbols: `relu`, `tanh`
#' - Character strings (simple): `"relu"`, `"tanh"`
#' - Character strings (with params): `"softshrink(lambda = 0.1)"`, `"rrelu(lower = 1/5, upper = 1/4)"`
#' - Named with parameters: `softmax = args(dim = 2L)`
#' - Indexed syntax (named): `softshrink[lambd = 0.2]`, `rrelu[lower = 1/5, upper = 1/4]`
#' - Indexed syntax (unnamed): `softshrink[0.5]`, `elu[0.5]`
#'
#' @return A `vctrs` vector with class "activation_spec" containing validated
#' activation specifications.
#'
#' @importFrom rlang enquos quo_get_expr is_call call_name as_string eval_tidy
#' @importFrom cli cli_abort
#' @importFrom vctrs new_vctr
#'
#' @export
act_funs = function(...) {
    dots = enquos(..., .ignore_empty = "all")
    mask = list(args = args)
    
    out = purrr::imap(dots, function(quo, name) {
        expr = quo_get_expr(quo)
        
        # ---- Indexed syntax ----
        # Ex. softshrink[lambd = 0.2] or softshrink[0.5]
        if (is_call(expr) && identical(call_name(expr), "[")) {
            calls = as.list(expr)
            act_name = as_string(calls[[2]])
            params = calls[-(1:2)]
            
            # Keep params as-is (named or unnamed)
            # Names will be NULL for unnamed params, which is fine
            validate_activation(act_name, prefix = "nnf_")
            if (!is.null(names(params)) && any(names(params) != "")) {
                validate_args_formals(act_name, params, prefix = "nnf_")
            }
            structure(params, act_name = act_name, class = "parameterized_activation")
            
        } else if (!is.null(name) && name != "") {
            # ---- Named parameter ---- 
            # This is a superseded behavior after v0.3.x, in favor of indexed syntax
            # Ex. softshrink = args(lambd = 0.5)
            validate_activation(name, prefix = "nnf_")
            extract_call = eval_tidy(quo, data = mask)
            
            if (inherits(extract_call, "activation_args")) {
                params = unclass(extract_call)
                validate_args_formals(name, params, prefix = "nnf_")
                structure(params, act_name = name, class = "parameterized_activation")
            } else if (is.character(extract_call) && length(extract_call) == 1 && extract_call == "") {
                structure(list(), act_name = name, class = "parameterized_activation")
            } else {
                cli_abort(c(
                    "Invalid syntax for parameterized activation '{name}'.",
                    i = "Use: {.code {name} = args(param = value)}."
                ), class = "activation_syntax_error")
            }
        } else if (is.symbol(expr)) {
            # ---- Bare symbol ----
            # Ex. relu, tanh
            act_name = as_string(expr)
            validate_activation(act_name, prefix = "nnf_")
            act_name
        } else if (is.character(expr)) {
            # ---- Character string ----
            # Note: Stringly typed expression is a bad practice, but tolerable
            # Ex. "relu" or "softshrink(lambd = 0.1)"
            parsed = parse_activation_string(expr)
            validate_activation(parsed$name, prefix = "nnf_")
            
            if (length(parsed$params) > 0) {
                validate_args_formals(parsed$name, parsed$params, prefix = "nnf_")
                structure(parsed$params, act_name = parsed$name, class = "parameterized_activation")
            } else {
                parsed$name
            }
        } else if (is_call(expr) && identical(call_name(expr), "new_act_fn")) {
            eval_tidy(quo, data = list(new_act_fn = new_act_fn))
        } else {
            cli_abort(c(
                "Invalid activation specification.",
                i = "Use bare names like {.code relu}, strings like {.code 'relu'},",
                i = "bracket syntax like {.code softshrink[lambd = 0.1]},",
                i = "parameterized strings like {.code 'softshrink(lambd = 0.1)'},",
                i = "or DSL like {.code softmax = args(dim = 2L)}."
            ), class = "activation_syntax_error")
        }
    })
    
    new_vctr(out, class = "activation_spec")
}

#' Activation Function Arguments Helper
#' 
#' @description
#' `r lifecycle::badge("superseded")`
#' 
#' This is superseded in v0.3.0 in favour of indexed syntax, e.g. `<act_fn[param = 0]>` type. 
#' Type-safe helper to specify parameters for activation functions.
#' All parameters must be named and match the formal arguments of the
#' corresponding `{torch}` activation function.
#'
#' @param ... Named arguments for the activation function.
#' @return A list with class "activation_args" containing the parameters.
#'
#' @importFrom cli cli_abort
#'
#' @export
args = function(...) {
    lifecycle::deprecate_soft(
        when = "0.3.0",
        what = "args()",
        details = "Use indexed syntax for parametric activation functions, e.g. `<softplus[beta = 0.5]>`."
    )
    
    params = list(...)
    
    if (length(params) == 0) {
        structure(list(), class = "activation_args")
    } else {
        param_names = names(params)
        
        if (is.null(param_names) || any(param_names == "")) {
            cli_abort(c(
                "{.fn args} requires all arguments to be named.",
                i = "Use named arguments like {.code args(dim = 2L)}."
            ), class = "activation_args_error")
        }
        
        structure(params, class = "activation_args")
    }
}

#' Parse activation function string with parameters
#'
#' Parse strings like "softshrink(lambda = 0.1)" or "rrelu(lower = 1/5, upper = 1/4)"
#' into activation name and parameter list.
#'
#' @param act_str Character string with activation and optional parameters.
#' @return List with `name` and `params` elements.
#'
#' @importFrom cli cli_abort
#'
#' @noRd
parse_activation_string = function(act_str) {
    if (!is.character(act_str) || length(act_str) != 1) {
        cli_abort("Activation string must be a single character value.")
    }
    
    # Check if it has parameters (contains parentheses)
    if (!grepl("\\(", act_str)) {
        return(list(name = act_str, params = list()))
    }
    
    match = regexec("^([a-z_0-9]+)\\((.*)\\)$", act_str, ignore.case = TRUE)
    matches = regmatches(act_str, match)[[1]]
    
    if (length(matches) != 3) {
        cli_abort(c(
            "Invalid activation string format: {.val {act_str}}",
            i = "Expected format: {.code 'function_name(param1 = value1, param2 = value2)'}"
        ))
    }
    
    act_name = matches[2]
    params_str = matches[3]
    
    if (nchar(trimws(params_str)) == 0) {
        params_list = list()
    } else {
        expr_str = paste0("list(", params_str, ")")
        
        tryCatch({
            params_list = eval(parse(text = expr_str))
        }, error = function(e) {
            cli_abort(c(
                "Failed to parse parameters in: {.val {act_str}}",
                i = "Error: {e$message}",
                i = "Parameters must be valid R expressions."
            ))
        })
    }
    
    list(name = act_name, params = params_list)
}

#' Validate Activation Function Exists
#'
#' @param act_name Character. Activation function name (without prefix).
#' @param prefix Character. Prefix for the function name (default: "nnf_").
#'
#' @importFrom cli cli_abort
#' @importFrom rlang is_installed
#'
#' @noRd
validate_activation = function(act_name, prefix = "nnf_") {
    if (!is_installed("torch")) {
        cli_abort(c(
            "{.pkg torch} package is required but not installed.",
            i = "Install it with: {.code install.packages('torch')}"
        ), class = "torch_missing_error")
    }
    
    fn_name = paste0(prefix, act_name)
    
    if (!exists(fn_name, where = asNamespace("torch"), mode = "function")) {
        cli_abort(c(
            "Activation function {.fn {fn_name}} does not exist in {.pkg torch}.",
            i = "Available functions start with {.code {prefix}}.",
            i = "Check {.code ?torch::nnf_relu} for examples."
        ), class = "activation_not_found_error")
    }
    
    fn_name
}

#' Validate Arguments Match Function Formals
#'
#' @param act_name Character. Activation function name (without prefix).
#' @param params List. Parameters to validate.
#' @param prefix Character. Prefix for the function name (default: "nnf_").
#'
#' @importFrom cli cli_abort
#'
#' @noRd
validate_args_formals = function(act_name, params, prefix = "nnf_") {
    fn_name = paste0(prefix, act_name)
    
    if (!requireNamespace("torch", quietly = TRUE)) {
        invisible(NULL)
    } else {
        fn = get(fn_name, envir = asNamespace("torch"))
        fn_formals = names(formals(fn))
        fn_formals = fn_formals[!fn_formals %in% c("input", "x", "...")]
        
        param_names = names(params)
        # Only check named parameters
        if (!is.null(param_names)) {
            named_params = param_names[param_names != ""]
            invalid_params = setdiff(named_params, fn_formals)
            
            if (length(invalid_params) > 0) {
                valid_params_str = paste(fn_formals, collapse = ", ")
                cli_abort(c(
                    "Invalid parameter{?s} for {.fn {fn_name}}: {.arg {invalid_params}}.",
                    i = "Valid parameters are: {.code {valid_params_str}}."
                ), class = "activation_invalid_params_error")
            }
        }
        
        invisible(NULL)
    }
}

#' Activation Function Specifications Parser
#'
#' This function parses activation function specifications provided in various formats
#'
#' @param activations Activation function specifications. Can be:
#' - `NULL`: No activation functions.
#' - Character vector: e.g., `c("relu", "tanh")` or `c("relu", "softshrink(lambda = 0.1)")`.
#' - List: e.g., `list(relu, tanh, softmax = args(dim = 2L))`.
#' - `activation_spec` object from `act_funs()`.
#' @param n_layers Number of layers to apply activations to.
#'
#' @return A list with two elements:
#' - `names`: Character vector of activation function names.
#' - `params`: List of parameter lists for each activation function.
#'
#' @importFrom cli cli_abort
#' @importFrom rlang is_call call_name as_string
#' @importFrom purrr imap set_names transpose
#'
#' @noRd
parse_activation_spec = function(activations, n_layers) {
    if (is.null(activations)) {
        list(
            names = rep(NA_character_, n_layers),
            params = vector("list", n_layers)
        )
    } else if (inherits(activations, "activation_spec")) {
        if (length(activations) == 1L) {
            activations = rep(activations, n_layers)
        }
        
        if (length(activations) != n_layers) {
            cli_abort(c(
                "{.arg activations} must specify 1 or {n_layers} activation(s).",
                x = "You provided {length(activations)}."
            ), class = "activation_spec_length_error")
        }
        
        parsed_activation =
            imap(activations, function(elem, i) {
                
                if (inherits(elem, "custom_activation")) {
                    list(
                        name = "<custom>",
                        param = list(fn = attr(elem, "act_fn"))
                    )
                } else if (inherits(elem, "parameterized_activation")) {
                    params = unclass(elem)
                    attr(params, "act_name") = NULL
                    list(
                        name = attr(elem, "act_name"),
                        param = params
                    )
                } else if (is.character(elem)) {
                    list(
                        name = elem,
                        param = list()
                    )
                } else {
                    cli_abort(paste("Unsupported activation type at index", i))
                }
            })
        
        set_names(transpose(parsed_activation), c("names", "params"))
    } else if (is.character(activations)) {
        if (length(activations) == 1L) {
            activations = rep(activations, n_layers)
        }
        
        if (length(activations) != n_layers) {
            cli_abort(c(
                "{.arg activations} must be length 1 or {n_layers}.",
                x = "You provided length {length(activations)}."
            ), class = "activation_spec_length_error")
        }
        
        parsed_activation = lapply(activations, function(act_str) {
            parsed = parse_activation_string(act_str)
            list(name = parsed$name, param = parsed$params)
        })
        
        set_names(transpose(parsed_activation), c("names", "params"))
    } else if (is.list(activations)) {
        if (length(activations) == 1L) {
            activations = rep(activations, n_layers)
        }
        
        if (length(activations) != n_layers) {
            cli_abort(c(
                "{.arg activations} must be length 1 or {n_layers}.",
                x = "You provided length {length(activations)}."
            ), class = "activation_spec_length_error")
        }
        
        parsed_activation =
            imap(activations, function(elem, i) {
                elem_name = if (!is.null(names(activations))) names(activations)[i] else ""
                
                if (elem_name != "" && is.list(elem)) {
                    list(name = elem_name, param = elem)
                    
                } else if (elem_name == "" && is.character(elem) && length(elem) == 1L) {
                    parsed = parse_activation_string(elem)
                    list(name = parsed$name, param = parsed$params)
                    
                } else if (elem_name == "" && is.name(elem)) {
                    list(name = as.character(elem), param = list())
                    
                } else if (elem_name != "" && is.character(elem) && elem == "") {
                    list(name = elem_name, param = list())
                    
                } else {
                    cli_abort(c(
                        "Invalid activation specification at position {i}.",
                        i = "Use: {.code relu} or {.code 'relu'} for simple activations",
                        i = "Use: {.code 'softshrink(lambda = 0.1)'} for parameterized strings",
                        i = "Use: {.code softmax = args(dim = 2L)} for DSL"
                    ), class = "activation_syntax_error")
                }
            })
        
        set_names(transpose(parsed_activation), c("names", "params"))
    } else {
        cli_abort(c(
            "{.arg activations} must be character vector, list, or {.code act_funs()}.",
            i = "Examples:",
            "*" = "Character: {.code c('relu', 'tanh')}",
            "*" = "Parameterized: {.code c('relu', 'softshrink(lambda = 0.1)')}",
            "*" = "Bracket syntax: {.code act_funs(relu, softshrink[lambd = 0.2])}",
            "*" = "List: {.code list(relu, tanh, softmax = args(dim = 2L))}",
            "*" = "DSL: {.code act_funs(relu, tanh, softmax = args(dim = 2L))}"
        ), class = "activation_type_error")
    }
}

#' Activation Functions Processor
#'
#' This function processes activation function specifications into callable
#' expressions with proper torch namespace handling.
#'
#' @param activation_spec A list with two elements:
#' - `names`: Character vector of activation function names.
#' - `params`: List of parameter lists for each activation function.
#' @param prefix Prefix for activation function names (default: "nnf_").
#'
#' @return A list of functions that generate activation function calls with
#' proper torch:: namespace prefixing.
#'
#' @importFrom rlang call2 sym exec
#' @importFrom purrr map2
#'
#' @noRd
process_activations = function(activation_spec, prefix = "nnf_") {
    act_names = activation_spec$names
    act_params = activation_spec$params
    
    map2(act_names, act_params, function(name, params) {
        if (is.na(name)) {
            NULL
        } else if (name == "<custom>") {
            # fn = attr(params, "act_fn")
            # function(x_expr) call2(fn, x_expr)
            fn = params$fn
            local({
                fn = fn
                function(x_expr) {
                    rlang::expr((!!fn)(!!x_expr))
                }
            })
        } else {
            fn_name = paste0(prefix, name)
            fn_call = call2("::", sym("torch"), sym(fn_name))
            
            if (is.null(params) || length(params) == 0) {
                function(x_expr) {
                    # rlang::call_standardise(call2(fn_call, x_expr))
                    call2(fn_call, x_expr)
                }
            } else {
                function(x_expr) {
                    # rlang::call_standardise(exec(call2, fn_call, x_expr, !!!params))
                    exec(call2, fn_call, x_expr, !!!params)
                }
            }
        }
    })
}

#' Activation Function Specs Evaluation
#' 
#' Helper function for `act_funcs()` argument.
#' 
#' @param activations Quosure containing the activations expression
#' @param output_activation Quosure containing the output_activation expression
#' 
#' @return A list with two elements: `activations` and `output_activation`
#' 
#' @importFrom rlang enquo quo_is_null eval_tidy
#' @keywords internal
eval_act_funs = function(activations, output_activation) {
    activations_quo = enquo(activations)
    output_activation_quo = enquo(output_activation)
    
    env_mask = list(
        act_funs = act_funs,
        args = args
    )
    
    activations = if (!quo_is_null(activations_quo)) {
        eval_tidy(activations_quo, data = env_mask)
    } else {
        NULL
    }
    
    output_activation = if (!quo_is_null(output_activation_quo)) {
        eval_tidy(output_activation_quo, data = env_mask)
    } else {
        NULL
    }
    
    list(
        activations = activations,
        output_activation = output_activation
    )
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("FFNN generator uses torch namespace explicitly", {
    skip_if_no_torch()

    expr = ffnn_generator(
        nn_name = "TestFFNN",
        hd_neurons = c(10, 5),
        no_x = 3,
        no_y = 1,
        activations = "relu"
    )
    code_str = paste(deparse(expr), collapse = " ")

    expect_match(code_str, "torch::nn_module", fixed = TRUE)
    expect_match(code_str, "torch::nn_linear", fixed = TRUE)
    expect_match(code_str, "torch::nnf_relu", fixed = TRUE)
    expect_false(grepl("(?<!torch::)\\bnn_linear\\s*\\(", code_str, perl = TRUE))
    expect_false(grepl("(?<!torch::)\\bnnf_relu\\s*\\(", code_str, perl = TRUE))
})

test_that("RNN generator uses torch namespace explicitly", {
    skip_if_no_torch()

    expr = rnn_generator(
        nn_name = "TestRNN",
        hd_neurons = c(20, 10, 15, 8),
        no_x = 5,
        no_y = 2,
        rnn_type = "lstm",
        activations = act_funs(relu, selu, softshrink[lambd = 0.5], gelu)
    )

    code_str = paste(deparse(expr), collapse = " ")

    expect_match(code_str, "torch::nn_module", fixed = TRUE)
    expect_match(code_str, "torch::nn_lstm", fixed = TRUE)
    expect_match(code_str, "torch::nn_linear", fixed = TRUE)
    expect_match(code_str, "torch::nnf_relu", fixed = TRUE)
    expect_match(code_str, "torch::nnf_selu", fixed = TRUE)
    expect_match(code_str, "torch::nnf_softshrink", fixed = TRUE)
    expect_match(code_str, "torch::nnf_gelu", fixed = TRUE)
})

test_that("Invalid activation 'tanh' throws error in FFNN generator", {
    skip_if_no_torch()

    expect_error(
        ffnn_generator(
            hd_neurons = c(10, 5),
            no_x = 3,
            no_y = 1,
            activations = act_funs(relu, tanh)
        ),
        class = "activation_not_found_error"
    )
})

test_that("Invalid activation 'tanh' throws error in RNN generator", {
    skip_if_no_torch()

    expect_error(
        rnn_generator(
            hd_neurons = c(20, 10),
            no_x = 5,
            no_y = 2,
            rnn_type = "lstm",
            activations = act_funs(relu, tanh)
        ),
        class = "activation_not_found_error"
    )
})

test_that("Generated model can be evaluated without library(torch)", {
    skip_if_no_torch()

    test_env = new.env()
    expr = ffnn_generator(
        hd_neurons = c(10),
        no_x = 5,
        no_y = 1,
        activations = "relu"
    )

    expect_error({
        model_class = eval(expr, envir = test_env)
        model = model_class()
    }, NA)
})

test_that("Activation functions are properly namespaced", {
    skip_if_no_torch()

    spec = parse_activation_spec(
        act_funs(relu, selu, leaky_relu[negative_slope = 0.01]),
        n_layers = 3
    )

    calls = process_activations(spec)
    relu_call = calls[[1]](quote(x))

    expect_equal(as.character(relu_call[[1]][[1]]), "::")
    expect_equal(as.character(relu_call[[1]][[2]]), "torch")
    expect_equal(as.character(relu_call[[1]][[3]]), "nnf_relu")
})

test_that("Complex activations maintain namespace", {
    skip_if_no_torch()

    expr = ffnn_generator(
        hd_neurons = c(128, 64, 32, 16),
        no_x = 20,
        no_y = 10,
        activations = act_funs(
            relu,
            leaky_relu[negative_slope = 0.01],
            elu[alpha = 1.0],
            softshrink[lambd = 0.5]
        )
    )

    code_str = paste(deparse(expr), collapse = " ")

    expect_match(code_str, "torch::nnf_relu")
    expect_match(code_str, "torch::nnf_leaky_relu")
    expect_match(code_str, "torch::nnf_elu")
    expect_match(code_str, "torch::nnf_softshrink")
})

test_that("Generated model works in isolated environment", {
    skip_if_no_torch()

    isolated_env = new.env(parent = asNamespace("kindling"))
    expr = ffnn_generator(
        hd_neurons = c(10, 5),
        no_x = 4,
        no_y = 1,
        activations = "relu"
    )

    model_class = eval(expr, envir = isolated_env)

    expect_s3_class(model_class, "nn_module_generator")

    model = model_class()
    expect_s3_class(model, "nn_module")
})

test_that("torch functions not in kindling namespace", {
    kindling_exports = getNamespaceExports("kindling")

    expect_false("nn_module" %in% kindling_exports)
    expect_false("nn_linear" %in% kindling_exports)
    expect_false("nn_lstm" %in% kindling_exports)
    expect_false("nnf_relu" %in% kindling_exports)
    expect_false("nnf_tanh" %in% kindling_exports)
})

test_that("Generated code doesn't require torch in global env", {
    skip_if_no_torch()

    if ("package:torch" %in% search()) {
        detach("package:torch", unload = FALSE)
    }

    expr = ffnn_generator(
        hd_neurons = c(10),
        no_x = 5,
        no_y = 1
    )
    expect_error({
        model_class = eval(expr)
        model = model_class()
    }, NA)
    expect_false("package:torch" %in% search())
})

test_that("RNN types use correct namespace", {
    skip_if_no_torch()

    for (rnn_type in c("rnn", "lstm", "gru")) {
        expr = rnn_generator(
            hd_neurons = c(10),
            no_x = 5,
            no_y = 1,
            rnn_type = rnn_type
        )

        code_str = paste(deparse(expr), collapse = " ")
        expected_call = paste0("torch::nn_", rnn_type)

        expect_match(code_str, expected_call, fixed = TRUE,
                    info = paste("RNN type:", rnn_type))
    }
})

test_that("Bidirectional RNN maintains namespace", {
    skip_if_no_torch()

    expr = rnn_generator(
        hd_neurons = c(20, 10),
        no_x = 5,
        no_y = 1,
        rnn_type = "lstm",
        bidirectional = TRUE
    )

    code_str = paste(deparse(expr), collapse = " ")
    expect_match(code_str, "bidirectional = TRUE")
    expect_match(code_str, "torch::nn_lstm")
})

test_that("Dropout in RNN maintains namespace", {
    skip_if_no_torch()

    expr = rnn_generator(
        hd_neurons = c(30, 20, 10),
        no_x = 5,
        no_y = 1,
        rnn_type = "lstm",
        dropout = 0.3
    )

    code_str = paste(deparse(expr), collapse = " ")
    expect_match(code_str, "dropout")
    expect_match(code_str, "torch::nn_lstm")
})

test_that("No torch imports in generated forward pass", {
    skip_if_no_torch()

    expr = ffnn_generator(
        hd_neurons = c(10, 5),
        no_x = 3,
        no_y = 1,
        activations = c("relu", "elu")
    )

    code_str = paste(deparse(expr), collapse = "\n")
    expect_false(grepl("library\\(torch\\)", code_str))
    expect_false(grepl("require\\(torch\\)", code_str))
    expect_false(grepl("@import torch", code_str))
    expect_false(grepl("@importFrom torch", code_str))
})

test_that("Real model training works without explicit torch load", {
    skip_if_no_torch()
    if ("package:torch" %in% search()) {
        detach("package:torch", unload = FALSE)
    }
    expect_error({
        model = ffnn(
            Sepal.Length ~ Petal.Length + Petal.Width,
            data = iris[1:50, ],
            hidden_neurons = c(10, 5),
            activations = "relu",
            epochs = 5,
            verbose = FALSE
        )
    }, NA)
    expect_s3_class(model, "ffnn_fit")
    expect_false("package:torch" %in% search())
})

test_that("Multiple models don't interfere", {
    skip_if_no_torch()

    expr1 = ffnn_generator(
        nn_name = "Model1",
        hd_neurons = c(10),
        no_x = 5,
        no_y = 1
    )

    expr2 = ffnn_generator(
        nn_name = "Model2",
        hd_neurons = c(20, 10),
        no_x = 8,
        no_y = 3
    )

    model1_class = eval(expr1)
    model2_class = eval(expr2)

    model1 = model1_class()
    model2 = model2_class()

    expect_s3_class(model1, "nn_module")
    expect_s3_class(model2, "nn_module")

    expect_false(identical(model1, model2))
})

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

test_that("args() returns activation_args class", {
    skip_if_no_torch()
    
    expect_warning({ res = kindling::args(dim = 2L) }, class = "lifecycle_warning_deprecated")
    expect_s3_class(res, "activation_args")
})

test_that("args() is deprecated", {
    skip_if_no_torch()
    
    expect_warning(args(dim = 2L), class = "lifecycle_warning_deprecated")
})

test_that("args() errors on unnamed params", {
    skip_if_no_torch()
    
    expect_warning(
        expect_error(args(1L), class = "activation_args_error"),
        class = "lifecycle_warning_deprecated"
    )
})

# ---- Custom activation function ----

test_that("new_act_fn() accepts a valid function", {
    skip_if_no_torch()
    fn = new_act_fn(\(x) torch::torch_tanh(x))
    expect_s3_class(fn, "custom_activation")
    expect_s3_class(fn, "parameterized_activation")
})

test_that("new_act_fn() errors if fn is not a function", {
    expect_error(new_act_fn("not a fn"), class = "custom_activation_type_error")
})

test_that("new_act_fn() errors if fn has no arguments", {
    expect_error(new_act_fn(function() 1), class = "custom_activation_arity_error")
})

test_that("new_act_fn() works inside act_funs()", {
    skip_if_no_torch()
    spec = act_funs(relu, new_act_fn(\(x) torch::torch_tanh(x)))
    expect_length(spec, 2)
    expect_s3_class(spec[[2]], "custom_activation")
})

test_that("new_act_fn() skips probe when probe = FALSE", {
    fn = new_act_fn(\(x) x, probe = FALSE)
    expect_s3_class(fn, "custom_activation")
})

test_that("new_act_fn() stores custom .name attribute", {
    skip_if_no_torch()
    fn = new_act_fn(\(x) torch::torch_tanh(x), .name = "my_tanh")
    expect_equal(attr(fn, "act_name"), "my_tanh")
})

test_that(".assert_tensor_output() errors on non-tensor output", {
    skip_if_no_torch()
    expect_error(
        new_act_fn(\(x) as.numeric(x))(),
        class = "custom_activation_output_error"
    )
})

test_that("new_act_fn() call-time guard errors on non-tensor", {
    skip_if_no_torch()
    fn = new_act_fn(\(x) x, probe = FALSE)
    guarded = attr(fn, "act_fn")
    expect_error(
        guarded(42),
        class = "custom_activation_output_error"
    )
})

# ---- eval_act_funs() ----

test_that("eval_act_funs() returns NULL for both when not supplied", {
    res = eval_act_funs(NULL, NULL)
    expect_null(res$activations)
    expect_null(res$output_activation)
})

test_that("eval_act_funs() passes through act_funs() spec", {
    skip_if_no_torch()
    res = eval_act_funs(act_funs(relu), NULL)
    expect_s3_class(res$activations, "activation_spec")
    expect_null(res$output_activation)
})

test_that("eval_act_funs() handles output_activation", {
    skip_if_no_torch()
    res = eval_act_funs(NULL, act_funs(relu))
    expect_null(res$activations)
    expect_s3_class(res$output_activation, "activation_spec")
})

#' Validate device and get default device
#'
#' Check if requested device is available. And auto-detect available GPU device
#' or fallback to CPU.
#'
#' @param device Character. Requested device.
#' @return Character string of validated device.
#' @keywords internal
validate_device = function(device) {
    device = tolower(device)

    if (device == "cuda" && !torch::cuda_is_available()) {
        cli::cli_warn("CUDA not available. Falling back to CPU.")
        return("cpu")
    }

    if (device == "mps" && !torch::backends_mps_is_available()) {
        cli::cli_warn("MPS not available. Falling back to CPU.")
        return("cpu")
    }

    if (!device %in% c("cpu", "cuda", "mps")) {
        cli::cli_abort("Invalid device: {device}. Must be 'cpu', 'cuda', or 'mps'.")
    }

    device
}

get_default_device = function() {
    if (torch::cuda_is_available()) {
        return("cuda")
    } else if (torch::backends_mps_is_available()) {
        return("mps")
    } else {
        return("cpu")
    }
}

#' Summarize and Display a Two-Column Data Frame as a Formatted Table
#'
#' This function takes a two-column data frame and formats it into a summary-like table.
#' The table can be optionally split into two parts, centered, and given a title.
#' It is useful for displaying summary information in a clean, tabular format.
#' The function also supports styling with ANSI colors and text formatting through
#' the `{cli}` package and column alignment options.
#'
#' @param data A data frame with exactly two columns. The data to be summarized and displayed.
#' @param title A character string. An optional title to be displayed above the table.
#' @param l An integer. The number of rows to include in the left part of a split table.
#'           If `NULL`, the table is not split.
#' @param header A logical value. If `TRUE`, the column names of `data` are displayed as a header.
#' @param center_table A logical value. If `TRUE`, the table is centered in the terminal.
#' @param border_char Character used for borders. Default is `"\u2500"`.
#' @param style A list controlling the visual styling of table elements using ANSI formatting.
#'   Can include the following components:
#'   - `left_col`: Styling for the left column values.
#'   - `right_col`: Styling for the right column values.
#'   - `border_text`: Styling for the border.
#'   - `title`: Styling for the title.
#'   - `sep`: Separator character between left and right column.
#'
#'   Each style component can be either a predefined style string (e.g., "blue", "red_italic", "bold")
#'   or a function that takes a context list with/without a `value` element and returns the styled text.
#' @param align Controls the alignment of column values.
#'   Can be specified in three ways:
#'   - A single string: affects only the left column (e.g., "left", "center", "right").
#'   - A vector of two strings: affects both columns in order (e.g., c("left", "right")).
#'   - A list with named components: explicitly specifies alignment for each column
#'
#' @param ... Additional arguments (currently unused).
#'
#' @return This function does not return a value. It prints the formatted table to the console.
#'
#' @examples
#' # Create a sample data frame
#' df = data.frame(
#'     Category = c("A", "B", "C", "D", "E"),
#'     Value = c(10, 20, 30, 40, 50)
#' )
#'
#' # Display the table with a title and header
#' table_summary(df, title = "Sample Table", header = TRUE)
#'
#' # Split the table after the second row and center it
#' table_summary(df, l = 2, center_table = TRUE)
#'
#' # Use styling and alignment
#' table_summary(
#'     df, header = TRUE,
#'     style = list(
#'         left_col = "blue_bold",
#'         right_col = "red",
#'         title = "green",
#'         border_text = "yellow"
#'     ),
#'     align = c("center", "right")
#' )
#'
#' # Use custom styling with lambda functions
#' table_summary(
#'     df, header = TRUE,
#'     style = list(
#'         left_col = \(ctx) cli::col_red(ctx), # ctx$value is another option
#'         right_col = \(ctx) cli::col_blue(ctx)
#'     ),
#'     align = list(left_col = "left", right_col = "right")
#' )
#'
#' @export
table_summary = function(
    data,
    title = NULL,
    l = NULL,
    header = FALSE,
    center_table = FALSE,
    border_char = "-",
    style = list(),
    align = NULL,
    ...
) {
    if (!is.data.frame(data) || ncol(data) != 2) {
        stop("Input must be a data frame with exactly 2 columns")
    }

    col_names = colnames(data)
    data_matrix = as.matrix(data)

    n_rows = nrow(data)
    is_split = !is.null(l) && l < n_rows

    sep_width = 4
    if (!is.null(style) && !is.null(style$sep)) {
        sep_width = nchar(paste0(" ", style$sep, " "))
    }

    if (is_split) {
        left_table = data_matrix[1:l, ]
        right_table = data_matrix[(l+1):n_rows, ]

        left_left_width = max(nchar(left_table[, 1]), nchar(col_names[1]))
        left_right_width = max(nchar(left_table[, 2]), nchar(col_names[2]))
        right_left_width = max(nchar(right_table[, 1]), nchar(col_names[1]))
        right_right_width = max(nchar(right_table[, 2]), nchar(col_names[2]))

        left_table_width = left_left_width + left_right_width + 4 + sep_width
        right_table_width = right_left_width + right_right_width + 4 + sep_width
        full_width = left_table_width + right_table_width
    } else {
        left_width = max(nchar(data_matrix[, 1]), nchar(col_names[1]))
        right_width = max(nchar(data_matrix[, 2]), nchar(col_names[2]))
        full_width = left_width + right_width + 4 + sep_width
    }

    horizontal_line = strrep(border_char, full_width)
    styled_horizontal_line = horizontal_line

    if (!is.null(style) && !is.null(style$border_text)) {
        if (is.function(style$border_text)) {
            styled_horizontal_line = style$border_text(horizontal_line)
        } else if (is.character(style$border_text)) {
            styled_horizontal_line = switch(
                style$border_text,
                bold = cli::style_bold(horizontal_line),
                italic = cli::style_italic(horizontal_line),
                blue = cli::col_blue(horizontal_line),
                red = cli::col_red(horizontal_line),
                green = cli::col_green(horizontal_line),
                yellow = cli::col_yellow(horizontal_line),
                blue_bold = cli::col_blue(cli::style_bold(horizontal_line)),
                red_italic = cli::col_red(cli::style_italic(horizontal_line)),
                horizontal_line
            )
        }
    }

    prefix = ""
    if (center_table) {
        term_width = tryCatch({
            as.numeric(system("tput cols", intern = TRUE))
        }, error = function(e) {
            as.double(options("width"))
        })

        left_padding = max(0, floor((term_width - full_width) / 2))
        prefix = strrep(" ", left_padding)
    }

    if (!is.null(title)) {
        formatted_title = align_test(title, full_width)

        if (!is.null(style) && !is.null(style$title)) {
            if (is.function(style$title)) {
                formatted_title = style$title(formatted_title)
            } else if (is.character(style$title)) {
                formatted_title = switch(
                    style$title,
                    bold = cli::style_bold(formatted_title),
                    italic = cli::style_italic(formatted_title),
                    blue = cli::col_blue(formatted_title),
                    red = cli::col_red(formatted_title),
                    green = cli::col_green(formatted_title),
                    yellow = cli::col_yellow(formatted_title),
                    blue_bold = cli::col_blue(cli::style_bold(formatted_title)),
                    formatted_title
                )
            }
        }

        if (center_table) {
            cat("\n", prefix, formatted_title, "\n", sep = "")
        } else {
            cat("\n", formatted_title, "\n", sep = "")
        }
    }

    cat(prefix, styled_horizontal_line, "\n", sep = "")

    if (header) {
        if (is_split) {
            header_row = paste0(
                format_row_summary(col_names[1], col_names[2], left_left_width, left_right_width, style = style),
                "  ",
                format_row_summary(col_names[1], col_names[2], right_left_width, right_right_width, style = style)
            )
        } else {
            header_row = format_row_summary(col_names[1], col_names[2], left_width, right_width, style = style)
        }
        cat(prefix, header_row, "\n", sep = "")
        cat(prefix, styled_horizontal_line, "\n", sep = "")
    }

    if (!is_split) {
        for (i in 1:n_rows) {
            row_output = format_row_summary(
                data_matrix[i, 1],
                data_matrix[i, 2],
                left_width,
                right_width,
                align = align,
                style = style
            )
            cat(prefix, row_output, "\n", sep = "")
        }
    } else {
        for (i in 1:max(l, nrow(right_table))) {
            left_row = if (i <= l) left_table[i, ] else c("", "")
            right_row = if (i <= nrow(right_table)) right_table[i, ] else c("", "")

            left_output = format_row_summary(
                left_row[1],
                left_row[2],
                left_left_width,
                left_right_width,
                align = align,
                style = style
            )

            right_output = format_row_summary(
                right_row[1],
                right_row[2],
                right_left_width,
                right_right_width,
                align = align,
                style = style
            )

            row_output = paste0(left_output, "  ", right_output)
            cat(prefix, row_output, "\n", sep = "")
        }
    }

    cat(prefix, styled_horizontal_line, "\n", sep = "")
}

align_test = function(text, width) {
    padding = width - nchar(text)
    left_pad = floor(padding / 2)
    right_pad = ceiling(padding / 2)
    paste0(strrep(" ", left_pad), text, strrep(" ", right_pad))
}

format_row_summary = function(
    left,
    right,
    left_width,
    right_width,
    align = NULL,
    style = NULL
) {

    format_aligned = function(text, width, alignment) {
        switch(
            alignment,
            left = sprintf("%-*s", width, text),
            right = sprintf("%*s",  width, text),
            center = align_test(text, width),
            sprintf("%-*s", width, text)
        )
    }

    apply_cli_style = function(text, style_name) {
        switch(
            style_name,
            bold = cli::style_bold(text),
            italic = cli::style_italic(text),
            blue = cli::col_blue(text),
            red = cli::col_red(text),
            green = cli::col_green(text),
            yellow = cli::col_yellow(text),
            blue_bold = cli::col_blue(cli::style_bold(text)),
            red_italic = cli::col_red(cli::style_italic(text)),
            text
        )
    }

    left_align  = "left"
    right_align = "right"

    if (!is.null(align)) {
        if (is.character(align) && length(align) == 1) {
            left_align = right_align = align
        } else if (is.character(align) && length(align) == 2) {
            left_align  = align[1]
            right_align = align[2]
        } else if (is.list(align)) {
            if (!is.null(align$left_col))  left_align  = align$left_col
            if (!is.null(align$right_col)) right_align = align$right_col
        }
    }

    left_formatted = format_aligned(left,  left_width,  left_align)
    right_formatted = format_aligned(right, right_width, right_align)

    if (!is.null(style)) {
        if (!is.null(style$left_col)) {
            if (is.function(style$left_col)) {
                left_formatted = style$left_col(list(value = left_formatted))
            } else if (is.character(style$left_col)) {
                left_formatted = apply_cli_style(left_formatted, style$left_col)
            }
        }

        if (!is.null(style$right_col)) {
            if (is.function(style$right_col)) {
                right_formatted = style$right_col(list(value = right_formatted))
            } else if (is.character(style$right_col)) {
                right_formatted = apply_cli_style(right_formatted, style$right_col)
            }
        }
    }

    sep = "    "
    if (!is.null(style) && !is.null(style$sep)) {
        sep = paste0(" ", style$sep, " ")
    }

    paste0("  ", left_formatted, sep, right_formatted)
}

#' Generalized Neural Network Trainer
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `train_nn()` is a generic function for training neural networks with a
#' user-defined architecture via [nn_arch()]. Dispatch is based on the class
#' of `x`.
#'
#' Recommended workflow:
#' 1. Define architecture with [nn_arch()] (optional).
#' 2. Train with `train_nn()`.
#' 3. Predict with [predict.nn_fit()].
#'
#' All methods delegate to a shared implementation core after preprocessing.
#' When `architecture = NULL`, the model falls back to a plain feed-forward neural network
#' (`nn_linear`) architecture.
#'
#' @param x Dispatch is based on its current class:
#'   - `matrix`: used directly, no preprocessing applied.
#'   - `data.frame`: preprocessed via `hardhat::mold()`. `y` may be a vector /
#'     factor / matrix of outcomes, or a formula describing the outcome–predictor
#'     relationship within `x`.
#'   - `formula`: combined with `data` and preprocessed via `hardhat::mold()`.
#'   - `dataset`: a `torch` dataset object; batched via `torch::dataloader()`.
#'     This is the recommended interface for sequence/time-series and image data.
#' @param y Outcome data. Interpretation depends on the method:
#'   - For the `matrix` and `data.frame` methods: a numeric vector, factor, or
#'     matrix of outcomes.
#'   - For the `data.frame` method only: alternatively a formula of the form
#'     `outcome ~ predictors`, evaluated against `x`.
#'   - Ignored when `x` is a formula (outcome is taken from the formula) or a
#'     `dataset` (labels come from the dataset itself).
#' @param data A data frame. Required when `x` is a formula.
#' @param hidden_neurons Integer vector specifying the number of neurons in each
#'   hidden layer, e.g. `c(128, 64)` for two hidden layers. When `NULL` or missing,
#'   no hidden layers are used and the model reduces to a single linear mapping
#'   from inputs to outputs.
#' @param activations Activation function specification(s) for the hidden layers.
#'   See [act_funs()] for supported values. Recycled if a single value is given.
#' @param output_activation Optional activation function for the output layer.
#'   Defaults to `NULL` (no activation / linear output).
#' @param bias Logical. Whether to include bias terms in each layer. Default `TRUE`.
#' @param architecture An [nn_arch()] object specifying a custom architecture. Default
#'   `NULL`, which falls back to a standard feed-forward network.
#' @param arch Backward-compatible alias for `architecture`. If both are supplied,
#'   they must be identical.
#' @param flatten_input Logical or `NULL` (dataset method only). Controls whether
#'   each batch/sample is flattened to 2D before entering the model. `NULL`
#'   (default) auto-selects: `TRUE` when `architecture = NULL`, otherwise `FALSE`.
#' @param early_stopping An [early_stop()] object specifying early stopping
#'   behaviour, or `NULL` (default) to disable early stopping. When supplied,
#'   training halts if the monitored metric does not improve by at least
#'   `min_delta` for `patience` consecutive epochs.
#'   Example: `early_stopping = early_stop(patience = 10)`.
#' @param epochs Positive integer. Number of full passes over the training data.
#'   Default `100`.
#' @param batch_size Positive integer. Number of samples per mini-batch. Default `32`.
#' @param penalty Non-negative numeric. L1/L2 regularization strength (lambda).
#'   Default `0` (no regularization).
#' @param mixture Numeric in \[0, 1\]. Elastic net mixing parameter: `0` = pure
#'   ridge (L2), `1` = pure lasso (L1). Default `0`.
#' @param learn_rate Positive numeric. Step size for the optimizer. Default `0.001`.
#' @param optimizer Character. Optimizer algorithm. One of `"adam"` (default),
#'   `"sgd"`, or `"rmsprop"`.
#' @param optimizer_args Named list of additional arguments forwarded to the
#'   optimizer constructor (e.g. `list(momentum = 0.9)` for SGD). Default `list()`.
#' @param loss Character or function. Loss function used during training. Built-in
#'   options: `"mse"` (default), `"mae"`, `"cross_entropy"`, or `"bce"`. For
#'   classification tasks with a scalar label, `"cross_entropy"` is set
#'   automatically. Alternatively, supply a custom function or formula with
#'   signature `function(input, target)` returning a scalar `torch_tensor`.
#' @param validation_split Numeric in \[0, 1). Proportion of training data held
#'   out for validation. Default `0` (no validation set).
#' @param device Character. Compute device: `"cpu"`, `"cuda"`, or `"mps"`.
#'   Default `NULL`, which auto-detects the best available device.
#' @param verbose Logical. If `TRUE`, prints loss (and validation loss) at regular
#'   intervals during training. Default `FALSE`.
#' @param cache_weights Logical. If `TRUE`, stores a copy of the trained weight
#'   matrices in the returned object under `$cached_weights`. Default `FALSE`.
#' @param ... Additional arguments passed to specific methods.
#'
#' @return An object of class `"nn_fit"`, or one of its subclasses:
#'   - `c("nn_fit_tab", "nn_fit")` — returned by the `data.frame` and `formula` methods
#'   - `c("nn_fit_ds", "nn_fit")` — returned by the `dataset` method
#'
#'   All subclasses share a common structure. See **Details** for the list of
#'   components.
#'
#' @details
#' The returned `"nn_fit"` object is a named list with the following components:
#'
#' - `model` — the trained `torch::nn_module` object
#' - `fitted` — fitted values on the training data (or `NULL` for dataset fits)
#' - `loss_history` — numeric vector of per-epoch training loss, trimmed to
#'   actual epochs run (relevant when early stopping is active)
#' - `val_loss_history` — per-epoch validation loss, or `NULL` if
#'   `validation_split = 0`
#' - `n_epochs` — number of epochs actually trained
#' - `stopped_epoch` — epoch at which early stopping triggered, or `NA` if
#'   training ran to completion
#' - `hidden_neurons`, `activations`, `output_activation` — architecture spec
#' - `penalty`, `mixture` — regularization settings
#' - `feature_names`, `response_name` — variable names (tabular methods only)
#' - `no_x`, `no_y` — number of input features and output nodes
#' - `is_classification` — logical flag
#' - `y_levels`, `n_classes` — class labels and count (classification only)
#' - `device` — device the model is on
#' - `cached_weights` — list of weight matrices, or `NULL`
#' - `arch` — the `nn_arch` object used, or `NULL`
#'
#' @section Supported tasks and input formats:
#' `train_nn()` is task-agnostic by design (no explicit `task` argument).
#' Task behavior is determined by your input interface and architecture:
#' - **Tabular data**: use `matrix`, `data.frame`, or `formula` methods.
#' - **Time series**: use the `dataset` method with per-item tensors shaped as
#'   `[time, features]` (or your preferred convention) and a recurrent
#'   architecture via [nn_arch()].
#' - **Image classification**: use the `dataset` method with per-item tensors
#'   shaped for your first layer (commonly `[channels, height, width]` for
#'   `torch::nn_conv2d`). If your source arrays are channel-last, reorder in the
#'   dataset or via `input_transform`.
#'
#' @examples
#' \donttest{
#' if (torch::torch_is_installed()) {
#'     # Matrix method — no preprocessing
#'     model = train_nn(
#'         x = as.matrix(iris[, 2:4]),
#'         y = iris$Sepal.Length,
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 50
#'     )
#'
#'     # Data frame method — y as a vector
#'     model = train_nn(
#'         x = iris[, 2:4],
#'         y = iris$Sepal.Length,
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 50
#'     )
#'
#'     # Data frame method — y as a formula evaluated against x
#'     model = train_nn(
#'         x = iris,
#'         y = Sepal.Length ~ . - Species,
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 50
#'     )
#'
#'     # Formula method — outcome derived from formula
#'     model = train_nn(
#'         x = Sepal.Length ~ .,
#'         data = iris[, 1:4],
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 50
#'     )
#'
#'     # No hidden layers — linear model
#'     model = train_nn(
#'         x = Sepal.Length ~ .,
#'         data = iris[, 1:4],
#'         epochs = 50
#'     )
#'
#'     # Architecture object (nn_arch -> train_nn)
#'     mlp_arch = nn_arch(nn_name = "mlp_model")
#'     model = train_nn(
#'         x = Sepal.Length ~ .,
#'         data = iris[, 1:4],
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         architecture = mlp_arch,
#'         epochs = 50
#'     )
#'
#'     # Custom layer architecture
#'     custom_linear = torch::nn_module(
#'         "CustomLinear",
#'         initialize = function(in_features, out_features, bias = TRUE) {
#'             self$layer = torch::nn_linear(in_features, out_features, bias = bias)
#'         },
#'         forward = function(x) self$layer(x)
#'     )
#'     custom_arch = nn_arch(
#'         nn_name = "custom_linear_mlp",
#'         nn_layer = ~ custom_linear
#'     )
#'     model = train_nn(
#'         x = Sepal.Length ~ .,
#'         data = iris[, 1:4],
#'         hidden_neurons = c(16, 8),
#'         activations = "relu",
#'         architecture = custom_arch,
#'         epochs = 50
#'     )
#'
#'     # With early stopping
#'     model = train_nn(
#'         x = Sepal.Length ~ .,
#'         data = iris[, 1:4],
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 200,
#'         validation_split = 0.2,
#'         early_stopping = early_stop(patience = 10)
#'     )
#' }
#' }
#'
#' @seealso [predict.nn_fit()], [nn_arch()], [act_funs()], [early_stop()]
#' @name gen-nn-train
#' @export
train_nn = function(x, ...) {
    UseMethod("train_nn")
}


#' Normalize architecture argument for train_nn()
#' @noRd
.resolve_train_architecture = function(architecture = NULL, arch = NULL) {
    if (!is.null(architecture) && !is.null(arch) && !identical(architecture, arch)) {
        cli::cli_abort(c(
            "{.arg architecture} and {.arg arch} were both supplied with different values.",
            i = "Supply only {.arg architecture}, or provide identical values for both."
        ))
    }

    out = architecture %||% arch

    if (!is.null(out) && !inherits(out, "nn_arch")) {
        cli::cli_abort("{.arg architecture} must be an {.cls nn_arch} object created with {.fn nn_arch}.")
    }

    out
}


#' Normalize predict newdata argument
#' @noRd
.resolve_predict_newdata = function(newdata = NULL, new_data = NULL) {
    if (!is.null(new_data) && is.null(newdata)) {
        cli::cli_warn("{.arg new_data} is a legacy alias. Prefer {.arg newdata}.")
        return(new_data)
    }

    newdata
}

#' @rdname gen-nn-train
#'
#' @section Matrix method:
#' When `x` is supplied as a raw numeric matrix, no preprocessing is applied.
#' Data is passed directly to the shared `train_nn_impl` core.
#'
#' @export
train_nn.matrix =
    function(
        x,
        y,
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = TRUE,
        arch = NULL,
        architecture = NULL,
        early_stopping = NULL,
        epochs = 100,
        batch_size = 32,
        penalty = 0,
        mixture = 0,
        learn_rate = 0.001,
        optimizer = "adam",
        optimizer_args = list(),
        loss = "mse",
        validation_split = 0,
        device = NULL,
        verbose = FALSE,
        cache_weights = FALSE,
        ...
    ) {
        arch = .resolve_train_architecture(architecture = architecture, arch = arch)

        act_specs = eval_act_funs({{ activations }}, {{ output_activation }})
        activations = act_specs$activations
        output_activation = act_specs$output_activation

        train_nn_impl(
            x = x,
            y = y,
            hidden_neurons = hidden_neurons,
            activations = activations,
            output_activation = output_activation,
            bias = bias,
            arch = arch,
            early_stopping = early_stopping,
            epochs = epochs,
            batch_size = batch_size,
            penalty = penalty,
            mixture = mixture,
            learn_rate = learn_rate,
            optimizer = optimizer,
            optimizer_args = optimizer_args,
            loss = loss,
            validation_split = validation_split,
            device = device,
            verbose = verbose,
            cache_weights = cache_weights,
            fit_class = "nn_fit"
        )
    }


#' @rdname gen-nn-train
#'
#' @section Data frame method:
#' When `x` is a data frame, `y` can be either a vector / factor / matrix of
#' outcomes, or a formula of the form `outcome ~ predictors` evaluated against
#' `x`. Preprocessing is handled by `hardhat::mold()`.
#'
#' @export
train_nn.data.frame =
    function(
        x,
        y,
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = TRUE,
        arch = NULL,
        architecture = NULL,
        early_stopping = NULL,
        epochs = 100,
        batch_size = 32,
        penalty = 0,
        mixture = 0,
        learn_rate = 0.001,
        optimizer = "adam",
        optimizer_args = list(),
        loss = "mse",
        validation_split = 0,
        device = NULL,
        verbose = FALSE,
        cache_weights = FALSE,
        ...
    ) {
        arch = .resolve_train_architecture(architecture = architecture, arch = arch)

        act_specs = eval_act_funs({{ activations }}, {{ output_activation }})
        activations = act_specs$activations
        output_activation = act_specs$output_activation

        processed = if (rlang::is_formula(y)) hardhat::mold(y, x) else hardhat::mold(x, y)

        .train_nn_tab_impl(
            processed = processed,
            formula = NULL,
            hidden_neurons = hidden_neurons,
            activations = activations,
            output_activation = output_activation,
            bias = bias,
            arch = arch,
            early_stopping = early_stopping,
            epochs = epochs,
            batch_size = batch_size,
            penalty = penalty,
            mixture = mixture,
            learn_rate = learn_rate,
            optimizer = optimizer,
            optimizer_args = optimizer_args,
            loss = loss,
            validation_split = validation_split,
            device = device,
            verbose = verbose,
            cache_weights = cache_weights
        )
    }


#' @rdname gen-nn-train
#'
#' @section Formula method:
#' When `x` is a formula, `data` must be supplied as the data frame against
#' which the formula is evaluated. Preprocessing is handled by `hardhat::mold()`.
#'
#' @export
train_nn.formula =
    function(
        x,
        data,
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = TRUE,
        arch = NULL,
        architecture = NULL,
        early_stopping = NULL,
        epochs = 100,
        batch_size = 32,
        penalty = 0,
        mixture = 0,
        learn_rate = 0.001,
        optimizer = "adam",
        optimizer_args = list(),
        loss = "mse",
        validation_split = 0,
        device = NULL,
        verbose = FALSE,
        cache_weights = FALSE,
        ...
    ) {
        arch = .resolve_train_architecture(architecture = architecture, arch = arch)

        if (missing(data) || is.null(data)) {
            cli::cli_abort("{.arg data} must be provided when using the formula interface.")
        }

        act_specs = eval_act_funs({{ activations }}, {{ output_activation }})
        activations = act_specs$activations
        output_activation = act_specs$output_activation

        processed = hardhat::mold(x, data)

        .train_nn_tab_impl(
            processed = processed,
            formula = x,
            hidden_neurons = hidden_neurons,
            activations = activations,
            output_activation = output_activation,
            bias = bias,
            arch = arch,
            early_stopping = early_stopping,
            epochs = epochs,
            batch_size = batch_size,
            penalty = penalty,
            mixture = mixture,
            learn_rate = learn_rate,
            optimizer = optimizer,
            optimizer_args = optimizer_args,
            loss = loss,
            validation_split = validation_split,
            device = device,
            verbose = verbose,
            cache_weights = cache_weights
        )
    }


#' @rdname gen-nn-train
#' @export
train_nn.default = function(x, ...) {
    cli::cli_abort(c(
        "No {.fn train_nn} method for class {.cls {class(x)}}.",
        i = "Supported classes: {.cls matrix}, {.cls data.frame}, {.cls formula}, {.cls dataset}."
    ))
}


#' Preprocessing bridge for data.frame and formula methods
#' @keywords internal
.train_nn_tab_impl =
    function(
        processed,
        formula,
        hidden_neurons,
        activations,
        output_activation,
        bias,
        arch,
        early_stopping,
        epochs,
        batch_size,
        penalty,
        mixture,
        learn_rate,
        optimizer,
        optimizer_args,
        loss,
        validation_split,
        device,
        verbose,
        cache_weights
    ) {
        predictors = processed$predictors
        outcomes = processed$outcomes

        if (!is.matrix(predictors)) predictors = as.matrix(predictors)

        if (is.data.frame(outcomes)) {
            outcomes = if (ncol(outcomes) == 1L) outcomes[[1L]] else as.matrix(outcomes)
        } else if (!is.null(ncol(outcomes)) && ncol(outcomes) == 1L) {
            outcomes = outcomes[[1L]]
        }

        fit = train_nn_impl(
            x = predictors,
            y = outcomes,
            hidden_neurons = hidden_neurons,
            activations = activations,
            output_activation = output_activation,
            bias = bias,
            arch = arch,
            early_stopping = early_stopping,
            epochs = epochs,
            batch_size = batch_size,
            penalty = penalty,
            mixture = mixture,
            learn_rate = learn_rate,
            optimizer = optimizer,
            optimizer_args = optimizer_args,
            loss = loss,
            validation_split = validation_split,
            device = device,
            verbose = verbose,
            cache_weights = cache_weights,
            fit_class = "nn_fit_tab"
        )

        fit$blueprint = processed$blueprint
        if (!is.null(formula)) fit$formula = formula

        fit
    }


#' Shared core implementation
#' @keywords internal
train_nn_impl =
    function(
        x,
        y,
        hidden_neurons,
        activations = NULL,
        output_activation = NULL,
        bias = TRUE,
        arch = NULL,
        early_stopping = NULL,
        epochs = 100,
        batch_size = 32,
        penalty = 0,
        mixture = 0,
        learn_rate = 0.001,
        optimizer = "adam",
        optimizer_args = list(),
        loss = "mse",
        validation_split = 0,
        device = NULL,
        verbose = FALSE,
        cache_weights = FALSE,
        fit_class = "nn_fit"
    ) {
        if (!requireNamespace("torch", quietly = TRUE)) {
            cli::cli_abort("Package {.pkg torch} is required but not installed.")
        }

        if (missing(hidden_neurons) || is.null(hidden_neurons) || length(hidden_neurons) == 0L) {
            hidden_neurons = integer(0L)
        }

        # ---- Device ----
        device = if (is.null(device)) get_default_device() else validate_device(device)
        if (verbose) cli::cli_alert_info("Using device: {device}")

        validate_regularization(penalty, mixture)

        # ---- Input transform ----
        input_fn = if (!is.null(arch) && !is.null(arch$input_transform)) {
            rlang::as_function(arch$input_transform)
        } else {
            identity
        }

        # ---- Metadata ----
        feature_names = colnames(x) %||% paste0("V", seq_len(ncol(x)))
        response_name = names(y)[1L] %||% "y"
        is_classification = is.factor(y) || is.character(y)

        # ---- y encoding ----
        if (is_classification) {
            if (is.character(y)) y = as.factor(y)
            y_levels = levels(y)
            n_classes = length(y_levels)
            y_numeric = as.integer(y)
            no_y = n_classes
        } else {
            y_levels = NULL
            n_classes = NULL
            y_numeric = if (is.matrix(y)) y else as.numeric(y)
            no_y = if (is.matrix(y)) ncol(y) else 1L
        }

        no_x = ncol(x)
        n_obs = nrow(x)

        # ---- Validation split ----
        if (validation_split > 0 && validation_split < 1) {
            n_val = floor(n_obs * validation_split)
            val_idx = sample(n_obs, n_val)
            train_idx = setdiff(seq_len(n_obs), val_idx)

            x_train = x[train_idx, , drop = FALSE]
            y_train = y_numeric[train_idx]
            x_val = x[val_idx, , drop = FALSE]
            y_val = y_numeric[val_idx]
        } else {
            x_train = x
            y_train = y_numeric
            x_val = NULL
            y_val = NULL
        }

        # ---- Build model ----
        arch_args = if (!is.null(arch)) {
            args = unclass(arch)
            args$input_transform = NULL
            args
        } else {
            list()
        }

        arch_env = if (!is.null(arch)) attr(arch, "env") else parent.frame()
        # model_expr = do.call(
        #     nn_module_generator,
        #     c(
        #         list(
        #             hd_neurons = hidden_neurons,
        #             no_x = no_x,
        #             no_y = no_y,
        #             activations = activations,
        #             output_activation = output_activation,
        #             bias = bias
        #         ),
        #         arch_args,
        #         .env = arch_env
        #     )
        # )
        model_expr = rlang::exec(
            nn_module_generator,

            # Arguments
            hd_neurons = hidden_neurons,
            no_x = no_x,
            no_y = no_y,
            activations = activations,
            output_activation = output_activation,
            bias = bias,

            # Executing arguments from `nn_arch()`
            # To sync up with the current environment
            !!!arch_args,
            .env = arch_env
        )

        model = rlang::eval_tidy(model_expr)()
        model$to(device = device)

        # ---- Early stopping state ----
        es_state = if (!is.null(early_stopping)) {
            if (!inherits(early_stopping, "early_stop_spec")) {
                cli::cli_abort(c(
                    "{.arg early_stopping} must be created with {.fn early_stop}.",
                    i = "Example: {.code early_stopping = early_stop(patience = 10)}."
                ))
            }
            if (early_stopping$monitor == "val_loss" && validation_split == 0) {
                cli::cli_abort(c(
                    "Early stopping on {.val val_loss} requires a validation set.",
                    i = "Set {.arg validation_split} > 0, or use {.code monitor = 'train_loss'}."
                ))
            }
            list(
                spec = early_stopping,
                best_loss = Inf,
                best_weights = NULL,
                wait = 0L,
                stopped_epoch = NA_integer_
            )
        } else {
            NULL
        }

        # ---- Tensors ----
        .make_tensor = function(mat, is_y = FALSE) {
            t = torch::torch_tensor(
                mat,
                device = device,
                dtype = if (is_y && is_classification) torch::torch_long()
                else torch::torch_float32()
            )
            if (!is_y) input_fn(t) else t
        }

        x_train_t = .make_tensor(x_train)
        y_train_t = .make_tensor(
            if (is_classification || is.matrix(y_train)) y_train else matrix(y_train, ncol = 1L),
            is_y = TRUE
        )

        if (!is.null(x_val)) {
            x_val_t = .make_tensor(x_val)
            y_val_t = .make_tensor(
                if (is_classification || is.matrix(y_val)) y_val else matrix(y_val, ncol = 1L),
                is_y = TRUE
            )
        }

        # ---- Optimizer ----
        validate_optimizer(tolower(optimizer))
        optimizer_fn = get(paste0("optim_", tolower(optimizer)), envir = asNamespace("torch"))
        # opt = do.call(optimizer_fn, c(list(params = model$parameters, lr = learn_rate), optimizer_args))
        opt = rlang::exec(
            optimizer_fn,

            # Arguments
            params = model$parameters,
            lr = learn_rate,

            # Extra arguments from `optimizer_args`
            !!!optimizer_args
        )

        # ---- Loss function ----
        loss_fn = if (is.function(loss)) {
            .validate_loss_fn(loss)
        } else if (rlang::is_formula(loss)) {
            .validate_loss_fn(rlang::as_function(loss))
        } else {
            loss_name = tolower(loss)

            if (is_classification && loss_name == "mse") {
                loss_name = "cross_entropy"
                if (verbose) cli::cli_alert("Auto-detected classification task. Using {.val cross_entropy} loss.")
            }

            switch(
                loss_name,
                mse = function(input, target) torch::nnf_mse_loss(input, target),
                mae = function(input, target) torch::nnf_l1_loss(input, target),
                cross_entropy = function(input, target) torch::nnf_cross_entropy(input, target),
                bce = function(input, target) torch::nnf_binary_cross_entropy_with_logits(input, target),
                cli::cli_abort(
                    "Unknown loss: {.val {loss_name}}.",
                    i = "Use one of {.val mse}, {.val mae}, {.val cross_entropy}, {.val bce}, or supply a {.cls function}."
                )
            )
        }

        # ---- Training loop ----
        loss_history = numeric(epochs)
        val_loss_history = if (!is.null(x_val)) numeric(epochs) else NULL
        n_batches = ceiling(nrow(x_train) / batch_size)

        for (epoch in seq_len(epochs)) {
            model$train()
            epoch_loss = 0
            idx = sample(nrow(x_train))

            for (batch in seq_len(n_batches)) {
                start_i = (batch - 1L) * batch_size + 1L
                end_i = min(batch * batch_size, nrow(x_train))
                batch_idx = idx[start_i:end_i]

                x_batch = x_train_t[batch_idx, ]
                y_batch = y_train_t[batch_idx]

                opt$zero_grad()
                y_pred = model(x_batch)
                batch_loss = loss_fn(y_pred, y_batch)
                reg_loss = regularizer(model, penalty, mixture)
                total_loss = batch_loss + reg_loss
                total_loss$backward()
                opt$step()

                epoch_loss = epoch_loss + total_loss$item()
            }

            loss_history[epoch] = epoch_loss / n_batches

            # ---- Validation ----
            if (!is.null(x_val)) {
                model$eval()
                torch::with_no_grad({
                    y_val_pred = model(x_val_t)
                    val_loss = loss_fn(y_val_pred, y_val_t)
                    val_loss_history[epoch] = val_loss$item()
                })
            }

            # ---- Verbose ----
            if (verbose && (epoch %% max(1L, epochs %/% 10L) == 0L || epoch == epochs)) {
                msg = glue::glue(
                    "Epoch {epoch}/{epochs} - Loss: {round(loss_history[epoch], 4)}"
                )
                if (!is.null(val_loss_history))
                    msg = glue::glue("{msg} - Val Loss: {round(val_loss_history[epoch], 4)}")

                message(msg)
            }

            # ---- Early stopping check ----
            if (!is.null(es_state)) {
                monitored = if (es_state$spec$monitor == "val_loss") {
                    val_loss_history[epoch]
                } else {
                    loss_history[epoch]
                }

                if (monitored < es_state$best_loss - es_state$spec$min_delta) {
                    es_state$best_loss = monitored
                    es_state$wait = 0L
                    if (es_state$spec$restore_best_weights) {
                        es_state$best_weights = lapply(
                            model$parameters,
                            function(p) p$detach()$clone()
                        )
                    }
                } else {
                    es_state$wait = es_state$wait + 1L
                    if (es_state$wait >= es_state$spec$patience) {
                        es_state$stopped_epoch = epoch
                        if (verbose) cli::cli_alert_warning(
                            "Early stopping at epoch {epoch}. Best {es_state$spec$monitor}: {round(es_state$best_loss, 4)}."
                        )
                        break
                    }
                }
            }
        }

        # ---- Restore best weights ----
        if (!is.null(es_state) && es_state$spec$restore_best_weights && !is.null(es_state$best_weights)) {
            params = model$parameters
            torch::with_no_grad({
                for (i in seq_along(params)) {
                    params[[i]]$set_data(es_state$best_weights[[i]])
                }
            })
            if (verbose) {
                best_ep = if (!is.na(es_state$stopped_epoch)) {
                    es_state$stopped_epoch - es_state$spec$patience
                } else {
                    epoch
                }
                cli::cli_alert_info("Restored best weights from epoch {best_ep}.")
            }
        }

        # ---- Trim histories to actual epochs run ----
        actual_epochs = if (!is.null(es_state) && !is.na(es_state$stopped_epoch)) {
            es_state$stopped_epoch
        } else {
            epochs
        }
        loss_history = loss_history[seq_len(actual_epochs)]
        val_loss_history = if (!is.null(val_loss_history)) val_loss_history[seq_len(actual_epochs)] else NULL
        stopped_epoch = if (!is.null(es_state) && !is.na(es_state$stopped_epoch)) {
            es_state$stopped_epoch
        } else {
            NA_integer_
        }

        # ---- Fitted values ----
        model$eval()
        x_full_t = .make_tensor(x)
        fitted_tensor = torch::with_no_grad(model(x_full_t))

        fitted_values = if (is_classification) {
            fitted_probs = torch::nnf_softmax(fitted_tensor, dim = 2L)
            fitted_classes = torch::torch_argmax(fitted_probs, dim = 2L)
            factor(
                as.integer(fitted_classes$cpu()),
                levels = seq_along(y_levels),
                labels = y_levels
            )
        } else {
            convert = as.matrix(fitted_tensor$cpu())
            if (no_y == 1L) {
                as.vector(convert)
            } else {
                convert
            }
        }

        # ---- Weight caching ----
        # For visualizing weight distribution
        # Allow `cache_weights` to generate
        cached_weights = if (cache_weights) {
            tryCatch(
                lapply(model$parameters, function(p) as.matrix(p$cpu())),
                error = function(e) {
                    cli::cli_warn("Weight caching failed: {conditionMessage(e)}")
                    NULL
                }
            )
        } else {
            NULL
        }

        structure(
            list(
                model = model,
                fitted = fitted_values,
                loss_history = loss_history,
                val_loss_history = val_loss_history,
                n_epochs = actual_epochs,
                stopped_epoch = stopped_epoch,
                hidden_neurons = hidden_neurons,
                activations = activations,
                output_activation = output_activation,
                penalty = penalty,
                mixture = mixture,
                feature_names = feature_names,
                response_name = response_name,
                no_x = no_x,
                no_y = no_y,
                is_classification = is_classification,
                y_levels = y_levels,
                n_classes = n_classes,
                device = device,
                cached_weights = cached_weights,
                arch = arch
            ),
            class = unique(c(fit_class, "nn_fit"))
        )
    }

#' Predict from a trained neural network
#'
#' @description
#' Generate predictions from an `"nn_fit"` object produced by [train_nn()].
#'
#' Three S3 methods are registered:
#'
#' - `predict.nn_fit()` — base method for `matrix`-trained models.
#' - `predict.nn_fit_tab()` — extends the base method for tabular fits; runs new
#'   data through `hardhat::forge()` before predicting.
#' - `predict.nn_fit_ds()` — extends the base method for torch `dataset` fits.
#'
#' @param object A fitted model object returned by [train_nn()].
#' @param newdata New predictor data. Accepted forms depend on the method:
#'   - `predict.nn_fit()`: a numeric `matrix` or coercible object.
#'   - `predict.nn_fit_tab()`: a `data.frame` with the same columns used during
#'     training; preprocessing is applied automatically via `hardhat::forge()`.
#'   - `predict.nn_fit_ds()`: a `torch` `dataset`, numeric `array`, `matrix`, or
#'     `data.frame`.
#'   If `NULL`, the cached fitted values from training are returned (not
#'   available for `type = "prob"`).
#' @param new_data Legacy alias for `newdata`. Retained for compatibility.
#' @param type Character. Output type:
#'   - `"response"` (default): predicted class labels (factor) for
#'     classification, or a numeric vector / matrix for regression.
#'   - `"prob"`: a numeric matrix of class probabilities (classification only).
#' @param ... Currently unused; reserved for future extensions.
#'
#' @return
#' - **Regression**: a numeric vector (single output) or matrix (multiple outputs).
#' - **Classification**, `type = "response"`: a factor with levels matching those
#'   seen during training.
#' - **Classification**, `type = "prob"`: a numeric matrix with one column per
#'   class, columns named by class label.
#'
#' @seealso [train_nn()]
#' @name gen-nn-predict
#' @export
predict.nn_fit =
    function(
        object,
        newdata = NULL,
        new_data = NULL,
        type = "response",
        ...
    ) {
        if (!requireNamespace("torch", quietly = TRUE)) {
            cli::cli_abort("Package {.pkg torch} is required but not installed.")
        }

        if (!type %in% c("response", "prob")) {
            cli::cli_abort("{.arg type} must be one of {.val response} or {.val prob}.")
        }

        newdata = .resolve_predict_newdata(newdata = newdata, new_data = new_data)

        device = object$device
        input_fn = if (!is.null(object$arch) && !is.null(object$arch$input_transform)) {
            rlang::as_function(object$arch$input_transform)
        } else {
            identity
        }

        # Early return is a must, when `newdata` / `new_data` is not used
        # Do not attempt to use `type = "prob"` when classification is not applied
        if (is.null(newdata)) {
            if (type == "prob" && object$is_classification) {
                cli::cli_abort("Cannot compute probabilities without {.arg newdata}. Use fitted values instead.")
            }
            if (type == "prob" && !object$is_classification) {
                cli::cli_abort("{.arg type = 'prob'} is only available for classification models.")
            }
            return(object$fitted)
        }

        if (!is.matrix(newdata)) newdata = as.matrix(newdata)

        x_new_t = input_fn(torch::torch_tensor(newdata, dtype = torch::torch_float32(), device = device))
        object$model$eval()
        pred_tensor = torch::with_no_grad(object$model(x_new_t))

        predictions = if (object$is_classification) {
            probs = torch::nnf_softmax(pred_tensor, dim = 2L)

            if (type == "prob") {
                prob_matrix = as.matrix(probs$cpu())
                colnames(prob_matrix) = object$y_levels
                prob_matrix
            } else {
                factor(
                    as.integer(torch::torch_argmax(probs, dim = 2L)$cpu()),
                    levels = seq_along(object$y_levels),
                    labels = object$y_levels
                )
            }
        } else {
            if (type == "prob") {
                cli::cli_abort("{.arg type = 'prob'} is only available for classification models.")
            }

            convert = as.matrix(pred_tensor$cpu())
            if (object$no_y == 1L) {
                as.vector(convert)
            } else {
                convert
            }
        }

        predictions
    }


#' @rdname gen-nn-predict
#' @export
predict.nn_fit_tab =
    function(
        object,
        newdata = NULL,
        new_data = NULL,
        type = "response",
        ...
    ) {
        newdata = .resolve_predict_newdata(newdata = newdata, new_data = new_data)

        if (!is.null(newdata) && !is.null(object$blueprint)) {
            processed = hardhat::forge(newdata, object$blueprint)
            newdata = as.matrix(processed$predictors)
        }

        NextMethod()
    }


#' @keywords internal
#' @export
`$.nn_fit` = function(x, name) {
    if (name %in% names(x)) return(x[[name]])
    attr(x, name, exact = TRUE)
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

iris_x = as.matrix(iris[, 2:4])
iris_y = iris$Sepal.Length
iris_cls_x = as.matrix(iris[, 1:4])
iris_cls_y = iris$Species

test_that("train_nn() dispatches on matrix, data.frame, and formula", {
    skip_if_no_torch()
    expect_s3_class(
        train_nn(iris_x, iris_y, epochs = 5),
        "nn_fit"
    )
    expect_s3_class(
        train_nn(iris[, 2:4], iris_y, epochs = 5),
        c("nn_fit_tab", "nn_fit"), exact = TRUE
    )
    expect_s3_class(
        train_nn(Sepal.Length ~ ., data = iris[, 1:4], epochs = 5),
        c("nn_fit_tab", "nn_fit"), exact = TRUE
    )
})

# ---- Classification -----

test_that("train_nn() handles classification correctly", {
    skip_if_no_torch()
    m = train_nn(iris_cls_x, iris_cls_y, epochs = 5)
    expect_true(m$is_classification)
    expect_equal(m$y_levels, levels(iris_cls_y))
    expect_s3_class(m$fitted, "factor")
    expect_equal(levels(m$fitted), levels(iris_cls_y))
})

test_that("train_nn() errors on unsupported input types", {
    skip_if_no_torch()
    expect_error(train_nn("not valid"), class = "rlang_error")
    expect_error(train_nn(Sepal.Length ~ ., data = NULL), class = "rlang_error")
    expect_error(
        train_nn(iris_x, iris_y, arch = list(nn_name = "bad"), epochs = 5),
        class = "rlang_error"
    )
})

test_that("train_nn() return object has correct structure", {
    skip_if_no_torch()
    m = train_nn(iris_x, iris_y, epochs = 10, validation_split = 0.2)
    expect_named(m, c(
        "model", "fitted", "loss_history", "val_loss_history",
        "n_epochs", "stopped_epoch", "hidden_neurons", "activations",
        "output_activation", "penalty", "mixture", "feature_names",
        "response_name", "no_x", "no_y", "is_classification",
        "y_levels", "n_classes", "device", "cached_weights", "arch"
    ), ignore.order = TRUE)
    expect_length(m$loss_history, 10)
    expect_length(m$val_loss_history, 10)
    expect_equal(m$no_x, ncol(iris_x))
    expect_null(m$cached_weights)
    expect_true(is.na(m$stopped_epoch))
})

test_that("cache_weights stores weight matrices when TRUE", {
    skip_if_no_torch()
    m = train_nn(iris_x, iris_y, epochs = 5, cache_weights = TRUE)
    expect_type(m$cached_weights, "list")
})

test_that("train_nn() accepts various act_funs() syntaxes", {
    skip_if_no_torch()
    expect_no_error(
        train_nn(
            iris_x,
            iris_y,
            hidden_neurons = c(16, 8),
            activations = act_funs(relu, ),
            epochs = 5
        )
    )
    expect_no_error(
        train_nn(
            iris_x,
            iris_y,
            hidden_neurons = 16,
            activations = act_funs(elu[alpha = 0.5]),
            epochs = 5
        )
    )
    expect_no_error(
        train_nn(
            iris_x,
            iris_y,
            hidden_neurons = 16,
            activations = act_funs(new_act_fn(\(x) torch::torch_tanh(x))),
            epochs = 5
        )
    )
})

test_that("act_funs() errors on bad activation specs", {
    skip_if_no_torch()
    expect_error(act_funs(not_a_real_fn), class = "activation_not_found_error")
    expect_error(act_funs(relu[bad_param = 1]), class = "purrr_error_indexed")
})

# ---- Loss functions ----

test_that("train_nn() accepts built-in and custom loss functions", {
    skip_if_no_torch()
    expect_no_error(train_nn(iris_x, iris_y, loss = "mae", epochs = 5))
    expect_no_error(
        train_nn(
            iris_x,
            iris_y,
            loss = \(input, target) torch::nnf_mse_loss(input, target),
            epochs = 5
        )
    )
    expect_error(train_nn(iris_x, iris_y, loss = "not_a_loss", epochs = 5))
    expect_error(
        train_nn(iris_x, iris_y, loss = \(input, target) 42, epochs = 5),
        class = "loss_fn_output_error"
    )
    expect_error(
        train_nn(iris_x, iris_y, loss = \(x) torch::nnf_mse_loss(x, x), epochs = 5),
        class = "loss_fn_arity_error"
    )
})

# ---- Early stopping ----

describe("train_nn() early stopping", {
    it("runs cleanly and trims loss_history when triggered", {
        skip_if_no_torch()
        # min_delta = 1e10 forces early stopping to fire reliably
        es = early_stop(patience = 2, min_delta = 1e10, monitor = "val_loss")
        m = train_nn(iris_x, iris_y, epochs = 50,
                     validation_split = 0.2, early_stopping = es)
        expect_lt(length(m$loss_history), 50)
        expect_false(is.na(m$stopped_epoch))
    })

    it("errors when val_loss monitor is used without validation_split", {
        skip_if_no_torch()
        expect_error(
            train_nn(
                iris_x,
                iris_y,
                epochs = 10,
                early_stopping = early_stop(patience = 3, monitor = "val_loss")
            ),
            class = "rlang_error"
        )
    })

    it("errors when early_stopping is not an early_stop_spec", {
        skip_if_no_torch()
        expect_error(
            train_nn(iris_x, iris_y, epochs = 5, early_stopping = list(patience = 5)),
            class = "rlang_error"
        )
    })
})

test_that("predict.nn_fit() returns correct output types", {
    skip_if_no_torch()
    m_reg = train_nn(iris_x, iris_y, epochs = 5)
    m_cls = train_nn(iris_cls_x, iris_cls_y, epochs = 5)

    expect_equal(predict(m_reg), m_reg$fitted)
    expect_type(predict(m_reg, newdata = iris_x), "double")
    expect_s3_class(predict(m_cls, newdata = iris_cls_x), "factor")

    probs = predict(m_cls, newdata = iris_cls_x, type = "prob")
    expect_true(is.matrix(probs))
    expect_equal(rowSums(probs), rep(1, nrow(iris_cls_x)), tolerance = 1e-5)

    expect_error(predict(m_reg, newdata = iris_x, type = "prob"), class = "rlang_error")
    expect_error(predict(m_reg, newdata = iris_x, type = "bad"), class = "rlang_error")
    expect_error(predict(m_reg, newdata = iris_x, type = "good"), class = "rlang_error")
})

test_that("new_data is accepted as alias for newdata", {
    skip_if_no_torch()
    m = train_nn(iris_x, iris_y, epochs = 5)
    expect_warning(predict(m, new_data = iris_x))
    expect_equal(predict(m, newdata = iris_x), suppressWarnings(predict(m, new_data = iris_x)))
})

test_that("train_nn() handles edge case inputs", {
    skip_if_no_torch()
    m = train_nn(iris_x, iris_y, epochs = 5)
    expect_length(predict(m, newdata = iris_x[1, , drop = FALSE]), 1)

    expect_no_error(
        train_nn(iris_x[1:10, ], iris_y[1:10], batch_size = 50, epochs = 5)
    )

    expect_length(train_nn(iris_x, iris_y, epochs = 1)$loss_history, 1)

    m_multi = train_nn(as.matrix(iris[, 3:4]), as.matrix(iris[, 1:2]), epochs = 5)
    expect_equal(m_multi$no_y, 2L)
})

skip_if_no_torch = function() {
    testthat::skip_if_not_installed("torch")
    testthat::skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

make_dataset = function(n = 10L) {
    force(n)

    torch::dataset(
        name = "toy_cls_dataset",
        initialize = function() {
            self$x = torch::torch_tensor(
                matrix(rnorm(n * 4L), ncol = 4L),
                dtype = torch::torch_float32()
            )
            self$y = torch::torch_tensor(
                rep(c(1L, 2L), length.out = n),
                dtype = torch::torch_long()
            )
        },
        .getitem = function(i) {
            list(self$x[i, ], self$y[i])
        },
        .length = function() {
            self$x$size(1)
        }
    )()
}

test_that("dataset method trains and predicts", {
    skip_if_no_torch()

    set.seed(1)
    ds = make_dataset(10L)

    fit = kindling::train_nn(
        x = ds,
        hidden_neurons = c(8L),
        epochs = 2,
        batch_size = 4,
        learn_rate = 0.01,
        n_classes = 2,
        verbose = FALSE
    )

    expect_s3_class(fit, "nn_fit_ds")

    cls_pred = stats::predict(fit, ds)
    prob_pred = stats::predict(fit, ds, type = "prob")

    expect_s3_class(cls_pred, "factor")
    expect_length(cls_pred, length(ds))
    expect_equal(dim(prob_pred), c(length(ds), 2L))
    expect_equal(colnames(prob_pred), c("1", "2"))
})

test_that("dataset classification requires n_classes", {
    skip_if_no_torch()

    set.seed(2)
    ds = make_dataset(8L)

    expect_error(
        kindling::train_nn(
            x = ds,
            hidden_neurons = c(4L),
            epochs = 1,
            batch_size = 4,
            learn_rate = 0.01,
            verbose = FALSE
        ),
        "n_classes"
    )
})

test_that("dataset validation split guards against empty partitions", {
    skip_if_no_torch()

    set.seed(3)
    ds = make_dataset(1L)

    expect_error(
        kindling::train_nn(
            x = ds,
            hidden_neurons = c(4L),
            epochs = 1,
            batch_size = 1,
            learn_rate = 0.01,
            n_classes = 2,
            validation_split = 0.5,
            verbose = FALSE
        ),
        "empty train/validation"
    )
})

#' @param n_classes Positive integer. Number of output classes. Required when
#'   `x` is a `dataset` with scalar (classification) labels; ignored otherwise.
#'
#' @section Dataset method (`train_nn.dataset()`):
#' Trains a neural network directly on a `torch` dataset object. Batching and
#' lazy loading are handled by `torch::dataloader()`, making this method
#' well-suited for large datasets that do not fit entirely in memory.
#'
#' Architecture configuration follows the same contract as other `train_nn()`
#' methods via `architecture = nn_arch(...)` (or legacy `arch = ...`).
#' For non-tabular inputs (time series, images), set `flatten_input = FALSE` to
#' preserve tensor dimensions expected by recurrent or convolutional layers.
#'
#' Labels are taken from the second element of each dataset item (i.e.
#' `dataset[[i]][[2]]`), so `y` is ignored. When the label is a scalar tensor,
#' a classification task is assumed and `n_classes` must be supplied. The loss
#' is automatically switched to `"cross_entropy"` in that case.
#'
#' Fitted values are **not** cached in the returned object. Use
#' [predict.nn_fit_ds()] with `newdata` to obtain predictions after training.
#'
#' @examples
#' \donttest{
#' if (torch::torch_is_installed()) {
#'     # torch dataset method — labels come from the dataset itself
#'     iris_cls_dataset = torch::dataset(
#'         name = "iris_cls_dataset",
#'         
#'         initialize = function(data = iris) {
#'             self$x = torch::torch_tensor(
#'                 as.matrix(data[, 1:4]),
#'                 dtype = torch::torch_float32()
#'             )
#'             # Species is a factor; convert to integer (1-indexed -> keep as-is for cross_entropy)
#'             self$y = torch::torch_tensor(
#'                 as.integer(data$Species),
#'                 dtype = torch::torch_long()
#'             )
#'         },
#'         
#'         .getitem = function(i) {
#'             list(self$x[i, ], self$y[i])
#'         },
#'         
#'         .length = function() {
#'             self$x$size(1)
#'         }
#'     )()
#'     
#'     model_nn_ds = train_nn(
#'         x = iris_cls_dataset,
#'         hidden_neurons = c(32, 10),
#'         activations = "relu",
#'         epochs = 80,
#'         batch_size = 16,
#'         learn_rate = 0.01,
#'         n_classes = 3, # Iris dataset has only 3 species
#'         validation_split = 0.2,
#'         verbose = TRUE
#'     )
#'     
#'     pred_nn = predict(model_nn_ds, iris_cls_dataset)
#'     class_preds = c("Setosa", "Versicolor", "Virginica")[predict(model_nn_ds, iris_cls_dataset)]
#'     
#'     # Confusion Matrix
#'     table(actual = iris$Species, pred = class_preds)
#' }
#' }
#'
#' @rdname gen-nn-train
#' @export
train_nn.dataset = 
    function(
        x,
        y = NULL,
        hidden_neurons = NULL,
        activations = NULL,
        output_activation = NULL,
        bias = TRUE,
        arch = NULL,
        architecture = NULL,
        flatten_input = NULL,
        epochs = 100,
        batch_size = 32,
        penalty = 0,
        mixture = 0,
        learn_rate = 0.001,
        optimizer = "adam",
        optimizer_args = list(),
        loss = "mse",
        validation_split = 0,
        device = NULL,
        verbose = FALSE,
        cache_weights = FALSE,
        n_classes = NULL,
        ...
    ) 
{
    if (!requireNamespace("torch", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg torch} is required but not installed.")
    }

    if (length(x) < 1L) {
        cli::cli_abort("{.arg x} dataset is empty. Provide a dataset with at least one observation.")
    }

    arch = .resolve_train_architecture(architecture = architecture, arch = arch)
    flatten_input = .resolve_flatten_input(flatten_input = flatten_input, arch = arch)

    if (!flatten_input && is.null(arch)) {
        cli::cli_abort(c(
            "{.arg flatten_input = FALSE} requires a custom {.cls nn_arch}.",
            i = "Use {.arg architecture} to provide a layer stack that supports non-flattened tensors."
        ))
    }

    if (!is.null(y)) {
        cli::cli_warn("{.arg y} is ignored when {.arg x} is a dataset. Labels come from the dataset itself.")
    }

    act_specs = eval_act_funs({{ activations }}, {{ output_activation }})
    activations = act_specs$activations
    output_activation = act_specs$output_activation

    # ---- Infer dimensions from first batch ----
    first_item = .dataset_get_item(x, 1L)
    x_sample = first_item[[1]]
    y_sample = first_item[[2]]

    if (!inherits(x_sample, "torch_tensor") || !inherits(y_sample, "torch_tensor")) {
        cli::cli_abort(c(
            "Dataset items must be lists of two torch tensors: {.code list(x, y)}.",
            i = "Found an incompatible structure at {.code dataset[[1]]}."
        ))
    }

    no_x = as.integer(prod(x_sample$size()))

    is_classification = FALSE
    if (inherits(y_sample, "torch_tensor")) {
        y_dims = as.integer(y_sample$size())
        is_scalar_label = length(y_dims) == 0L || prod(y_dims) == 1L

        if (y_sample$dtype == torch::torch_long() && is_scalar_label) {
            is_classification = TRUE
            if (
                is.null(n_classes) ||
                    !is.numeric(n_classes) ||
                    length(n_classes) != 1L ||
                    is.na(n_classes) ||
                    n_classes < 2L ||
                    (n_classes %% 1L) != 0
            ) {
                cli::cli_abort("{.arg n_classes} must be a single integer >= 2 for classification datasets.")
            }

            no_y = as.integer(n_classes)
        } else {
            no_y = as.integer(if (length(y_dims) == 0L) 1L else prod(y_dims))

            if (!is.null(n_classes)) {
                cli::cli_warn("{.arg n_classes} is ignored when dataset labels are not scalar integer class IDs.")
            }
        }
    } else {
        cli::cli_abort("Dataset labels must be torch tensors.")
    }

    if (is_classification && loss == "mse") {
        loss = "cross_entropy"
        if (verbose) cli::cli_alert("Auto-detected classification task. Using cross_entropy loss.")
    }

    train_nn_impl_dataset(
        dataset = x,
        no_x = no_x,
        no_y = no_y,
        is_classification = is_classification,
        hidden_neurons = hidden_neurons,
        activations = activations,
        output_activation = output_activation,
        bias = bias,
        epochs = epochs,
        batch_size = batch_size,
        penalty = penalty,
        mixture = mixture,
        learn_rate = learn_rate,
        optimizer = optimizer,
        optimizer_args = optimizer_args,
        loss = loss,
        validation_split = validation_split,
        device = device,
        verbose = verbose,
        cache_weights = cache_weights,
        flatten_input = flatten_input,
        arch = arch,
        fit_class = "nn_fit_ds"
    )
}


#' Retrieve and validate one item from a torch dataset
#' @noRd
.dataset_get_item = function(dataset, i) {
    item = tryCatch(
        dataset[i],
        error = function(e) {
            cli::cli_abort(c(
                "Failed to read item {.val {i}} from dataset.",
                i = "Ensure {.arg x} implements {.code .getitem(i)} and {.code .length()}.",
                x = conditionMessage(e)
            ))
        }
    )

    if (!is.list(item) || length(item) < 2L) {
        cli::cli_abort(c(
            "Each dataset item must be a list with at least two elements: {.code list(x, y)}.",
            i = "Found {.cls {class(item)[1]}} at {.code dataset[[{i}]]}."
        ))
    }

    item
}


#' Normalize flatten_input argument for dataset training
#' @noRd
.resolve_flatten_input = function(flatten_input = NULL, arch = NULL) {
    if (is.null(flatten_input)) {
        return(is.null(arch))
    }

    if (!is.logical(flatten_input) || length(flatten_input) != 1L || is.na(flatten_input)) {
        cli::cli_abort("{.arg flatten_input} must be a single TRUE or FALSE value.")
    }

    flatten_input
}


#' train_nn implementation for torch datasets
#' @keywords internal
train_nn_impl_dataset = 
    function(
        dataset,
        no_x,
        no_y,
        is_classification,
        hidden_neurons,
        activations = NULL,
        output_activation = NULL,
        bias = TRUE,
        epochs = 100,
        batch_size = 32,
        penalty = 0,
        mixture = 0,
        learn_rate = 0.001,
        optimizer = "adam",
        optimizer_args = list(),
        loss = "mse",
        validation_split = 0,
        device = NULL,
        verbose = FALSE,
        cache_weights = FALSE,
        flatten_input = TRUE,
        arch = NULL,
        fit_class = "nn_fit_ds"
    ) 
{
    if (missing(hidden_neurons) || is.null(hidden_neurons) || length(hidden_neurons) == 0L) {
        hidden_neurons = integer(0L)
    }

    # ---- Device ----
    if (is.null(device)) {
        device = get_default_device()
    } else {
        device = validate_device(device)
    }

    if (verbose) cli::cli_alert_info("Using device: {device}")

    validate_regularization(penalty, mixture)

    if (!is.numeric(validation_split) || length(validation_split) != 1L || is.na(validation_split) ||
        validation_split < 0 || validation_split >= 1) {
        cli::cli_abort("{.arg validation_split} must be a single number in [0, 1).")
    }

    # ---- Input transform ----
    input_fn = if (!is.null(arch) && !is.null(arch$input_transform)) {
        rlang::as_function(arch$input_transform)
    } else {
        identity
    }

    # ---- Validation split ----
    n_obs = length(dataset)
    if (n_obs < 1L) {
        cli::cli_abort("{.arg dataset} is empty. Provide at least one observation.")
    }

    if (validation_split > 0 && validation_split < 1) {
        n_val = floor(n_obs * validation_split)

        if (n_val < 1L || n_val >= n_obs) {
            cli::cli_abort(c(
                "{.arg validation_split} yields an empty train/validation partition.",
                i = "Use more observations or a different split ratio."
            ))
        }

        val_idx = sample(n_obs, n_val)
        train_idx = setdiff(seq_len(n_obs), val_idx)

        train_ds = torch::dataset_subset(dataset, train_idx)
        val_ds = torch::dataset_subset(dataset, val_idx)
    } else {
        train_ds = dataset
        val_ds = NULL
    }

    # ---- Build model via nn_module_generator() ----
    arch_args = if (!is.null(arch)) {
        args = unclass(arch)
        args$input_transform = NULL
        args
    } else {
        list()
    }

    arch_env = if (!is.null(arch)) attr(arch, "env") else parent.frame()
    
    model_expr = do.call(
        nn_module_generator,
        c(
            list(
                hd_neurons = hidden_neurons,
                no_x = no_x,
                no_y = no_y,
                activations = activations,
                output_activation = output_activation,
                bias = bias
            ),
            arch_args,
            .env = arch_env
        )
    )
    # model = eval(model_expr)()
    model = rlang::eval_tidy(model_expr)()
    model$to(device = device)

    # ---- Dataloaders ----
    train_dl = torch::dataloader(train_ds, batch_size = batch_size, shuffle = TRUE)
    val_dl = if (!is.null(val_ds)) torch::dataloader(val_ds, batch_size = batch_size, shuffle = FALSE) else NULL

    # ---- Optimizer ----
    validate_optimizer(tolower(optimizer))
    optimizer_fn = get(paste0("optim_", tolower(optimizer)), envir = asNamespace("torch"))
    opt = do.call(
        optimizer_fn,
        c(list(params = model$parameters, lr = learn_rate), optimizer_args)
    )

    # ---- Loss function ----
    loss_name = tolower(loss)
    loss_fn = switch(
        loss_name,
        mse = function(input, target) torch::nnf_mse_loss(input, target),
        mae = function(input, target) torch::nnf_l1_loss(input, target),
        cross_entropy = function(input, target) torch::nnf_cross_entropy(input, target),
        bce = function(input, target) torch::nnf_binary_cross_entropy_with_logits(input, target),
        cli::cli_abort("Unknown loss function: {loss}")
    )
    is_cross_entropy = identical(loss_name, "cross_entropy")

    # ---- Training loop ----
    loss_history = numeric(epochs)
    val_loss_history = if (!is.null(val_dl)) numeric(epochs) else NULL

    for (epoch in seq_len(epochs)) {
        model$train()
        epoch_loss = 0
        n_batches = 0

        coro::loop(for (batch in train_dl) {
            x_batch = batch[[1]]$to(device = device)
            y_batch = batch[[2]]$to(device = device)

            if (is_classification && is_cross_entropy) {
                y_batch = y_batch$to(dtype = torch::torch_long())$view(c(-1))
            }

            if (flatten_input && length(x_batch$size()) > 2) {
                x_batch = x_batch$view(c(x_batch$size(1), -1))
            }

            x_batch = input_fn(x_batch)

            opt$zero_grad()
            y_pred = model(x_batch)
            loss = loss_fn(y_pred, y_batch)
            reg_loss = regularizer(model, penalty, mixture)
            total_loss = loss + reg_loss
            total_loss$backward()
            opt$step()

            epoch_loss = epoch_loss + total_loss$item()
            n_batches = n_batches + 1
        })

        loss_history[epoch] = epoch_loss / n_batches

        if (!is.null(val_dl)) {
            model$eval()
            val_epoch_loss = 0
            val_n_batches = 0

            torch::with_no_grad({
                coro::loop(for (batch in val_dl) {
                    x_batch = batch[[1]]$to(device = device)
                    y_batch = batch[[2]]$to(device = device)

                    if (is_classification && is_cross_entropy) {
                        y_batch = y_batch$to(dtype = torch::torch_long())$view(c(-1))
                    }

                    if (flatten_input && length(x_batch$size()) > 2) {
                        x_batch = x_batch$view(c(x_batch$size(1), -1))
                    }

                    x_batch = input_fn(x_batch)

                    y_pred = model(x_batch)
                    val_loss = loss_fn(y_pred, y_batch)
                    val_epoch_loss = val_epoch_loss + val_loss$item()
                    val_n_batches = val_n_batches + 1
                })
            })

            val_loss_history[epoch] = val_epoch_loss / val_n_batches
        }

        if (verbose && (epoch %% max(1L, epochs %/% 10L) == 0L || epoch == epochs)) {
            msg = sprintf("Epoch %d/%d - Loss: %.4f", epoch, epochs, loss_history[epoch])
            if (!is.null(val_loss_history)) {
                msg = paste0(msg, sprintf(" - Val Loss: %.4f", val_loss_history[epoch]))
            }
            message(msg)
        }
    }

    # ---- No fitted values for datasets (too expensive to materialize) ----
    fitted = NULL

    # ---- Weight caching ----
    cached_weights = NULL
    if (cache_weights) {
        cached_weights = tryCatch(
            lapply(model$parameters, function(p) as.matrix(p$cpu())),
            error = function(e) {
                cli::cli_warn("Weight caching failed: {conditionMessage(e)}")
                NULL
            }
        )
    }

    structure(
        list(
            model = model,
            fitted = fitted,
            loss_history = loss_history,
            val_loss_history = val_loss_history,
            n_epochs = epochs,
            hidden_neurons = hidden_neurons,
            activations = activations,
            output_activation = output_activation,
            penalty = penalty,
            mixture = mixture,
            feature_names = NULL,
            response_name = NULL,
            no_x = no_x,
            no_y = no_y,
            is_classification = is_classification,
            y_levels = if (is_classification) as.character(1:no_y) else NULL,
            n_classes = if (is_classification) no_y else NULL,
            device = device,
            cached_weights = cached_weights,
            flatten_input = flatten_input,
            arch = arch
        ),
        class = unique(c(fit_class, "nn_fit"))
    )
}


#' Predict method for dataset-trained neural networks
#'
#' @rdname gen-nn-predict
#' @keywords internal
#' @export
predict.nn_fit_ds = 
    function(
        object,
        newdata = NULL,
        new_data = NULL,
        type = "response",
        ...
    ) 
{
    newdata = .resolve_predict_newdata(newdata = newdata, new_data = new_data)

    if (!type %in% c("response", "prob")) {
        cli::cli_abort("{.arg type} must be one of {.val response} or {.val prob}.")
    }

    if (!object$is_classification && type == "prob") {
        cli::cli_abort("{.arg type = 'prob'} is only available for classification models.")
    }

    if (is.null(newdata)) {
        cli::cli_abort("Cannot compute fitted values for dataset fits. Provide {.arg newdata}.")
    }

    flatten_input = object$flatten_input %||% TRUE
    device = object$device
    input_fn = if (!is.null(object$arch) && !is.null(object$arch$input_transform)) {
        rlang::as_function(object$arch$input_transform)
    } else {
        identity
    }

    if (inherits(newdata, "dataset")) {
        dl = torch::dataloader(newdata, batch_size = 32, shuffle = FALSE)
        all_preds = list()

        object$model$eval()
        torch::with_no_grad({
            coro::loop(for (batch in dl) {
                x_batch = batch[[1]]$to(device = device)

                if (flatten_input && length(x_batch$size()) > 2) {
                    x_batch = x_batch$view(c(x_batch$size(1), -1))
                }

                x_batch = input_fn(x_batch)
                pred_tensor = object$model(x_batch)

                if (object$is_classification) {
                    probs = torch::nnf_softmax(pred_tensor, dim = 2L)
                    if (type == "prob") {
                        all_preds[[length(all_preds) + 1]] = as.matrix(probs$cpu())
                    } else {
                        pred_classes = torch::torch_argmax(probs, dim = 2L)
                        all_preds[[length(all_preds) + 1]] = as.integer(pred_classes$cpu())
                    }
                } else {
                    all_preds[[length(all_preds) + 1]] = as.matrix(pred_tensor$cpu())
                }
            })
        })

        # Combine batches
        if (object$is_classification) {
            if (type == "prob") {
                prob_matrix = do.call(rbind, all_preds)
                colnames(prob_matrix) = object$y_levels
                return(prob_matrix)
            } else {
                predictions = unlist(all_preds)
                predictions = factor(predictions,
                    levels = seq_along(object$y_levels),
                    labels = object$y_levels)
                return(predictions)
            }
        } else {
            predictions = do.call(rbind, all_preds)
            if (object$no_y == 1L) predictions = as.vector(predictions)
            return(predictions)
        }
    }

    if (is.data.frame(newdata)) {
        x_in = as.matrix(newdata)
    } else if (is.matrix(newdata) || is.array(newdata)) {
        x_in = newdata
    } else {
        cli::cli_abort(c(
            "Unsupported {.arg newdata} type for {.fn predict.nn_fit_ds}.",
            i = "Use a {.cls dataset}, matrix, array, or data.frame."
        ))
    }

    x_new_t = torch::torch_tensor(x_in, dtype = torch::torch_float32(), device = device)
    if (flatten_input && length(x_new_t$size()) > 2) {
        x_new_t = x_new_t$view(c(x_new_t$size(1), -1))
    }
    x_new_t = input_fn(x_new_t)

    object$model$eval()
    pred_tensor = torch::with_no_grad(object$model(x_new_t))

    if (object$is_classification) {
        probs = torch::nnf_softmax(pred_tensor, dim = 2L)

        if (type == "prob") {
            prob_matrix = as.matrix(probs$cpu())
            colnames(prob_matrix) = object$y_levels
            return(prob_matrix)
        }

        pred_classes = torch::torch_argmax(probs, dim = 2L)
        predictions = as.integer(pred_classes$cpu())
        predictions = factor(predictions,
            levels = seq_along(object$y_levels),
            labels = object$y_levels)
        return(predictions)
    }

    predictions = as.matrix(pred_tensor$cpu())
    if (object$no_y == 1L) predictions = as.vector(predictions)
    predictions
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

# ---- torch dataset ----

make_iris_dataset = function() {
    torch::dataset(
        initialize = function() {
            self$x = torch::torch_tensor(
                as.matrix(iris[, 1:4]),
                dtype = torch::torch_float32()
            )
            self$y = torch::torch_tensor(
                as.integer(iris$Species),
                dtype = torch::torch_long()
            )
        },
        .getitem = function(i) list(self$x[i, ], self$y[i]),
        .length = function() self$x$size(1)
    )()
}

make_reg_dataset = function() {
    torch::dataset(
        initialize = function() {
            self$x = torch::torch_tensor(
                as.matrix(iris[, 2:4]),
                dtype = torch::torch_float32()
            )
            self$y = torch::torch_tensor(
                as.matrix(iris[, 1, drop = FALSE]),
                dtype = torch::torch_float32()
            )
        },
        .getitem = function(i) list(self$x[i, ], self$y[i]),
        .length = function() self$x$size(1)
    )()
}

# ---- More for nn_arch() ----

test_that("nn_arch() returns correct classes and defaults", {
    arch = nn_arch()
    expect_s3_class(arch, "nn_arch")
    expect_s3_class(arch, "kindling_arch")
    expect_equal(arch$nn_name, "nnModule")
    expect_null(arch$nn_layer)
    expect_null(arch$out_nn_layer)
    expect_null(arch$input_transform)
    expect_true(is.environment(attr(arch, "env")))
})

test_that("nn_arch() stores all supplied arguments", {
    arch = nn_arch(
        nn_name = "MyGRU",
        nn_layer = "torch::nn_gru",
        out_nn_layer = "torch::nn_linear",
        nn_layer_args = list(batch_first = TRUE),
        input_transform = ~ .$unsqueeze(2)
    )
    expect_equal(arch$nn_name, "MyGRU")
    expect_equal(arch$nn_layer, "torch::nn_gru")
    expect_equal(arch$out_nn_layer, "torch::nn_linear")
    expect_equal(arch$nn_layer_args, list(batch_first = TRUE))
    expect_false(is.null(arch$input_transform))
})

test_that("nn_arch() captures caller environment", {
    my_env = environment()
    arch = nn_arch()
    expect_true(is.environment(attr(arch, "env")))
})

# ---- train_nn.dataset() ----

test_that("train_nn.dataset() trains a classification model", {
    skip_if_no_torch()
    ds = make_iris_dataset()
    m = train_nn(ds, hidden_neurons = c(16L, 8L), activations = "relu",
                 epochs = 5, n_classes = 3)
    expect_s3_class(m, "nn_fit_ds")
    expect_s3_class(m, "nn_fit")
    expect_true(m$is_classification)
    expect_equal(m$n_classes, 3L)
    expect_length(m$loss_history, 5)
})

test_that("train_nn.dataset() trains a regression model", {
    skip_if_no_torch()
    ds = make_reg_dataset()
    m = train_nn(ds, hidden_neurons = 16L, activations = "relu", epochs = 5)
    expect_s3_class(m, "nn_fit_ds")
    expect_false(m$is_classification)
})

test_that("train_nn.dataset() auto-switches loss to cross_entropy", {
    skip_if_no_torch()
    ds = make_iris_dataset()
    m = train_nn(ds, epochs = 5, n_classes = 3, loss = "mse")
    expect_s3_class(m, "nn_fit_ds")
})

test_that("train_nn.dataset() errors without n_classes for classification", {
    skip_if_no_torch()
    ds = make_iris_dataset()
    expect_error(train_nn(ds, epochs = 5), class = "rlang_error")
})

test_that("train_nn.dataset() warns when y is supplied", {
    skip_if_no_torch()
    ds = make_iris_dataset()
    expect_warning(
        train_nn(ds, y = 1:150, epochs = 5, n_classes = 3),
        class = "rlang_warning"
    )
})

test_that("train_nn.dataset() supports validation_split", {
    skip_if_no_torch()
    ds = make_reg_dataset()
    m = train_nn(ds, epochs = 5, validation_split = 0.2)
    expect_length(m$val_loss_history, 5)
})

test_that("train_nn.dataset() supports cache_weights", {
    skip_if_no_torch()
    ds = make_reg_dataset()
    m = train_nn(ds, epochs = 5, cache_weights = TRUE)
    expect_type(m$cached_weights, "list")
})

test_that("predict.nn_fit_ds() works with a dataset", {
    skip_if_no_torch()
    ds = make_iris_dataset()
    m = train_nn(ds, hidden_neurons = 16L, epochs = 5, n_classes = 3)
    preds = predict(m, newdata = ds)
    expect_s3_class(preds, "factor")
    expect_length(preds, 150)
})

test_that("predict.nn_fit_ds() type = 'prob' returns valid probability matrix", {
    skip_if_no_torch()
    ds = make_iris_dataset()
    m = train_nn(ds, hidden_neurons = 16L, epochs = 5, n_classes = 3)
    probs = predict(m, newdata = ds, type = "prob")
    expect_true(is.matrix(probs))
    expect_equal(ncol(probs), 3L)
    expect_equal(rowSums(probs), rep(1, 150), tolerance = 1e-5)
})

test_that("predict.nn_fit_ds() works with a matrix as newdata", {
    skip_if_no_torch()
    ds = make_reg_dataset()
    m = train_nn(ds, epochs = 5)
    preds = predict(m, newdata = as.matrix(iris[, 2:4]))
    expect_length(preds, 150)
})

test_that("predict.nn_fit_ds() errors when newdata is NULL", {
    skip_if_no_torch()
    ds = make_reg_dataset()
    m = train_nn(ds, epochs = 5)
    expect_error(predict(m), class = "rlang_error")
})

test_that("predict.nn_fit_ds() errors on type = 'prob' for regression", {
    skip_if_no_torch()
    ds = make_reg_dataset()
    m = train_nn(ds, epochs = 5)
    expect_error(
        predict(m, newdata = make_reg_dataset(), type = "prob"),
        class = "rlang_error"
    )
})

test_that("predict.nn_fit_ds() errors on invalid type", {
    skip_if_no_torch()
    ds = make_iris_dataset()
    m = train_nn(ds, epochs = 5, n_classes = 3)
    expect_error(
        predict(m, newdata = ds, type = "bad"),
        class = "rlang_error"
    )
})

test_that("train_nn.dataset() with nn_arch and flatten_input = FALSE", {
    skip_if_no_torch()
    gru_arch = nn_arch(
        nn_name = "GRU",
        nn_layer = "torch::nn_gru",
        layer_arg_fn = ~ if (.is_output) {
            list(.in, .out)
        } else {
            list(input_size = .in, hidden_size = .out, batch_first = TRUE)
        },
        out_nn_layer = "torch::nn_linear",
        forward_extract = ~ .[[1]],
        before_output_transform = ~ .[, .$size(2), ],
        input_transform = ~ .$unsqueeze(2)
    )
    ds = make_reg_dataset()
    m = train_nn(ds, hidden_neurons = 16L, epochs = 3,
                 architecture = gru_arch, flatten_input = FALSE)
    expect_s3_class(m, "nn_fit_ds")
})

test_that("train_nn.dataset() errors with flatten_input = FALSE and no arch", {
    skip_if_no_torch()
    ds = make_reg_dataset()
    expect_error(
        train_nn(ds, epochs = 3, flatten_input = FALSE),
        class = "rlang_error"
    )
})

#' `{kindling}`: Higher-level interface of torch package to auto-train neural networks
#'
#' The `{kindling}` package provides a unified, high-level interface that bridges
#' the `{torch}` and `{tidymodels}` ecosystems, making it easy to define, train,
#' and tune deep learning models using the familiar `tidymodels` workflow.
#'
#' @description
#' `{kindling}` enables R users to build and train deep neural networks such as:
#'
#' - Deep Neural Networks / (Deep) Feedforward Neural Networks (DNN / FFNN)
#' - Recurrent Neural Networks (RNN)
#'
#' It is mainly designed to generate code expressions for current architectures,
#' which reduces boilerplate `{torch}` code.
#' It also integrates seamlessly with machine learning workflows in R,
#' currently through `{tidymodels}` and components like `{parsnip}`,
#' `{recipes}`, and `{workflows}`, allowing an ergonomic interface for model
#' specification, training, and evaluation.
#'
#' Thus, the package supports hyperparameter tuning for:
#'
#' -  Number of hidden layers
#' -  Number of units per layer
#' -  Choice of activation functions
#'
#' Clarification: The hyperparameter tuning support is supported in version 0.1.0,
#' but `n_hlayer()` dial parameter is not supported. This changes after 0.2.0 release.
#'
#' @section How to use:
#' This package can be used at 3 levels:
#'
#' Level 1: Code generation
#'
#' ``` r
#' ffnn_generator(
#'     nn_name = "MyFFNN",
#'     hd_neurons = c(64, 32, 16),
#'     no_x = 10,
#'     no_y = 1,
#'     activations = 'relu'
#' )
#' ```
#'
#' Level 2: Direct Execution
#'
#' ``` r
#' ffnn(
#'     Species ~ .,
#'     data = iris,
#'     hidden_neurons = c(128, 64, 32),
#'     activations = 'relu',
#'     loss = "cross_entropy",
#'     epochs = 100
#' )
#' ```
#'
#' Level 3: Conventional tidymodels interface
#'
#' ``` r
#' # library(parsnip)
#' # library(kindling)
#' box::use(
#'    kindling[mlp_kindling, rnn_kindling, act_funs, args],
#'    parsnip[fit, augment],
#'    yardstick[metrics],
#'    mlbench[Ionosphere] # data(Ionosphere, package = "mlbench")
#' )
#'
#' # Remove V2 as it's all zeros
#' ionosphere_data = Ionosphere[, -2]
#'
#' # MLP example
#' mlp_kindling(
#'     mode = "classification",
#'     hidden_neurons = c(128, 64),
#'     activations = act_funs(relu, softshrink = args(lambd = 0.5)),
#'     epochs = 100
#' ) |>
#'     fit(Class ~ ., data = ionosphere_data) |>
#'     augment(new_data = ionosphere_data) |>
#'     metrics(truth = Class, estimate = .pred_class)
#' #> A tibble: 2 × 3
#' #>   .metric  .estimator .estimate
#' #>   <chr>    <chr>          <dbl>
#' #> 1 accuracy binary         0.989
#' #> 2 kap      binary         0.975
#'
#' # RNN example (toy usage on non-sequential data)
#' rnn_kindling(
#'     mode = "classification",
#'     hidden_neurons = c(128, 64),
#'     activations = act_funs(relu, elu),
#'     epochs = 100,
#'     rnn_type = "gru"
#' ) |>
#'     fit(Class ~ ., data = ionosphere_data) |>
#'     augment(new_data = ionosphere_data) |>
#'     metrics(truth = Class, estimate = .pred_class)
#' #> A tibble: 2 × 3
#' #>   .metric  .estimator .estimate
#' #>   <chr>    <chr>          <dbl>
#' #> 1 accuracy binary         0.641
#' #> 2 kap      binary         0
#' ```
#'
#' @section Main Features:
#' -   Code generation of `{torch}` expression
#' -   Multiple architectures available: feedforward networks (MLP/DNN/FFNN) and recurrent variants (RNN, LSTM, GRU)
#' -   Native support for `{tidymodels}` workflows and pipelines
#' -   Fine-grained control over network depth, layer sizes, and activation functions
#' -   GPU acceleration support via `{torch}` tensors
#'
#' @references
#'
#' Falbel D, Luraschi J (2023). _torch: Tensors and Neural Networks with 'GPU'
#' Acceleration_. R package version 0.13.0,
#' \url{https://torch.mlverse.org}, \url{https://github.com/mlverse/torch}.
#'
#' Wickham H (2019). _Advanced R_, 2nd edition. Chapman and Hall/CRC.
#' ISBN 978-0815384571, \url{https://adv-r.hadley.nz/}.
#'
#' Goodfellow I, Bengio Y, Courville A (2016). _Deep Learning_. MIT Press.
#' \url{https://www.deeplearningbook.org/}.
#'
#' @section License:
#' MIT + file LICENSE
#'
#' @docType package
#' @name kindling
#' @keywords internal
"_PACKAGE"

# Suppress R CMD check notes for NSE variables used in torch modules
# These variables (self, x) are used in non-standard evaluation contexts
# within torch::nn_module definitions and are intentional
utils::globalVariables(c("self", "x"))

#' Architecture specification for train_nn()
#'
#' @description
#' `nn_arch()` defines an architecture specification object consumed by
#' [train_nn()] via `architecture` (or legacy `arch`).
#'
#' Conceptual workflow:
#' 1. Define architecture with `nn_arch()`.
#' 2. Train with `train_nn(..., architecture = <nn_arch>)`.
#' 3. Predict with `predict()`.
#'
#' Architecture fields mirror `nn_module_generator()` and are passed through
#' without additional branching logic.
#'
#' @param nn_name Character. Name of the generated module class. Default `"nnModule"`.
#' @param nn_layer Layer type. See `nn_module_generator()`. Default `NULL` (`nn_linear`).
#' @param out_nn_layer Optional. Layer type forced on the last layer. Default `NULL`.
#' @param nn_layer_args Named list. Additional arguments passed to every layer constructor.
#'   Default `list()`.
#' @param layer_arg_fn Formula or function. Generates per-layer constructor arguments.
#'   Default `NULL` (FFNN-style: `list(in_dim, out_dim, bias = bias)`).
#' @param forward_extract Formula or function. Processes layer output in the forward pass.
#'   Default `NULL`.
#' @param before_output_transform Formula or function. Transforms input before the output
#'   layer. Default `NULL`.
#' @param after_output_transform Formula or function. Transforms output after the output
#'   layer. Default `NULL`.
#' @param last_layer_args Named list or formula. Extra arguments for the output layer only.
#'   Default `list()`.
#' @param input_transform Formula or function. Transforms the entire input tensor
#'   before training begins. Applied once to the full dataset tensor, not per-batch.
#'   Transforms must therefore be independent of batch size. Safe examples:
#'   `~ .$unsqueeze(2)` (RNN sequence dim), `~ .$unsqueeze(1)` (CNN channel dim).
#'   Avoid transforms that reshape based on `.$size(1)` as this will reflect the
#'   full dataset size, not the mini-batch size.
#'
#' @return An object of class `c("nn_arch", "kindling_arch")`, implemented as a
#'   named list of `nn_module_generator()` arguments with an `"env"` attribute
#'   capturing the calling environment for custom layer resolution.
#'
#' @examples
#' \donttest{
#' if (torch::torch_is_installed()) {
#'     # Standard architecture object for train_nn()
#'     std_arch = nn_arch(nn_name = "mlp_model")
#'
#'     # GRU architecture spec
#'     gru_arch = nn_arch(
#'         nn_name = "GRU",
#'         nn_layer = "torch::nn_gru",
#'         layer_arg_fn = ~ if (.is_output) {
#'             list(.in, .out)
#'         } else {
#'             list(input_size = .in, hidden_size = .out, batch_first = TRUE)
#'         },
#'         out_nn_layer = "torch::nn_linear",
#'         forward_extract = ~ .[[1]],
#'         before_output_transform = ~ .[, .$size(2), ],
#'         input_transform = ~ .$unsqueeze(2)
#'     )
#'
#'     # Custom layer architecture (resolved from calling environment)
#'     custom_linear = torch::nn_module(
#'         "CustomLinear",
#'         initialize = function(in_features, out_features, bias = TRUE) {
#'             self$layer = torch::nn_linear(in_features, out_features, bias = bias)
#'         },
#'         forward = function(x) self$layer(x)
#'     )
#'     custom_arch = nn_arch(
#'         nn_name = "CustomMLP",
#'         nn_layer = ~ custom_linear
#'     )
#'
#'     model = train_nn(
#'         Sepal.Length ~ .,
#'         data = iris[, 1:4],
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 50,
#'         architecture = gru_arch
#'     )
#' }
#' }
#'
#' @export
nn_arch = function(
        nn_name = "nnModule",
        nn_layer = NULL,
        out_nn_layer = NULL,
        nn_layer_args = list(),
        layer_arg_fn = NULL,
        forward_extract = NULL,
        before_output_transform = NULL,
        after_output_transform = NULL,
        last_layer_args = list(),
        input_transform = NULL
) {
    struc = structure(
        list(
            nn_name = nn_name,
            nn_layer = nn_layer,
            out_nn_layer = out_nn_layer,
            nn_layer_args = nn_layer_args,
            layer_arg_fn = layer_arg_fn,
            forward_extract = forward_extract,
            before_output_transform = before_output_transform,
            after_output_transform = after_output_transform,
            last_layer_args = last_layer_args,
            input_transform = input_transform
        ),
        class = c("nn_arch", "kindling_arch")
    )
    attr(struc, "env") = rlang::caller_env()
    struc
}

#' Display `nn_arch()` configuration
#'
#' @param x An object of class "nn_arch"
#' @param ... Additional arguments (unused)
#'
#' @return No return value, just the information
#'
#' @keywords internal
#' @export
print.nn_arch = function(x, ...) {
    cli::cli_h3("Neural Network Architecture Spec")
    cli::cli_bullets(c(
        "*" = "Name:            {x$nn_name}",
        "*" = "Layer:           {x$nn_layer %||% 'nn_linear (default)'}",
        "*" = "Out layer:       {x$out_nn_layer %||% 'same as nn_layer'}",
        "*" = "Input transform: {if (is.null(x$input_transform)) 'none' else 'yes'}"
    ))
    invisible(x)
}

#' Early Stopping Specification
#' 
#' @description
#' `early_stop()` is a helper function to be supplied on `early_stopping` arguments. 
#' 
#' @param patience Integer. Epochs to wait after last improvement. Default `5`.
#' @param min_delta Numeric. Minimum improvement to qualify as better. Default `1e-4`.
#' @param restore_best_weights Logical. Restore weights from best epoch. Default `TRUE`.
#' @param monitor Character. Metric to monitor. One of `"val_loss"` (default) or `"train_loss"`.
#'
#' @return An object of class `"early_stop_spec"`.
#' @export
early_stop = 
    function(
        patience = 5L,
        min_delta = 1e-4,
        restore_best_weights = TRUE,
        monitor = "val_loss"
    ) 
{
    monitor = rlang::arg_match(monitor, c("val_loss", "train_loss"))
    
    if (!is.numeric(patience) || patience < 1L) {
        cli::cli_abort("{.arg patience} must be a positive integer.")
    }
    if (!is.numeric(min_delta) || min_delta < 0) {
        cli::cli_abort("{.arg min_delta} must be non-negative.")
    }
    
    structure(
        list(
            patience = as.integer(patience),
            min_delta = min_delta,
            restore_best_weights = restore_best_weights,
            monitor = monitor
        ),
        class = "early_stop_spec"
    )
}

#' @export
print.early_stop_spec = function(x, ...) {
    cli::cli_h3("Early Stopping Spec")
    cli::cli_bullets(c(
        "*" = "Monitor: {x$monitor}",
        "*" = "Patience: {x$patience}",
        "*" = "Min delta: {x$min_delta}",
        "*" = "Restore best weights: {x$restore_best_weights}"
    ))
    invisible(x)
}

#' Plot training loss history for a fitted neural network
#'
#' Creates a line plot of training (and optionally validation) loss over epochs
#' for a model of class `nn_fit`.
#'
#' @param object A fitted model of class `nn_fit`, as returned by
#'   [train_nn()], [ffnn()], or [rnn()].
#' @param ... Additional arguments (currently unused).
#'
#' @return A [ggplot2::ggplot()] object showing loss vs epoch.
#'
#' @examples
#' \donttest{
#'   if (torch::torch_is_installed()) {
#'     m = train_nn(
#'       as.matrix(iris[, 2:4]), iris$Sepal.Length,
#'       epochs = 5
#'     )
#'     ggplot2::autoplot(m)
#'   }
#' }
#'
#' @keywords internal
#' @export
#' @method autoplot nn_fit
autoplot.nn_fit = function(object, ...) {
    n = length(object$loss_history)
    epochs = seq_len(n)

    if (!is.null(object$val_loss_history)) {
        long = data.frame(
            epoch = c(epochs, epochs),
            loss = c(object$loss_history, object$val_loss_history),
            set = rep(c("Training", "Validation"), each = n)
        )
    } else {
        long = data.frame(
            epoch = epochs,
            loss = object$loss_history,
            set = "Training"
        )
    }

    p = ggplot2::ggplot(long, ggplot2::aes(x = epoch, y = loss, colour = set)) +
        ggplot2::geom_line() +
        ggplot2::labs(
            x = "Epoch",
            y = "Loss",
            title = "Training History",
            colour = NULL
        ) +
        ggplot2::theme_minimal()

    if (is.null(object$val_loss_history)) {
        p = p + ggplot2::guides(colour = "none")
    }

    if (!is.na(object$stopped_epoch)) {
        p = p +
            ggplot2::geom_vline(
                xintercept = object$stopped_epoch,
                linetype = "dashed",
                colour = "grey40"
            ) +
            ggplot2::annotate(
                "text",
                x = object$stopped_epoch,
                y = max(long$loss, na.rm = TRUE),
                label = paste("early stop:", object$stopped_epoch),
                hjust = -0.05,
                size = 3,
                colour = "grey40"
            )
    }

    p
}

#' @rdname autoplot.nn_fit
#' @param x A fitted model of class `nn_fit`, as returned by
#'   [train_nn()], [ffnn()], or [rnn()].
#' @export
#' @method plot nn_fit
plot.nn_fit = function(x, ...) {
    print(ggplot2::autoplot(x, ...))
    invisible(x)
}

#' Plot prediction diagnostics for a fitted neural network
#'
#' @description
#' Produces diagnostic plots comparing fitted values against actual (true)
#' response values for a fitted `nn_fit` model.
#'
#' - **Regression (single output)**: returns a named list with two panels --
#'   residuals vs fitted and actual vs fitted.
#' - **Regression (multi-output)**: returns a named list with one actual vs
#'   fitted panel per output column.
#' - **Classification**: returns a single confusion matrix heatmap.
#'
#' @param object A fitted model of class `nn_fit`, as returned by
#'   [train_nn()], [ffnn()], or [rnn()].
#' @param actual A vector of true response values, the same length as the
#'   training data used to fit `object`.
#' @param ... Additional arguments (currently unused).
#'
#' @return For regression, a named list of [ggplot2::ggplot()] objects (one per
#'   diagnostic panel). For classification, a single [ggplot2::ggplot()]
#'   confusion matrix heatmap.
#'
#' @examples
#' \donttest{
#'   if (torch::torch_is_installed()) {
#'     # Regression
#'     m = train_nn(
#'       as.matrix(iris[, 2:4]), iris$Sepal.Length,
#'       epochs = 5
#'     )
#'     autoplot_diagnostics(m, actual = iris$Sepal.Length)
#'
#'     # Classification
#'     m_cls = train_nn(
#'       as.matrix(iris[, 1:4]), iris$Species,
#'       epochs = 5
#'     )
#'     autoplot_diagnostics(m_cls, actual = iris$Species)
#'   }
#' }
#'
#' @export
autoplot_diagnostics = function(object, actual, ...) {
    if (missing(actual)) {
        cli::cli_abort(
            "Argument {.arg actual} is required: supply the vector of true response values."
        )
    }
    if (is.null(object$fitted)) {
        cli::cli_abort(c(
            "Prediction diagnostics require fitted values, which are not available for dataset fits ({.cls nn_fit_ds}).",
            "i" = "Use {.fn predict} with held-out data instead."
        ))
    }

    n_fitted = if (is.matrix(object$fitted)) nrow(object$fitted) else length(object$fitted)
    if (length(actual) != n_fitted) {
        cli::cli_abort(
            "{.arg actual} has length {length(actual)} but {.field fitted} has length {n_fitted}: they must match."
        )
    }

    if (object$is_classification) {
        actual_fac = factor(actual, levels = object$y_levels)
        tbl = table(actual = actual_fac, predicted = object$fitted)
        df = as.data.frame(tbl, stringsAsFactors = FALSE)
        names(df)[names(df) == "Freq"] = "n"
        row_totals = tapply(df$n, df$actual, sum)
        df$prop = df$n / row_totals[as.character(df$actual)]

        ggplot2::ggplot(df, ggplot2::aes(x = predicted, y = actual, fill = prop)) +
            ggplot2::geom_tile() +
            ggplot2::geom_text(ggplot2::aes(label = n), colour = "black") +
            ggplot2::scale_fill_gradient(low = "white", high = "steelblue", guide = "none") +
            ggplot2::labs(x = "Predicted", y = "Actual", title = "Confusion Matrix") +
            ggplot2::theme_minimal()
    } else if (object$no_y > 1L) {
        actual_mat = if (is.matrix(actual)) actual else matrix(as.numeric(actual), ncol = 1L)
        col_labels = colnames(object$fitted)
        if (is.null(col_labels)) col_labels = paste0("output_", seq_len(object$no_y))

        plots = lapply(seq_len(object$no_y), function(j) {
            df = data.frame(
                fitted_val = as.numeric(object$fitted[, j]),
                actual_val = as.numeric(actual_mat[, j])
            )
            ggplot2::ggplot(df, ggplot2::aes(x = fitted_val, y = actual_val)) +
                ggplot2::geom_point() +
                ggplot2::geom_abline(slope = 1, intercept = 0, linetype = "dashed") +
                ggplot2::labs(
                    x = "Fitted values",
                    y = "Actual values",
                    title = col_labels[j]
                ) +
                ggplot2::theme_minimal()
        })
        names(plots) = col_labels

        cli::cli_inform(
            "Returning one plot per output column. Install {.pkg patchwork} to combine them."
        )
        plots
    } else {
        fitted_vals = as.numeric(object$fitted)
        df = data.frame(
            fitted_val = fitted_vals,
            residuals = as.numeric(actual) - fitted_vals,
            actual_val = as.numeric(actual)
        )

        p1 = ggplot2::ggplot(df, ggplot2::aes(x = fitted_val, y = residuals)) +
            ggplot2::geom_point() +
            ggplot2::geom_hline(yintercept = 0, linetype = "dashed") +
            ggplot2::labs(x = "Fitted values", y = "Residuals") +
            ggplot2::theme_minimal()

        p2 = ggplot2::ggplot(df, ggplot2::aes(x = fitted_val, y = actual_val)) +
            ggplot2::geom_point() +
            ggplot2::geom_abline(slope = 1, intercept = 0, linetype = "dashed") +
            ggplot2::labs(x = "Fitted values", y = "Actual values") +
            ggplot2::theme_minimal()

        cli::cli_inform(
            "Returning two plots. Install {.pkg patchwork} and use {.code p[[1]] + p[[2]]} to combine them."
        )
        list(residuals_vs_fitted = p1, actual_vs_fitted = p2)
    }
}

#' @rdname autoplot_diagnostics
#' @export
plot_diagnostics = function(object, actual, ...) {
    result = autoplot_diagnostics(object, actual, ...)
    if (inherits(result, "gg")) {
        print(result)
    } else if (is.list(result)) {
        lapply(result, print)
    }
    invisible(object)
}

skip_if_no_torch = function() {
    skip_if_not_installed("torch")
    skip_if_not(torch::torch_is_installed(), "Torch backend not available")
}

iris_x = as.matrix(iris[, 2:4])
iris_y = iris$Sepal.Length
iris_cls_x = as.matrix(iris[, 1:4])
iris_cls_y = iris$Species

make_reg_ds = function() {
    torch::dataset(
        initialize = function() {
            self$x = torch::torch_tensor(
                as.matrix(iris[, 2:4]),
                dtype = torch::torch_float32()
            )
            self$y = torch::torch_tensor(
                as.matrix(iris[, 1, drop = FALSE]),
                dtype = torch::torch_float32()
            )
        },
        .getitem = function(i) list(self$x[i, ], self$y[i]),
        .length = function() self$x$size(1)
    )()
}

test_that("autoplot() returns a ggplot for a basic fit (no validation)", {
    skip_if_no_torch()
    skip_if_not_installed("ggplot2")
    m = train_nn(iris_x, iris_y, epochs = 5)
    p = ggplot2::autoplot(m)
    expect_s3_class(p, "ggplot")
})

test_that("autoplot() returns a ggplot for a fit with validation_split", {
    skip_if_no_torch()
    skip_if_not_installed("ggplot2")
    m = train_nn(iris_x, iris_y, epochs = 5, validation_split = 0.2)
    p = ggplot2::autoplot(m)
    expect_s3_class(p, "ggplot")
})

test_that("autoplot() includes a geom_vline layer when early stopping fires", {
    skip_if_no_torch()
    skip_if_not_installed("ggplot2")
    es = early_stop(patience = 2, min_delta = 1e10, monitor = "val_loss")
    m = train_nn(iris_x, iris_y, epochs = 50, validation_split = 0.2, early_stopping = es)
    expect_false(is.na(m$stopped_epoch))
    p = ggplot2::autoplot(m)
    layer_classes = vapply(p$layers, function(l) class(l$geom)[1], character(1))
    expect_true("GeomVline" %in% layer_classes)
})

test_that("plot.nn_fit() returns the fit object invisibly", {
    skip_if_no_torch()
    skip_if_not_installed("ggplot2")
    m = train_nn(iris_x, iris_y, epochs = 5)
    pdf(NULL)
    on.exit(dev.off(), add = TRUE)
    result = plot(m)
    expect_identical(result, m)
})

# ---- autoplot_diagnostics() ----

test_that("autoplot_diagnostics() returns a named list of ggplots for regression", {
    skip_if_no_torch()
    skip_if_not_installed("ggplot2")
    m = train_nn(iris_x, iris_y, epochs = 5)
    result = suppressMessages(autoplot_diagnostics(m, actual = iris_y))
    expect_type(result, "list")
    expect_named(result, c("residuals_vs_fitted", "actual_vs_fitted"))
    expect_s3_class(result$residuals_vs_fitted, "ggplot")
    expect_s3_class(result$actual_vs_fitted, "ggplot")
})

test_that("autoplot_diagnostics() returns a ggplot for classification", {
    skip_if_no_torch()
    skip_if_not_installed("ggplot2")
    m = train_nn(iris_cls_x, iris_cls_y, epochs = 5)
    p = autoplot_diagnostics(m, actual = iris_cls_y)
    expect_s3_class(p, "ggplot")
})

test_that("autoplot_diagnostics() errors cleanly on length mismatch", {
    skip_if_no_torch()
    skip_if_not_installed("ggplot2")
    m = train_nn(iris_x, iris_y, epochs = 5)
    expect_error(
        autoplot_diagnostics(m, actual = iris_y[1:10]),
        class = "rlang_error"
    )
})

test_that("autoplot_diagnostics() errors cleanly for nn_fit_ds (no fitted values)", {
    skip_if_no_torch()
    skip_if_not_installed("ggplot2")
    ds = make_reg_ds()
    m_ds = train_nn(ds, hidden_neurons = 8L, epochs = 5)
    expect_error(
        autoplot_diagnostics(m_ds, actual = iris_y),
        class = "rlang_error"
    )
})

test_that("plot_diagnostics() returns the fit object invisibly (regression)", {
    skip_if_no_torch()
    skip_if_not_installed("ggplot2")
    m = train_nn(iris_x, iris_y, epochs = 5)
    pdf(NULL)
    on.exit(dev.off(), add = TRUE)
    result = suppressMessages(plot_diagnostics(m, actual = iris_y))
    expect_identical(result, m)
})

#' Validate Optimizer Exists
#'
#' @param act_name Character. Activation function name (without prefix).
#'
#' @importFrom cli cli_abort
#' @importFrom rlang is_installed
#'
#' @noRd
validate_optimizer = function(optimizer) {
    if (!is_installed("torch")) {
        cli_abort(c(
            "{.pkg torch} package is required but not installed.",
            i = "Install it with: {.code install.packages('torch')}"
        ), class = "torch_missing_error")
    }

    prefix = "optim_"
    fn_name = paste0(prefix, optimizer)

    if (!exists(fn_name, where = asNamespace("torch"), mode = "function")) {
        cli_abort(c(
            "Optimizer {.fn {fn_name}} does not exist in {.pkg torch}.",
            i = "Available functions start with {.code {prefix}}.",
            i = "Check {.code ?torch::optim_adam} for examples."
        ), class = "optimizer_not_found_error")
    }
}

#' Depth-Aware Grid Generation for Neural Networks
#'
#' `grid_depth()` extends standard grid generation to support multi-layer
#' neural network architectures. It creates heterogeneous layer configurations
#' by generating list columns for `hidden_neurons` and `activations`.
#'
#' @param x A `parameters` object, list, workflow, or model spec. Can also be
#'   a single `param` object if `...` contains additional param objects.
#' @param ... One or more `param` objects (e.g., `hidden_neurons()`, `epochs()`).
#'   If `x` is a `parameters` object, `...` is ignored. None of the objects can
#'   have `unknown()` values.
#' @param n_hlayer Integer vector specifying number of hidden layers to generate
#'   (e.g., `2:4` for 2, 3, or 4 layers), or a `param` object created with `n_hlayers()`.
#'   Default is 2.
#' @param size Integer. Number of parameter combinations to generate.
#' @param type Character. Type of grid: "regular", "random", "latin_hypercube",
#'   "max_entropy", or "audze_eglais".
#' @param original Logical. Should original parameter ranges be used?
#' @param levels Integer. Levels per parameter for regular grids.
#' @param variogram_range Numeric. Range for audze_eglais design.
#' @param iter Integer. Iterations for max_entropy optimization.
#'
#' @details
#' This function is specifically for `{kindling}` models. The `n_hlayer` parameter
#' determines network depth and creates list columns for `hidden_neurons` and
#' `activations`, where each element is a vector of length matching the sampled depth.
#'
#' When `n_hlayer` is a parameter object (created with `n_hlayers()`), it will be
#' treated as a tunable parameter and sampled according to its defined range.
#'
#' @return A tibble with list columns for `hidden_neurons` and `activations`,
#'   where each element is a vector of length `n_hlayer`.
#'
#' @examples
#' \donttest{
#' \dontrun{
#' library(dials)
#' library(workflows)
#' library(tune)
#'
#' # Method 1: Fixed depth
#' grid = grid_depth(
#'     hidden_neurons(c(32L, 128L)),
#'     activations(c("relu", "elu")),
#'     epochs(c(50L, 200L)),
#'     n_hlayer = 2:3,
#'     type = "random",
#'     size = 20
#' )
#'
#' # Method 2: Tunable depth using parameter object
#' grid = grid_depth(
#'     hidden_neurons(c(32L, 128L)),
#'     activations(c("relu", "elu")),
#'     epochs(c(50L, 200L)),
#'     n_hlayer = n_hlayers(range = c(2L, 4L)),
#'     type = "random",
#'     size = 20
#' )
#'
#' # Method 3: From workflow
#' wf = workflow() |>
#'     add_model(mlp_kindling(hidden_neurons = tune(), activations = tune())) |>
#'     add_formula(y ~ .)
#' grid = grid_depth(wf, n_hlayer = 2:4, type = "latin_hypercube", size = 15)
#' }
#' }
#'
#' @rdname grid_depth
#' @export
grid_depth = 
    function(
        x,
        ...,
        n_hlayer = 2L,
        size = 5L,
        type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
        original = TRUE,
        levels = 3L,
        variogram_range = 0.5,
        iter = 1000L
    ) {
        UseMethod("grid_depth")
    }

#' @export
#' @rdname grid_depth
grid_depth.parameters = 
    function(
        x,
        ...,
        n_hlayer = 2L,
        size = 5L,
        type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
        original = TRUE,
        levels = 3L,
        variogram_range = 0.5,
        iter = 1000L
    ) {
        
        type = rlang::arg_match(type)
        param_list = setNames(x$object, x$name)
        
        # Extract depth parameter if present
        depth_info = extract_depth_param(n_hlayer, param_list, levels)
        n_hlayer_vals = depth_info$values
        
        # Remove n_hlayers from param_list if it was there
        param_list = param_list[names(param_list) != "n_hlayers"]
        
        has_neurons = "hidden_neurons" %in% names(param_list)
        has_activations = "activations" %in% names(param_list)
        
        neuron_param = if (has_neurons) param_list[["hidden_neurons"]] else NULL
        activation_param = if (has_activations) param_list[["activations"]] else NULL
        
        scalar_names = setdiff(names(param_list), c("hidden_neurons", "activations"))
        scalar_params = param_list[scalar_names]
        
        generate_grid(
            neuron_param = neuron_param,
            activation_param = activation_param,
            n_hlayer = n_hlayer_vals,
            scalar_params = scalar_params,
            type = type,
            size = size,
            levels = levels,
            original = original,
            variogram_range = variogram_range,
            iter = iter
        )
    }

#' @export
#' @rdname grid_depth
grid_depth.list = 
    function(
        x,
        ...,
        n_hlayer = 2L,
        size = 5L,
        type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
        original = TRUE,
        levels = 3L,
        variogram_range = 0.5,
        iter = 1000L
    ) {
        params = rlang::exec(dials::parameters, !!!x)
        grid_depth.parameters(
            params,
            n_hlayer = n_hlayer,
            size = size,
            type = type,
            original = original,
            levels = levels,
            variogram_range = variogram_range,
            iter = iter
        )
    }

#' @export
#' @rdname grid_depth
grid_depth.workflow = 
    function(
        x,
        ...,
        n_hlayer = 2L,
        size = 5L,
        type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
        original = TRUE,
        levels = 3L,
        variogram_range = 0.5,
        iter = 1000L
    ) {
        params = workflows::extract_parameter_set_dials(x)
        grid_depth.parameters(
            params,
            n_hlayer = n_hlayer,
            size = size,
            type = type,
            original = original,
            levels = levels,
            variogram_range = variogram_range,
            iter = iter
        )
    }

#' @export
#' @rdname grid_depth
grid_depth.model_spec = 
    function(
        x,
        ...,
        n_hlayer = 2L,
        size = 5L,
        type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
        original = TRUE,
        levels = 3L,
        variogram_range = 0.5,
        iter = 1000L
    ) {
        tunable_params = tune::tunable(x)
        param_list = purrr::map(seq_len(nrow(tunable_params)), function(i) {
            call_info = tunable_params$call_info[[i]]
            rlang::exec(call_info$fun, !!!call_info$args, .env = asNamespace(call_info$pkg))
        })
        names(param_list) = tunable_params$name
        params = rlang::exec(dials::parameters, !!!param_list)
        
        grid_depth.parameters(
            params,
            n_hlayer = n_hlayer,
            size = size,
            type = type,
            original = original,
            levels = levels,
            variogram_range = variogram_range,
            iter = iter
        )
    }

#' @export
#' @rdname grid_depth
grid_depth.param = 
    function(
        x,
        ...,
        n_hlayer = 2L,
        size = 5L,
        type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
        original = TRUE,
        levels = 3L,
        variogram_range = 0.5,
        iter = 1000L
    ) {
        
        dots = rlang::list2(...)
        all_params = c(list(x), dots)
        param_objects = purrr::keep(all_params, ~ inherits(.x, "param"))
        
        if (length(param_objects) == 0) {
            cli::cli_abort(
                c(
                    "Could not find any param objects.",
                    "i" = "Provide param objects like {.code hidden_neurons()}, {.code epochs()}, etc."
                )
            )
        }
        
        params = rlang::exec(dials::parameters, !!!param_objects)
        grid_depth.parameters(
            params,
            n_hlayer = n_hlayer,
            size = size,
            type = type,
            original = original,
            levels = levels,
            variogram_range = variogram_range,
            iter = iter
        )
    }

#' @export
#' @rdname grid_depth
grid_depth.default = 
    function(
        x,
        ...,
        n_hlayer = 2L,
        size = 5L,
        type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
        original = TRUE,
        levels = 3L,
        variogram_range = 0.5,
        iter = 1000L
    ) {
        
        cli::cli_abort(
            c(
                "No method for object of class {.cls {class(x)}}",
                "i" = "Provide param objects, a {.cls parameters} object, {.cls workflow}, or {.cls model_spec}."
            )
        )
    }

#' Extract depth parameter values from n_hlayer argument
#'
#' @param n_hlayer Either an integer vector or a param object
#' @param param_list List of parameters (for extracting n_hlayers if present)
#' @param levels Number of levels for regular grids
#'
#' @return List with `values` component containing integer vector of depths
#' @keywords internal
extract_depth_param = function(n_hlayer, param_list = list(), levels = 3L) {
    if ("n_hlayers" %in% names(param_list)) {
        depth_param = param_list[["n_hlayers"]]
        depth_vals = extract_param_range(depth_param, levels)
        return(list(values = as.integer(depth_vals)))
    }
    
    if (inherits(n_hlayer, "param")) {
        depth_vals = extract_param_range(n_hlayer, levels)
        return(list(values = as.integer(depth_vals)))
    }
    
    out = list(values = as.integer(n_hlayer))
    out
}

generate_grid = 
    function(
        neuron_param, activation_param, n_hlayer,
        scalar_params, type, size, levels, original,
        variogram_range, iter
    ) {
    if (is.null(neuron_param) && is.null(activation_param)) {
        if (length(scalar_params) == 0) {
            cli::cli_abort("No parameters provided for grid generation.")
        }
        return(make_scalar_grid(scalar_params, size, levels, type, original))
    }
    
    n_hlayer = as.integer(n_hlayer)
    out = switch(
        type,
        regular = generate_regular_grid(
            neuron_param, activation_param, n_hlayer,
            scalar_params, levels, original
        ),
        random = generate_random_grid(
            neuron_param, activation_param, n_hlayer,
            scalar_params, size, original
        ),
        latin_hypercube = generate_lhs_grid(
            neuron_param, activation_param, n_hlayer,
            scalar_params, size, original
        ),
        max_entropy = generate_sfd_grid(
            neuron_param, activation_param, n_hlayer,
            scalar_params, size, "max_entropy", variogram_range, iter, original
        ),
        audze_eglais = generate_sfd_grid(
            neuron_param, activation_param, n_hlayer,
            scalar_params, size, "audze_eglais", variogram_range, iter, original
        )
    )
    
    # Force scalar, not list
    if (all(n_hlayer == 1)) {
        if ("hidden_neurons" %in% names(out)) {
            out$hidden_neurons = purrr::map_int(out$hidden_neurons, 1)
        }
        if ("activations" %in% names(out)) {
            out$activations = purrr::map_chr(out$activations, 1)
        }
    }
    
    out
}

generate_regular_grid = 
    function(
        neuron_param, activation_param, n_hlayer,
        scalar_params, levels, original
    ) {
    neuron_vals = extract_param_range(neuron_param, levels, original)
    activation_vals = extract_param_values(activation_param)
    
    arch_grid = purrr::map_dfr(n_hlayer, function(depth) {
        if (!is.null(neuron_vals) && !is.null(activation_vals)) {
            expand_architecture(neuron_vals, activation_vals, depth)
        } else if (!is.null(neuron_vals)) {
            expand_neurons_only(neuron_vals, depth)
        } else {
            expand_activations_only(activation_vals, depth)
        }
    })
    
    if (length(scalar_params) > 0) {
        scalar_grid = dials::grid_regular(
            dials::parameters(scalar_params),
            levels = levels,
            original = original
        )
        tidyr::crossing(arch_grid, scalar_grid)
    } else {
        arch_grid
    }
}

generate_random_grid = 
    function(
        neuron_param, activation_param, n_hlayer,
        scalar_params, size, original
    ) {
    activation_vals = extract_param_values(activation_param)
    
    purrr::map_dfr(seq_len(size), function(i) {
        depth = safe_sample(n_hlayer, 1)
        row_data = list()
        
        if (!is.null(neuron_param)) {
            neurons = sample_from_param(neuron_param, depth, original)
            row_data$hidden_neurons = list(as.integer(neurons))
        }
        
        if (!is.null(activation_vals)) {
            activations = sample(activation_vals, depth, replace = TRUE)
            row_data$activations = list(as.character(activations))
        }
        
        if (length(scalar_params) > 0) {
            scalars = dials::grid_random(
                dials::parameters(scalar_params),
                size = 1,
                original = original
            )
            row_data = c(row_data, as.list(scalars))
        }
        
        tibble::as_tibble(row_data)
    })
}

generate_lhs_grid = 
    function(
        neuron_param, activation_param, n_hlayer,
        scalar_params, size, original
    ) {
    if (!requireNamespace("lhs", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg lhs} required for Latin Hypercube sampling.")
    }
    
    has_neurons = !is.null(neuron_param)
    activation_vals = extract_param_values(activation_param)
    max_depth = max(n_hlayer)
    
    n_numeric_arch = if (has_neurons) max_depth else 0
    n_numeric_scalars = count_numeric_params(scalar_params)
    n_dims = n_numeric_arch + n_numeric_scalars
    
    if (n_dims == 0) {
        return(generate_random_grid(
            neuron_param, activation_param, n_hlayer,
            scalar_params, size, original
        ))
    }
    
    design = lhs::randomLHS(size, n_dims)
    depths = safe_sample(n_hlayer, size, replace = TRUE)
    
    results = vector("list", size)
    
    for (i in seq_len(size)) {
        depth = depths[i]
        row = design[i, ]
        row_data = list()
        
        if (has_neurons) {
            neuron_indices = row[seq_len(n_numeric_arch)]
            neurons = decode_neurons_from_design(
                neuron_param, 
                neuron_indices[seq_len(depth)],
                original
            )
            row_data$hidden_neurons = list(as.integer(neurons))
        }
        
        if (!is.null(activation_vals)) {
            activations = sample(activation_vals, depth, replace = TRUE)
            row_data$activations = list(as.character(activations))
        }
        
        if (length(scalar_params) > 0) {
            if (n_numeric_scalars > 0) {
                scalar_indices = row[(n_numeric_arch + 1):n_dims]
                scalars = decode_scalars(scalar_params, scalar_indices, original)
            } else {
                scalars = decode_scalars(scalar_params, numeric(0), original)
            }
            row_data = c(row_data, as.list(scalars))
        }
        
        results[[i]] = tibble::as_tibble(row_data)
    }
    
    dplyr::bind_rows(results)
}

generate_sfd_grid = 
    function(
        neuron_param, activation_param, n_hlayer,
        scalar_params, size, sfd_type, variogram_range, iter, original
    ) {
    if (!requireNamespace("sfd", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg sfd} required for space-filling designs.")
    }
    
    has_neurons = !is.null(neuron_param)
    activation_vals = extract_param_values(activation_param)
    
    max_depth = max(n_hlayer)
    n_numeric_arch = if (has_neurons) max_depth else 0
    n_numeric_scalars = count_numeric_params(scalar_params)
    n_dims = n_numeric_arch + n_numeric_scalars
    
    if (n_dims == 0) {
        return(generate_random_grid(
            neuron_param, activation_param, n_hlayer,
            scalar_params, size, original
        ))
    }
    
    has_premade = sfd::sfd_available(n_dims, size, sfd_type)
    
    if (has_premade) {
        design = sfd::get_design(n_dims, num_points = size, type = sfd_type)
        design = apply(design, 2, function(col) {
            col_range = max(col) - min(col)
            if (col_range > 0) {
                (col - min(col)) / col_range
            } else {
                rep(0.5, length(col))
            }
        })
    } else {
        if (!requireNamespace("DiceDesign", quietly = TRUE)) {
            cli::cli_abort("Package {.pkg DiceDesign} required when pre-made designs are not available.")
        }
        design = DiceDesign::dmaxDesign(
            size,
            n_dims,
            range = 1,
            niter_max = iter
        )$design
    }
    
    depths = safe_sample(n_hlayer, size, replace = TRUE)
    
    results = vector("list", size)
    
    for (i in seq_len(size)) {
        depth = depths[i]
        row = design[i, ]
        row_data = list()
        
        if (has_neurons) {
            neuron_indices = row[seq_len(n_numeric_arch)]
            neurons = decode_neurons_from_design(
                neuron_param, 
                neuron_indices[seq_len(depth)],
                original
            )
            row_data$hidden_neurons = list(as.integer(neurons))
        }
        
        if (!is.null(activation_vals)) {
            activations = sample(activation_vals, depth, replace = TRUE)
            row_data$activations = list(as.character(activations))
        }
        
        if (length(scalar_params) > 0) {
            if (n_numeric_scalars > 0) {
                scalar_indices = row[(n_numeric_arch + 1):n_dims]
                scalars = decode_scalars(scalar_params, scalar_indices, original)
            } else {
                scalars = decode_scalars(scalar_params, numeric(0), original)
            }
            row_data = c(row_data, as.list(scalars))
        }
        
        results[[i]] = tibble::as_tibble(row_data)
    }
    
    dplyr::bind_rows(results)
}

extract_param_range = function(param, levels, original = TRUE) {
    if (is.null(param)) return(NULL)
    
    if (!is.null(param$values)) {
        return(param$values)
    }
    
    if (param$type %in% c("integer", "double")) {
        lower = param$range$lower
        upper = param$range$upper
        
        has_trans = !is.null(param$trans)
        
        if (!is.null(levels)) {
            if (has_trans) {
                vals_trans = seq(lower, upper, length.out = levels)
                vals = param$trans$inverse(vals_trans)
            } else {
                vals = seq(lower, upper, length.out = levels)
            }
            
            if (param$type == "integer") {
                unique(as.integer(round(vals)))
            } else {
                vals
            }
        } else {
            if (param$type == "integer") {
                seq.int(lower, upper)
            } else {
                c(lower, upper)
            }
        }
    } else {
        NULL
    }
}

extract_param_bounds = function(param) {
    if (is.null(param)) return(NULL)
    
    if (param$type %in% c("integer", "double")) {
        c(param$range$lower, param$range$upper)
    } else {
        NULL
    }
}

extract_param_values = function(param) {
    if (is.null(param)) return(NULL)
    
    if (param$type == "character" || param$type == "logical") {
        param$values
    } else {
        NULL
    }
}

sample_from_param = function(param, n, original = TRUE) {
    if (is.null(param)) return(NULL)
    if (!is.null(param$values)) {
        return(sample(param$values, n, replace = TRUE))
    }
    
    lower = param$range$lower
    upper = param$range$upper
    
    if (!is.null(param$trans)) {
        vals_trans = stats::runif(n, lower, upper)
        vals = param$trans$inverse(vals_trans)
    } else {
        vals = stats::runif(n, lower, upper)
    }
    
    if (param$type == "integer") {
        as.integer(round(vals))
    } else {
        vals
    }
}

decode_neurons_from_design = function(param, design_vals, original = TRUE) {
    lower = param$range$lower
    upper = param$range$upper
    
    if (!is.null(param$values)) {
        values = param$values
        indices = pmax(
            1L, 
            pmin(length(values), ceiling(design_vals * length(values)))
        )
        return(values[indices])
    }
    # discrete_vals = attr(param, "discrete_values")
    # if (!is.null(discrete_vals)) {
    #     indices = pmax(
    #         1, 
    #         pmin(length(discrete_vals), ceiling(design_vals * length(discrete_vals)))
    #     )
    #     return(discrete_vals[indices])
    # }
    
    if (!is.null(param$trans)) {
        vals_trans = lower + design_vals * (upper - lower)
        vals = param$trans$inverse(vals_trans)
    } else {
        vals = lower + design_vals * (upper - lower)
    }
    
    if (param$type == "integer") {
        as.integer(round(vals))
    } else {
        vals
    }
}

expand_architecture = function(neuron_vals, activation_vals, depth) {
    neuron_cols = stats::setNames(
        rep(list(neuron_vals), depth),
        paste0("n", seq_len(depth))
    )
    activation_cols = stats::setNames(
        rep(list(activation_vals), depth),
        paste0("a", seq_len(depth))
    )
    
    # Generate all combinations
    grid = tidyr::expand_grid(!!!neuron_cols, !!!activation_cols)
    
    # Convert to list columns
    neuron_col_names = paste0("n", seq_len(depth))
    activation_col_names = paste0("a", seq_len(depth))
    
    tibble::tibble(
        hidden_neurons = purrr::pmap(
            dplyr::select(grid, dplyr::all_of(neuron_col_names)),
            ~ as.integer(c(...))
        ),
        activations = purrr::pmap(
            dplyr::select(grid, dplyr::all_of(activation_col_names)),
            ~ as.character(c(...))
        )
    )
}

expand_neurons_only = function(neuron_vals, depth) {
    neuron_cols = stats::setNames(
        rep(list(neuron_vals), depth),
        paste0("n", seq_len(depth))
    )
    
    grid = tidyr::expand_grid(!!!neuron_cols)
    
    tibble::tibble(
        hidden_neurons = purrr::pmap(grid, ~ as.integer(c(...)))
    )
}

expand_activations_only = function(activation_vals, depth) {
    activation_cols = stats::setNames(
        rep(list(activation_vals), depth),
        paste0("a", seq_len(depth))
    )
    
    grid = tidyr::expand_grid(!!!activation_cols)
    
    tibble::tibble(
        activations = purrr::pmap(grid, ~ as.character(c(...)))
    )
}

make_scalar_grid = function(scalar_params, size, levels, type, original) {
    if (length(scalar_params) == 0) {
        return(tibble::tibble())
    }
    
    params_obj = dials::parameters(scalar_params)
    
    if (type == "regular") {
        dials::grid_regular(params_obj, levels = levels, original = original)
    } else {
        dials::grid_random(params_obj, size = size, original = original)
    }
}

count_numeric_params = function(scalar_params) {
    sum(purrr::map_lgl(scalar_params, ~ .x$type %in% c("double", "integer")))
}

decode_scalars = function(scalar_params, design_vals, original = TRUE) {
    if (length(scalar_params) == 0) {
        return(tibble::tibble())
    }
    
    numeric_params = purrr::keep(scalar_params, ~ .x$type %in% c("double", "integer"))
    categorical_params = purrr::keep(scalar_params, ~ .x$type %in% c("character", "logical"))
    
    decoded_numeric = if (length(numeric_params) > 0 && length(design_vals) > 0) {
        purrr::imap_dfc(numeric_params, function(param, param_name) {
            idx = which(names(numeric_params) == param_name)
            
            lower = param$range$lower
            upper = param$range$upper
            
            if (!is.null(param$trans)) {
                val_trans = lower + design_vals[idx] * (upper - lower)
                val = param$trans$inverse(val_trans)
            } else {
                val = lower + design_vals[idx] * (upper - lower)
            }
            
            if (param$type == "integer") {
                val = as.integer(round(val))
            }
            
            tibble::tibble(!!param_name := val)
        })
    } else {
        NULL
    }
    
    decoded_categorical = if (length(categorical_params) > 0) {
        purrr::imap_dfc(categorical_params, function(param, param_name) {
            tibble::tibble(!!param_name := sample(param$values, 1))
        })
    } else {
        NULL
    }
    
    if (!is.null(decoded_numeric) && !is.null(decoded_categorical)) {
        dplyr::bind_cols(decoded_numeric, decoded_categorical)
    } else if (!is.null(decoded_numeric)) {
        decoded_numeric
    } else if (!is.null(decoded_categorical)) {
        decoded_categorical
    } else {
        tibble::tibble()
    }
}

`%||%` = function(x, y) if (is.null(x)) y else x

#' Safe sampling function
#' 
#' R's sample() has quirky behavior: sample(5, 1) samples from 1:5, not from c(5).
#' This function ensures we sample from the actual vector provided.
#'
#' @param x Vector to sample from
#' @param size Number of samples
#' @param replace Sample with replacement?
#' @keywords internal
safe_sample = function(x, size, replace = FALSE) {
    if (length(x) == 1) {
        rep(x, size)
    } else {
        sample(x, size, replace = replace)
    }
}

#' @keywords internal
.validate_loss_fn = function(fn) {
    # ---- Arity check ----
    if (length(formals(fn)) < 2L) {
        cli::cli_abort(c(
            "Custom loss function must accept at least 2 arguments: {.arg input} and {.arg target}.",
            i = "Use a lambda like {.code \\(input, target) torch::nnf_mse_loss(input, target)}."
        ), class = "loss_fn_arity_error")
    }

    # ---- Dry-run probe ----
    if (requireNamespace("torch", quietly = TRUE)) {
        dummy_input  = torch::torch_randn(c(2L, 1L))
        dummy_target = torch::torch_randn(c(2L, 1L))

        out = tryCatch(
            fn(dummy_input, dummy_target),
            error = function(e) {
                cli::cli_abort(c(
                    "Dry-run of custom loss function failed.",
                    x = "{e$message}",
                    i = "Ensure your function accepts two {.cls torch_tensor}s and returns a scalar tensor."
                ), class = "loss_fn_probe_error")
            }
        )

        if (!inherits(out, "torch_tensor")) {
            cli::cli_abort(c(
                "Custom loss function must return a {.cls torch_tensor}.",
                x = "Got {.cls {class(out)}}.",
                i = "Ensure your function returns the result of a {.pkg torch} operation."
            ), class = "loss_fn_output_error")
        }

        if (out$numel() != 1L) {
            cli::cli_warn(c(
                "Custom loss function returned a non-scalar tensor with {out$numel()} elements.",
                i = "Loss functions should return a scalar. Did you forget to reduce (e.g. {.code $mean()})?"
            ), class = "loss_fn_shape_warning")
        }
    }

    # ---- Build the call-time type-guarded wrapper ----
    function(input, target) {
        out = fn(input, target)
        if (!inherits(out, "torch_tensor")) {
            cli::cli_abort(c(
                "Custom loss function must return a {.cls torch_tensor}.",
                x = "Got {.cls {class(out)}}."
            ), class = "loss_fn_output_error")
        }
        out
    }
}

#' Regularize Loss Function
#'
#' @param model Torch module
#' @param penalty Numeric. Regularization strength
#' @param penalty_type Character. Type of regularization
#' @param mixture Numeric. Elastic net mixing parameter
#'
#' @return Torch tensor with regularization loss
#' @noRd
regularizer = function(model, penalty, mixture = 0.5) {
    # Default: When regularization wasn't applied
    # Early return
    if (penalty == 0) {
        return(torch::torch_tensor(
            0,
            device = model$parameters[[1]]$device,
            dtype = torch::torch_float32()
        ))
    }
    
    reg_loss = torch::torch_tensor(
        0,
        device = model$parameters[[1]]$device,
        dtype = torch::torch_float32()
    )
    
    for (param in model$parameters) {
        if (length(param$size()) <= 1) next 
        
        if (mixture == 1) {
            reg_loss = reg_loss + torch::torch_sum(torch::torch_abs(param))
        } else if (mixture == 0) {
            reg_loss = reg_loss + 0.5 * torch::torch_sum(param$pow(2))
        } else {
            l1 = torch::torch_sum(torch::torch_abs(param))
            l2 = 0.5 * torch::torch_sum(param$pow(2))
            reg_loss = reg_loss + mixture * l1 + (1 - mixture) * l2
        }
    }
    
    penalty * reg_loss
}

#' Validate Regularization Parameters
#'
#' @param penalty Numeric
#' @param penalty_type Character
#' @param mixture Numeric
#'
#' @noRd
validate_regularization = function(penalty, mixture) {
    if (penalty < 0) {
        cli::cli_abort(c(
            "{.arg penalty} must be non-negative.",
            x = "You provided {.val {penalty}}."
        ))
    }
    
    if (mixture < 0 || mixture > 1) {
        cli::cli_abort(c(
            "{.arg mixture} must be between 0 and 1.",
            x = "You provided {.val {mixture}}."
        ))
    }
    
    invisible(NULL)
}

#' Variable Importance Methods for kindling Models
#'
#' This file implements methods for variable importance generics from
#' NeuralNetTools and vip packages.
#'
#' @name kindling-varimp
#'
#' @references
#'
#' Beck, M.W. 2018. NeuralNetTools: Visualization and Analysis Tools for Neural Networks. Journal of Statistical Software. 85(11):1-20.
#'
#' Garson, G.D. 1991. Interpreting neural network connection weights. Artificial Intelligence Expert. 6(4):46-51.
#'
#' Goh, A.T.C. 1995. Back-propagation neural networks for modeling complex systems. Artificial Intelligence in Engineering. 9(3):143-151.
#'
#' Olden, J.D., Jackson, D.A. 2002. Illuminating the 'black-box': a randomization approach for understanding variable contributions in artificial neural networks. Ecological Modelling. 154:135-150.
#'
#' Olden, J.D., Joy, M.K., Death, R.G. 2004. An accurate comparison of methods for quantifying variable importance in artificial neural networks using simulated data. Ecological Modelling. 178:389-397.
NULL

#' @importFrom NeuralNetTools garson
#' @export
NeuralNetTools::garson

#' @importFrom NeuralNetTools olden
#' @export
NeuralNetTools::olden

#' Extract Weight Matrices from FFNN Model
#'
#' @param mod_in A fitted ffnn_fit object
#'
#' @return List with input_weights and output_weights matrices
#' @noRd
extract_ffnn_weights = function(mod_in) {
    if (!is.null(mod_in$cached_weights)) {
        return(mod_in$cached_weights)
    }
    
    model = mod_in$model
    n_hidden = length(mod_in$hidden_neurons)
    input_layer = model$fc1
    W_input = as.matrix(input_layer$weight$cpu())
    output_layer = model$out
    W_output = as.matrix(output_layer$weight$cpu())
    intermediate_weights = list()
    if (n_hidden > 1) {
        for (i in seq_len(n_hidden - 1)) {
            layer = model[[paste0("fc", i + 1)]]
            intermediate_weights[[i]] = as.matrix(layer$weight$cpu())
        }
    }
    
    list(
        input = W_input,
        output = W_output,
        intermediate = intermediate_weights
    )
}

#' @section Garson's Algorithm for FFNN Models:
#' `{kindling}` inherits `NeuralNetTools::garson` to extract the variable
#' importance from the fitted `ffnn()` model.
#'
#' @rdname kindling-varimp
#'
#' @param mod_in A fitted model object of class "ffnn_fit".
#' @param bar_plot Logical. Whether to plot variable importance (default FALSE).
#' @param ... Additional arguments passed to NeuralNetTools plotting.
#'
#' @return A data frame for both "garson" and "olden" classes with columns:
#'     \item{x_names}{Character vector of predictor variable names}
#'     \item{y_names}{Character string of response variable name}
#'     \item{rel_imp}{Numeric vector of relative importance scores (percentage)}
#'     The data frame is sorted by importance in descending order.
#'
#' @examples
#' \donttest{
#' if (torch::torch_is_installed()) {
#'     model_mlp = ffnn(
#'         Species ~ .,
#'         data = iris,
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 100,
#'         verbose = FALSE,
#'         cache_weights = TRUE
#'     )
#'     
#'     # Directly use `NeuralNetTools::garson`
#'     model_mlp |>
#'         garson()
#'     
#'     # Directly use `NeuralNetTools::olden`    
#'     model_mlp |>
#'         olden()
#' } else {
#'     message("Torch not fully installed — skipping example")
#' }
#' }
#'
#' @export
#' @method garson ffnn_fit
garson.ffnn_fit = function(mod_in, bar_plot = FALSE, ...) {
    if (!requireNamespace("torch", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg torch} is required but not installed.")
    }
    
    weights = extract_ffnn_weights(mod_in)
    
    W_in = abs(t(weights$input))
    W_out = abs(weights$output)
    
    n_features = nrow(W_in)
    n_outputs = nrow(W_out)
    
    if (length(weights$intermediate) > 0) {
        W_propagated = W_in
        for (W_layer in weights$intermediate) {
            W_propagated = W_propagated %*% t(abs(W_layer))
        }
        W_combined = W_propagated
    } else {
        W_combined = W_in
    }
    
    importance = numeric(n_features)
    for (i in seq_len(n_features)) {
        total = 0
        for (j in seq_len(n_outputs)) {
            for (k in seq_len(ncol(W_combined))) {
                total = total + W_combined[i, k] * W_out[j, k]
            }
        }
        importance[i] = total
    }
    
    total_importance = sum(importance)
    if (total_importance > 0) {
        importance = (importance / total_importance) * 100
    }
    
    out_gar = data.frame(
        x_names = mod_in$feature_names,
        y_names = mod_in$response_name,
        rel_imp = importance,
        stringsAsFactors = FALSE
    )
    
    out_gar = out_gar[order(out_gar$rel_imp, decreasing = TRUE), ]
    rownames(out_gar) = NULL
    
    class(out_gar) = c("garson", "data.frame")
    
    if (bar_plot && requireNamespace("ggplot2", quietly = TRUE)) {
        p = ggplot2::ggplot(out_gar) +
            ggplot2::aes(x = reorder(x_names, rel_imp), y = rel_imp) +
            ggplot2::geom_col(fill = "steelblue") +
            ggplot2::coord_flip() +
            ggplot2::labs(
                x = "Features",
                y = "Relative Importance",
                title = "Variable Importance (Olden Method)"
            ) +
            ggplot2::theme_minimal()
        
        print(p)
    }
    
    out_gar
}

#' @section Olden's Algorithm for FFNN Models:
#' `{kindling}` inherits `NeuralNetTools::olden` to extract the variable
#' importance from the fitted `ffnn()` model.
#'
#' @rdname kindling-varimp
#'
#' @param mod_in A fitted model object of class "ffnn_fit".
#' @param bar_plot Logical. Whether to plot variable importance (default TRUE).
#' @param ... Additional arguments passed to NeuralNetTools plotting.
#'
#' @export
#' @method olden ffnn_fit
olden.ffnn_fit = function(mod_in, bar_plot = TRUE, ...) {
    if (!requireNamespace("torch", quietly = TRUE)) {
        cli::cli_abort("Package {.pkg torch} is required but not installed.")
    }
    
    weights = extract_ffnn_weights(mod_in)
    
    W_in = t(weights$input)
    W_out = weights$output
    
    n_features = nrow(W_in)
    n_outputs = nrow(W_out)
    
    if (length(weights$intermediate) > 0) {
        W_propagated = W_in
        for (W_layer in weights$intermediate) {
            W_propagated = W_propagated %*% t(W_layer)
        }
        W_combined = W_propagated
    } else {
        W_combined = W_in
    }
    
    importance_matrix = matrix(0, nrow = n_features, ncol = n_outputs)
    
    for (i in seq_len(n_features)) {
        for (o in seq_len(n_outputs)) {
            importance_matrix[i, o] = sum(W_combined[i, ] * W_out[o, ])
        }
    }
    
    importance = if (n_outputs > 1) {
        rowMeans(importance_matrix)
    } else {
        importance_matrix[, 1]
    }
    
    out_old = data.frame(
        x_names = mod_in$feature_names,
        y_names = mod_in$response_name,
        rel_imp = importance,
        stringsAsFactors = FALSE
    )
    
    out_old = out_old[order(abs(out_old$rel_imp), decreasing = TRUE), ]
    rownames(out_old) = NULL
    
    class(out_old) = c("olden", "data.frame")
    
    if (bar_plot && requireNamespace("ggplot2", quietly = TRUE)) {
        p = ggplot2::ggplot(out_old) +
            ggplot2::aes(x = reorder(x_names, rel_imp), y = rel_imp) +
            ggplot2::geom_col(fill = "steelblue") +
            ggplot2::coord_flip() +
            ggplot2::labs(
                x = "Features",
                y = "Relative Importance",
                title = "Variable Importance (Olden Method)"
            ) +
            ggplot2::theme_minimal()
        
        print(p)
    }
    
    out_old
}

#' @section Variable Importance via `{vip}` Package:
#' If the `{vip}` package (Suggests only) is installed, `kindling` registers a
#' `vip::vi_model()` method for `"ffnn_fit"` objects, so `vip::vi()` and
#' `vip::vi_model()` work out of the box.
#'
#' @rdname kindling-varimp
#'
#' @param object A fitted model object of class "ffnn_fit".
#' @param type Type of algorithm to extract the variable importance.
#'  This must be one of the strings:
#'  - 'olden'
#'  - 'garson'
#' @param ... Additional arguments passed to methods.
#'
#' @return A tibble with columns "Variable" and "Importance"
#'    (via `vip::vi()` / `vip::vi_model()` only).
#'
#' @examples
#' \donttest{
#' # kindling also supports `vip::vi()` / `vip::vi_model()` when {vip} is installed
#' if (torch::torch_is_installed() && requireNamespace("vip", quietly = TRUE)) {
#'     model_mlp = ffnn(
#'         Species ~ .,
#'         data = iris,
#'         hidden_neurons = c(64, 32),
#'         activations = "relu",
#'         epochs = 100,
#'         verbose = FALSE,
#'         cache_weights = TRUE
#'     )
#'
#'     model_mlp |>
#'         vip::vi(type = 'garson') |>
#'         vip::vip()
#' } else {
#'     message("Torch and/or {vip} not installed — skipping example")
#' }
#' }
vi_model.ffnn_fit = function(object, type = c("olden", "garson"),  ...) {
    type = match.arg(type)
    result = switch(
        type,
        olden = olden(object, bar_plot = FALSE, ...),
        garson = garson(object, bar_plot = FALSE, ...)
    )
    
    if (requireNamespace("tibble", quietly = TRUE)) {
        tibble::tibble(
            Variable = result$x_names,
            Importance = abs(result$rel_imp)
        )
    } else {
        data.frame(
            Variable = result$x_names,
            Importance = abs(result$rel_imp),
            stringsAsFactors = FALSE
        )
    }
}

---
title: "Tuning Capabilities"
output:
    rmarkdown::html_vignette:
        toc: true
vignette: >
  %\VignetteIndexEntry{Tuning Capabilities}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    # eval = identical(Sys.getenv("BUILD_VIGNETTES"), "true"), 
    eval = identical(Sys.getenv("NOT_CRAN"), "true"), 
    fig.width = 7,
    fig.height = 5,
    warning = FALSE,
    message = FALSE
)
```

## Rationale

How capable this package when tuning neural networks? One of the package's capabilities is the ability to fine-tune the whole architecture, and this includes the depth of the architecture — not limited to the number of hidden neurons, also includes the number of layers. Neural networks with `{torch}` natively supports different activation functions for different layers, thus `{kindling}` supports: 

- The **number of hidden layers** (depth)
- The **number of neurons per layer** (width)
- The **activation function per layer**, including parametric variants (e.g. `softshrink(lambd = 0.2)`)

## Custom grid creation

`{kindling}` has its own function to define the grid which includes the depth of the architecture: `grid_depth()`, an analogue function to `dials::grid_space_filling()`, except it creates `"regular"` grid. You can tweak `n_hlayer` parameter, and you can define the grid that has the depth. This parameter can be scalar (e.g. `2`), integer vector (e.g. `1:2`), and/or using a `{dials}` function called `n_hlayer()`. When `n_hlayer` is greater than 2, the certain parameters `hidden_neurons` and `activations` creates a list-column, which contains vectors for each parameter grid, depending on `n_hlayer` you defined. 

## Setup

We won't stop you from using `library()` function, but we strongly recommend using `box::use()` and explicitly import the names from the namespaces you want to attach.

```{r}
# library(kindling)
# library(tidymodels)
# library(modeldata)

box::use(
    kindling[mlp_kindling, act_funs, args, hidden_neurons, activations, grid_depth],
    dplyr[select, ends_with, mutate, slice_sample],
    tidyr[drop_na],
    rsample[initial_split, training, testing, vfold_cv],
    recipes[
        recipe, step_dummy, step_normalize,
        all_nominal_predictors, all_numeric_predictors
    ],
    modeldata[penguins],
    parsnip[tune, set_mode, fit, augment],
    workflows[workflow, add_recipe, add_model],
    dials[learn_rate],
    tune[tune_grid, show_best, collect_metrics, select_best, finalize_workflow, last_fit],
    yardstick[metric_set, rmse, rsq],
    ggplot2[autoplot]
)
```

We'll use the `penguins` dataset from `{modeldata}` to predict body mass (in kilograms) from physical measurements — a straightforward regression task that lets us focus on the tuning workflow.

## Usage

`{kindling}` provides the `mlp_kindling()` model spec. Parameters you want to search over are marked with `tune()`.

```{r spec}
spec = mlp_kindling(
    hidden_neurons = tune(),
    activations = tune(),
    epochs = 50,
    learn_rate = tune()
) |>
    set_mode("regression")
```

Note that `n_hlayer` is not listed here — it is handled inside `grid_depth()` rather than the model spec directly.

### Data Preparation

We sample 30 rows per species to keep the example lightweight, and stratify splits on `species` to preserve class balance. The target variable is `body_mass_kg`, derived from the original `body_mass_g` column.

```{r data}
penguins_clean = penguins |>
    drop_na() |>
    select(body_mass_g, ends_with("_mm"), sex, species) |>
    mutate(body_mass_kg = body_mass_g / 1000) |>
    slice_sample(n = 30, by = species)

set.seed(123)
split = initial_split(penguins_clean, prop = 0.8, strata = species)
train = training(split)
test = testing(split)
folds = vfold_cv(train, v = 5, strata = body_mass_kg)


rec = recipe(body_mass_kg ~ ., data = train) |>
    step_dummy(all_nominal_predictors()) |>
    step_normalize(all_numeric_predictors())
```

### Using grid_depth()

You still can use standard `{dials}` grids but the limitation is that they don't know about network depth, so `{kindling}` provides `grid_depth()`. The `n_hlayer` argument controls which depths to search over. Remember, it accepts:

- A scalar: `n_hlayer = 2`
- An integer vector: `n_hlayer = 1:3`
- A `{dials}` range object: `n_hlayer = n_hlayer(c(1, 3))`

When `n_hlayer > 1`, the `hidden_neurons` and `activations` columns become list-columns, where each row holds a vector of per-layer values.

```{r grid}
set.seed(42)
depth_grid = grid_depth(
    hidden_neurons(c(16, 32)),
    activations(c("relu", "elu", "softshrink(lambd = 0.2)")),
    learn_rate(),
    n_hlayer = 1:3,
    size = 10,
    type = "latin_hypercube"
)

depth_grid
```

Here we constrain `hidden_neurons` to the range `[16, 32]` and limit activations to three candidates — including the parametric `softshrink`. Latin hypercube sampling spreads the 10 candidates more evenly across the search space compared to a random grid.

### Tuning

What happens to the tuning part? The solution is easy: the parameters induced into list-columns and it becomes something like `list(c(1, 2))`, so internally the configured argument unlisted through `list(c(1, 2))[[1]]` (it always produces only 1 element). 

```{r tune}
wflow = workflow() |>
    add_recipe(rec) |>
    add_model(spec)

tune_res = tune_grid(
    wflow,
    resamples = folds,
    grid = depth_grid,
    metrics = metric_set(rmse, rsq)
)
```

### Inspect

Even with the list-columns, it still normally produces the output we want to produce. Use functions to extract the metrics output after grid search, e.g. `collect_metrics()` and `show_best()`. 

```{r results}
collect_metrics(tune_res)
show_best(tune_res, metric = "rmse", n = 5)
```

## Visualizing Results

<!-- ### Metric distributions across the grid -->

<!-- `autoplot()` gives a quick overview of how each hyperparameter relates to model performance across all resamples. -->

<!-- ```{r autoplot} -->
<!-- tune::autoplot(tune_res) -->
<!-- ``` -->

## Finalizing the Model

Once we've identified the best configuration, we finalize the workflow and fit
it on the full training set.

```{r final}
best_params = select_best(tune_res, metric = "rmse")
final_wflow = wflow |>
    finalize_workflow(best_params)

final_model = fit(final_wflow, data = train)
final_model
```

### Evaluating on the test set

```{r eval}
final_model |>
    augment(new_data = test) |>
    metric_set(rmse, rsq)(
        truth = body_mass_kg,
        estimate = .pred
    )
```

## A Note on Parametric Activations

`{kindling}` supports parametric activation functions, meaning each layer's activation can carry its own tunable parameter. When passed as a string such as `"softshrink(lambd = 0.2)"`, `{kindling}` parses and constructs the activation automatically. This means you can include them directly in the `activations()` candidate list inside `grid_depth()` without any extra setup, as shown above.

For manual (non-tuned) use, you can also specify activations per layer
explicitly:

```{r parametric}
spec_manual = mlp_kindling(
    hidden_neurons = c(50, 15),
    activations = act_funs(
        softshrink[lambd = 0.5],
        relu
    ),
    epochs = 150,
    learn_rate = 0.01
) |>
    set_mode("regression")
```

---
title: "Similar packages and comparison"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Similar packages and comparison}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    fig.width = 7,
    fig.height = 5
)
```

## Similar packages

All packages discussed here are built on top of `{torch}`, R's native implementation of PyTorch. The torch package provides low-level tensor operations and neural network building blocks, but requires substantial boilerplate code for training. Higher-level packages like `{kindling}`, `{brulee}`, `{cito}`, and `{luz}` simplify this process while offering different features and design philosophies.

`{kindling}` distinguishes itself through its unique code generation approach, versatile neural architecture support (can be expanded more in the future), and three-level API design. While `{brulee}` focuses on production-ready statistical models, `{cito}` emphasizes explainability and statistical inference, and `{luz}` provides adaptable training loops. `{kindling}` is different but not mutually exclusive to them: it offers deep architectural control and bridges the gap between torch code and tidymodels workflows.

### Package Comparison

```{r comparison_table, echo=FALSE}
comparison = data.frame(
    Feature = c(
        "**Primary Focus**",
        "**Design Philosophy**",
        "**Architectures**",
        "**Code Generation**",
        "**tidymodels Integration**",
        "**Formula Syntax**",
        "**Layer-specific Activations**",
        "**GPU Support**",
        "**Explainability/xAI**",
        "**Statistical Inference**",
        "**Custom Loss Functions**",
        "**For whom?**"
    ),
    kindling = c(
        "Architectural versatility & flexibility, statistical modelling, and code generation",
        "Three-level API (code gen, training, ML framework (currently tidymodels) integration)",
        "Versatile — Feedforward Neural Networks (DNN/FFNN/MLP), Recurrent Neural Networks (RNN, LSTM, GRU), and more (in the future)",
        "Yes (inspect & modify torch code)",
        "Full (parsnip models & tuning)",
        "Yes",
        "Yes",
        "Yes",
        "Garson's & Olden's algorithms, vip integration, and more in the future",
        "Not yet implemented",
        "Yes",
        "Wanted versatile architectures (more in the future), fine-grained control, tidymodels users"
    ),
    brulee = c(
        "Production-ready statistical models",
        "Batteries-included with sensible defaults",
        "MLP, Linear/Logistic/Multinomial regression",
        "No",
        "Full (official tidymodels package)",
        "Yes",
        "No",
        "Yes",
        "Limited",
        "No",
        "No",
        "Wants standard supervised learning, stable production models"
    ),
    cito = c(
        "Statistical inference & interpretation",
        "User-friendly with comprehensive xAI pipeline",
        "Fully-connected networks, CNNs",
        "No",
        "No (standalone package)",
        "Yes",
        "No",
        "Yes (CPU, GPU, MacOS)",
        "Extensive (PDP, ALE, variable importance, etc.)",
        "Yes (confidence intervals, p-values via bootstrap)",
        "Yes",
        "Do ecological modeling, interpretable models, statistical inference"
    ),
    luz = c(
        "Training loop abstraction",
        "High-level API reducing boilerplate",
        "Any torch nn_module",
        "No",
        "No (standalone package)",
        "No (uses torch modules directly)",
        "No (also uses torch modules directly)",
        "Yes (automatic device placement)",
        "No",
        "No",
        "Yes",
        "Wants custom architectures, users needing human-friendly training loop control"
    ),
    stringsAsFactors = FALSE
)

knitr::kable(
    comparison, 
    col.names = c("Feature", "kindling", "brulee", "cito", "luz"),
    label = "Table of comparison"
)
```

## Complementary Use

These packages aren't mutually exclusive. You can use `{kindling}` and `{brulee}` for production MLPs, except `{kindling}` provides RNNs. The `{cito}` package is when you need oriented model interpretation. Then, the `{luz}` package is ideal when you want less verbose training loops. Despite the difference between the philosophies and main usage, all integrate with or build upon the torch ecosystem, allowing you to switch between them as your modeling needs evolve.

For instance, prototyping with `{kindling}` to explore different network architectures is much easier, as well as deploying its models for production, just like `{brulee}`, but `{cito}` is when stakeholders need detailed explanations of model predictions.

---
title: "Custom Activation Function"
output:
    rmarkdown::html_vignette:
        toc: true
vignette: >
  %\VignetteIndexEntry{Custom Activation Function}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE, warning=FALSE, message=FALSE}
knitr::opts_chunk$set(
    eval = identical(Sys.getenv("NOT_CRAN"), "true"),
    fig.width = 7,
    fig.height = 5,
    warning = FALSE,
    message = FALSE
)
library(kindling)
```

# Rationale

The biggest strength of `{kindling}` when modelling neural networks is its versatility — it inherits `{torch}`'s versatility while being human friendly, including the ability to apply custom optimizer functions, loss functions, and per-layer activation functions. Learn more: [https://kindling.joshuamarie.com/articles/special-cases](https://kindling.joshuamarie.com/articles/special-cases).

With `act_funs()`, you are not limited to the activation functions available in `{torch}`'s namespace. Use `new_act_fn()` to wrap any compatible function into a validated custom activation. This feature, however, is only available on version 0.3.0 and above. 

# Function to use

To do this, use `new_act_fn()`. It takes a user-supplied function, validates it against a small dummy tensor at *definition time* (a dry-run probe), and wraps it in a call-time type guard. This means errors surface early — before your model ever starts training.

The function you supply must:

- Accept at least one argument (the input tensor).
- Return a `torch_tensor`.

## Basic Usage

Currently, `nnf_tanh` doesn't exist in the `{torch}` namespace, so `tanh` is not a valid argument to `act_funs()`. With `new_act_fn()`, you can wrap `torch::torch_tanh()` to make it usable. 

Here's a basic example that wraps `torch::torch_tanh()` as a custom activation:

``` r
hyper_tan = new_act_fn(\(x) torch::torch_tanh(x))
```

You can also pass it directly into `act_funs()`, just like any built-in activation:

``` r
act_funs(relu, elu, new_act_fn(\(x) torch::torch_tanh(x)))
```

## Using Custom Activations in a Model

Naturally, functions for modelling, like `ffnn()`, accept `act_funs()` into the `activations` argument. Again, you can pass a custom activation function within `new_act_fn()`, then pass it through `act_funs()`. 

Here's a basic example: 

```{r}
model = ffnn(
    Sepal.Length ~ .,
    data = iris[, 1:4],
    hidden_neurons = c(64, 32, 16),
    activations = act_funs(
        relu,
        silu,
        new_act_fn(\(x) torch::torch_tanh(x))
    ),
    epochs = 50
)
model
```

Each element of `act_funs()` corresponds to one hidden layer, in order. Here, the first hidden layer uses ReLU, the second uses SiLU (Swish), and the third uses Tanh.

You can also use a single custom activation recycled across all layers:

```{r}
ffnn(
    Sepal.Length ~ .,
    data = iris[, 1:4],
    hidden_neurons = c(64, 32),
    activations = act_funs(new_act_fn(\(x) torch::torch_tanh(x))),
    epochs = 50
)
```

## Skipping the Dry-Run Probe

By default, `new_act_fn()` runs a quick dry-run with a small dummy tensor to validate your function before training. You can disable this with `probe = FALSE`, though this is generally not recommended:

```{r, eval = FALSE}
my_act = new_act_fn(\(x) torch::torch_tanh(x), probe = FALSE)
```

## Naming Your Custom Activation

You can provide a human-readable name via `.name`, which is used in print output and diagnostics:

```{r}
my_act = new_act_fn(\(x) torch::torch_tanh(x), .name = "my_tanh")
```

Here's a simple application:

```{r}
ffnn(
    Sepal.Length ~ .,
    data = iris[, 1:4],
    hidden_neurons = c(64, 32),
    activations = act_funs(
        relu, 
        new_act_fn(\(x) torch::torch_tanh(x), .name = "hyper_tanh")
    ),
    epochs = 50
)
```

# Error Handling

`new_act_fn()` is designed to fail loudly and early. Common errors include:

1.  Function returns a non-tensor. This will error at definition time:

    ```{r, error = TRUE}
    new_act_fn(\(x) as.numeric(x))
    ```

2.  Function accepts no arguments. This will error immediately:

    ```{r, error = TRUE}
    new_act_fn(function() torch::torch_zeros(2))
    ```

These checks ensure your model's architecture is valid before any data ever flows through it.

# Summary

| Feature | Details |
|---|---|
| Wraps any R function | Must accept a tensor, return a tensor |
| Dry-run probe | Validates at definition time (`probe = TRUE` by default) |
| Call-time guard | Type-checks output on every forward pass |
| Compatible with `act_funs()` | Use alongside built-in activations freely |
| Closures supported | Parametric activations work naturally |

---
title: "Special Cases: Linear and Logistic Regression"
output:
    rmarkdown::html_vignette:
        toc: true
vignette: >
  %\VignetteIndexEntry{Special Cases: Linear and Logistic Regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    eval = identical(Sys.getenv("NOT_CRAN"), "true"), 
    fig.width = 7,
    fig.height = 5,
    warning = FALSE,
    message = FALSE
)

# Sys.setenv("_R_USE_PIPEBIND_" = TRUE)
```

## What's so special about {kindling}

This package is planned to make it compatible for any machine learning task, even time series and image classification cam be supported. Yes, you can do both linear regression and logistic regression with extra steps: heavily customized optimizer and loss functions. The `train_nn()` function (available on >v0.3.x) supports this { `optimizer` $\leftrightarrow$ `optimizer_args` } and { `loss` }. For both cases, the key is to remove all hidden layers and rely entirely on the output layer and the appropriate loss function to recover the classical model's behavior.

## Setup

```{r}
box::use(
    kindling[train_nn, act_funs, args],
    recipes[
        recipe, step_dummy, step_normalize,
        all_nominal_predictors, all_numeric_predictors
    ],
    rsample[initial_split, training, testing],
    yardstick[metric_set, rmse, rsq, accuracy, mn_log_loss],
    dplyr[mutate, select],
    tibble[tibble]
)
```

## Linear Regression as a Special Case

A standard linear regression model predicts a continuous outcome as a weighted sum of inputs — no nonlinearity, no hidden layers. A neural network recovers this exactly when:

- There are *no hidden layers* (`hidden_neurons = integer(0)` or simply omit it),
- The *output activation is the identity* (i.e., no activation), and
- The common loss function is MSE, but we can choose different loss function:  (`loss = "mse"`).

Under these conditions, gradient descent minimizes the same objective as ordinary least squares, and the learned weights converge to the OLS solution given sufficient epochs and a small learning rate.

### Data

We use `mtcars` to predict fuel efficiency (`mpg`) from the other variables.

```{r linear-data}
set.seed(42)
split = initial_split(mtcars, prop = 0.8)
train = training(split)
test = testing(split)

rec = recipe(mpg ~ ., data = train) |>
    step_normalize(all_numeric_predictors())
```

### Fitting the model

To create no hidden units, the `hidden_neuron` parameter from `train_nn()` considers the following to achieve:

1. `NULL`
2.  Empty `c()`
3.  No arguments at all

In this example, the empty vector `c()` is used and will collapse the network to a single linear layer from inputs to output. The `optimizer = "rmsprop"` with a small `learn_rate` mirrors classical gradient descent for OLS.

```{r linear-fit}
lm_nn = train_nn(
    mpg ~ .,
    data = train,
    hidden_neurons = c(),
    loss = torch::nnf_l1_loss,
    optimizer = "rmsprop", 
    learn_rate = 0.01,
    epochs = 200,
    verbose = FALSE
)

lm_nn
```

### Evaluation

```{r linear-eval}
preds = predict(lm_nn, newdata = test)

tibble(
    truth = test$mpg,
    estimate = preds
) |>
    metric_set(rmse, rsq)(truth = truth, estimate = estimate)
```

### Comparison with `lm()`

```{r linear-compare}
lm_fit = lm(mpg ~ ., data = train)

tibble(
    truth = test$mpg,
    estimate = predict(lm_fit, newdata = test)
) |>
    metric_set(rmse, rsq)(truth = truth, estimate = estimate)
```

The two models should produce very similar RMSE and $R^2$ values. Any small gap reflects that gradient descent is an iterative approximation, while `lm()` solves for the exact OLS coefficients directly. Increasing `epochs` or switching to `optimizer = "lbfgs"` (if supported) will close the gap further.

## Logistic Regression as a Special Case

Logistic regression models a binary or multiclass outcome by passing a linear combination of inputs through a sigmoid or softmax activation. A neural network with:

- **No hidden layers**,
- A **sigmoid output** for binary classification (or softmax for multiclass), and
- **Cross-entropy** (`loss = "cross_entropy"`) for the loss function

is mathematically equivalent to logistic regression.

### Binary Logistic Regression

We use the `Sonar` dataset from `{mlbench}` to distinguish rocks from mines (binary outcome).

```{r binary-data}
data("Sonar", package = "mlbench")

sonar = Sonar
set.seed(42)
split_s = initial_split(sonar, prop = 0.8, strata = Class)
train_s = training(split_s)
test_s = testing(split_s)

rec_s = recipe(Class ~ ., data = train_s) |>
    step_normalize(all_numeric_predictors())
```

```{r binary-fit}
logit_nn = train_nn(
    Class ~ .,
    data = train_s,
    hidden_neurons = c(),
    loss = "cross_entropy",
    optimizer = "adam",
    learn_rate = 0.01,
    epochs = 200,
    verbose = FALSE
)

logit_nn
```

```{r binary-eval}
preds_s = predict(logit_nn, newdata = test_s, type = "response")

tibble(
    truth = test_s$Class,
    estimate = preds_s
) |>
    accuracy(truth = truth, estimate = estimate)
```

<!-- ### Multiclass Logistic Regression -->

<!-- For more than two classes, cross-entropy loss with a softmax output generalizes logistic regression. We demonstrate on `iris` with three species. -->

<!-- ```{r multiclass-data} -->
<!-- set.seed(42) -->
<!-- split_i = initial_split(iris, prop = 0.8, strata = Species) -->
<!-- train_i = training(split_i) -->
<!-- test_i = testing(split_i) -->

<!-- rec_i = recipe(Species ~ ., data = train_i) |> -->
<!--     step_normalize(all_numeric_predictors()) -->
<!-- ``` -->

<!-- ```{r multiclass-fit} -->
<!-- softmax_nn = train_nn( -->
<!--     Species ~ ., -->
<!--     data = train_i, -->
<!--     hidden_neurons = integer(0), -->
<!--     loss = "cross_entropy", -->
<!--     optimizer = "adam", -->
<!--     learn_rate = 0.01, -->
<!--     epochs = 300, -->
<!--     verbose = FALSE -->
<!-- ) -->

<!-- softmax_nn -->
<!-- ``` -->

<!-- ```{r multiclass-eval} -->
<!-- preds_i = predict(softmax_nn, newdata = test_i, type = "class") -->

<!-- tibble( -->
<!--     truth = test_i$Species, -->
<!--     estimate = preds_i -->
<!-- ) |> -->
<!--     accuracy(truth = truth, estimate = estimate) -->
<!-- ``` -->

### Comparison with `glm()` / `nnet::multinom()`

```{r logit-compare}
box::use(nnet[multinom])

glm_fit = glm(Class ~ ., data = train_s, family = binomial())

tibble(
    truth = test_s$Class,
    estimate = {
        as.factor({
            preds = predict(glm_fit, newdata = test_s, type = "response")
            ifelse(preds < 0.5, "M", "R")
        })
    }
) |>
    accuracy(truth = truth, estimate = estimate)
```

<!-- > To learn more about `=>`, it is a pipe-bind operator by running this on your R console: `Sys.setenv("_R_USE_PIPEBIND_" = TRUE)` -->

Again, accuracy should be comparable between the two approaches. The neural network version converges iteratively, so the match is not guaranteed to be exact, but both are optimizing the same cross-entropy objective over a linear model.

---
title: "Getting Started with kindling"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with kindling}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    eval = identical(Sys.getenv("NOT_CRAN"), "true"),
    fig.width = 7,
    fig.height = 5,
    warning = FALSE,
    message = FALSE
)

fansi::set_knit_hooks(knitr::knit_hooks)
```

## Introduction

`{kindling}` bridges the gap between `{torch}` and `{tidymodels}`, providing a streamlined interface for building, training, and tuning deep learning models. This vignette will guide you through the basic usage.

## Installation

You can install `{kindling}` on CRAN:

``` r
install.packages('kindling')
```

Or install the development version from GitHub:

```{r eval = FALSE}
# install.packages("pak")
pak::pak("joshuamarie/kindling")
## devtools::install_github("joshuamarie/kindling") 
```

## Before using {kindling}

```{r}
library(kindling)
```

Before starting, you need to install LibTorch first, the backend of PyTorch, which is also the backend of `{torch}` R package:

``` r
torch::install_torch()
```

## Main Features

Current `{kindling}` supports the following: 

-   Code generation of `{torch}` expression
-   Multiple architectures available

    -  Base models interface: feedforward networks (MLP/DNN/FFNN) and recurrent variants (RNN, LSTM, GRU)
    -  Generalized neural network trainer that has the same topology as MLPs

-   Native support for R ML workflows and pipelines (currently `{tidymodels}`; `{mlr3}` planned)
-   Fine-grained control over network depth, layer sizes, and activation functions
-   GPU acceleration support via `{torch}` tensors

### What it doesn't support

As of `{kindling}` >0.3.0, it supports most of NN architectures thanks to its versatility, as long as they follow typical MLP's topology. This package, however, does not support the following:

1.  Residual Networks (ResNet)
2.  Automatic Integration (AutoInt)
3.  Self-Attention and Inter-sample Attention Transformer (Saint)

To use all of these, you might want to take an interest towards `{brulee}` package instead. The said NN architectures above are available on version 1.0.0 (and later) release. 

## Usage: Three Levels of Interaction

`{kindling}` is powered by R's metaprogramming capabilities through *code generation*. Generated `torch::nn_module()` expressions power the training functions, which in turn serve as engines for `{tidymodels}` integration. This architecture gives you flexibility to work at whatever abstraction level suits your task.

### Level 1: Code Generation for `torch::nn_module`

At the lowest level, you can generate raw `torch::nn_module` code for maximum customization. Functions ending with `_generator` return unevaluated expressions you can inspect, modify, or execute.

Here's how to generate a feedforward network specification:

```{r, collapse=FALSE, comment=""}
ffnn_generator(
    nn_name = "MyFFNN",
    hd_neurons = c(64, 32, 16),
    no_x = 10,
    no_y = 1,
    activations = 'relu'
)
```

This creates a three-hidden-layer network (64 - 32 - 16 neurons) that takes 10 inputs and produces 1 output. Each hidden layer uses ReLU activation, while the output layer remains "untransformed".

### Level 2: Direct Training Interface

Skip the code generation and train models directly with your data. This approach handles all the `{torch}` boilerplate when training the models internally.

Let's classify iris species:

```{r, comment="", collapse=FALSE} 
model = ffnn(
    Species ~ .,
    data = iris,
    hidden_neurons = c(10, 15, 7),
    activations = act_funs(relu, softshrink[lambd = 0.5], elu), 
    loss = "cross_entropy",
    epochs = 100
)

model
```

> For parametric activation functions like softshrink, which contains `"lambd"` ($\lambda$) as its parameter (the default is 1), use indexed syntax (available on v0.3.x+) e.g. `softshrink[lambd = 0.5]` or `softshrink[0.5]`, or a string literal expression e.g. `"softshrink(lambd = 0.5)"`, to transmute the parameter value. See `?kindling::act_funs()` for more details. 

Evaluate the prediction through `predict()`. The `predict()` method is extended for fitted models through its `newdata` argument. 

Two kinds of `predict()` usage: 

1.  **Without `newdata`** predictions default to the training data. 

    ```{r}
    predict(model) |>
        (\(x) table(actual = iris$Species, predicted = x))()
    ```


2.  **With `newdata`** simply pass the new data frame as the new reference. 

    ```{r}
    sample_iris = dplyr::slice_sample(iris, n = 10, by = Species)

    predict(model, newdata = sample_iris) |>
        (\(x) table(actual = sample_iris$Species, predicted = x))()
    ```


### Level 3: Conventional tidymodels Integration

Work with neural networks just like any other `{parsnip}` model. This unlocks the entire `{tidymodels}` toolkit for preprocessing, cross-validation, and model evaluation.

```{r}
# library(kindling)
# library(parsnip)
# library(yardstick)
box::use(
    kindling[mlp_kindling, rnn_kindling, act_funs, args],
    parsnip[fit, augment],
    yardstick[metrics]
)
data(Ionosphere, package = "mlbench")

ionosphere_data = Ionosphere[, -2]

# Train a feedforward network with parsnip
mlp_kindling(
    mode = "classification",
    hidden_neurons = c(128, 64),
    activations = act_funs(relu, softshrink[lambd = 0.5]),
    epochs = 100
) |>
    fit(Class ~ ., data = ionosphere_data) |>
    augment(new_data = ionosphere_data) |>
    metrics(truth = Class, estimate = .pred_class)

# Or try a recurrent architecture (demonstrative example with tabular data)
rnn_kindling(
    mode = "classification",
    hidden_neurons = c(128, 64),
    activations = act_funs(relu, elu),
    epochs = 100,
    rnn_type = "gru"
) |>
    fit(Class ~ ., data = ionosphere_data) |>
    augment(new_data = ionosphere_data) |>
    metrics(truth = Class, estimate = .pred_class)
```

## Hyperparameter Tuning & Resampling

The package has integration with `{tidymodels}`, so it supports hyperparameter tuning via `{tune}` with searchable parameters.

The current searchable parameters under `{kindling}`: 

-   Layer widths (neurons per layer)
-   Network depth (number of hidden layers)
-   Activation function combinations
-   Output activation
-   Optimizer (Type of optimization algorithm)
-   Bias (choose between the presence and the absence of the bias term)
-   Validation Split Proportion
-   Bidirectional (boolean; only for RNN)

The searchable parameters outside from `{kindling}`, i.e. under `{dials}` package such as `learn_rate()` also supported. 

Here's an example: 

```{r, collapse=FALSE, comment=""}
# library(tidymodels)
box::use(
    kindling[
        mlp_kindling, hidden_neurons, activations, output_activation, grid_depth
    ],
    parsnip[fit, augment],
    recipes[recipe],
    workflows[workflow, add_recipe, add_model],
    rsample[vfold_cv],
    tune[tune_grid, tune, select_best, finalize_workflow],
    dials[grid_random],
    yardstick[accuracy, roc_auc, metric_set, metrics]
)

mlp_tune_spec = mlp_kindling(
    mode = "classification",
    hidden_neurons = tune(),
    activations = tune(),
    output_activation = tune()
)

iris_folds = vfold_cv(iris, v = 3)
nn_wf = workflow() |>
    add_recipe(recipe(Species ~ ., data = iris)) |>
    add_model(mlp_tune_spec)

nn_grid_depth = grid_depth(
    hidden_neurons(c(32L, 128L)),
    activations(c("relu", "elu")),
    output_activation(c("sigmoid", "linear")),
    n_hlayer = 2,
    size = 10,
    type = "latin_hypercube"
)

# This is supported but limited to 1 hidden layer only
## nn_grid = grid_random(
##     hidden_neurons(c(32L, 128L)),
##     activations(c("relu", "elu")),
##     output_activation(c("sigmoid", "linear")),
##     size = 10
## )

nn_tunes = tune::tune_grid(
    nn_wf,
    iris_folds,
    grid = nn_grid_depth
)

best_nn = select_best(nn_tunes)
best_nn
final_nn = finalize_workflow(nn_wf, best_nn)
final_nn_model = fit(final_nn, data = iris)
final_nn_model

final_nn_model |>
    augment(new_data = iris) |>
    metrics(truth = Species, estimate = .pred_class)
```

Resampling strategies from `{rsample}` will enable robust cross-validation workflows, orchestrated through the `{tune}` and `{dials}` APIs.

## Variable Importance

`{kindling}` integrates with established variable importance methods from `{NeuralNetTools}` and `{vip}` to interpret trained neural networks. Two primary algorithms are available:

1.  Garson's Algorithm

    ```{r}
    garson(model, bar_plot = FALSE)
    ```

2.  Olden's Algorithm

    ```{r}
    olden(model, bar_plot = FALSE)
    ```

### Integration with {vip}

For users working within the `{tidymodels}` ecosystem, `{kindling}` models work seamlessly with the `{vip}` package:

```{r, fig.height=6, fig.width=8}
#| label: example
#| fig.alt: "Variable Importance Plot"
box::use(
    vip[vi, vip]
)

vi(model) |> 
    vip()
```

*Note: Weight caching increases memory usage proportional to network size. Only enable it when you plan to compute variable importance multiple times on the same model.*

## Learn More

- Visit the package website: https://kindling.joshuamarie.com
- Report issues: https://github.com/joshuamarie/kindling/issues

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print-gen-nn.R
\name{print.nn_fit}
\alias{print.nn_fit}
\title{Print method for nn_fit objects}
\usage{
\method{print}{nn_fit}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"nn_fit"}}

\item{...}{Additional arguments (unused)}
}
\value{
No return value, called for side effects (printing model summary)
}
\description{
Print method for nn_fit objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/act-fun.R
\name{args}
\alias{args}
\title{Activation Function Arguments Helper}
\usage{
args(...)
}
\arguments{
\item{...}{Named arguments for the activation function.}
}
\value{
A list with class "activation_args" containing the parameters.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

This is superseded in v0.3.0 in favour of indexed syntax, e.g. \verb{<act_fn[param = 0]>} type.
Type-safe helper to specify parameters for activation functions.
All parameters must be named and match the formal arguments of the
corresponding \code{{torch}} activation function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen-train-nn-parsnip.R, R/mlp_kindling.R,
%   R/rnn_kindling.R
\name{train_nn_wrapper}
\alias{train_nn_wrapper}
\alias{ffnn_wrapper}
\alias{rnn_wrapper}
\title{kindling-tidymodels wrapper}
\usage{
train_nn_wrapper(formula, data, ...)

ffnn_wrapper(formula, data, ...)

rnn_wrapper(formula, data, ...)
}
\arguments{
\item{formula}{A formula specifying the model (e.g., \code{y ~ x1 + x2})}

\item{data}{A data frame containing the training data}

\item{...}{Additional arguments passed to the underlying training function}
}
\value{
\code{train_nn_wrapper()} returns an \code{"nn_fit_tab"} object. See \code{\link[=train_nn]{train_nn()}} for details.

\itemize{
\item \code{ffnn_wrapper()} returns an object of class \code{"ffnn_fit"} containing the trained feedforward neural network model and metadata. See \code{\link[=ffnn]{ffnn()}} for details.
\item \code{rnn_wrapper()} returns an object of class \code{"rnn_fit"} containing the trained recurrent neural network model and metadata. See \code{\link[=rnn]{rnn()}} for details.
}
}
\description{
kindling-tidymodels wrapper

Basemodels-tidymodels wrappers
}
\details{
This wrapper function is designed to interface with the \code{{tidymodels}}
ecosystem, particularly for use with \code{\link[tune:tune_grid]{tune::tune_grid()}} and workflows.
It handles the conversion of tuning parameters (especially list-column
parameters from \code{\link[=grid_depth]{grid_depth()}}) into the format expected by \code{\link[=train_nn]{train_nn()}}.

These wrapper functions are designed to interface with the \code{{tidymodels}}
ecosystem, particularly for use with \code{\link[tune:tune_grid]{tune::tune_grid()}} and workflows.
They handle the conversion of tuning parameters (especially list-column
parameters from \code{\link[=grid_depth]{grid_depth()}}) into the format expected by \code{\link[=ffnn]{ffnn()}} and \code{\link[=rnn]{rnn()}}.
}
\section{MLP Wrapper for \code{{tidymodels}} interface}{

Internal wrapper — use \code{mlp_kindling()} + \code{fit()} instead.
}

\section{FFNN (MLP) Wrapper for \code{{tidymodels}} interface}{

This is a function to interface into \code{{tidymodels}}
(do not use this, use \code{kindling::ffnn()} instead).
}

\section{RNN Wrapper for \code{{tidymodels}} interface}{

Internal wrapper — use \code{rnn_kindling()} + \code{fit()} instead.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/torch-code-gen.R
\name{layer-attributes}
\alias{layer-attributes}
\alias{$.layer_pr}
\title{"Layer" attributes}
\usage{
\method{$}{layer_pr}(x, name)
}
\arguments{
\item{x}{The .layer itself}

\item{name}{It could be the following:
\itemize{
\item \code{i}: Layer index (1-based integer)
\item \code{ind}: Input dimension for the layer
\item \code{out}: Output dimension for the layer
\item \code{is_output}: Logical indicating if this is the output layer
}}
}
\value{
A pronoun, it returns nothing
}
\description{
"Layer" attributes
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_depth.R
\name{grid_depth}
\alias{grid_depth}
\alias{grid_depth.parameters}
\alias{grid_depth.list}
\alias{grid_depth.workflow}
\alias{grid_depth.model_spec}
\alias{grid_depth.param}
\alias{grid_depth.default}
\title{Depth-Aware Grid Generation for Neural Networks}
\usage{
grid_depth(
  x,
  ...,
  n_hlayer = 2L,
  size = 5L,
  type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
  original = TRUE,
  levels = 3L,
  variogram_range = 0.5,
  iter = 1000L
)

\method{grid_depth}{parameters}(
  x,
  ...,
  n_hlayer = 2L,
  size = 5L,
  type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
  original = TRUE,
  levels = 3L,
  variogram_range = 0.5,
  iter = 1000L
)

\method{grid_depth}{list}(
  x,
  ...,
  n_hlayer = 2L,
  size = 5L,
  type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
  original = TRUE,
  levels = 3L,
  variogram_range = 0.5,
  iter = 1000L
)

\method{grid_depth}{workflow}(
  x,
  ...,
  n_hlayer = 2L,
  size = 5L,
  type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
  original = TRUE,
  levels = 3L,
  variogram_range = 0.5,
  iter = 1000L
)

\method{grid_depth}{model_spec}(
  x,
  ...,
  n_hlayer = 2L,
  size = 5L,
  type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
  original = TRUE,
  levels = 3L,
  variogram_range = 0.5,
  iter = 1000L
)

\method{grid_depth}{param}(
  x,
  ...,
  n_hlayer = 2L,
  size = 5L,
  type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
  original = TRUE,
  levels = 3L,
  variogram_range = 0.5,
  iter = 1000L
)

\method{grid_depth}{default}(
  x,
  ...,
  n_hlayer = 2L,
  size = 5L,
  type = c("regular", "random", "latin_hypercube", "max_entropy", "audze_eglais"),
  original = TRUE,
  levels = 3L,
  variogram_range = 0.5,
  iter = 1000L
)
}
\arguments{
\item{x}{A \code{parameters} object, list, workflow, or model spec. Can also be
a single \code{param} object if \code{...} contains additional param objects.}

\item{...}{One or more \code{param} objects (e.g., \code{hidden_neurons()}, \code{epochs()}).
If \code{x} is a \code{parameters} object, \code{...} is ignored. None of the objects can
have \code{unknown()} values.}

\item{n_hlayer}{Integer vector specifying number of hidden layers to generate
(e.g., \code{2:4} for 2, 3, or 4 layers), or a \code{param} object created with \code{n_hlayers()}.
Default is 2.}

\item{size}{Integer. Number of parameter combinations to generate.}

\item{type}{Character. Type of grid: "regular", "random", "latin_hypercube",
"max_entropy", or "audze_eglais".}

\item{original}{Logical. Should original parameter ranges be used?}

\item{levels}{Integer. Levels per parameter for regular grids.}

\item{variogram_range}{Numeric. Range for audze_eglais design.}

\item{iter}{Integer. Iterations for max_entropy optimization.}
}
\value{
A tibble with list columns for \code{hidden_neurons} and \code{activations},
where each element is a vector of length \code{n_hlayer}.
}
\description{
\code{grid_depth()} extends standard grid generation to support multi-layer
neural network architectures. It creates heterogeneous layer configurations
by generating list columns for \code{hidden_neurons} and \code{activations}.
}
\details{
This function is specifically for \code{{kindling}} models. The \code{n_hlayer} parameter
determines network depth and creates list columns for \code{hidden_neurons} and
\code{activations}, where each element is a vector of length matching the sampled depth.

When \code{n_hlayer} is a parameter object (created with \code{n_hlayers()}), it will be
treated as a tunable parameter and sampled according to its defined range.
}
\examples{
\donttest{
\dontrun{
library(dials)
library(workflows)
library(tune)

# Method 1: Fixed depth
grid = grid_depth(
    hidden_neurons(c(32L, 128L)),
    activations(c("relu", "elu")),
    epochs(c(50L, 200L)),
    n_hlayer = 2:3,
    type = "random",
    size = 20
)

# Method 2: Tunable depth using parameter object
grid = grid_depth(
    hidden_neurons(c(32L, 128L)),
    activations(c("relu", "elu")),
    epochs(c(50L, 200L)),
    n_hlayer = n_hlayers(range = c(2L, 4L)),
    type = "random",
    size = 20
)

# Method 3: From workflow
wf = workflow() |>
    add_model(mlp_kindling(hidden_neurons = tune(), activations = tune())) |>
    add_formula(y ~ .)
grid = grid_depth(wf, n_hlayer = 2:4, type = "latin_hypercube", size = 15)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_depth.R
\name{extract_depth_param}
\alias{extract_depth_param}
\title{Extract depth parameter values from n_hlayer argument}
\usage{
extract_depth_param(n_hlayer, param_list = list(), levels = 3L)
}
\arguments{
\item{n_hlayer}{Either an integer vector or a param object}

\item{param_list}{List of parameters (for extracting n_hlayers if present)}

\item{levels}{Number of levels for regular grids}
}
\value{
List with \code{values} component containing integer vector of depths
}
\description{
Extract depth parameter values from n_hlayer argument
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalized-nn-fit.R, R/generalized-nn-fitds.R
\name{gen-nn-train}
\alias{gen-nn-train}
\alias{train_nn}
\alias{train_nn.matrix}
\alias{train_nn.data.frame}
\alias{train_nn.formula}
\alias{train_nn.default}
\alias{train_nn.dataset}
\title{Generalized Neural Network Trainer}
\usage{
train_nn(x, ...)

\method{train_nn}{matrix}(
  x,
  y,
  hidden_neurons = NULL,
  activations = NULL,
  output_activation = NULL,
  bias = TRUE,
  arch = NULL,
  architecture = NULL,
  early_stopping = NULL,
  epochs = 100,
  batch_size = 32,
  penalty = 0,
  mixture = 0,
  learn_rate = 0.001,
  optimizer = "adam",
  optimizer_args = list(),
  loss = "mse",
  validation_split = 0,
  device = NULL,
  verbose = FALSE,
  cache_weights = FALSE,
  ...
)

\method{train_nn}{data.frame}(
  x,
  y,
  hidden_neurons = NULL,
  activations = NULL,
  output_activation = NULL,
  bias = TRUE,
  arch = NULL,
  architecture = NULL,
  early_stopping = NULL,
  epochs = 100,
  batch_size = 32,
  penalty = 0,
  mixture = 0,
  learn_rate = 0.001,
  optimizer = "adam",
  optimizer_args = list(),
  loss = "mse",
  validation_split = 0,
  device = NULL,
  verbose = FALSE,
  cache_weights = FALSE,
  ...
)

\method{train_nn}{formula}(
  x,
  data,
  hidden_neurons = NULL,
  activations = NULL,
  output_activation = NULL,
  bias = TRUE,
  arch = NULL,
  architecture = NULL,
  early_stopping = NULL,
  epochs = 100,
  batch_size = 32,
  penalty = 0,
  mixture = 0,
  learn_rate = 0.001,
  optimizer = "adam",
  optimizer_args = list(),
  loss = "mse",
  validation_split = 0,
  device = NULL,
  verbose = FALSE,
  cache_weights = FALSE,
  ...
)

\method{train_nn}{default}(x, ...)

\method{train_nn}{dataset}(
  x,
  y = NULL,
  hidden_neurons = NULL,
  activations = NULL,
  output_activation = NULL,
  bias = TRUE,
  arch = NULL,
  architecture = NULL,
  flatten_input = NULL,
  epochs = 100,
  batch_size = 32,
  penalty = 0,
  mixture = 0,
  learn_rate = 0.001,
  optimizer = "adam",
  optimizer_args = list(),
  loss = "mse",
  validation_split = 0,
  device = NULL,
  verbose = FALSE,
  cache_weights = FALSE,
  n_classes = NULL,
  ...
)
}
\arguments{
\item{x}{Dispatch is based on its current class:
\itemize{
\item \code{matrix}: used directly, no preprocessing applied.
\item \code{data.frame}: preprocessed via \code{hardhat::mold()}. \code{y} may be a vector /
factor / matrix of outcomes, or a formula describing the outcome–predictor
relationship within \code{x}.
\item \code{formula}: combined with \code{data} and preprocessed via \code{hardhat::mold()}.
\item \code{dataset}: a \code{torch} dataset object; batched via \code{torch::dataloader()}.
This is the recommended interface for sequence/time-series and image data.
}}

\item{...}{Additional arguments passed to specific methods.}

\item{y}{Outcome data. Interpretation depends on the method:
\itemize{
\item For the \code{matrix} and \code{data.frame} methods: a numeric vector, factor, or
matrix of outcomes.
\item For the \code{data.frame} method only: alternatively a formula of the form
\code{outcome ~ predictors}, evaluated against \code{x}.
\item Ignored when \code{x} is a formula (outcome is taken from the formula) or a
\code{dataset} (labels come from the dataset itself).
}}

\item{hidden_neurons}{Integer vector specifying the number of neurons in each
hidden layer, e.g. \code{c(128, 64)} for two hidden layers. When \code{NULL} or missing,
no hidden layers are used and the model reduces to a single linear mapping
from inputs to outputs.}

\item{activations}{Activation function specification(s) for the hidden layers.
See \code{\link[=act_funs]{act_funs()}} for supported values. Recycled if a single value is given.}

\item{output_activation}{Optional activation function for the output layer.
Defaults to \code{NULL} (no activation / linear output).}

\item{bias}{Logical. Whether to include bias terms in each layer. Default \code{TRUE}.}

\item{arch}{Backward-compatible alias for \code{architecture}. If both are supplied,
they must be identical.}

\item{architecture}{An \code{\link[=nn_arch]{nn_arch()}} object specifying a custom architecture. Default
\code{NULL}, which falls back to a standard feed-forward network.}

\item{early_stopping}{An \code{\link[=early_stop]{early_stop()}} object specifying early stopping
behaviour, or \code{NULL} (default) to disable early stopping. When supplied,
training halts if the monitored metric does not improve by at least
\code{min_delta} for \code{patience} consecutive epochs.
Example: \code{early_stopping = early_stop(patience = 10)}.}

\item{epochs}{Positive integer. Number of full passes over the training data.
Default \code{100}.}

\item{batch_size}{Positive integer. Number of samples per mini-batch. Default \code{32}.}

\item{penalty}{Non-negative numeric. L1/L2 regularization strength (lambda).
Default \code{0} (no regularization).}

\item{mixture}{Numeric in [0, 1]. Elastic net mixing parameter: \code{0} = pure
ridge (L2), \code{1} = pure lasso (L1). Default \code{0}.}

\item{learn_rate}{Positive numeric. Step size for the optimizer. Default \code{0.001}.}

\item{optimizer}{Character. Optimizer algorithm. One of \code{"adam"} (default),
\code{"sgd"}, or \code{"rmsprop"}.}

\item{optimizer_args}{Named list of additional arguments forwarded to the
optimizer constructor (e.g. \code{list(momentum = 0.9)} for SGD). Default \code{list()}.}

\item{loss}{Character or function. Loss function used during training. Built-in
options: \code{"mse"} (default), \code{"mae"}, \code{"cross_entropy"}, or \code{"bce"}. For
classification tasks with a scalar label, \code{"cross_entropy"} is set
automatically. Alternatively, supply a custom function or formula with
signature \verb{function(input, target)} returning a scalar \code{torch_tensor}.}

\item{validation_split}{Numeric in [0, 1). Proportion of training data held
out for validation. Default \code{0} (no validation set).}

\item{device}{Character. Compute device: \code{"cpu"}, \code{"cuda"}, or \code{"mps"}.
Default \code{NULL}, which auto-detects the best available device.}

\item{verbose}{Logical. If \code{TRUE}, prints loss (and validation loss) at regular
intervals during training. Default \code{FALSE}.}

\item{cache_weights}{Logical. If \code{TRUE}, stores a copy of the trained weight
matrices in the returned object under \verb{$cached_weights}. Default \code{FALSE}.}

\item{data}{A data frame. Required when \code{x} is a formula.}

\item{flatten_input}{Logical or \code{NULL} (dataset method only). Controls whether
each batch/sample is flattened to 2D before entering the model. \code{NULL}
(default) auto-selects: \code{TRUE} when \code{architecture = NULL}, otherwise \code{FALSE}.}

\item{n_classes}{Positive integer. Number of output classes. Required when
\code{x} is a \code{dataset} with scalar (classification) labels; ignored otherwise.}
}
\value{
An object of class \code{"nn_fit"}, or one of its subclasses:
\itemize{
\item \code{c("nn_fit_tab", "nn_fit")} — returned by the \code{data.frame} and \code{formula} methods
\item \code{c("nn_fit_ds", "nn_fit")} — returned by the \code{dataset} method
}

All subclasses share a common structure. See \strong{Details} for the list of
components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{train_nn()} is a generic function for training neural networks with a
user-defined architecture via \code{\link[=nn_arch]{nn_arch()}}. Dispatch is based on the class
of \code{x}.

Recommended workflow:
\enumerate{
\item Define architecture with \code{\link[=nn_arch]{nn_arch()}} (optional).
\item Train with \code{train_nn()}.
\item Predict with \code{\link[=predict.nn_fit]{predict.nn_fit()}}.
}

All methods delegate to a shared implementation core after preprocessing.
When \code{architecture = NULL}, the model falls back to a plain feed-forward neural network
(\code{nn_linear}) architecture.
}
\details{
The returned \code{"nn_fit"} object is a named list with the following components:
\itemize{
\item \code{model} — the trained \code{torch::nn_module} object
\item \code{fitted} — fitted values on the training data (or \code{NULL} for dataset fits)
\item \code{loss_history} — numeric vector of per-epoch training loss, trimmed to
actual epochs run (relevant when early stopping is active)
\item \code{val_loss_history} — per-epoch validation loss, or \code{NULL} if
\code{validation_split = 0}
\item \code{n_epochs} — number of epochs actually trained
\item \code{stopped_epoch} — epoch at which early stopping triggered, or \code{NA} if
training ran to completion
\item \code{hidden_neurons}, \code{activations}, \code{output_activation} — architecture spec
\item \code{penalty}, \code{mixture} — regularization settings
\item \code{feature_names}, \code{response_name} — variable names (tabular methods only)
\item \code{no_x}, \code{no_y} — number of input features and output nodes
\item \code{is_classification} — logical flag
\item \code{y_levels}, \code{n_classes} — class labels and count (classification only)
\item \code{device} — device the model is on
\item \code{cached_weights} — list of weight matrices, or \code{NULL}
\item \code{arch} — the \code{nn_arch} object used, or \code{NULL}
}
}
\section{Supported tasks and input formats}{

\code{train_nn()} is task-agnostic by design (no explicit \code{task} argument).
Task behavior is determined by your input interface and architecture:
\itemize{
\item \strong{Tabular data}: use \code{matrix}, \code{data.frame}, or \code{formula} methods.
\item \strong{Time series}: use the \code{dataset} method with per-item tensors shaped as
\verb{[time, features]} (or your preferred convention) and a recurrent
architecture via \code{\link[=nn_arch]{nn_arch()}}.
\item \strong{Image classification}: use the \code{dataset} method with per-item tensors
shaped for your first layer (commonly \verb{[channels, height, width]} for
\code{torch::nn_conv2d}). If your source arrays are channel-last, reorder in the
dataset or via \code{input_transform}.
}
}

\section{Matrix method}{

When \code{x} is supplied as a raw numeric matrix, no preprocessing is applied.
Data is passed directly to the shared \code{train_nn_impl} core.
}

\section{Data frame method}{

When \code{x} is a data frame, \code{y} can be either a vector / factor / matrix of
outcomes, or a formula of the form \code{outcome ~ predictors} evaluated against
\code{x}. Preprocessing is handled by \code{hardhat::mold()}.
}

\section{Formula method}{

When \code{x} is a formula, \code{data} must be supplied as the data frame against
which the formula is evaluated. Preprocessing is handled by \code{hardhat::mold()}.
}

\section{Dataset method (\code{train_nn.dataset()})}{

Trains a neural network directly on a \code{torch} dataset object. Batching and
lazy loading are handled by \code{torch::dataloader()}, making this method
well-suited for large datasets that do not fit entirely in memory.

Architecture configuration follows the same contract as other \code{train_nn()}
methods via \code{architecture = nn_arch(...)} (or legacy \code{arch = ...}).
For non-tabular inputs (time series, images), set \code{flatten_input = FALSE} to
preserve tensor dimensions expected by recurrent or convolutional layers.

Labels are taken from the second element of each dataset item (i.e.
\code{dataset[[i]][[2]]}), so \code{y} is ignored. When the label is a scalar tensor,
a classification task is assumed and \code{n_classes} must be supplied. The loss
is automatically switched to \code{"cross_entropy"} in that case.

Fitted values are \strong{not} cached in the returned object. Use
\code{\link[=predict.nn_fit_ds]{predict.nn_fit_ds()}} with \code{newdata} to obtain predictions after training.
}

\examples{
\donttest{
if (torch::torch_is_installed()) {
    # Matrix method — no preprocessing
    model = train_nn(
        x = as.matrix(iris[, 2:4]),
        y = iris$Sepal.Length,
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 50
    )

    # Data frame method — y as a vector
    model = train_nn(
        x = iris[, 2:4],
        y = iris$Sepal.Length,
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 50
    )

    # Data frame method — y as a formula evaluated against x
    model = train_nn(
        x = iris,
        y = Sepal.Length ~ . - Species,
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 50
    )

    # Formula method — outcome derived from formula
    model = train_nn(
        x = Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 50
    )

    # No hidden layers — linear model
    model = train_nn(
        x = Sepal.Length ~ .,
        data = iris[, 1:4],
        epochs = 50
    )

    # Architecture object (nn_arch -> train_nn)
    mlp_arch = nn_arch(nn_name = "mlp_model")
    model = train_nn(
        x = Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(64, 32),
        activations = "relu",
        architecture = mlp_arch,
        epochs = 50
    )

    # Custom layer architecture
    custom_linear = torch::nn_module(
        "CustomLinear",
        initialize = function(in_features, out_features, bias = TRUE) {
            self$layer = torch::nn_linear(in_features, out_features, bias = bias)
        },
        forward = function(x) self$layer(x)
    )
    custom_arch = nn_arch(
        nn_name = "custom_linear_mlp",
        nn_layer = ~ custom_linear
    )
    model = train_nn(
        x = Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(16, 8),
        activations = "relu",
        architecture = custom_arch,
        epochs = 50
    )

    # With early stopping
    model = train_nn(
        x = Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 200,
        validation_split = 0.2,
        early_stopping = early_stop(patience = 10)
    )
}
}

\donttest{
if (torch::torch_is_installed()) {
    # torch dataset method — labels come from the dataset itself
    iris_cls_dataset = torch::dataset(
        name = "iris_cls_dataset",
        
        initialize = function(data = iris) {
            self$x = torch::torch_tensor(
                as.matrix(data[, 1:4]),
                dtype = torch::torch_float32()
            )
            # Species is a factor; convert to integer (1-indexed -> keep as-is for cross_entropy)
            self$y = torch::torch_tensor(
                as.integer(data$Species),
                dtype = torch::torch_long()
            )
        },
        
        .getitem = function(i) {
            list(self$x[i, ], self$y[i])
        },
        
        .length = function() {
            self$x$size(1)
        }
    )()
    
    model_nn_ds = train_nn(
        x = iris_cls_dataset,
        hidden_neurons = c(32, 10),
        activations = "relu",
        epochs = 80,
        batch_size = 16,
        learn_rate = 0.01,
        n_classes = 3, # Iris dataset has only 3 species
        validation_split = 0.2,
        verbose = TRUE
    )
    
    pred_nn = predict(model_nn_ds, iris_cls_dataset)
    class_preds = c("Setosa", "Versicolor", "Virginica")[predict(model_nn_ds, iris_cls_dataset)]
    
    # Confusion Matrix
    table(actual = iris$Species, pred = class_preds)
}
}

}
\seealso{
\code{\link[=predict.nn_fit]{predict.nn_fit()}}, \code{\link[=nn_arch]{nn_arch()}}, \code{\link[=act_funs]{act_funs()}}, \code{\link[=early_stop]{early_stop()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/act-fun.R
\name{act_funs}
\alias{act_funs}
\title{Activation Functions Specification Helper}
\usage{
act_funs(...)
}
\arguments{
\item{...}{Activation function specifications. Can be:
\itemize{
\item Bare symbols: \code{relu}, \code{tanh}
\item Character strings (simple): \code{"relu"}, \code{"tanh"}
\item Character strings (with params): \code{"softshrink(lambda = 0.1)"}, \code{"rrelu(lower = 1/5, upper = 1/4)"}
\item Named with parameters: \code{softmax = args(dim = 2L)}
\item Indexed syntax (named): \code{softshrink[lambd = 0.2]}, \code{rrelu[lower = 1/5, upper = 1/4]}
\item Indexed syntax (unnamed): \code{softshrink[0.5]}, \code{elu[0.5]}
}}
}
\value{
A \code{vctrs} vector with class "activation_spec" containing validated
activation specifications.
}
\description{
This function is a DSL function, kind of like \code{ggplot2::aes()}, that helps to
specify activation functions for neural network layers. It validates that
activation functions exist in \code{torch} and that any parameters match the
function's formal arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlp.R, R/rnn.R
\name{kindling-basemodels}
\alias{kindling-basemodels}
\alias{ffnn}
\alias{rnn}
\title{Base models for Neural Network Training in kindling}
\usage{
ffnn(
  formula = NULL,
  data = NULL,
  hidden_neurons,
  activations = NULL,
  output_activation = NULL,
  bias = TRUE,
  epochs = 100,
  batch_size = 32,
  penalty = 0,
  mixture = 0,
  learn_rate = 0.001,
  optimizer = "adam",
  optimizer_args = list(),
  loss = "mse",
  validation_split = 0,
  device = NULL,
  verbose = FALSE,
  cache_weights = FALSE,
  ...,
  x = NULL,
  y = NULL
)

rnn(
  formula = NULL,
  data = NULL,
  hidden_neurons,
  rnn_type = "lstm",
  activations = NULL,
  output_activation = NULL,
  bias = TRUE,
  bidirectional = TRUE,
  dropout = 0,
  epochs = 100,
  batch_size = 32,
  penalty = 0,
  mixture = 0,
  learn_rate = 0.001,
  optimizer = "adam",
  optimizer_args = list(),
  loss = "mse",
  validation_split = 0,
  device = NULL,
  verbose = FALSE,
  cache_weights = FALSE,
  ...,
  x = NULL,
  y = NULL
)
}
\arguments{
\item{formula}{Formula. Model formula (e.g., y ~ x1 + x2).}

\item{data}{Data frame. Training data.}

\item{hidden_neurons}{Integer vector. Number of neurons in each hidden layer.}

\item{activations}{Activation function specifications. See \code{act_funs()}.}

\item{output_activation}{Optional. Activation for output layer.}

\item{bias}{Logical. Use bias weights. Default \code{TRUE}.}

\item{epochs}{Integer. Number of training epochs. Default \code{100}.}

\item{batch_size}{Integer. Batch size for training. Default \code{32}.}

\item{penalty}{Numeric. Regularization penalty (lambda). Default \code{0} (no regularization).}

\item{mixture}{Numeric. Elastic net mixing parameter (0-1). Default \code{0}.}

\item{learn_rate}{Numeric. Learning rate for optimizer. Default \code{0.001}.}

\item{optimizer}{Character. Optimizer type ("adam", "sgd", "rmsprop"). Default \code{"adam"}.}

\item{optimizer_args}{Named list. Additional arguments passed to the optimizer. Default \code{list()}.}

\item{loss}{Character. Loss function ("mse", "mae", "cross_entropy", "bce"). Default \code{"mse"}.}

\item{validation_split}{Numeric. Proportion of data for validation (0-1). Default \code{0}.}

\item{device}{Character. Device to use ("cpu", "cuda", "mps"). Default \code{NULL} (auto-detect).}

\item{verbose}{Logical. Print training progress. Default \code{FALSE}.}

\item{cache_weights}{Logical. Cache weight matrices for faster variable importance. Default \code{FALSE}.}

\item{...}{Additional arguments. Can be used to pass \code{x} and \code{y} for direct interface.}

\item{x}{When not using formula: predictor data (data.frame or matrix).}

\item{y}{When not using formula: outcome data (vector, factor, or matrix).}

\item{rnn_type}{Character. Type of RNN ("rnn", "lstm", "gru"). Default \code{"lstm"}.}

\item{bidirectional}{Logical. Use bidirectional RNN. Default \code{TRUE}.}

\item{dropout}{Numeric. Dropout rate between layers. Default \code{0}.}
}
\value{
An object of class "ffnn_fit" containing the trained model and metadata.
}
\description{
Base models for Neural Network Training in kindling
}
\section{FFNN}{

Train a feed-forward neural network using the torch package.
}

\section{RNN}{

Train a recurrent neural network using the torch package.
}

\examples{
\donttest{
if (torch::torch_is_installed()) {
    # Formula interface (original)
    model_reg = ffnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 50
    )

    # XY interface (new)
    model_xy = ffnn(
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 50,
        x = iris[, 2:4],
        y = iris$Sepal.Length
    )
}
}

\donttest{
if (torch::torch_is_installed()) {
    # Formula interface (original)
    model_rnn = rnn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(64, 32),
        rnn_type = "lstm",
        activations = "relu",
        epochs = 50
    )

    # XY interface (new)
    model_xy = rnn(
        hidden_neurons = c(64, 32),
        rnn_type = "gru",
        epochs = 50,
        x = iris[, 2:4],
        y = iris$Sepal.Length
    )
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-device.R
\name{validate_device}
\alias{validate_device}
\title{Validate device and get default device}
\usage{
validate_device(device)
}
\arguments{
\item{device}{Character. Requested device.}
}
\value{
Character string of validated device.
}
\description{
Check if requested device is available. And auto-detect available GPU device
or fallback to CPU.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{autoplot.nn_fit}
\alias{autoplot.nn_fit}
\alias{plot.nn_fit}
\title{Plot training loss history for a fitted neural network}
\usage{
\method{autoplot}{nn_fit}(object, ...)

\method{plot}{nn_fit}(x, ...)
}
\arguments{
\item{object}{A fitted model of class \code{nn_fit}, as returned by
\code{\link[=train_nn]{train_nn()}}, \code{\link[=ffnn]{ffnn()}}, or \code{\link[=rnn]{rnn()}}.}

\item{...}{Additional arguments (currently unused).}

\item{x}{A fitted model of class \code{nn_fit}, as returned by
\code{\link[=train_nn]{train_nn()}}, \code{\link[=ffnn]{ffnn()}}, or \code{\link[=rnn]{rnn()}}.}
}
\value{
A \code{\link[ggplot2:ggplot]{ggplot2::ggplot()}} object showing loss vs epoch.
}
\description{
Creates a line plot of training (and optionally validation) loss over epochs
for a model of class \code{nn_fit}.
}
\examples{
\donttest{
  if (torch::torch_is_installed()) {
    m = train_nn(
      as.matrix(iris[, 2:4]), iris$Sepal.Length,
      epochs = 5
    )
    ggplot2::autoplot(m)
  }
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_depth.R
\name{safe_sample}
\alias{safe_sample}
\title{Safe sampling function}
\usage{
safe_sample(x, size, replace = FALSE)
}
\arguments{
\item{x}{Vector to sample from}

\item{size}{Number of samples}

\item{replace}{Sample with replacement?}
}
\description{
R's sample() has quirky behavior: sample(5, 1) samples from 1:5, not from c(5).
This function ensures we sample from the actual vector provided.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/torch-code-gen.R
\docType{data}
\name{layer_prs}
\alias{layer_prs}
\alias{.layer}
\alias{.i}
\alias{.in}
\alias{.out}
\alias{.is_output}
\title{Layer argument pronouns for formula-based specifications}
\format{
An object of class \code{layer_pr} (inherits from \code{list}) of length 0.

An object of class \code{layer_index_pr} (inherits from \code{layer_pr}, \code{list}) of length 0.

An object of class \code{layer_input_pr} (inherits from \code{layer_pr}, \code{list}) of length 0.

An object of class \code{layer_output_pr} (inherits from \code{layer_pr}, \code{list}) of length 0.

An object of class \code{layer_is_output_pr} (inherits from \code{layer_pr}, \code{list}) of length 0.
}
\usage{
.layer

.i

.in

.out

.is_output
}
\description{
These pronouns provide a cleaner, more readable way to reference layer parameters
in formula-based specifications for \code{nn_module_generator()} and related functions.
They work similarly to \code{rlang::.data} and \code{rlang::.env}.
}
\details{
Available pronouns:
\itemize{
\item \code{.layer}: Access all layer parameters as a list-like object
\item \code{.i}: Layer index (1-based integer)
\item \code{.in}: Input dimension for the layer
\item \code{.out}: Output dimension for the layer
\item \code{.is_output}: Logical indicating if this is the output layer
}

These pronouns can be used in formulas passed to:
\itemize{
\item \code{layer_arg_fn} parameter
\item Custom layer configuration functions
}
}
\section{Usage}{


\if{html}{\out{<div class="sourceCode r">}}\preformatted{# Using individual pronouns
layer_arg_fn = ~ list(
    input_size = .in,
    hidden_size = .out,
    num_layers = if (.i == 1) 2L else 1L
)

# Using .layer pronoun (alternative syntax)
layer_arg_fn = ~ list(
    input_size = .layer$ind,
    hidden_size = .layer$out,
    is_first = .layer$i == 1
)
}\if{html}{\out{</div>}}
}

\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nn_arch.R
\name{nn_arch}
\alias{nn_arch}
\title{Architecture specification for train_nn()}
\usage{
nn_arch(
  nn_name = "nnModule",
  nn_layer = NULL,
  out_nn_layer = NULL,
  nn_layer_args = list(),
  layer_arg_fn = NULL,
  forward_extract = NULL,
  before_output_transform = NULL,
  after_output_transform = NULL,
  last_layer_args = list(),
  input_transform = NULL
)
}
\arguments{
\item{nn_name}{Character. Name of the generated module class. Default \code{"nnModule"}.}

\item{nn_layer}{Layer type. See \code{nn_module_generator()}. Default \code{NULL} (\code{nn_linear}).}

\item{out_nn_layer}{Optional. Layer type forced on the last layer. Default \code{NULL}.}

\item{nn_layer_args}{Named list. Additional arguments passed to every layer constructor.
Default \code{list()}.}

\item{layer_arg_fn}{Formula or function. Generates per-layer constructor arguments.
Default \code{NULL} (FFNN-style: \code{list(in_dim, out_dim, bias = bias)}).}

\item{forward_extract}{Formula or function. Processes layer output in the forward pass.
Default \code{NULL}.}

\item{before_output_transform}{Formula or function. Transforms input before the output
layer. Default \code{NULL}.}

\item{after_output_transform}{Formula or function. Transforms output after the output
layer. Default \code{NULL}.}

\item{last_layer_args}{Named list or formula. Extra arguments for the output layer only.
Default \code{list()}.}

\item{input_transform}{Formula or function. Transforms the entire input tensor
before training begins. Applied once to the full dataset tensor, not per-batch.
Transforms must therefore be independent of batch size. Safe examples:
\code{~ .$unsqueeze(2)} (RNN sequence dim), \code{~ .$unsqueeze(1)} (CNN channel dim).
Avoid transforms that reshape based on \code{.$size(1)} as this will reflect the
full dataset size, not the mini-batch size.}
}
\value{
An object of class \code{c("nn_arch", "kindling_arch")}, implemented as a
named list of \code{nn_module_generator()} arguments with an \code{"env"} attribute
capturing the calling environment for custom layer resolution.
}
\description{
\code{nn_arch()} defines an architecture specification object consumed by
\code{\link[=train_nn]{train_nn()}} via \code{architecture} (or legacy \code{arch}).

Conceptual workflow:
\enumerate{
\item Define architecture with \code{nn_arch()}.
\item Train with \verb{train_nn(..., architecture = <nn_arch>)}.
\item Predict with \code{predict()}.
}

Architecture fields mirror \code{nn_module_generator()} and are passed through
without additional branching logic.
}
\examples{
\donttest{
if (torch::torch_is_installed()) {
    # Standard architecture object for train_nn()
    std_arch = nn_arch(nn_name = "mlp_model")

    # GRU architecture spec
    gru_arch = nn_arch(
        nn_name = "GRU",
        nn_layer = "torch::nn_gru",
        layer_arg_fn = ~ if (.is_output) {
            list(.in, .out)
        } else {
            list(input_size = .in, hidden_size = .out, batch_first = TRUE)
        },
        out_nn_layer = "torch::nn_linear",
        forward_extract = ~ .[[1]],
        before_output_transform = ~ .[, .$size(2), ],
        input_transform = ~ .$unsqueeze(2)
    )

    # Custom layer architecture (resolved from calling environment)
    custom_linear = torch::nn_module(
        "CustomLinear",
        initialize = function(in_features, out_features, bias = TRUE) {
            self$layer = torch::nn_linear(in_features, out_features, bias = bias)
        },
        forward = function(x) self$layer(x)
    )
    custom_arch = nn_arch(
        nn_name = "CustomMLP",
        nn_layer = ~ custom_linear
    )

    model = train_nn(
        Sepal.Length ~ .,
        data = iris[, 1:4],
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 50,
        architecture = gru_arch
    )
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print-nn-exec.R
\name{print.ffnn_fit}
\alias{print.ffnn_fit}
\title{Print method for ffnn_fit objects}
\usage{
\method{print}{ffnn_fit}(x, ...)
}
\arguments{
\item{x}{An object of class "ffnn_fit"}

\item{...}{Additional arguments (unused)}
}
\value{
No return value, called for side effects (printing model summary)
}
\description{
Print method for ffnn_fit objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nn_arch.R
\name{print.nn_arch}
\alias{print.nn_arch}
\title{Display \code{nn_arch()} configuration}
\usage{
\method{print}{nn_arch}(x, ...)
}
\arguments{
\item{x}{An object of class "nn_arch"}

\item{...}{Additional arguments (unused)}
}
\value{
No return value, just the information
}
\description{
Display \code{nn_arch()} configuration
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlp_kindling.R
\name{mlp_kindling}
\alias{mlp_kindling}
\title{Multi-Layer Perceptron (Feedforward Neural Network) via kindling}
\usage{
mlp_kindling(
  mode = "unknown",
  engine = "kindling",
  hidden_neurons = NULL,
  activations = NULL,
  output_activation = NULL,
  bias = NULL,
  epochs = NULL,
  batch_size = NULL,
  penalty = NULL,
  mixture = NULL,
  learn_rate = NULL,
  optimizer = NULL,
  validation_split = NULL,
  optimizer_args = NULL,
  loss = NULL,
  architecture = NULL,
  flatten_input = NULL,
  early_stopping = NULL,
  device = NULL,
  verbose = NULL,
  cache_weights = NULL
)
}
\arguments{
\item{mode}{A single character string for the type of model. Possible values
are "unknown", "regression", or "classification".}

\item{engine}{A single character string specifying what computational engine
to use for fitting. Currently only "kindling" is supported.}

\item{hidden_neurons}{An integer vector for the number of units in each hidden
layer. Can be tuned.}

\item{activations}{A character vector of activation function names for each
hidden layer (e.g., "relu", "tanh", "sigmoid"). Can be tuned.}

\item{output_activation}{A character string for the output activation function.
Can be tuned.}

\item{bias}{Logical for whether to include bias terms. Can be tuned.}

\item{epochs}{An integer for the number of training iterations. Can be tuned.}

\item{batch_size}{An integer for the batch size during training. Can be tuned.}

\item{penalty}{A number for the regularization penalty (lambda). Default \code{0}
(no regularization). Higher values increase regularization strength. Can be tuned.}

\item{mixture}{A number between 0 and 1 for the elastic net mixing parameter.
Default \code{0} (pure L2/Ridge regularization).
\itemize{
\item \code{0}: Pure L2 regularization (Ridge)
\item \code{1}: Pure L1 regularization (Lasso)
\item \verb{0 < mixture < 1}: Elastic net (combination of L1 and L2)
Only relevant when \code{penalty > 0}. Can be tuned.
}}

\item{learn_rate}{A number for the learning rate. Can be tuned.}

\item{optimizer}{A character string for the optimizer type ("adam", "sgd",
"rmsprop"). Can be tuned.}

\item{validation_split}{A number between 0 and 1 for the proportion of data
used for validation. Can be tuned.}

\item{optimizer_args}{A named list of additional arguments passed to the
optimizer. Cannot be tuned — pass via \code{set_engine()}.}

\item{loss}{A character string for the loss function ("mse", "mae",
"cross_entropy", "bce"). Cannot be tuned — pass via \code{set_engine()}.}

\item{architecture}{An \code{\link[=nn_arch]{nn_arch()}} object for a custom architecture. Cannot
be tuned — pass via \code{set_engine()}.}

\item{flatten_input}{Logical or \code{NULL}. Controls input flattening. Cannot be
tuned — pass via \code{set_engine()}.}

\item{early_stopping}{An \code{\link[=early_stop]{early_stop()}} object or \code{NULL}. Cannot be tuned —
pass via \code{set_engine()}.}

\item{device}{A character string for the device ("cpu", "cuda", "mps").
Cannot be tuned — pass via \code{set_engine()}.}

\item{verbose}{Logical for whether to print training progress. Cannot be
tuned — pass via \code{set_engine()}.}

\item{cache_weights}{Logical. If \code{TRUE}, stores trained weight matrices in
the returned object. Cannot be tuned — pass via \code{set_engine()}.}
}
\value{
A model specification object with class \code{mlp_kindling}.
}
\description{
\code{mlp_kindling()} defines a feedforward neural network model that can be used
for classification or regression. It integrates with the tidymodels ecosystem
and uses the torch backend via kindling.
}
\details{
This function creates a model specification for a feedforward neural network
that can be used within tidymodels workflows. The model supports:
\itemize{
\item Multiple hidden layers with configurable units
\item Various activation functions per layer
\item GPU acceleration (CUDA, MPS, or CPU)
\item Hyperparameter tuning integration
\item Both regression and classification tasks
}

Parameters that cannot be tuned (\code{architecture}, \code{flatten_input},
\code{early_stopping}, \code{device}, \code{verbose}, \code{cache_weights}, \code{optimizer_args},
\code{loss}) must be set via \code{set_engine()}, not as arguments to \code{mlp_kindling()}.
}
\examples{
\donttest{
if (torch::torch_is_installed()) {
    box::use(
        recipes[recipe],
        workflows[workflow, add_recipe, add_model],
        tune[tune],
        parsnip[fit]
    )

    # library(recipes)
    # library(workflows)
    # library(parsnip)
    # library(tune)

    # Model specs
    mlp_spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = c(128, 64, 32),
        activation = c("relu", "relu", "relu"),
        epochs = 100
    )

    # If you want to tune
    mlp_tune_spec = mlp_kindling(
        mode = "classification",
        hidden_neurons = tune(),
        activation = tune(),
        epochs = tune(),
        learn_rate = tune()
    )
     wf = workflow() |>
        add_recipe(recipe(Species ~ ., data = iris)) |>
        add_model(mlp_spec)

     fit_wf = fit(wf, data = iris)
} else {
    message("Torch not fully installed — skipping example")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/torch-code-gen.R
\name{substitute_dot}
\alias{substitute_dot}
\title{Recursively Substitute . with Expression}
\usage{
substitute_dot(expr, replacement)
}
\arguments{
\item{expr}{Expression containing \code{.} placeholders}

\item{replacement}{Expression to substitute for \code{.}}
}
\value{
Modified expression
}
\description{
Recursively Substitute . with Expression
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/torch-code-gen.R
\name{print.layer_pr}
\alias{print.layer_pr}
\alias{print.layer_index_pr}
\alias{print.layer_input_pr}
\alias{print.layer_output_pr}
\alias{print.layer_is_output_pr}
\title{Print method for the pronouns}
\usage{
\method{print}{layer_pr}(x, ...)

\method{print}{layer_index_pr}(x, ...)

\method{print}{layer_input_pr}(x, ...)

\method{print}{layer_output_pr}(x, ...)

\method{print}{layer_is_output_pr}(x, ...)
}
\arguments{
\item{x}{An object of class "ffnn_fit"}

\item{...}{Additional arguments (unused)}
}
\value{
No return value, prints out the type of pronoun to be used
}
\description{
Print method for the pronouns
}
\section{For \code{.layer}}{

It displays what fields to be accessed by \code{$}.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/torch-code-gen.R
\name{nn_module_generator}
\alias{nn_module_generator}
\title{Generalized Neural Network Module Expression Generator}
\usage{
nn_module_generator(
  nn_name = "nnModule",
  nn_layer = NULL,
  out_nn_layer = NULL,
  nn_layer_args = list(),
  layer_arg_fn = NULL,
  forward_extract = NULL,
  before_output_transform = NULL,
  after_output_transform = NULL,
  last_layer_args = list(),
  hd_neurons,
  no_x,
  no_y,
  activations = NULL,
  output_activation = NULL,
  bias = TRUE,
  eval = FALSE,
  .env = parent.frame(),
  ...
)
}
\arguments{
\item{nn_name}{Character string specifying the name of the generated neural network module class.
Default is \code{"nnModule"}.}

\item{nn_layer}{The type of neural network layer to use. Can be specified as:
\itemize{
\item \code{NULL} (default): Uses \code{nn_linear()} from \code{{torch}}
\item Character string: e.g., \code{"nn_linear"}, \code{"nn_gru"}, \code{"nn_lstm"}, \code{"some_custom_layer"}
\item Named function: A function object that constructs the layer
\item Anonymous function: e.g., \verb{\\() nn_linear()} or \code{function() nn_linear()}
}

The layer constructor is first searched in the current environment, then in parent
environments, and finally falls back to the \code{{torch}} namespace. This allows you to
use custom layer implementations alongside standard torch layers.}

\item{out_nn_layer}{Default \code{NULL}. If supplied, it forces to be the neural network layer to be used
on the last layer. Can be specified as:
\itemize{
\item Character string, e.g. \code{"nn_linear"}, \code{"nn_gru"}, \code{"nn_lstm"}, \code{"some_custom_layer"}
\item Named function: A function object that constructs the layer
\item Formula interface, e.g. \code{~torch::nn_linear}, \code{~some_custom_layer}
}

Internally, it almost works the same as \code{nn_layer} parameter.}

\item{nn_layer_args}{Named list of additional arguments passed to the layer constructor
specified by \code{nn_layer}. These arguments are applied to all layers. For layer-specific
arguments, use \code{layer_arg_fn}. Default is an empty list.}

\item{layer_arg_fn}{Optional function or formula that generates layer-specific construction arguments.
Can be specified as:
\itemize{
\item \strong{Formula}: \code{~ list(input_size = .in, hidden_size = .out)} where \code{.in}, \code{.out}, \code{.i}, and \code{.is_output} are available
\item \strong{Function}: \verb{function(i, in_dim, out_dim, is_output)} with signature as before
}

The formula/function should return a named list of arguments to pass to the layer constructor.
Available variables in formula context:
\itemize{
\item \code{.i} or \code{i}: Integer, the layer index (1-based)
\item \code{.in} or \code{in_dim}: Integer, input dimension for this layer
\item \code{.out} or \code{out_dim}: Integer, output dimension for this layer
\item \code{.is_output} or \code{is_output}: Logical, whether this is the final output layer
}

If \code{NULL}, defaults to FFNN-style arguments: \code{list(in_dim, out_dim, bias = bias)}.}

\item{forward_extract}{Optional formula or function that processes layer outputs in the forward pass.
Useful for layers that return complex structures (e.g., RNNs return \code{list(output, hidden)}).
Can be specified as:
\itemize{
\item \strong{Formula}: \code{~ .[[1]]} or \code{~ .$output} where \code{.} represents the layer output
\item \strong{Function}: \verb{function(expr)} that accepts/returns a language object
}

Common patterns:
\itemize{
\item Extract first element: \code{~ .[[1]]}
\item Extract named element: \code{~ .$output}
\item Extract with method: \code{~ .$get_output()}
}

If \code{NULL}, layer outputs are used directly.}

\item{before_output_transform}{Optional formula or function that transforms input before the output layer.
This is applied after the last hidden layer (and its activation) but before the output layer.
Can be specified as:
\itemize{
\item \strong{Formula}: \code{~ .[, .$size(2), ]} where \code{.} represents the current tensor
\item \strong{Function}: \verb{function(expr)} that accepts/returns a language object
}

Common patterns:
\itemize{
\item Extract last timestep: \code{~ .[, .$size(2), ]}
\item Flatten: \code{~ .$flatten(start_dim = 1)}
\item Global pooling: \code{~ .$mean(dim = 2)}
\item Extract token: \code{~ .[, 1, ]}
}

If \code{NULL}, no transformation is applied.}

\item{after_output_transform}{Optional formula or function that transforms the output after the output layer.
This is applied after \code{self$out(x)} (the final layer) but before returning the result.
Can be specified as:
\itemize{
\item \strong{Formula}: \code{~ .$mean(dim = 2)} where \code{.} represents the output tensor
\item \strong{Function}: \verb{function(expr)} that accepts/returns a language object
}

Common patterns:
\itemize{
\item Global average pooling: \code{~ .$mean(dim = 2)}
\item Squeeze dimensions: \code{~ .$squeeze()}
\item Reshape output: \code{~ .$view(c(-1, 10))}
\item Extract specific outputs: \code{~ .[, , 1:5]}
}

If \code{NULL}, no transformation is applied.}

\item{last_layer_args}{Optional named list or formula specifying additional arguments
for the output layer only. These arguments are appended to the output layer constructor
after the arguments from \code{layer_arg_fn}. Can be specified as:
\itemize{
\item \strong{Formula}: \code{~ list(kernel_size = 2L, bias = FALSE)}
\item \strong{Named list}: \code{list(kernel_size = 2L, bias = FALSE)}
}

This is useful when you need to override or add specific parameters to the final layer
without affecting hidden layers. For example, in CNNs you might want a different kernel
size for the output layer, or in RNNs you might want to disable bias in the final linear
projection. Arguments in \code{last_layer_args} will override any conflicting arguments from
\code{layer_arg_fn} when \code{.is_output = TRUE}. Default is an empty list.}

\item{hd_neurons}{Integer vector specifying the number of neurons (hidden units) in each
hidden layer. The length determines the number of hidden layers in the network.
Must contain at least one element.}

\item{no_x}{Integer specifying the number of input features (input dimension).}

\item{no_y}{Integer specifying the number of output features (output dimension).}

\item{activations}{Activation function specifications for hidden layers. Can be:
\itemize{
\item \code{NULL}: No activation functions applied
\item Character vector: e.g., \code{c("relu", "sigmoid", "tanh")}
\item \code{activation_spec} object: Created using \code{act_funs()}, which allows
specifying custom arguments. See examples.
}

If a single activation is provided, it will be replicated across all hidden layers.
Otherwise, the length should match the number of hidden layers.}

\item{output_activation}{Optional activation function for the output layer.
Same format as \code{activations}, but should specify only a single activation.
Common choices include \code{"softmax"} for classification or \code{"sigmoid"} for
binary outcomes. Default is \code{NULL} (no output activation).}

\item{bias}{Logical indicating whether to include bias terms in layers.
Default is \code{TRUE}. Note that this is passed to \code{layer_arg_fn} if provided,
so custom layer argument functions should handle this parameter appropriately.}

\item{eval}{Logical indicating whether to evaluate the generated expression immediately.
If \code{TRUE}, returns an instantiated \code{nn_module} class that can be called directly
(e.g., \code{model()}). If \code{FALSE} (default), returns the unevaluated language expression
that can be inspected or evaluated later with \code{eval()}. Default is \code{FALSE}.}

\item{.env}{Default is \code{\link[=parent.frame]{parent.frame()}}. The environment in which the generated expression is to be evaluated}

\item{...}{Additional arguments passed to layer constructors or for future extensions.}
}
\value{
If \code{eval = FALSE} (default): A language object (unevaluated expression) representing
a \code{torch::nn_module} definition. This expression can be evaluated with \code{eval()} to
create the module class, which can then be instantiated with \code{eval(result)()} to
create a model instance.

If \code{eval = TRUE}: An instantiated \code{nn_module} class constructor that can be called
directly to create model instances (e.g., \code{result()}).
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{nn_module_generator()} is a generalized function that generates neural network
module expressions for various architectures. It provides a flexible framework for creating
custom neural network modules by parameterizing layer types, construction arguments, and
forward pass behavior.

While designed primarily for \code{{torch}} modules, it can work with custom layer implementations
from the current environment, including user-defined layers like RBF networks, custom
attention mechanisms, or other novel architectures.

This function serves as the foundation for specialized generators like \code{ffnn_generator()}
and \code{rnn_generator()}, but can be used directly to create custom architectures.
}
\examples{
\dontrun{
\donttest{
if (torch::torch_is_installed()) {
    # Basic usage with formula interface
    nn_module_generator(
        nn_name = "MyGRU",
        nn_layer = "nn_gru",
        layer_arg_fn = ~ if (.is_output) {
            list(.in, .out)
        } else {
            list(input_size = .in, hidden_size = .out, 
                 num_layers = 1L, batch_first = TRUE)
        },
        forward_extract = ~ .[[1]],
        before_output_transform = ~ .[, .$size(2), ],
        hd_neurons = c(128, 64, 32),
        no_x = 20,
        no_y = 5,
        activations = "relu"
    )
    
    # LSTM with cleaner syntax
    nn_module_generator(
        nn_name = "MyLSTM",
        nn_layer = "nn_lstm",
        layer_arg_fn = ~ list(
            input_size = .in,
            hidden_size = .out,
            batch_first = TRUE
        ),
        forward_extract = ~ .[[1]],
        before_output_transform = ~ .[, .$size(2), ],
        hd_neurons = c(64, 32),
        no_x = 10,
        no_y = 2
    )
    
    # CNN with global average pooling
    nn_module_generator(
        nn_name = "SimpleCNN",
        nn_layer = "nn_conv1d",
        layer_arg_fn = ~ list(
            in_channels = .in,
            out_channels = .out,
            kernel_size = 3L,
            padding = 1L
        ),
        before_output_transform = ~ .$mean(dim = 2),
        hd_neurons = c(16, 32, 64),
        no_x = 1,
        no_y = 10,
        activations = "relu"
    )
    
    # CNN with after_output_transform (pooling applied AFTER output layer)
    nn_module_generator(
        nn_name = "CNN1DClassifier",
        nn_layer = "nn_conv1d",
        layer_arg_fn = ~ if (.is_output) {
            list(.in, .out)
        } else {
            list(
                in_channels = .in,
                out_channels = .out,
                kernel_size = 3L,
                stride = 1L,
                padding = 1L 
            )
        },
        after_output_transform = ~ .$mean(dim = 2),
        last_layer_args = list(kernel_size = 1, stride = 2),
        hd_neurons = c(16, 32, 64),
        no_x = 1,
        no_y = 10,
        activations = "relu"
    )
    
} else {
  message("torch not installed - skipping examples")
}
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/torch-nn-generator.R
\name{nn_gens}
\alias{nn_gens}
\alias{ffnn_generator}
\alias{rnn_generator}
\title{Functions to generate \code{nn_module} (language) expression}
\usage{
ffnn_generator(
  nn_name = "DeepFFN",
  hd_neurons,
  no_x,
  no_y,
  activations = NULL,
  output_activation = NULL,
  bias = TRUE
)

rnn_generator(
  nn_name = "DeepRNN",
  hd_neurons,
  no_x,
  no_y,
  rnn_type = "lstm",
  bias = TRUE,
  activations = NULL,
  output_activation = NULL,
  bidirectional = TRUE,
  dropout = 0,
  ...
)
}
\arguments{
\item{nn_name}{Character. Name of the generated RNN module class. Default is \code{"DeepRNN"}.}

\item{hd_neurons}{Integer vector. Number of neurons in each hidden RNN layer.}

\item{no_x}{Integer. Number of input features.}

\item{no_y}{Integer. Number of output features.}

\item{activations}{Activation function specifications for each hidden layer.
Can be:
\itemize{
\item \code{NULL}: No activation functions.
\item Character vector: e.g., \code{c("relu", "sigmoid")}.
\item List: e.g., \code{act_funs(relu, elu, softshrink = args(lambd = 0.5))}.
\item \code{activation_spec} object from \code{act_funs()}.
}

If the length of \code{activations} is \code{1L}, this will be the activation throughout the architecture.}

\item{output_activation}{Optional. Activation function for the output layer.
Same format as \code{activations} but should be a single activation.}

\item{bias}{Logical. Whether to use bias weights. Default is \code{TRUE}}

\item{rnn_type}{Character. Type of RNN to use. Must be one of \code{"rnn"}, \code{"lstm"}, or \code{"gru"}. Default is \code{"lstm"}.}

\item{bidirectional}{Logical. Whether to use bidirectional RNN layers. Default is \code{TRUE}.}

\item{dropout}{Numeric. Dropout rate between RNN layers. Default is \code{0}.}

\item{...}{Additional arguments (currently unused).}
}
\value{
A \code{torch} module expression representing the FFNN.

A \code{torch} module expression representing the RNN.
}
\description{
Functions to generate \code{nn_module} (language) expression
}
\details{
The generated FFNN module will have the specified number of hidden layers,
with each layer containing the specified number of neurons. Activation functions
can be applied after each hidden layer as specified.
This can be used for both classification and regression tasks.

The generated module properly namespaces all torch functions to avoid
polluting the global namespace.

The generated RNN module will have the specified number of recurrent layers,
with each layer containing the specified number of hidden units. Activation functions
can be applied after each RNN layer as specified. The final output is taken from the
last time step and passed through a linear layer.

The generated module properly namespaces all torch functions to avoid
polluting the global namespace.
}
\section{Feed-Forward Neural Network Module Generator}{

The \code{ffnn_generator()} function generates a feed-forward neural network (FFNN) module expression
from the \code{torch} package in R. It allows customization of the FFNN architecture,
including the number of hidden layers, neurons, and activation functions.
}

\section{Recurrent Neural Network Module Generator}{

The \code{rnn_generator()} function generates a recurrent neural network (RNN) module expression
from the \code{torch} package in R. It allows customization of the RNN architecture,
including the number of hidden layers, neurons, RNN type, activation functions,
and other parameters.
}

\examples{
\donttest{
# FFNN
if (torch::torch_is_installed()) {
    # Generate an MLP module with 3 hidden layers
    ffnn_mod = ffnn_generator(
        nn_name = "MyFFNN",
        hd_neurons = c(64, 32, 16),
        no_x = 10,
        no_y = 1,
        activations = 'relu'
    )

    # Evaluate and instantiate
    model = eval(ffnn_mod)()

    # More complex: With different activations
    ffnn_mod2 = ffnn_generator(
        nn_name = "MyFFNN2",
        hd_neurons = c(128, 64, 32),
        no_x = 20,
        no_y = 5,
        activations = act_funs(
            relu,
            selu,
            sigmoid
        )
    )

    # Even more complex: Different activations and customized argument
    # for the specific activation function
    ffnn_mod2 = ffnn_generator(
        nn_name = "MyFFNN2",
        hd_neurons = c(128, 64, 32),
        no_x = 20,
        no_y = 5,
        activations = act_funs(
            relu,
            selu,
            softshrink = args(lambd = 0.5)
        )
    )

    # Customize output activation (softmax is useful for classification tasks)
    ffnn_mod3 = ffnn_generator(
        hd_neurons = c(64, 32),
        no_x = 10,
        no_y = 3,
        activations = 'relu',
        output_activation = act_funs(softmax = args(dim = 2L))
    )
} else {
    message("Torch not fully installed — skipping example")
}
}

\donttest{
## RNN
if (torch::torch_is_installed()) {
    # Basic LSTM with 2 layers
    rnn_mod = rnn_generator(
        nn_name = "MyLSTM",
        hd_neurons = c(64, 32),
        no_x = 10,
        no_y = 1,
        rnn_type = "lstm",
        activations = 'relu'
    )

    # Evaluate and instantiate
    model = eval(rnn_mod)()

    # GRU with different activations
    rnn_mod2 = rnn_generator(
        nn_name = "MyGRU",
        hd_neurons = c(128, 64, 32),
        no_x = 20,
        no_y = 5,
        rnn_type = "gru",
        activations = act_funs(relu, elu, relu),
        bidirectional = FALSE
    )

} else {
    message("Torch not fully installed — skipping example")
}
}

\dontrun{
## Parameterized activation and dropout
# (Will throw an error due to `nnf_tanh()` not being available in `{torch}`)
# rnn_mod3 = rnn_generator(
#     hd_neurons = c(100, 50, 25),
#     no_x = 15,
#     no_y = 3,
#     rnn_type = "lstm",
#     activations = act_funs(
#         relu,
#         leaky_relu = args(negative_slope = 0.01),
#         tanh
#     ),
#     bidirectional = TRUE,
#     dropout = 0.3
# )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlp.R
\name{ffnn_impl}
\alias{ffnn_impl}
\title{FFNN Implementation}
\usage{
ffnn_impl(
  x,
  y,
  hidden_neurons,
  activations = NULL,
  output_activation = NULL,
  bias = TRUE,
  epochs = 100,
  batch_size = 32,
  penalty = 0,
  mixture = 0,
  learn_rate = 0.001,
  optimizer = "adam",
  optimizer_args = list(),
  loss = "mse",
  validation_split = 0,
  device = NULL,
  verbose = FALSE,
  cache_weights = FALSE
)
}
\description{
FFNN Implementation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/act-fun.R
\name{eval_act_funs}
\alias{eval_act_funs}
\title{Activation Function Specs Evaluation}
\usage{
eval_act_funs(activations, output_activation)
}
\arguments{
\item{activations}{Quosure containing the activations expression}

\item{output_activation}{Quosure containing the output_activation expression}
}
\value{
A list with two elements: \code{activations} and \code{output_activation}
}
\description{
Helper function for \code{act_funcs()} argument.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variable-imp.R
\name{kindling-varimp}
\alias{kindling-varimp}
\alias{garson.ffnn_fit}
\alias{olden.ffnn_fit}
\alias{vi_model.ffnn_fit}
\title{Variable Importance Methods for kindling Models}
\usage{
\method{garson}{ffnn_fit}(mod_in, bar_plot = FALSE, ...)

\method{olden}{ffnn_fit}(mod_in, bar_plot = TRUE, ...)

vi_model.ffnn_fit(object, type = c("olden", "garson"), ...)
}
\arguments{
\item{mod_in}{A fitted model object of class "ffnn_fit".}

\item{bar_plot}{Logical. Whether to plot variable importance (default TRUE).}

\item{...}{Additional arguments passed to methods.}

\item{object}{A fitted model object of class "ffnn_fit".}

\item{type}{Type of algorithm to extract the variable importance.
This must be one of the strings:
\itemize{
\item 'olden'
\item 'garson'
}}
}
\value{
A data frame for both "garson" and "olden" classes with columns:
\item{x_names}{Character vector of predictor variable names}
\item{y_names}{Character string of response variable name}
\item{rel_imp}{Numeric vector of relative importance scores (percentage)}
The data frame is sorted by importance in descending order.

A tibble with columns "Variable" and "Importance"
(via \code{vip::vi()} / \code{vip::vi_model()} only).
}
\description{
This file implements methods for variable importance generics from
NeuralNetTools and vip packages.
}
\section{Garson's Algorithm for FFNN Models}{

\code{{kindling}} inherits \code{NeuralNetTools::garson} to extract the variable
importance from the fitted \code{ffnn()} model.
}

\section{Olden's Algorithm for FFNN Models}{

\code{{kindling}} inherits \code{NeuralNetTools::olden} to extract the variable
importance from the fitted \code{ffnn()} model.
}

\section{Variable Importance via \code{{vip}} Package}{

If the \code{{vip}} package (Suggests only) is installed, \code{kindling} registers a
\code{vip::vi_model()} method for \code{"ffnn_fit"} objects, so \code{vip::vi()} and
\code{vip::vi_model()} work out of the box.
}

\examples{
\donttest{
if (torch::torch_is_installed()) {
    model_mlp = ffnn(
        Species ~ .,
        data = iris,
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 100,
        verbose = FALSE,
        cache_weights = TRUE
    )
    
    # Directly use `NeuralNetTools::garson`
    model_mlp |>
        garson()
    
    # Directly use `NeuralNetTools::olden`    
    model_mlp |>
        olden()
} else {
    message("Torch not fully installed — skipping example")
}
}

\donttest{
# kindling also supports `vip::vi()` / `vip::vi_model()` when {vip} is installed
if (torch::torch_is_installed() && requireNamespace("vip", quietly = TRUE)) {
    model_mlp = ffnn(
        Species ~ .,
        data = iris,
        hidden_neurons = c(64, 32),
        activations = "relu",
        epochs = 100,
        verbose = FALSE,
        cache_weights = TRUE
    )

    model_mlp |>
        vip::vi(type = 'garson') |>
        vip::vip()
} else {
    message("Torch and/or {vip} not installed — skipping example")
}
}
}
\references{
Beck, M.W. 2018. NeuralNetTools: Visualization and Analysis Tools for Neural Networks. Journal of Statistical Software. 85(11):1-20.

Garson, G.D. 1991. Interpreting neural network connection weights. Artificial Intelligence Expert. 6(4):46-51.

Goh, A.T.C. 1995. Back-propagation neural networks for modeling complex systems. Artificial Intelligence in Engineering. 9(3):143-151.

Olden, J.D., Jackson, D.A. 2002. Illuminating the 'black-box': a randomization approach for understanding variable contributions in artificial neural networks. Ecological Modelling. 154:135-150.

Olden, J.D., Joy, M.K., Death, R.G. 2004. An accurate comparison of methods for quantifying variable importance in artificial neural networks using simulated data. Ecological Modelling. 178:389-397.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/act-funnew.R
\name{new_act_fn}
\alias{new_act_fn}
\title{Custom Activation Function Constructor}
\usage{
new_act_fn(fn, probe = TRUE, .name = "<custom>")
}
\arguments{
\item{fn}{A function taking a single tensor argument and returning a tensor.
E.g. \verb{\\(x) torch::torch_tanh(x)}.}

\item{probe}{Logical. If \code{TRUE} (default), runs a dry-run with a small
dummy tensor at definition time to catch obvious errors early.}

\item{.name}{A string to be stored in an attribute. Nothing special, except it is
used when displaying the info of a trained neural network model.
Default is \code{"<custom>"}.}
}
\value{
An object of class \code{c("custom_activation", "parameterized_activation")},
compatible with \code{act_funs()}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Wraps a user-supplied function into a validated custom activation,
ensuring it accepts and returns a \code{torch_tensor}. Performs an eager
dry-run probe at \emph{definition time} so errors surface early, and
wraps the function with a \emph{call-time} type guard for safety.
}
\examples{
\dontrun{
\donttest{
act_funs(relu, elu, new_act_fn(\(x) torch::torch_tanh(x)))
act_funs(new_act_fn(\(x) torch::nnf_silu(x)))
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kindling-package.R
\docType{package}
\name{kindling}
\alias{kindling-package}
\alias{kindling}
\title{\code{{kindling}}: Higher-level interface of torch package to auto-train neural networks}
\description{
\code{{kindling}} enables R users to build and train deep neural networks such as:
\itemize{
\item Deep Neural Networks / (Deep) Feedforward Neural Networks (DNN / FFNN)
\item Recurrent Neural Networks (RNN)
}

It is mainly designed to generate code expressions for current architectures,
which reduces boilerplate \code{{torch}} code.
It also integrates seamlessly with machine learning workflows in R,
currently through \code{{tidymodels}} and components like \code{{parsnip}},
\code{{recipes}}, and \code{{workflows}}, allowing an ergonomic interface for model
specification, training, and evaluation.

Thus, the package supports hyperparameter tuning for:
\itemize{
\item Number of hidden layers
\item Number of units per layer
\item Choice of activation functions
}

Clarification: The hyperparameter tuning support is supported in version 0.1.0,
but \code{n_hlayer()} dial parameter is not supported. This changes after 0.2.0 release.
}
\details{
The \code{{kindling}} package provides a unified, high-level interface that bridges
the \code{{torch}} and \code{{tidymodels}} ecosystems, making it easy to define, train,
and tune deep learning models using the familiar \code{tidymodels} workflow.
}
\section{How to use}{

This package can be used at 3 levels:

Level 1: Code generation

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ffnn_generator(
    nn_name = "MyFFNN",
    hd_neurons = c(64, 32, 16),
    no_x = 10,
    no_y = 1,
    activations = 'relu'
)
}\if{html}{\out{</div>}}

Level 2: Direct Execution

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ffnn(
    Species ~ .,
    data = iris,
    hidden_neurons = c(128, 64, 32),
    activations = 'relu',
    loss = "cross_entropy",
    epochs = 100
)
}\if{html}{\out{</div>}}

Level 3: Conventional tidymodels interface

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# library(parsnip)
# library(kindling)
box::use(
   kindling[mlp_kindling, rnn_kindling, act_funs, args],
   parsnip[fit, augment],
   yardstick[metrics],
   mlbench[Ionosphere] # data(Ionosphere, package = "mlbench")
)

# Remove V2 as it's all zeros
ionosphere_data = Ionosphere[, -2]

# MLP example
mlp_kindling(
    mode = "classification",
    hidden_neurons = c(128, 64),
    activations = act_funs(relu, softshrink = args(lambd = 0.5)),
    epochs = 100
) |>
    fit(Class ~ ., data = ionosphere_data) |>
    augment(new_data = ionosphere_data) |>
    metrics(truth = Class, estimate = .pred_class)
#> A tibble: 2 × 3
#>   .metric  .estimator .estimate
#>   <chr>    <chr>          <dbl>
#> 1 accuracy binary         0.989
#> 2 kap      binary         0.975

# RNN example (toy usage on non-sequential data)
rnn_kindling(
    mode = "classification",
    hidden_neurons = c(128, 64),
    activations = act_funs(relu, elu),
    epochs = 100,
    rnn_type = "gru"
) |>
    fit(Class ~ ., data = ionosphere_data) |>
    augment(new_data = ionosphere_data) |>
    metrics(truth = Class, estimate = .pred_class)
#> A tibble: 2 × 3
#>   .metric  .estimator .estimate
#>   <chr>    <chr>          <dbl>
#> 1 accuracy binary         0.641
#> 2 kap      binary         0
}\if{html}{\out{</div>}}
}

\section{Main Features}{

\itemize{
\item Code generation of \code{{torch}} expression
\item Multiple architectures available: feedforward networks (MLP/DNN/FFNN) and recurrent variants (RNN, LSTM, GRU)
\item Native support for \code{{tidymodels}} workflows and pipelines
\item Fine-grained control over network depth, layer sizes, and activation functions
\item GPU acceleration support via \code{{torch}} tensors
}
}

\section{License}{

MIT + file LICENSE
}

\references{
Falbel D, Luraschi J (2023). \emph{torch: Tensors and Neural Networks with 'GPU'
Acceleration}. R package version 0.13.0,
\url{https://torch.mlverse.org}, \url{https://github.com/mlverse/torch}.

Wickham H (2019). \emph{Advanced R}, 2nd edition. Chapman and Hall/CRC.
ISBN 978-0815384571, \url{https://adv-r.hadley.nz/}.

Goodfellow I, Bengio Y, Courville A (2016). \emph{Deep Learning}. MIT Press.
\url{https://www.deeplearningbook.org/}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://kindling.joshuamarie.com}
  \item \url{https://github.com/joshuamarie/kindling}
  \item Report bugs at \url{https://github.com/joshuamarie/kindling/issues}
}

}
\author{
\strong{Maintainer}: Joshua Marie \email{joshua.marie.k@gmail.com}

Authors:
\itemize{
  \item Antoine Soetewey \email{ant.soetewey@gmail.com} (\href{https://orcid.org/0000-0001-8159-0804}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/register-parsnip.R
\name{make_kindling}
\alias{make_kindling}
\title{Register kindling engines with parsnip}
\usage{
make_kindling()
}
\description{
This function registers the kindling engine for MLP, RNN, and train_nn models
with parsnip. It should be called when the package is loaded.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsnip-helpers.R
\name{prepare_kindling_args}
\alias{prepare_kindling_args}
\title{Prepare arguments for kindling models}
\usage{
prepare_kindling_args(args)
}
\description{
Prepare arguments for kindling models
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_summary.R
\name{table_summary}
\alias{table_summary}
\title{Summarize and Display a Two-Column Data Frame as a Formatted Table}
\usage{
table_summary(
  data,
  title = NULL,
  l = NULL,
  header = FALSE,
  center_table = FALSE,
  border_char = "-",
  style = list(),
  align = NULL,
  ...
)
}
\arguments{
\item{data}{A data frame with exactly two columns. The data to be summarized and displayed.}

\item{title}{A character string. An optional title to be displayed above the table.}

\item{l}{An integer. The number of rows to include in the left part of a split table.
If \code{NULL}, the table is not split.}

\item{header}{A logical value. If \code{TRUE}, the column names of \code{data} are displayed as a header.}

\item{center_table}{A logical value. If \code{TRUE}, the table is centered in the terminal.}

\item{border_char}{Character used for borders. Default is \code{"\\u2500"}.}

\item{style}{A list controlling the visual styling of table elements using ANSI formatting.
Can include the following components:
\itemize{
\item \code{left_col}: Styling for the left column values.
\item \code{right_col}: Styling for the right column values.
\item \code{border_text}: Styling for the border.
\item \code{title}: Styling for the title.
\item \code{sep}: Separator character between left and right column.
}

Each style component can be either a predefined style string (e.g., "blue", "red_italic", "bold")
or a function that takes a context list with/without a \code{value} element and returns the styled text.}

\item{align}{Controls the alignment of column values.
Can be specified in three ways:
\itemize{
\item A single string: affects only the left column (e.g., "left", "center", "right").
\item A vector of two strings: affects both columns in order (e.g., c("left", "right")).
\item A list with named components: explicitly specifies alignment for each column
}}

\item{...}{Additional arguments (currently unused).}
}
\value{
This function does not return a value. It prints the formatted table to the console.
}
\description{
This function takes a two-column data frame and formats it into a summary-like table.
The table can be optionally split into two parts, centered, and given a title.
It is useful for displaying summary information in a clean, tabular format.
The function also supports styling with ANSI colors and text formatting through
the \code{{cli}} package and column alignment options.
}
\examples{
# Create a sample data frame
df = data.frame(
    Category = c("A", "B", "C", "D", "E"),
    Value = c(10, 20, 30, 40, 50)
)

# Display the table with a title and header
table_summary(df, title = "Sample Table", header = TRUE)

# Split the table after the second row and center it
table_summary(df, l = 2, center_table = TRUE)

# Use styling and alignment
table_summary(
    df, header = TRUE,
    style = list(
        left_col = "blue_bold",
        right_col = "red",
        title = "green",
        border_text = "yellow"
    ),
    align = c("center", "right")
)

# Use custom styling with lambda functions
table_summary(
    df, header = TRUE,
    style = list(
        left_col = \(ctx) cli::col_red(ctx), # ctx$value is another option
        right_col = \(ctx) cli::col_blue(ctx)
    ),
    align = list(left_col = "left", right_col = "right")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print-nn-exec.R
\name{ordinal_gen}
\alias{ordinal_gen}
\title{Ordinal Suffixes Generator}
\usage{
ordinal_gen(x)
}
\arguments{
\item{x}{Vector of numbers. Could be a string equivalent}
}
\value{
Returns a string vector with ordinal suffixes.
}
\description{
This function is originally from \code{numform::f_ordinal()}.
}
\section{This is how you use it}{


\if{html}{\out{<div class="sourceCode r">}}\preformatted{kindling:::ordinal_gen(1:10)
}\if{html}{\out{</div>}}

Note: This is not exported into public namespace.
So please, refer to \code{numform::f_ordinal()} instead.
}

\references{
Rinker, T. W. (2021). numform: A publication style number and plot formatter
version 0.7.0. \url{https://github.com/trinker/numform}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalized-nn-fit.R
\name{train_nn_impl}
\alias{train_nn_impl}
\title{Shared core implementation}
\usage{
train_nn_impl(
  x,
  y,
  hidden_neurons,
  activations = NULL,
  output_activation = NULL,
  bias = TRUE,
  arch = NULL,
  early_stopping = NULL,
  epochs = 100,
  batch_size = 32,
  penalty = 0,
  mixture = 0,
  learn_rate = 0.001,
  optimizer = "adam",
  optimizer_args = list(),
  loss = "mse",
  validation_split = 0,
  device = NULL,
  verbose = FALSE,
  cache_weights = FALSE,
  fit_class = "nn_fit"
)
}
\description{
Shared core implementation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/early_stop.R
\name{early_stop}
\alias{early_stop}
\title{Early Stopping Specification}
\usage{
early_stop(
  patience = 5L,
  min_delta = 1e-04,
  restore_best_weights = TRUE,
  monitor = "val_loss"
)
}
\arguments{
\item{patience}{Integer. Epochs to wait after last improvement. Default \code{5}.}

\item{min_delta}{Numeric. Minimum improvement to qualify as better. Default \code{1e-4}.}

\item{restore_best_weights}{Logical. Restore weights from best epoch. Default \code{TRUE}.}

\item{monitor}{Character. Metric to monitor. One of \code{"val_loss"} (default) or \code{"train_loss"}.}
}
\value{
An object of class \code{"early_stop_spec"}.
}
\description{
\code{early_stop()} is a helper function to be supplied on \code{early_stopping} arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{autoplot_diagnostics}
\alias{autoplot_diagnostics}
\alias{plot_diagnostics}
\title{Plot prediction diagnostics for a fitted neural network}
\usage{
autoplot_diagnostics(object, actual, ...)

plot_diagnostics(object, actual, ...)
}
\arguments{
\item{object}{A fitted model of class \code{nn_fit}, as returned by
\code{\link[=train_nn]{train_nn()}}, \code{\link[=ffnn]{ffnn()}}, or \code{\link[=rnn]{rnn()}}.}

\item{actual}{A vector of true response values, the same length as the
training data used to fit \code{object}.}

\item{...}{Additional arguments (currently unused).}
}
\value{
For regression, a named list of \code{\link[ggplot2:ggplot]{ggplot2::ggplot()}} objects (one per
diagnostic panel). For classification, a single \code{\link[ggplot2:ggplot]{ggplot2::ggplot()}}
confusion matrix heatmap.
}
\description{
Produces diagnostic plots comparing fitted values against actual (true)
response values for a fitted \code{nn_fit} model.
\itemize{
\item \strong{Regression (single output)}: returns a named list with two panels --
residuals vs fitted and actual vs fitted.
\item \strong{Regression (multi-output)}: returns a named list with one actual vs
fitted panel per output column.
\item \strong{Classification}: returns a single confusion matrix heatmap.
}
}
\examples{
\donttest{
  if (torch::torch_is_installed()) {
    # Regression
    m = train_nn(
      as.matrix(iris[, 2:4]), iris$Sepal.Length,
      epochs = 5
    )
    autoplot_diagnostics(m, actual = iris$Sepal.Length)

    # Classification
    m_cls = train_nn(
      as.matrix(iris[, 1:4]), iris$Species,
      epochs = 5
    )
    autoplot_diagnostics(m_cls, actual = iris$Species)
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/torch-code-gen.R
\name{formula_to_expr_transformer}
\alias{formula_to_expr_transformer}
\title{Convert Formula to Expression Transformer}
\usage{
formula_to_expr_transformer(formula_or_fn)
}
\arguments{
\item{formula_or_fn}{A formula like \code{~ .[[1]]} or a function that transforms expressions}
}
\value{
A function that takes an expression and returns a transformed expression, or NULL
}
\description{
Convert Formula to Expression Transformer
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dials-kindling.R
\name{dials-kindling}
\alias{dials-kindling}
\alias{n_hlayers}
\alias{hidden_neurons}
\alias{activations}
\alias{output_activation}
\alias{optimizer}
\alias{bias}
\alias{validation_split}
\alias{bidirectional}
\title{Tunable hyperparameters for \code{kindling} models}
\usage{
n_hlayers(range = c(1L, 2L), trans = NULL)

hidden_neurons(range = c(8L, 512L), disc_values = NULL, trans = NULL)

activations(
  values = c("relu", "relu6", "elu", "selu", "celu", "leaky_relu", "gelu", "softplus",
    "softshrink", "softsign", "tanhshrink", "hardtanh", "hardshrink", "hardswish",
    "hardsigmoid", "silu", "mish", "logsigmoid")
)

output_activation(
  values = c("relu", "elu", "selu", "softplus", "softmax", "log_softmax", "logsigmoid",
    "hardtanh", "hardsigmoid", "silu")
)

optimizer(values = c("adam", "sgd", "rmsprop", "adamw"))

bias(values = c(TRUE, FALSE))

validation_split(range = c(0, 0.5), trans = NULL)

bidirectional(values = c(TRUE, FALSE))
}
\arguments{
\item{range}{A two-element numeric vector with the default lower and upper bounds.}

\item{trans}{An optional transformation; \code{NULL} for none.}

\item{disc_values}{\code{NULL} (default) or an integer vector of specific possible disc_values
(e.g., \code{c(32L, 64L, 128L, 256L)}). When provided, tuning will be restricted to
these discrete values. The range is automatically derived from these values if not explicitly given.
The \code{trans} parameter would still be ignored by this parameter when supplied.}

\item{values}{Logical vector of possible values.}
}
\value{
Each function returns a \code{dials} parameter object:
\describe{
\item{\code{n_hlayers()}}{A quantitative parameter for the number of hidden layers}
\item{\code{hidden_neurons()}}{A quantitative parameter for hidden units per layer}
\item{\code{activations()}}{A qualitative parameter for activation function names}
\item{\code{output_activation()}}{A qualitative parameter for output activation}
\item{\code{optimizer()}}{A qualitative parameter for optimizer type}
\item{\code{bias()}}{A qualitative parameter for bias inclusion}
\item{\code{validation_split()}}{A quantitative parameter for validation proportion}
\item{\code{bidirectional()}}{A qualitative parameter for bidirectional RNN}
}
}
\description{
These parameters extend the \strong{dials} framework to support hyperparameter
tuning of neural networks built with the \code{{kindling}} package. They control
network architecture, activation functions, optimization, and training
behavior.
}
\section{Architecture Strategy}{

Since tidymodels tuning works with independent parameters, we use a simplified
approach where:
\itemize{
\item \code{hidden_neurons} specifies a single value that will be used for ALL layers
\item \code{activations} specifies a single activation that will be used for ALL layers
\item \code{n_hlayers} controls the depth
}

For more complex architectures with different neurons/activations per layer,
users should manually specify these after tuning or use custom tuning logic.
}

\section{Parameters}{


\describe{
\item{\code{n_hlayers}}{Number of hidden layers in the network.}
\item{\code{hidden_neurons}}{Number of units per hidden layer (applied to all layers).}
\item{\code{activation}}{Single activation function applied to all hidden layers.}
\item{\code{output_activation}}{Activation function for the output layer.}
\item{\code{optimizer}}{Optimizer algorithm.}
\item{\code{bias}}{Whether to include bias terms in layers.}
\item{\code{validation_split}}{Proportion of training data held out for validation.}
\item{\code{bidirectional}}{Whether RNN layers are bidirectional.}
}
}

\section{Number of Hidden Layers}{

Controls the depth of the network. When tuning, this will determine how many
layers are created, each with \code{hidden_neurons} units and \code{activations} function.
}

\section{Hidden Units per Layer}{

Specifies the number of units per hidden layer.
}

\section{Activation Function (Hidden Layers)}{

Activation functions for hidden layers.
}

\section{Output Activation Function}{

Activation function applied to the output layer. Values must correspond to
\verb{torch::nnf_*} functions.
}

\section{Optimizer Type}{

The optimization algorithm used during training.
}

\section{Include Bias Terms}{

Whether layers should include bias parameters.
}

\section{Validation Split Proportion}{

Fraction of the training data to use as a validation set during training.
}

\section{Bidirectional RNN}{

Whether recurrent layers should process sequences in both directions.
}

\examples{
\donttest{
library(dials)
library(tune)

# Create a tuning grid
grid = grid_regular(
    n_hlayers(range = c(1L, 4L)),
    hidden_neurons(range = c(32L, 256L)),
    activations(c('relu', 'elu', 'selu')),
    levels = c(4, 5, 3)
)

# Use in a model spec
mlp_spec = mlp_kindling(
    mode = "classification",
    hidden_neurons = tune(),
    activations = tune(),
    epochs = tune(),
    learn_rate = tune()
)
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalized-nn-fit.R
\name{.train_nn_tab_impl}
\alias{.train_nn_tab_impl}
\title{Preprocessing bridge for data.frame and formula methods}
\usage{
.train_nn_tab_impl(
  processed,
  formula,
  hidden_neurons,
  activations,
  output_activation,
  bias,
  arch,
  early_stopping,
  epochs,
  batch_size,
  penalty,
  mixture,
  learn_rate,
  optimizer,
  optimizer_args,
  loss,
  validation_split,
  device,
  verbose,
  cache_weights
)
}
\description{
Preprocessing bridge for data.frame and formula methods
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlp.R, R/rnn.R
\name{predict-basemodel}
\alias{predict-basemodel}
\alias{predict.ffnn_fit}
\alias{predict.rnn_fit}
\title{Predict method for kindling basemodel fits}
\usage{
\method{predict}{ffnn_fit}(object, newdata = NULL, new_data = NULL, type = "response", ...)

\method{predict}{rnn_fit}(object, newdata = NULL, new_data = NULL, type = "response", ...)
}
\arguments{
\item{object}{An object of class \code{"ffnn_fit"} or \code{"rnn_fit"}.}

\item{newdata}{Data frame. New data for predictions. If \code{NULL}, uses the
original training data (if available).}

\item{new_data}{Alternative to newdata (for consistency with hardhat).}

\item{type}{Character. Type of prediction:
\itemize{
\item \code{"response"} (default) – predicted values or predicted classes
\item \code{"prob"} – class probabilities (only for classification models)
}}

\item{...}{Currently unused.}
}
\value{
\itemize{
\item For \strong{regression} models: a numeric vector (single output) or matrix
(multiple outputs) of predicted values.
\item For \strong{classification} models:
\itemize{
\item \code{type = "response"}: a factor vector of predicted class labels
\item \code{type = "prob"}: a numeric matrix of class probabilities, with columns
named after the class levels.
}
}
}
\description{
Predict method for kindling basemodel fits
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalized-nn-fit.R, R/generalized-nn-fitds.R
\name{gen-nn-predict}
\alias{gen-nn-predict}
\alias{predict.nn_fit}
\alias{predict.nn_fit_tab}
\alias{predict.nn_fit_ds}
\title{Predict from a trained neural network}
\usage{
\method{predict}{nn_fit}(object, newdata = NULL, new_data = NULL, type = "response", ...)

\method{predict}{nn_fit_tab}(object, newdata = NULL, new_data = NULL, type = "response", ...)

\method{predict}{nn_fit_ds}(object, newdata = NULL, new_data = NULL, type = "response", ...)
}
\arguments{
\item{object}{A fitted model object returned by \code{\link[=train_nn]{train_nn()}}.}

\item{newdata}{New predictor data. Accepted forms depend on the method:
\itemize{
\item \code{predict.nn_fit()}: a numeric \code{matrix} or coercible object.
\item \code{predict.nn_fit_tab()}: a \code{data.frame} with the same columns used during
training; preprocessing is applied automatically via \code{hardhat::forge()}.
\item \code{predict.nn_fit_ds()}: a \code{torch} \code{dataset}, numeric \code{array}, \code{matrix}, or
\code{data.frame}.
If \code{NULL}, the cached fitted values from training are returned (not
available for \code{type = "prob"}).
}}

\item{new_data}{Legacy alias for \code{newdata}. Retained for compatibility.}

\item{type}{Character. Output type:
\itemize{
\item \code{"response"} (default): predicted class labels (factor) for
classification, or a numeric vector / matrix for regression.
\item \code{"prob"}: a numeric matrix of class probabilities (classification only).
}}

\item{...}{Currently unused; reserved for future extensions.}
}
\value{
\itemize{
\item \strong{Regression}: a numeric vector (single output) or matrix (multiple outputs).
\item \strong{Classification}, \code{type = "response"}: a factor with levels matching those
seen during training.
\item \strong{Classification}, \code{type = "prob"}: a numeric matrix with one column per
class, columns named by class label.
}
}
\description{
Generate predictions from an \code{"nn_fit"} object produced by \code{\link[=train_nn]{train_nn()}}.

Three S3 methods are registered:
\itemize{
\item \code{predict.nn_fit()} — base method for \code{matrix}-trained models.
\item \code{predict.nn_fit_tab()} — extends the base method for tabular fits; runs new
data through \code{hardhat::forge()} before predicting.
\item \code{predict.nn_fit_ds()} — extends the base method for torch \code{dataset} fits.
}
}
\seealso{
\code{\link[=train_nn]{train_nn()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/torch-code-gen.R
\name{formula_to_function}
\alias{formula_to_function}
\title{Formula to Function with Named Arguments}
\usage{
formula_to_function(
  formula_or_fn,
  default_fn = NULL,
  arg_names = NULL,
  alias_map = NULL
)
}
\arguments{
\item{formula_or_fn}{A formula or function}

\item{default_fn}{Default function if \code{formula_or_fn} is \code{NULL}}

\item{arg_names}{Character vector of formal argument names}

\item{alias_map}{Named list mapping arg_names to formula aliases (e.g., list(in_dim = ".in"))}
}
\value{
A function
}
\description{
Formula to Function with Named Arguments
}
\keyword{internal}
